// Generated by the protocol buffer compiler.  DO NOT EDIT!
// NO CHECKED-IN PROTOBUF GENCODE
// source: context/v1/stream.proto
// Protobuf C++ Version: 7.35.1

#ifndef context_2fv1_2fstream_2eproto_2epb_2eh
#define context_2fv1_2fstream_2eproto_2epb_2eh

#include <limits>
#include <string>
#include <type_traits>
#include <utility>

// clang-format off
#include "google/protobuf/runtime_version.h"
#if PROTOBUF_VERSION != 7035001
#error "Protobuf C++ gencode is built with an incompatible version of"
#error "Protobuf C++ headers/runtime. See"
#error "https://protobuf.dev/support/cross-version-runtime-guarantee/#cpp"
#endif
#include "google/protobuf/io/coded_stream.h"
#include "google/protobuf/arena.h"
#include "google/protobuf/arenastring.h"
#include "google/protobuf/generated_message_tctable_decl.h"
#include "google/protobuf/generated_message_util.h"
#include "google/protobuf/metadata_lite.h"
#include "google/protobuf/generated_message_reflection.h"
#include "google/protobuf/message.h"
#include "google/protobuf/message_lite.h"
#include "google/protobuf/repeated_field.h"  // IWYU pragma: export
#include "google/protobuf/extension_set.h"  // IWYU pragma: export
#include "google/protobuf/generated_enum_reflection.h"
#include "google/protobuf/unknown_field_set.h"
// @@protoc_insertion_point(includes)

// Must be included last.
#include "google/protobuf/port_def.inc"

#define PROTOBUF_INTERNAL_EXPORT_context_2fv1_2fstream_2eproto

namespace google {
namespace protobuf {
namespace internal {
template <typename T>
::absl::string_view GetAnyMessageName();
}  // namespace internal
}  // namespace protobuf
}  // namespace google

// Internal implementation detail -- do not use these members.
struct TableStruct_context_2fv1_2fstream_2eproto {
  static const ::uint32_t offsets[];
};
extern "C" {
extern const ::google::protobuf::internal::DescriptorTable descriptor_table_context_2fv1_2fstream_2eproto;
}  // extern "C"
namespace context {
namespace v1 {
enum StopStreamingContextsReason : int;
extern const uint32_t StopStreamingContextsReason_internal_data_[];
class ContextsCheckpoint;
struct ContextsCheckpointGlobalsTypeInternal;
#ifndef PROTOBUF_MESSAGE_GLOBALS
extern ContextsCheckpointGlobalsTypeInternal ContextsCheckpoint_globals_;
extern const ::google::protobuf::internal::ClassDataFull ContextsCheckpoint_class_data_;
#else
extern const ContextsCheckpointGlobalsTypeInternal ContextsCheckpoint_globals_;
#endif  // PROTOBUF_MESSAGE_GLOBALS
class StopStreamingContexts;
struct StopStreamingContextsGlobalsTypeInternal;
#ifndef PROTOBUF_MESSAGE_GLOBALS
extern StopStreamingContextsGlobalsTypeInternal StopStreamingContexts_globals_;
extern const ::google::protobuf::internal::ClassDataFull StopStreamingContexts_class_data_;
#else
extern const StopStreamingContextsGlobalsTypeInternal StopStreamingContexts_globals_;
#endif  // PROTOBUF_MESSAGE_GLOBALS
}  // namespace v1
}  // namespace context
namespace google {
namespace protobuf {
template <>
internal::EnumTraitsT<::context::v1::StopStreamingContextsReason_internal_data_>
    internal::EnumTraitsImpl::value<::context::v1::StopStreamingContextsReason>;
}  // namespace protobuf
}  // namespace google

namespace context {
namespace v1 {
enum StopStreamingContextsReason : int {
  RATE_LIMIT_EXCEEDED = 0,
  StopStreamingContextsReason_INT_MIN_SENTINEL_DO_NOT_USE_ =
      ::std::numeric_limits<::int32_t>::min(),
  StopStreamingContextsReason_INT_MAX_SENTINEL_DO_NOT_USE_ =
      ::std::numeric_limits<::int32_t>::max(),
};

extern const uint32_t StopStreamingContextsReason_internal_data_[];
inline constexpr StopStreamingContextsReason StopStreamingContextsReason_MIN =
    static_cast<StopStreamingContextsReason>(0);
inline constexpr StopStreamingContextsReason StopStreamingContextsReason_MAX =
    static_cast<StopStreamingContextsReason>(0);
[[nodiscard]] inline bool StopStreamingContextsReason_IsValid(int value) {
  return 0 <= value && value <= 0;
}
inline constexpr int StopStreamingContextsReason_ARRAYSIZE = 0 + 1;
[[nodiscard]] const ::google::protobuf::EnumDescriptor* PROTOBUF_NONNULL
StopStreamingContextsReason_descriptor();
[[nodiscard]] inline auto ProtobufInternalGetEnumDescriptor(StopStreamingContextsReason) {
  return StopStreamingContextsReason_descriptor();
}
template <typename T>
[[nodiscard]] const ::std::string& StopStreamingContextsReason_Name(T value) {
  static_assert(::std::is_same<T, StopStreamingContextsReason>::value ||
                    ::std::is_integral<T>::value,
                "Incorrect type passed to StopStreamingContextsReason_Name().");
  return StopStreamingContextsReason_Name(static_cast<StopStreamingContextsReason>(value));
}
template <>
[[nodiscard]] inline const ::std::string& StopStreamingContextsReason_Name(StopStreamingContextsReason value) {
  return ::google::protobuf::internal::NameOfDenseEnum<StopStreamingContextsReason_descriptor, 0, 0>(
      static_cast<int>(value));
}
[[nodiscard]] inline bool StopStreamingContextsReason_Parse(
    ::absl::string_view name, StopStreamingContextsReason* PROTOBUF_NONNULL value) {
  return ::google::protobuf::internal::ParseNamedEnum<StopStreamingContextsReason>(StopStreamingContextsReason_descriptor(), name,
                                           value);
}
using ::google::protobuf::internal::generated_enum::AbslParseFlag;
using ::google::protobuf::internal::generated_enum::AbslUnparseFlag;

// ===================================================================


// -------------------------------------------------------------------

class  PROTOBUF_FUTURE_ADD_EARLY_WARN_UNUSED StopStreamingContexts final : public ::google::protobuf::Message
/* @@protoc_insertion_point(class_definition:context.v1.StopStreamingContexts) */ {
 public:
  inline StopStreamingContexts() : StopStreamingContexts(nullptr) {}
  ~StopStreamingContexts() PROTOBUF_FINAL;

#if defined(PROTOBUF_CUSTOM_VTABLE)
  void operator delete(StopStreamingContexts* PROTOBUF_NONNULL msg, ::std::destroying_delete_t) {
    SharedDtor(*msg);
    ::google::protobuf::internal::SizedDelete(msg, sizeof(StopStreamingContexts));
  }
#endif

  template <typename = void>
  explicit constexpr StopStreamingContexts(::google::protobuf::internal::ConstantInitialized,
                           const ::google::protobuf::internal::ClassData* PROTOBUF_NONNULL
                               class_data);

  inline StopStreamingContexts(const StopStreamingContexts& from) : StopStreamingContexts(nullptr, from) {}
  inline StopStreamingContexts(StopStreamingContexts&& from) noexcept : StopStreamingContexts(nullptr, ::std::move(from)) {}
  inline StopStreamingContexts& operator=(const StopStreamingContexts& from) {
    CopyFrom(from);
    return *this;
  }
  inline StopStreamingContexts& operator=(StopStreamingContexts&& from) noexcept {
    if (this == &from) return *this;
    if (::google::protobuf::internal::CanMoveWithInternalSwap(GetArena(), from.GetArena())) {
      InternalSwap(&from);
    } else {
      CopyFrom(from);
    }
    return *this;
  }

  [[nodiscard]] inline const ::google::protobuf::UnknownFieldSet& unknown_fields() const
      ABSL_ATTRIBUTE_LIFETIME_BOUND {
    return _internal_metadata_.unknown_fields<::google::protobuf::UnknownFieldSet>(::google::protobuf::UnknownFieldSet::default_instance);
  }
  [[nodiscard]] inline ::google::protobuf::UnknownFieldSet* PROTOBUF_NONNULL
  mutable_unknown_fields() ABSL_ATTRIBUTE_LIFETIME_BOUND {
    return _internal_metadata_.mutable_unknown_fields<::google::protobuf::UnknownFieldSet>();
  }

  [[nodiscard]] static const ::google::protobuf::Descriptor* PROTOBUF_NONNULL descriptor() {
    return GetDescriptor();
  }
  [[nodiscard]] static const ::google::protobuf::Descriptor* PROTOBUF_NONNULL
  GetDescriptor() {
    return default_instance().GetMetadata().descriptor;
  }
  [[nodiscard]] static const ::google::protobuf::Reflection* PROTOBUF_NONNULL GetReflection() {
    return default_instance().GetMetadata().reflection;
  }
  [[nodiscard]] static const StopStreamingContexts& default_instance() {
    return *::google::protobuf::internal::MessageGlobalsBase::ToDefaultInstance<StopStreamingContexts>(&StopStreamingContexts_globals_);
  }
  static constexpr int kIndexInFileMessages = 1;
  friend void swap(StopStreamingContexts& a, StopStreamingContexts& b) { a.Swap(&b); }
  inline void Swap(StopStreamingContexts* PROTOBUF_NONNULL other) {
    if (other == this) return;
    if (::google::protobuf::internal::CanUseInternalSwap(GetArena(), other->GetArena())) {
      InternalSwap(other);
    } else {
      ::google::protobuf::internal::GenericSwap(this, other);
    }
  }
  void UnsafeArenaSwap(StopStreamingContexts* PROTOBUF_NONNULL other) {
    if (other == this) return;
    ABSL_DCHECK(GetArena() == other->GetArena());
    InternalSwap(other);
  }

  // implements Message ----------------------------------------------

  [[nodiscard]] StopStreamingContexts* PROTOBUF_NONNULL
  New(::google::protobuf::Arena* PROTOBUF_NULLABLE arena = nullptr) const {
    return ::google::protobuf::Message::DefaultConstruct<StopStreamingContexts>(arena);
  }
  using ::google::protobuf::Message::CopyFrom;
  void CopyFrom(const StopStreamingContexts& from);
  using ::google::protobuf::Message::MergeFrom;
  void MergeFrom(const StopStreamingContexts& from) { StopStreamingContexts::MergeImpl(*this, from); }

  private:
  static void MergeImpl(::google::protobuf::MessageLite& to_msg,
                        const ::google::protobuf::MessageLite& from_msg);

  public:
  [[nodiscard]] bool IsInitialized() const {
    return true;
  }
  ABSL_ATTRIBUTE_REINITIALIZES void Clear() PROTOBUF_FINAL;
  #if defined(PROTOBUF_CUSTOM_VTABLE)
  private:
  [[nodiscard]] static ::size_t ByteSizeLong(const ::google::protobuf::MessageLite& msg);
  [[nodiscard]] static ::uint8_t* PROTOBUF_NONNULL _InternalSerialize(
      const ::google::protobuf::MessageLite& msg, ::uint8_t* PROTOBUF_NONNULL target,
      ::google::protobuf::io::EpsCopyOutputStream* PROTOBUF_NONNULL stream);

  public:
  [[nodiscard]] ::size_t ByteSizeLong() const { return ByteSizeLong(*this); }
  [[nodiscard]] ::uint8_t* PROTOBUF_NONNULL _InternalSerialize(
      ::uint8_t* PROTOBUF_NONNULL target,
      ::google::protobuf::io::EpsCopyOutputStream* PROTOBUF_NONNULL stream) const {
    return _InternalSerialize(*this, target, stream);
  }
  #else   // PROTOBUF_CUSTOM_VTABLE
  [[nodiscard]] ::size_t ByteSizeLong() const final;
  [[nodiscard]] ::uint8_t* PROTOBUF_NONNULL _InternalSerialize(
      ::uint8_t* PROTOBUF_NONNULL target,
      ::google::protobuf::io::EpsCopyOutputStream* PROTOBUF_NONNULL stream) const final;
  #endif  // PROTOBUF_CUSTOM_VTABLE
  [[nodiscard]] int GetCachedSize() const {
    return _impl_._cached_size_.Get();
  }

  private:
  void SharedCtor(::google::protobuf::Arena* PROTOBUF_NULLABLE arena);
  static void SharedDtor(MessageLite& self);
  void InternalSwap(StopStreamingContexts* PROTOBUF_NONNULL other);
 private:
  template <typename T>
  friend ::absl::string_view(::google::protobuf::internal::GetAnyMessageName)();
  static ::absl::string_view FullMessageName() { return "context.v1.StopStreamingContexts"; }

  explicit StopStreamingContexts(::google::protobuf::Arena* PROTOBUF_NULLABLE arena);
  StopStreamingContexts(::google::protobuf::Arena* PROTOBUF_NULLABLE arena, const StopStreamingContexts& from);
  StopStreamingContexts(
      ::google::protobuf::Arena* PROTOBUF_NULLABLE arena, StopStreamingContexts&& from) noexcept
      : StopStreamingContexts(arena) {
    *this = ::std::move(from);
  }
  const ::google::protobuf::internal::ClassData* PROTOBUF_NONNULL GetClassData() const PROTOBUF_FINAL;
  static void* PROTOBUF_NONNULL PlacementNew_(
      const void* PROTOBUF_NONNULL, void* PROTOBUF_NONNULL mem,
      ::google::protobuf::Arena* PROTOBUF_NULLABLE arena);
  static constexpr auto InternalNewImpl_();

 public:
  static constexpr auto InternalGenerateClassData_(
      const MessageLite& prototype,
      const ::google::protobuf::internal::TcParseTableBase* PROTOBUF_NULLABLE tc_table = nullptr);

  [[nodiscard]] ::google::protobuf::Metadata GetMetadata() const;
  // nested types ----------------------------------------------------

  // accessors -------------------------------------------------------
  enum : int {
    kClaimIdFieldNumber = 1,
    kNodeIdFieldNumber = 2,
    kReasonFieldNumber = 3,
  };
  // string claim_id = 1;
  void clear_claim_id() ;
  [[nodiscard]] const ::std::string& claim_id() const;
  template <typename Arg_ = const ::std::string&, typename... Args_>
  void set_claim_id(Arg_&& arg, Args_... args);
  ::std::string* PROTOBUF_NONNULL mutable_claim_id();
  [[nodiscard]] ::std::string* PROTOBUF_NULLABLE release_claim_id();
  void set_allocated_claim_id(::std::string* PROTOBUF_NULLABLE value);

  private:
  const ::std::string& _internal_claim_id() const;
  PROTOBUF_ALWAYS_INLINE void _internal_set_claim_id(const ::std::string& value);
  ::std::string* PROTOBUF_NONNULL _internal_mutable_claim_id();

  public:
  // string node_id = 2;
  void clear_node_id() ;
  [[nodiscard]] const ::std::string& node_id() const;
  template <typename Arg_ = const ::std::string&, typename... Args_>
  void set_node_id(Arg_&& arg, Args_... args);
  ::std::string* PROTOBUF_NONNULL mutable_node_id();
  [[nodiscard]] ::std::string* PROTOBUF_NULLABLE release_node_id();
  void set_allocated_node_id(::std::string* PROTOBUF_NULLABLE value);

  private:
  const ::std::string& _internal_node_id() const;
  PROTOBUF_ALWAYS_INLINE void _internal_set_node_id(const ::std::string& value);
  ::std::string* PROTOBUF_NONNULL _internal_mutable_node_id();

  public:
  // .context.v1.StopStreamingContextsReason reason = 3;
  void clear_reason() ;
  [[nodiscard]] ::context::v1::StopStreamingContextsReason reason() const;
  void set_reason(::context::v1::StopStreamingContextsReason value);

  private:
  ::context::v1::StopStreamingContextsReason _internal_reason() const;
  void _internal_set_reason(::context::v1::StopStreamingContextsReason value);

  public:
  // @@protoc_insertion_point(class_scope:context.v1.StopStreamingContexts)
 private:
  class _Internal;
  using ParseTableT_ =
      ::google::protobuf::internal::TcParseTable<2, 3,
                          0, 56,
                          2>;
  static constexpr ParseTableT_ InternalGenerateParseTable_(
      const ::google::protobuf::internal::ClassData* PROTOBUF_NONNULL class_data);
  friend class ::google::protobuf::internal::TcParser;
  #ifndef PROTOBUF_MESSAGE_GLOBALS
  static const ParseTableT_ _table_;
  #endif

  friend class ::google::protobuf::MessageLite;
  friend class ::google::protobuf::Arena;
  friend ::google::protobuf::internal::PrivateAccess;
  template <typename T>
  friend class ::google::protobuf::Arena::InternalHelper;
  using InternalArenaConstructable_ = void;
  using DestructorSkippable_ = void;
  struct Impl_ {
    inline explicit constexpr Impl_(::google::protobuf::internal::InternalVisibility visibility,
                                    ::google::protobuf::internal::ConstantInitialized) noexcept;
    inline explicit Impl_(
        ::google::protobuf::internal::InternalVisibility visibility,
        ::google::protobuf::Arena* PROTOBUF_NULLABLE arena);
    inline explicit Impl_(
        ::google::protobuf::internal::InternalVisibility visibility,
        ::google::protobuf::Arena* PROTOBUF_NULLABLE arena, const Impl_& from,
        const StopStreamingContexts& from_msg);
    ::google::protobuf::internal::HasBits<1> _has_bits_;
    ::google::protobuf::internal::CachedSize _cached_size_;
    ::google::protobuf::internal::ArenaStringPtr claim_id_;
    ::google::protobuf::internal::ArenaStringPtr node_id_;
    int reason_;
    PROTOBUF_TSAN_DECLARE_MEMBER
  };
  union { Impl_ _impl_; };
  friend struct ::TableStruct_context_2fv1_2fstream_2eproto;
};
// -------------------------------------------------------------------

class  PROTOBUF_FUTURE_ADD_EARLY_WARN_UNUSED ContextsCheckpoint final : public ::google::protobuf::Message
/* @@protoc_insertion_point(class_definition:context.v1.ContextsCheckpoint) */ {
 public:
  inline ContextsCheckpoint() : ContextsCheckpoint(nullptr) {}
  ~ContextsCheckpoint() PROTOBUF_FINAL;

#if defined(PROTOBUF_CUSTOM_VTABLE)
  void operator delete(ContextsCheckpoint* PROTOBUF_NONNULL msg, ::std::destroying_delete_t) {
    SharedDtor(*msg);
    ::google::protobuf::internal::SizedDelete(msg, sizeof(ContextsCheckpoint));
  }
#endif

  template <typename = void>
  explicit constexpr ContextsCheckpoint(::google::protobuf::internal::ConstantInitialized,
                           const ::google::protobuf::internal::ClassData* PROTOBUF_NONNULL
                               class_data);

  inline ContextsCheckpoint(const ContextsCheckpoint& from) : ContextsCheckpoint(nullptr, from) {}
  inline ContextsCheckpoint(ContextsCheckpoint&& from) noexcept : ContextsCheckpoint(nullptr, ::std::move(from)) {}
  inline ContextsCheckpoint& operator=(const ContextsCheckpoint& from) {
    CopyFrom(from);
    return *this;
  }
  inline ContextsCheckpoint& operator=(ContextsCheckpoint&& from) noexcept {
    if (this == &from) return *this;
    if (::google::protobuf::internal::CanMoveWithInternalSwap(GetArena(), from.GetArena())) {
      InternalSwap(&from);
    } else {
      CopyFrom(from);
    }
    return *this;
  }

  [[nodiscard]] inline const ::google::protobuf::UnknownFieldSet& unknown_fields() const
      ABSL_ATTRIBUTE_LIFETIME_BOUND {
    return _internal_metadata_.unknown_fields<::google::protobuf::UnknownFieldSet>(::google::protobuf::UnknownFieldSet::default_instance);
  }
  [[nodiscard]] inline ::google::protobuf::UnknownFieldSet* PROTOBUF_NONNULL
  mutable_unknown_fields() ABSL_ATTRIBUTE_LIFETIME_BOUND {
    return _internal_metadata_.mutable_unknown_fields<::google::protobuf::UnknownFieldSet>();
  }

  [[nodiscard]] static const ::google::protobuf::Descriptor* PROTOBUF_NONNULL descriptor() {
    return GetDescriptor();
  }
  [[nodiscard]] static const ::google::protobuf::Descriptor* PROTOBUF_NONNULL
  GetDescriptor() {
    return default_instance().GetMetadata().descriptor;
  }
  [[nodiscard]] static const ::google::protobuf::Reflection* PROTOBUF_NONNULL GetReflection() {
    return default_instance().GetMetadata().reflection;
  }
  [[nodiscard]] static const ContextsCheckpoint& default_instance() {
    return *::google::protobuf::internal::MessageGlobalsBase::ToDefaultInstance<ContextsCheckpoint>(&ContextsCheckpoint_globals_);
  }
  static constexpr int kIndexInFileMessages = 0;
  friend void swap(ContextsCheckpoint& a, ContextsCheckpoint& b) { a.Swap(&b); }
  inline void Swap(ContextsCheckpoint* PROTOBUF_NONNULL other) {
    if (other == this) return;
    if (::google::protobuf::internal::CanUseInternalSwap(GetArena(), other->GetArena())) {
      InternalSwap(other);
    } else {
      ::google::protobuf::internal::GenericSwap(this, other);
    }
  }
  void UnsafeArenaSwap(ContextsCheckpoint* PROTOBUF_NONNULL other) {
    if (other == this) return;
    ABSL_DCHECK(GetArena() == other->GetArena());
    InternalSwap(other);
  }

  // implements Message ----------------------------------------------

  [[nodiscard]] ContextsCheckpoint* PROTOBUF_NONNULL
  New(::google::protobuf::Arena* PROTOBUF_NULLABLE arena = nullptr) const {
    return ::google::protobuf::Message::DefaultConstruct<ContextsCheckpoint>(arena);
  }
  using ::google::protobuf::Message::CopyFrom;
  void CopyFrom(const ContextsCheckpoint& from);
  using ::google::protobuf::Message::MergeFrom;
  void MergeFrom(const ContextsCheckpoint& from) { ContextsCheckpoint::MergeImpl(*this, from); }

  private:
  static void MergeImpl(::google::protobuf::MessageLite& to_msg,
                        const ::google::protobuf::MessageLite& from_msg);

  public:
  [[nodiscard]] bool IsInitialized() const {
    return true;
  }
  ABSL_ATTRIBUTE_REINITIALIZES void Clear() PROTOBUF_FINAL;
  #if defined(PROTOBUF_CUSTOM_VTABLE)
  private:
  [[nodiscard]] static ::size_t ByteSizeLong(const ::google::protobuf::MessageLite& msg);
  [[nodiscard]] static ::uint8_t* PROTOBUF_NONNULL _InternalSerialize(
      const ::google::protobuf::MessageLite& msg, ::uint8_t* PROTOBUF_NONNULL target,
      ::google::protobuf::io::EpsCopyOutputStream* PROTOBUF_NONNULL stream);

  public:
  [[nodiscard]] ::size_t ByteSizeLong() const { return ByteSizeLong(*this); }
  [[nodiscard]] ::uint8_t* PROTOBUF_NONNULL _InternalSerialize(
      ::uint8_t* PROTOBUF_NONNULL target,
      ::google::protobuf::io::EpsCopyOutputStream* PROTOBUF_NONNULL stream) const {
    return _InternalSerialize(*this, target, stream);
  }
  #else   // PROTOBUF_CUSTOM_VTABLE
  [[nodiscard]] ::size_t ByteSizeLong() const final;
  [[nodiscard]] ::uint8_t* PROTOBUF_NONNULL _InternalSerialize(
      ::uint8_t* PROTOBUF_NONNULL target,
      ::google::protobuf::io::EpsCopyOutputStream* PROTOBUF_NONNULL stream) const final;
  #endif  // PROTOBUF_CUSTOM_VTABLE
  [[nodiscard]] int GetCachedSize() const {
    return _impl_._cached_size_.Get();
  }

  private:
  void SharedCtor(::google::protobuf::Arena* PROTOBUF_NULLABLE arena);
  static void SharedDtor(MessageLite& self);
  void InternalSwap(ContextsCheckpoint* PROTOBUF_NONNULL other);
 private:
  template <typename T>
  friend ::absl::string_view(::google::protobuf::internal::GetAnyMessageName)();
  static ::absl::string_view FullMessageName() { return "context.v1.ContextsCheckpoint"; }

  explicit ContextsCheckpoint(::google::protobuf::Arena* PROTOBUF_NULLABLE arena);
  ContextsCheckpoint(::google::protobuf::Arena* PROTOBUF_NULLABLE arena, const ContextsCheckpoint& from);
  ContextsCheckpoint(
      ::google::protobuf::Arena* PROTOBUF_NULLABLE arena, ContextsCheckpoint&& from) noexcept
      : ContextsCheckpoint(arena) {
    *this = ::std::move(from);
  }
  const ::google::protobuf::internal::ClassData* PROTOBUF_NONNULL GetClassData() const PROTOBUF_FINAL;
  static void* PROTOBUF_NONNULL PlacementNew_(
      const void* PROTOBUF_NONNULL, void* PROTOBUF_NONNULL mem,
      ::google::protobuf::Arena* PROTOBUF_NULLABLE arena);
  static constexpr auto InternalNewImpl_();

 public:
  static constexpr auto InternalGenerateClassData_(
      const MessageLite& prototype,
      const ::google::protobuf::internal::TcParseTableBase* PROTOBUF_NULLABLE tc_table = nullptr);

  [[nodiscard]] ::google::protobuf::Metadata GetMetadata() const;
  // nested types ----------------------------------------------------

  // accessors -------------------------------------------------------
  enum : int {
    kClaimIdFieldNumber = 1,
    kNodeIdFieldNumber = 2,
    kVersionHashFieldNumber = 3,
  };
  // string claim_id = 1;
  void clear_claim_id() ;
  [[nodiscard]] const ::std::string& claim_id() const;
  template <typename Arg_ = const ::std::string&, typename... Args_>
  void set_claim_id(Arg_&& arg, Args_... args);
  ::std::string* PROTOBUF_NONNULL mutable_claim_id();
  [[nodiscard]] ::std::string* PROTOBUF_NULLABLE release_claim_id();
  void set_allocated_claim_id(::std::string* PROTOBUF_NULLABLE value);

  private:
  const ::std::string& _internal_claim_id() const;
  PROTOBUF_ALWAYS_INLINE void _internal_set_claim_id(const ::std::string& value);
  ::std::string* PROTOBUF_NONNULL _internal_mutable_claim_id();

  public:
  // string node_id = 2;
  void clear_node_id() ;
  [[nodiscard]] const ::std::string& node_id() const;
  template <typename Arg_ = const ::std::string&, typename... Args_>
  void set_node_id(Arg_&& arg, Args_... args);
  ::std::string* PROTOBUF_NONNULL mutable_node_id();
  [[nodiscard]] ::std::string* PROTOBUF_NULLABLE release_node_id();
  void set_allocated_node_id(::std::string* PROTOBUF_NULLABLE value);

  private:
  const ::std::string& _internal_node_id() const;
  PROTOBUF_ALWAYS_INLINE void _internal_set_node_id(const ::std::string& value);
  ::std::string* PROTOBUF_NONNULL _internal_mutable_node_id();

  public:
  // uint64 version_hash = 3;
  void clear_version_hash() ;
  [[nodiscard]] ::uint64_t version_hash() const;
  void set_version_hash(::uint64_t value);

  private:
  ::uint64_t _internal_version_hash() const;
  void _internal_set_version_hash(::uint64_t value);

  public:
  // @@protoc_insertion_point(class_scope:context.v1.ContextsCheckpoint)
 private:
  class _Internal;
  using ParseTableT_ =
      ::google::protobuf::internal::TcParseTable<2, 3,
                          0, 53,
                          2>;
  static constexpr ParseTableT_ InternalGenerateParseTable_(
      const ::google::protobuf::internal::ClassData* PROTOBUF_NONNULL class_data);
  friend class ::google::protobuf::internal::TcParser;
  #ifndef PROTOBUF_MESSAGE_GLOBALS
  static const ParseTableT_ _table_;
  #endif

  friend class ::google::protobuf::MessageLite;
  friend class ::google::protobuf::Arena;
  friend ::google::protobuf::internal::PrivateAccess;
  template <typename T>
  friend class ::google::protobuf::Arena::InternalHelper;
  using InternalArenaConstructable_ = void;
  using DestructorSkippable_ = void;
  struct Impl_ {
    inline explicit constexpr Impl_(::google::protobuf::internal::InternalVisibility visibility,
                                    ::google::protobuf::internal::ConstantInitialized) noexcept;
    inline explicit Impl_(
        ::google::protobuf::internal::InternalVisibility visibility,
        ::google::protobuf::Arena* PROTOBUF_NULLABLE arena);
    inline explicit Impl_(
        ::google::protobuf::internal::InternalVisibility visibility,
        ::google::protobuf::Arena* PROTOBUF_NULLABLE arena, const Impl_& from,
        const ContextsCheckpoint& from_msg);
    ::google::protobuf::internal::HasBits<1> _has_bits_;
    ::google::protobuf::internal::CachedSize _cached_size_;
    ::google::protobuf::internal::ArenaStringPtr claim_id_;
    ::google::protobuf::internal::ArenaStringPtr node_id_;
    ::uint64_t version_hash_;
    PROTOBUF_TSAN_DECLARE_MEMBER
  };
  union { Impl_ _impl_; };
  friend struct ::TableStruct_context_2fv1_2fstream_2eproto;
};

// ===================================================================




// ===================================================================


#ifdef __GNUC__
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wstrict-aliasing"
#endif  // __GNUC__
// -------------------------------------------------------------------

// ContextsCheckpoint

// string claim_id = 1;
inline void ContextsCheckpoint::clear_claim_id() {
  ::google::protobuf::internal::TSanWrite(&_impl_);
  _impl_.claim_id_.ClearToEmpty();
  ClearHasBit(_impl_._has_bits_[0], 0x00000001U);
}
inline const ::std::string& ContextsCheckpoint::claim_id() const
    ABSL_ATTRIBUTE_LIFETIME_BOUND {
  // @@protoc_insertion_point(field_get:context.v1.ContextsCheckpoint.claim_id)
  return _internal_claim_id();
}
template <typename Arg_, typename... Args_>
PROTOBUF_ALWAYS_INLINE void ContextsCheckpoint::set_claim_id(Arg_&& arg, Args_... args) {
  ::google::protobuf::internal::TSanWrite(&_impl_);
  SetHasBit(_impl_._has_bits_[0], 0x00000001U);
  _impl_.claim_id_.Set(static_cast<Arg_&&>(arg), args..., GetArena());
  // @@protoc_insertion_point(field_set:context.v1.ContextsCheckpoint.claim_id)
}
inline ::std::string* PROTOBUF_NONNULL ContextsCheckpoint::mutable_claim_id()
    ABSL_ATTRIBUTE_LIFETIME_BOUND {
  SetHasBit(_impl_._has_bits_[0], 0x00000001U);
  ::std::string* _s = _internal_mutable_claim_id();
  // @@protoc_insertion_point(field_mutable:context.v1.ContextsCheckpoint.claim_id)
  return _s;
}
inline const ::std::string& ContextsCheckpoint::_internal_claim_id() const {
  ::google::protobuf::internal::TSanRead(&_impl_);
  return _impl_.claim_id_.Get();
}
inline void ContextsCheckpoint::_internal_set_claim_id(const ::std::string& value) {
  ::google::protobuf::internal::TSanWrite(&_impl_);
  _impl_.claim_id_.Set(value, GetArena());
}
inline ::std::string* PROTOBUF_NONNULL ContextsCheckpoint::_internal_mutable_claim_id() {
  ::google::protobuf::internal::TSanWrite(&_impl_);
  return _impl_.claim_id_.Mutable( GetArena());
}
inline ::std::string* PROTOBUF_NULLABLE ContextsCheckpoint::release_claim_id() {
  ::google::protobuf::internal::TSanWrite(&_impl_);
  // @@protoc_insertion_point(field_release:context.v1.ContextsCheckpoint.claim_id)
  if (!CheckHasBit(_impl_._has_bits_[0], 0x00000001U)) {
    return nullptr;
  }
  ClearHasBit(_impl_._has_bits_[0], 0x00000001U);
  auto* released = _impl_.claim_id_.Release();
  if (::google::protobuf::internal::DebugHardenForceCopyDefaultString()) {
    _impl_.claim_id_.Set("", GetArena());
  }
  return released;
}
inline void ContextsCheckpoint::set_allocated_claim_id(::std::string* PROTOBUF_NULLABLE value) {
  ::google::protobuf::internal::TSanWrite(&_impl_);
  if (value != nullptr) {
    SetHasBit(_impl_._has_bits_[0], 0x00000001U);
  } else {
    ClearHasBit(_impl_._has_bits_[0], 0x00000001U);
  }
  _impl_.claim_id_.SetAllocated(value, GetArena());
  if (::google::protobuf::internal::DebugHardenForceCopyDefaultString() && _impl_.claim_id_.IsDefault()) {
    _impl_.claim_id_.Set("", GetArena());
  }
  // @@protoc_insertion_point(field_set_allocated:context.v1.ContextsCheckpoint.claim_id)
}

// string node_id = 2;
inline void ContextsCheckpoint::clear_node_id() {
  ::google::protobuf::internal::TSanWrite(&_impl_);
  _impl_.node_id_.ClearToEmpty();
  ClearHasBit(_impl_._has_bits_[0], 0x00000002U);
}
inline const ::std::string& ContextsCheckpoint::node_id() const
    ABSL_ATTRIBUTE_LIFETIME_BOUND {
  // @@protoc_insertion_point(field_get:context.v1.ContextsCheckpoint.node_id)
  return _internal_node_id();
}
template <typename Arg_, typename... Args_>
PROTOBUF_ALWAYS_INLINE void ContextsCheckpoint::set_node_id(Arg_&& arg, Args_... args) {
  ::google::protobuf::internal::TSanWrite(&_impl_);
  SetHasBit(_impl_._has_bits_[0], 0x00000002U);
  _impl_.node_id_.Set(static_cast<Arg_&&>(arg), args..., GetArena());
  // @@protoc_insertion_point(field_set:context.v1.ContextsCheckpoint.node_id)
}
inline ::std::string* PROTOBUF_NONNULL ContextsCheckpoint::mutable_node_id()
    ABSL_ATTRIBUTE_LIFETIME_BOUND {
  SetHasBit(_impl_._has_bits_[0], 0x00000002U);
  ::std::string* _s = _internal_mutable_node_id();
  // @@protoc_insertion_point(field_mutable:context.v1.ContextsCheckpoint.node_id)
  return _s;
}
inline const ::std::string& ContextsCheckpoint::_internal_node_id() const {
  ::google::protobuf::internal::TSanRead(&_impl_);
  return _impl_.node_id_.Get();
}
inline void ContextsCheckpoint::_internal_set_node_id(const ::std::string& value) {
  ::google::protobuf::internal::TSanWrite(&_impl_);
  _impl_.node_id_.Set(value, GetArena());
}
inline ::std::string* PROTOBUF_NONNULL ContextsCheckpoint::_internal_mutable_node_id() {
  ::google::protobuf::internal::TSanWrite(&_impl_);
  return _impl_.node_id_.Mutable( GetArena());
}
inline ::std::string* PROTOBUF_NULLABLE ContextsCheckpoint::release_node_id() {
  ::google::protobuf::internal::TSanWrite(&_impl_);
  // @@protoc_insertion_point(field_release:context.v1.ContextsCheckpoint.node_id)
  if (!CheckHasBit(_impl_._has_bits_[0], 0x00000002U)) {
    return nullptr;
  }
  ClearHasBit(_impl_._has_bits_[0], 0x00000002U);
  auto* released = _impl_.node_id_.Release();
  if (::google::protobuf::internal::DebugHardenForceCopyDefaultString()) {
    _impl_.node_id_.Set("", GetArena());
  }
  return released;
}
inline void ContextsCheckpoint::set_allocated_node_id(::std::string* PROTOBUF_NULLABLE value) {
  ::google::protobuf::internal::TSanWrite(&_impl_);
  if (value != nullptr) {
    SetHasBit(_impl_._has_bits_[0], 0x00000002U);
  } else {
    ClearHasBit(_impl_._has_bits_[0], 0x00000002U);
  }
  _impl_.node_id_.SetAllocated(value, GetArena());
  if (::google::protobuf::internal::DebugHardenForceCopyDefaultString() && _impl_.node_id_.IsDefault()) {
    _impl_.node_id_.Set("", GetArena());
  }
  // @@protoc_insertion_point(field_set_allocated:context.v1.ContextsCheckpoint.node_id)
}

// uint64 version_hash = 3;
inline void ContextsCheckpoint::clear_version_hash() {
  ::google::protobuf::internal::TSanWrite(&_impl_);
  _impl_.version_hash_ = ::uint64_t{0u};
  ClearHasBit(_impl_._has_bits_[0], 0x00000004U);
}
inline ::uint64_t ContextsCheckpoint::version_hash() const {
  // @@protoc_insertion_point(field_get:context.v1.ContextsCheckpoint.version_hash)
  return _internal_version_hash();
}
inline void ContextsCheckpoint::set_version_hash(::uint64_t value) {
  _internal_set_version_hash(value);
  SetHasBit(_impl_._has_bits_[0], 0x00000004U);
  // @@protoc_insertion_point(field_set:context.v1.ContextsCheckpoint.version_hash)
}
inline ::uint64_t ContextsCheckpoint::_internal_version_hash() const {
  ::google::protobuf::internal::TSanRead(&_impl_);
  return _impl_.version_hash_;
}
inline void ContextsCheckpoint::_internal_set_version_hash(::uint64_t value) {
  ::google::protobuf::internal::TSanWrite(&_impl_);
  _impl_.version_hash_ = value;
}

// -------------------------------------------------------------------

// StopStreamingContexts

// string claim_id = 1;
inline void StopStreamingContexts::clear_claim_id() {
  ::google::protobuf::internal::TSanWrite(&_impl_);
  _impl_.claim_id_.ClearToEmpty();
  ClearHasBit(_impl_._has_bits_[0], 0x00000001U);
}
inline const ::std::string& StopStreamingContexts::claim_id() const
    ABSL_ATTRIBUTE_LIFETIME_BOUND {
  // @@protoc_insertion_point(field_get:context.v1.StopStreamingContexts.claim_id)
  return _internal_claim_id();
}
template <typename Arg_, typename... Args_>
PROTOBUF_ALWAYS_INLINE void StopStreamingContexts::set_claim_id(Arg_&& arg, Args_... args) {
  ::google::protobuf::internal::TSanWrite(&_impl_);
  SetHasBit(_impl_._has_bits_[0], 0x00000001U);
  _impl_.claim_id_.Set(static_cast<Arg_&&>(arg), args..., GetArena());
  // @@protoc_insertion_point(field_set:context.v1.StopStreamingContexts.claim_id)
}
inline ::std::string* PROTOBUF_NONNULL StopStreamingContexts::mutable_claim_id()
    ABSL_ATTRIBUTE_LIFETIME_BOUND {
  SetHasBit(_impl_._has_bits_[0], 0x00000001U);
  ::std::string* _s = _internal_mutable_claim_id();
  // @@protoc_insertion_point(field_mutable:context.v1.StopStreamingContexts.claim_id)
  return _s;
}
inline const ::std::string& StopStreamingContexts::_internal_claim_id() const {
  ::google::protobuf::internal::TSanRead(&_impl_);
  return _impl_.claim_id_.Get();
}
inline void StopStreamingContexts::_internal_set_claim_id(const ::std::string& value) {
  ::google::protobuf::internal::TSanWrite(&_impl_);
  _impl_.claim_id_.Set(value, GetArena());
}
inline ::std::string* PROTOBUF_NONNULL StopStreamingContexts::_internal_mutable_claim_id() {
  ::google::protobuf::internal::TSanWrite(&_impl_);
  return _impl_.claim_id_.Mutable( GetArena());
}
inline ::std::string* PROTOBUF_NULLABLE StopStreamingContexts::release_claim_id() {
  ::google::protobuf::internal::TSanWrite(&_impl_);
  // @@protoc_insertion_point(field_release:context.v1.StopStreamingContexts.claim_id)
  if (!CheckHasBit(_impl_._has_bits_[0], 0x00000001U)) {
    return nullptr;
  }
  ClearHasBit(_impl_._has_bits_[0], 0x00000001U);
  auto* released = _impl_.claim_id_.Release();
  if (::google::protobuf::internal::DebugHardenForceCopyDefaultString()) {
    _impl_.claim_id_.Set("", GetArena());
  }
  return released;
}
inline void StopStreamingContexts::set_allocated_claim_id(::std::string* PROTOBUF_NULLABLE value) {
  ::google::protobuf::internal::TSanWrite(&_impl_);
  if (value != nullptr) {
    SetHasBit(_impl_._has_bits_[0], 0x00000001U);
  } else {
    ClearHasBit(_impl_._has_bits_[0], 0x00000001U);
  }
  _impl_.claim_id_.SetAllocated(value, GetArena());
  if (::google::protobuf::internal::DebugHardenForceCopyDefaultString() && _impl_.claim_id_.IsDefault()) {
    _impl_.claim_id_.Set("", GetArena());
  }
  // @@protoc_insertion_point(field_set_allocated:context.v1.StopStreamingContexts.claim_id)
}

// string node_id = 2;
inline void StopStreamingContexts::clear_node_id() {
  ::google::protobuf::internal::TSanWrite(&_impl_);
  _impl_.node_id_.ClearToEmpty();
  ClearHasBit(_impl_._has_bits_[0], 0x00000002U);
}
inline const ::std::string& StopStreamingContexts::node_id() const
    ABSL_ATTRIBUTE_LIFETIME_BOUND {
  // @@protoc_insertion_point(field_get:context.v1.StopStreamingContexts.node_id)
  return _internal_node_id();
}
template <typename Arg_, typename... Args_>
PROTOBUF_ALWAYS_INLINE void StopStreamingContexts::set_node_id(Arg_&& arg, Args_... args) {
  ::google::protobuf::internal::TSanWrite(&_impl_);
  SetHasBit(_impl_._has_bits_[0], 0x00000002U);
  _impl_.node_id_.Set(static_cast<Arg_&&>(arg), args..., GetArena());
  // @@protoc_insertion_point(field_set:context.v1.StopStreamingContexts.node_id)
}
inline ::std::string* PROTOBUF_NONNULL StopStreamingContexts::mutable_node_id()
    ABSL_ATTRIBUTE_LIFETIME_BOUND {
  SetHasBit(_impl_._has_bits_[0], 0x00000002U);
  ::std::string* _s = _internal_mutable_node_id();
  // @@protoc_insertion_point(field_mutable:context.v1.StopStreamingContexts.node_id)
  return _s;
}
inline const ::std::string& StopStreamingContexts::_internal_node_id() const {
  ::google::protobuf::internal::TSanRead(&_impl_);
  return _impl_.node_id_.Get();
}
inline void StopStreamingContexts::_internal_set_node_id(const ::std::string& value) {
  ::google::protobuf::internal::TSanWrite(&_impl_);
  _impl_.node_id_.Set(value, GetArena());
}
inline ::std::string* PROTOBUF_NONNULL StopStreamingContexts::_internal_mutable_node_id() {
  ::google::protobuf::internal::TSanWrite(&_impl_);
  return _impl_.node_id_.Mutable( GetArena());
}
inline ::std::string* PROTOBUF_NULLABLE StopStreamingContexts::release_node_id() {
  ::google::protobuf::internal::TSanWrite(&_impl_);
  // @@protoc_insertion_point(field_release:context.v1.StopStreamingContexts.node_id)
  if (!CheckHasBit(_impl_._has_bits_[0], 0x00000002U)) {
    return nullptr;
  }
  ClearHasBit(_impl_._has_bits_[0], 0x00000002U);
  auto* released = _impl_.node_id_.Release();
  if (::google::protobuf::internal::DebugHardenForceCopyDefaultString()) {
    _impl_.node_id_.Set("", GetArena());
  }
  return released;
}
inline void StopStreamingContexts::set_allocated_node_id(::std::string* PROTOBUF_NULLABLE value) {
  ::google::protobuf::internal::TSanWrite(&_impl_);
  if (value != nullptr) {
    SetHasBit(_impl_._has_bits_[0], 0x00000002U);
  } else {
    ClearHasBit(_impl_._has_bits_[0], 0x00000002U);
  }
  _impl_.node_id_.SetAllocated(value, GetArena());
  if (::google::protobuf::internal::DebugHardenForceCopyDefaultString() && _impl_.node_id_.IsDefault()) {
    _impl_.node_id_.Set("", GetArena());
  }
  // @@protoc_insertion_point(field_set_allocated:context.v1.StopStreamingContexts.node_id)
}

// .context.v1.StopStreamingContextsReason reason = 3;
inline void StopStreamingContexts::clear_reason() {
  ::google::protobuf::internal::TSanWrite(&_impl_);
  _impl_.reason_ = 0;
  ClearHasBit(_impl_._has_bits_[0], 0x00000004U);
}
inline ::context::v1::StopStreamingContextsReason StopStreamingContexts::reason() const {
  // @@protoc_insertion_point(field_get:context.v1.StopStreamingContexts.reason)
  return _internal_reason();
}
inline void StopStreamingContexts::set_reason(::context::v1::StopStreamingContextsReason value) {
  _internal_set_reason(value);
  SetHasBit(_impl_._has_bits_[0], 0x00000004U);
  // @@protoc_insertion_point(field_set:context.v1.StopStreamingContexts.reason)
}
inline ::context::v1::StopStreamingContextsReason StopStreamingContexts::_internal_reason() const {
  ::google::protobuf::internal::TSanRead(&_impl_);
  return static_cast<::context::v1::StopStreamingContextsReason>(_impl_.reason_);
}
inline void StopStreamingContexts::_internal_set_reason(::context::v1::StopStreamingContextsReason value) {
  ::google::protobuf::internal::TSanWrite(&_impl_);
  _impl_.reason_ = value;
}

#ifdef __GNUC__
#pragma GCC diagnostic pop
#endif  // __GNUC__

// @@protoc_insertion_point(namespace_scope)
}  // namespace v1
}  // namespace context


namespace google {
namespace protobuf {

template <>
struct is_proto_enum<::context::v1::StopStreamingContextsReason> : std::true_type {};
template <>
inline const EnumDescriptor* PROTOBUF_NONNULL GetEnumDescriptor<::context::v1::StopStreamingContextsReason>() {
  return ::context::v1::StopStreamingContextsReason_descriptor();
}

}  // namespace protobuf
}  // namespace google

// @@protoc_insertion_point(global_scope)

#include "google/protobuf/port_undef.inc"
// clang-format on

#endif  // context_2fv1_2fstream_2eproto_2epb_2eh
