// Generated by the protocol buffer compiler.  DO NOT EDIT!
// NO CHECKED-IN PROTOBUF GENCODE
// source: aclk/v1/lib.proto
// Protobuf C++ Version: 7.35.1

#include "aclk/v1/lib.pb.h"

#include <algorithm>
#include <type_traits>
#include "google/protobuf/io/coded_stream.h"
#include "google/protobuf/generated_message_tctable_impl.h"
#include "google/protobuf/internal_visibility.h"
#include "google/protobuf/extension_set.h"
#include "google/protobuf/generated_message_util.h"
#include "google/protobuf/wire_format_lite.h"
#include "google/protobuf/descriptor.h"
#include "google/protobuf/generated_message_reflection.h"
#include "google/protobuf/reflection_ops.h"
#include "google/protobuf/wire_format.h"
// @@protoc_insertion_point(includes)

// Must be included last.
#include "google/protobuf/port_def.inc"
PROTOBUF_PRAGMA_INIT_SEG
namespace _pb = ::google::protobuf;
namespace _pbi = ::google::protobuf::internal;
namespace _fl = ::google::protobuf::internal::field_layout;
#ifdef PROTOBUF_MESSAGE_GLOBALS
namespace {
PROTOBUF_CONSTINIT ::google::protobuf::internal::ReflectionData
    file_reflection_data[] = {
        // ::aclk_lib::v1::ACLKMessagePosition
        {&::_pbi::kDescriptorMethods, &::descriptor_table_aclk_2fv1_2flib_2eproto, /* tracker*/ nullptr,},
        // ::aclk_lib::v1::Capability
        {&::_pbi::kDescriptorMethods, &::descriptor_table_aclk_2fv1_2flib_2eproto, /* tracker*/ nullptr,},
};
}  // namespace
#endif
namespace aclk_lib {
namespace v1 {
class Capability::_Internal {
 public:
  using HasBits = decltype(::std::declval<Capability>()._impl_._has_bits_);
  static constexpr ::int32_t kHasBitsOffset =
      8 * PROTOBUF_FIELD_OFFSET(Capability, _impl_._has_bits_);
};

constexpr Capability::ParseTableT_ Capability::InternalGenerateParseTable_(const ::_pbi::ClassData* class_data) {
  return ParseTableT_{
    {
      PROTOBUF_FIELD_OFFSET(Capability, _impl_._has_bits_),
      0, // no _extensions_
      3, 24,  // max_field_number, fast_idx_mask
      offsetof(ParseTableT_, field_lookup_table),
      4294967288,  // skipmap
      offsetof(ParseTableT_, field_entries),
      3,  // num_field_entries
      0,  // num_aux_entries
      offsetof(ParseTableT_, field_names),  // no aux_entries
      class_data,
      nullptr,  // post_loop_handler
      ::_pbi::TcParser::GenericFallback,  // fallback
      #ifdef PROTOBUF_PREFETCH_PARSE_TABLE
      ::_pbi::TcParser::GetTable<::aclk_lib::v1::Capability>(),  // to_prefetch
      #endif  // PROTOBUF_PREFETCH_PARSE_TABLE
    }, {{
      {::_pbi::TcParser::MiniParse, {}},
      // string name = 1;
      {::_pbi::TcParser::FastUS1,
       {10, 0, 0,
        PROTOBUF_FIELD_OFFSET(Capability, _impl_.name_)}},
      // uint32 version = 2;
      {::_pbi::TcParser::SingularVarintNoZag1<::uint32_t, offsetof(Capability, _impl_.version_), 1>(),
       {16, 1, 0,
        PROTOBUF_FIELD_OFFSET(Capability, _impl_.version_)}},
      // bool enabled = 3;
      {::_pbi::TcParser::SingularVarintNoZag1<bool, offsetof(Capability, _impl_.enabled_), 2>(),
       {24, 2, 0,
        PROTOBUF_FIELD_OFFSET(Capability, _impl_.enabled_)}},
    }}, {{
      65535, 65535
    }}, {{
      // string name = 1;
      {PROTOBUF_FIELD_OFFSET(Capability, _impl_.name_), _Internal::kHasBitsOffset + 0, 0, (0 | ::_fl::kFcOptional | ::_fl::kUtf8String | ::_fl::kRepAString)},
      // uint32 version = 2;
      {PROTOBUF_FIELD_OFFSET(Capability, _impl_.version_), _Internal::kHasBitsOffset + 1, 0, (0 | ::_fl::kFcOptional | ::_fl::kUInt32)},
      // bool enabled = 3;
      {PROTOBUF_FIELD_OFFSET(Capability, _impl_.enabled_), _Internal::kHasBitsOffset + 2, 0, (0 | ::_fl::kFcOptional | ::_fl::kBool)},
    }},
    // no aux_entries
    {{
      "\26\4\0\0\0\0\0\0"
      "aclk_lib.v1.Capability"
      "name"
    }},
  };
}


inline constexpr Capability::Impl_::Impl_(
    [[maybe_unused]] ::google::protobuf::internal::InternalVisibility visibility,
    ::_pbi::ConstantInitialized) noexcept
      : _cached_size_{0},
        name_(
            &::google::protobuf::internal::fixed_address_empty_string,
            ::_pbi::ConstantInitialized()),
        version_{0u},
        enabled_{false} {}

template <typename>
constexpr Capability::Capability(::_pbi::ConstantInitialized,
                       const ::_pbi::ClassData* PROTOBUF_NONNULL class_data)
    : ::google::protobuf::Message(
#if defined(PROTOBUF_CUSTOM_VTABLE)
          class_data
#endif  // PROTOBUF_CUSTOM_VTABLE
          ),
      _impl_(internal_visibility(), ::_pbi::ConstantInitialized()) {
}
inline void* PROTOBUF_NONNULL Capability::PlacementNew_(
    const void* PROTOBUF_NONNULL, void* PROTOBUF_NONNULL mem,
    ::google::protobuf::Arena* PROTOBUF_NULLABLE arena) {
  return ::new (mem) Capability(arena);
}
constexpr auto Capability::InternalNewImpl_() {
  return ::google::protobuf::internal::MessageCreator::CopyInit(sizeof(Capability), alignof(Capability));
}
constexpr auto Capability::InternalGenerateClassData_(
    const MessageLite& prototype,
    const ::google::protobuf::internal::TcParseTableBase* tc_table) {
  return ::google::protobuf::internal::ClassDataFull{
      ::google::protobuf::internal::ClassData{
          &prototype,
#ifndef PROTOBUF_MESSAGE_GLOBALS
          &_table_.header,
#else
          tc_table,
#endif
          nullptr,  // IsInitialized
          &Capability::MergeImpl,
          ::google::protobuf::Message::GetNewImpl<Capability>(),
#if defined(PROTOBUF_CUSTOM_VTABLE)
          &Capability::SharedDtor,
          ::google::protobuf::Message::GetClearImpl<Capability>(), &Capability::ByteSizeLong,
              &Capability::_InternalSerialize,
#endif  // PROTOBUF_CUSTOM_VTABLE
          PROTOBUF_FIELD_OFFSET(Capability, _impl_._cached_size_),
          false,
      },
#ifdef PROTOBUF_MESSAGE_GLOBALS
      &file_reflection_data[1],
#else   // !PROTOBUF_MESSAGE_GLOBALS
      &::_pbi::kDescriptorMethods,
      &descriptor_table_aclk_2fv1_2flib_2eproto,
      nullptr,  // tracker
#endif  // PROTOBUF_MESSAGE_GLOBALS
  };
}
struct CapabilityGlobalsTypeInternal : ::_pbi::MessageGlobalsBase {
  constexpr CapabilityGlobalsTypeInternal()
      :
#ifndef PROTOBUF_MESSAGE_GLOBALS
        _default(::_pbi::ConstantInitialized{},
                 Capability_class_data_.base())
#else   // !PROTOBUF_MESSAGE_GLOBALS
        MessageGlobalsBase(Capability::InternalGenerateClassData_(
            _default, &Capability_globals_._table.header)),
        _default(::_pbi::ConstantInitialized{}, GetClassData()),
        _table(::_pbi::PrivateAccess::GenerateParseTable<Capability>(
            GetClassData()))
#endif  // PROTOBUF_MESSAGE_GLOBALS
  {
  }
  ~CapabilityGlobalsTypeInternal() {}
  union {
    alignas(::_pbi::kMaxMessageAlignment) Capability _default;
  };
#ifdef PROTOBUF_MESSAGE_GLOBALS
  decltype(::_pbi::PrivateAccess::GenerateParseTable<Capability>(
      ::std::declval<const ::_pbi::ClassData*>())) _table;
#endif
};
#ifdef PROTOBUF_MESSAGE_GLOBALS
static_assert(PROTOBUF_FIELD_OFFSET(CapabilityGlobalsTypeInternal, _default) ==
              ::_pbi::MessageGlobalsBase::OffsetToDefault());
#endif  // PROTOBUF_MESSAGE_GLOBALS

PROTOBUF_ATTRIBUTE_NO_DESTROY PROTOBUF_CONSTINIT
    PROTOBUF_ATTRIBUTE_INIT_PRIORITY1 PROTOBUF_MESSAGE_GLOBALS_CONST CapabilityGlobalsTypeInternal Capability_globals_
        PROTOBUF_MESSAGE_GLOBALS_SECTION(.data.rel.ro);
#if defined(PROTOBUF_CUSTOM_VTABLE)
namespace {
const ::_pbi::ClassData* Capability_get_class_data() {
#ifdef PROTOBUF_MESSAGE_GLOBALS
  return Capability_globals_.GetClassData();
#else
  return Capability_class_data_.base();
#endif  // PROTOBUF_MESSAGE_GLOBALS
}
}  // namespace
#endif  // PROTOBUF_CUSTOM_VTABLE
class ACLKMessagePosition::_Internal {
 public:
  using HasBits = decltype(::std::declval<ACLKMessagePosition>()._impl_._has_bits_);
  static constexpr ::int32_t kHasBitsOffset =
      8 * PROTOBUF_FIELD_OFFSET(ACLKMessagePosition, _impl_._has_bits_);
};

constexpr ACLKMessagePosition::ParseTableT_ ACLKMessagePosition::InternalGenerateParseTable_(const ::_pbi::ClassData* class_data) {
  return ParseTableT_{
    {
      PROTOBUF_FIELD_OFFSET(ACLKMessagePosition, _impl_._has_bits_),
      0, // no _extensions_
      3, 24,  // max_field_number, fast_idx_mask
      offsetof(ParseTableT_, field_lookup_table),
      4294967288,  // skipmap
      offsetof(ParseTableT_, field_entries),
      3,  // num_field_entries
      1,  // num_aux_entries
      offsetof(ParseTableT_, aux_entries),
      class_data,
      nullptr,  // post_loop_handler
      ::_pbi::TcParser::GenericFallback,  // fallback
      #ifdef PROTOBUF_PREFETCH_PARSE_TABLE
      ::_pbi::TcParser::GetTable<::aclk_lib::v1::ACLKMessagePosition>(),  // to_prefetch
      #endif  // PROTOBUF_PREFETCH_PARSE_TABLE
    }, {{
      {::_pbi::TcParser::MiniParse, {}},
      // uint64 sequence_id = 1;
      {::_pbi::TcParser::SingularVarintNoZag1<::uint64_t, offsetof(ACLKMessagePosition, _impl_.sequence_id_), 1>(),
       {8, 1, 0,
        PROTOBUF_FIELD_OFFSET(ACLKMessagePosition, _impl_.sequence_id_)}},
      // .google.protobuf.Timestamp seq_id_created_at = 2;
      {::_pbi::TcParser::FastMtS1,
       {18, 0, 0,
        PROTOBUF_FIELD_OFFSET(ACLKMessagePosition, _impl_.seq_id_created_at_)}},
      // uint64 previous_sequence_id = 3;
      {::_pbi::TcParser::SingularVarintNoZag1<::uint64_t, offsetof(ACLKMessagePosition, _impl_.previous_sequence_id_), 2>(),
       {24, 2, 0,
        PROTOBUF_FIELD_OFFSET(ACLKMessagePosition, _impl_.previous_sequence_id_)}},
    }}, {{
      65535, 65535
    }}, {{
      // uint64 sequence_id = 1;
      {PROTOBUF_FIELD_OFFSET(ACLKMessagePosition, _impl_.sequence_id_), _Internal::kHasBitsOffset + 1, 0, (0 | ::_fl::kFcOptional | ::_fl::kUInt64)},
      // .google.protobuf.Timestamp seq_id_created_at = 2;
      {PROTOBUF_FIELD_OFFSET(ACLKMessagePosition, _impl_.seq_id_created_at_), _Internal::kHasBitsOffset + 0, 0, (0 | ::_fl::kFcOptional | ::_fl::kMessage | ::_fl::kTvTable)},
      // uint64 previous_sequence_id = 3;
      {PROTOBUF_FIELD_OFFSET(ACLKMessagePosition, _impl_.previous_sequence_id_), _Internal::kHasBitsOffset + 2, 0, (0 | ::_fl::kFcOptional | ::_fl::kUInt64)},
    }},
    {{
        #ifndef PROTOBUF_MESSAGE_GLOBALS
        {::_pbi::TcParser::GetTable<::google::protobuf::Timestamp>()},
        #else
        {::_pbi::FieldAuxMessageGlobals(), &::google::protobuf::Timestamp_globals_},
        #endif
    }},
    {{
    }},
  };
}


inline constexpr ACLKMessagePosition::Impl_::Impl_(
    [[maybe_unused]] ::google::protobuf::internal::InternalVisibility visibility,
    ::_pbi::ConstantInitialized) noexcept
      : _cached_size_{0},
        seq_id_created_at_{nullptr},
        sequence_id_{::uint64_t{0u}},
        previous_sequence_id_{::uint64_t{0u}} {}

template <typename>
constexpr ACLKMessagePosition::ACLKMessagePosition(::_pbi::ConstantInitialized,
                       const ::_pbi::ClassData* PROTOBUF_NONNULL class_data)
    : ::google::protobuf::Message(
#if defined(PROTOBUF_CUSTOM_VTABLE)
          class_data
#endif  // PROTOBUF_CUSTOM_VTABLE
          ),
      _impl_(internal_visibility(), ::_pbi::ConstantInitialized()) {
}
inline void* PROTOBUF_NONNULL ACLKMessagePosition::PlacementNew_(
    const void* PROTOBUF_NONNULL, void* PROTOBUF_NONNULL mem,
    ::google::protobuf::Arena* PROTOBUF_NULLABLE arena) {
  return ::new (mem) ACLKMessagePosition(arena);
}
constexpr auto ACLKMessagePosition::InternalNewImpl_() {
  return ::google::protobuf::internal::MessageCreator::ZeroInit(sizeof(ACLKMessagePosition), alignof(ACLKMessagePosition));
}
constexpr auto ACLKMessagePosition::InternalGenerateClassData_(
    const MessageLite& prototype,
    const ::google::protobuf::internal::TcParseTableBase* tc_table) {
  return ::google::protobuf::internal::ClassDataFull{
      ::google::protobuf::internal::ClassData{
          &prototype,
#ifndef PROTOBUF_MESSAGE_GLOBALS
          &_table_.header,
#else
          tc_table,
#endif
          nullptr,  // IsInitialized
          &ACLKMessagePosition::MergeImpl,
          ::google::protobuf::Message::GetNewImpl<ACLKMessagePosition>(),
#if defined(PROTOBUF_CUSTOM_VTABLE)
          &ACLKMessagePosition::SharedDtor,
          ::google::protobuf::Message::GetClearImpl<ACLKMessagePosition>(), &ACLKMessagePosition::ByteSizeLong,
              &ACLKMessagePosition::_InternalSerialize,
#endif  // PROTOBUF_CUSTOM_VTABLE
          PROTOBUF_FIELD_OFFSET(ACLKMessagePosition, _impl_._cached_size_),
          false,
      },
#ifdef PROTOBUF_MESSAGE_GLOBALS
      &file_reflection_data[0],
#else   // !PROTOBUF_MESSAGE_GLOBALS
      &::_pbi::kDescriptorMethods,
      &descriptor_table_aclk_2fv1_2flib_2eproto,
      nullptr,  // tracker
#endif  // PROTOBUF_MESSAGE_GLOBALS
  };
}
struct ACLKMessagePositionGlobalsTypeInternal : ::_pbi::MessageGlobalsBase {
  constexpr ACLKMessagePositionGlobalsTypeInternal()
      :
#ifndef PROTOBUF_MESSAGE_GLOBALS
        _default(::_pbi::ConstantInitialized{},
                 ACLKMessagePosition_class_data_.base())
#else   // !PROTOBUF_MESSAGE_GLOBALS
        MessageGlobalsBase(ACLKMessagePosition::InternalGenerateClassData_(
            _default, &ACLKMessagePosition_globals_._table.header)),
        _default(::_pbi::ConstantInitialized{}, GetClassData()),
        _table(::_pbi::PrivateAccess::GenerateParseTable<ACLKMessagePosition>(
            GetClassData()))
#endif  // PROTOBUF_MESSAGE_GLOBALS
  {
  }
  ~ACLKMessagePositionGlobalsTypeInternal() {}
  union {
    alignas(::_pbi::kMaxMessageAlignment) ACLKMessagePosition _default;
  };
#ifdef PROTOBUF_MESSAGE_GLOBALS
  decltype(::_pbi::PrivateAccess::GenerateParseTable<ACLKMessagePosition>(
      ::std::declval<const ::_pbi::ClassData*>())) _table;
#endif
};
#ifdef PROTOBUF_MESSAGE_GLOBALS
static_assert(PROTOBUF_FIELD_OFFSET(ACLKMessagePositionGlobalsTypeInternal, _default) ==
              ::_pbi::MessageGlobalsBase::OffsetToDefault());
#endif  // PROTOBUF_MESSAGE_GLOBALS

PROTOBUF_ATTRIBUTE_NO_DESTROY PROTOBUF_CONSTINIT
    PROTOBUF_ATTRIBUTE_INIT_PRIORITY1 PROTOBUF_MESSAGE_GLOBALS_CONST ACLKMessagePositionGlobalsTypeInternal ACLKMessagePosition_globals_
        PROTOBUF_MESSAGE_GLOBALS_SECTION(.data.rel.ro);
#if defined(PROTOBUF_CUSTOM_VTABLE)
namespace {
const ::_pbi::ClassData* ACLKMessagePosition_get_class_data() {
#ifdef PROTOBUF_MESSAGE_GLOBALS
  return ACLKMessagePosition_globals_.GetClassData();
#else
  return ACLKMessagePosition_class_data_.base();
#endif  // PROTOBUF_MESSAGE_GLOBALS
}
}  // namespace
#endif  // PROTOBUF_CUSTOM_VTABLE
}  // namespace v1
}  // namespace aclk_lib
static constexpr const ::_pb::EnumDescriptor* PROTOBUF_NONNULL* PROTOBUF_NULLABLE
    file_level_enum_descriptors_aclk_2fv1_2flib_2eproto = nullptr;
static constexpr const ::_pb::ServiceDescriptor* PROTOBUF_NONNULL* PROTOBUF_NULLABLE
    file_level_service_descriptors_aclk_2fv1_2flib_2eproto = nullptr;
const ::uint32_t
    TableStruct_aclk_2fv1_2flib_2eproto::offsets[] ABSL_ATTRIBUTE_SECTION_VARIABLE(
        protodesc_cold) = {
        0x081, // bitmap
        PROTOBUF_FIELD_OFFSET(::aclk_lib::v1::ACLKMessagePosition, _impl_._has_bits_),
        6, // hasbit index offset
        PROTOBUF_FIELD_OFFSET(::aclk_lib::v1::ACLKMessagePosition, _impl_.sequence_id_),
        PROTOBUF_FIELD_OFFSET(::aclk_lib::v1::ACLKMessagePosition, _impl_.seq_id_created_at_),
        PROTOBUF_FIELD_OFFSET(::aclk_lib::v1::ACLKMessagePosition, _impl_.previous_sequence_id_),
        1,
        0,
        2,
        0x081, // bitmap
        PROTOBUF_FIELD_OFFSET(::aclk_lib::v1::Capability, _impl_._has_bits_),
        6, // hasbit index offset
        PROTOBUF_FIELD_OFFSET(::aclk_lib::v1::Capability, _impl_.name_),
        PROTOBUF_FIELD_OFFSET(::aclk_lib::v1::Capability, _impl_.version_),
        PROTOBUF_FIELD_OFFSET(::aclk_lib::v1::Capability, _impl_.enabled_),
        0,
        1,
        2,
};

static const ::_pbi::MigrationSchema
    schemas[] ABSL_ATTRIBUTE_SECTION_VARIABLE(protodesc_cold) = {
        {0, sizeof(::aclk_lib::v1::ACLKMessagePosition)},
        {9, sizeof(::aclk_lib::v1::Capability)},
};
static const ::_pbi::MessageGlobalsBase* PROTOBUF_NONNULL const
    file_message_globals[] = {
        &::aclk_lib::v1::ACLKMessagePosition_globals_,
        &::aclk_lib::v1::Capability_globals_,
};
const char descriptor_table_protodef_aclk_2fv1_2flib_2eproto[] ABSL_ATTRIBUTE_SECTION_VARIABLE(
    protodesc_cold) = {
    "\n\021aclk/v1/lib.proto\022\013aclk_lib.v1\032\037google"
    "/protobuf/timestamp.proto\"\177\n\023ACLKMessage"
    "Position\022\023\n\013sequence_id\030\001 \001(\004\0225\n\021seq_id_"
    "created_at\030\002 \001(\0132\032.google.protobuf.Times"
    "tamp\022\034\n\024previous_sequence_id\030\003 \001(\004\"<\n\nCa"
    "pability\022\014\n\004name\030\001 \001(\t\022\017\n\007version\030\002 \001(\r\022"
    "\017\n\007enabled\030\003 \001(\010B\025Z\023aclk_lib/v1;aclklibb"
    "\006proto3"
};
static const ::_pbi::DescriptorTable* PROTOBUF_NONNULL const
    descriptor_table_aclk_2fv1_2flib_2eproto_deps[1] = {
        &::descriptor_table_google_2fprotobuf_2ftimestamp_2eproto,
};
static ::absl::once_flag descriptor_table_aclk_2fv1_2flib_2eproto_once;
PROTOBUF_CONSTINIT const ::_pbi::DescriptorTable descriptor_table_aclk_2fv1_2flib_2eproto = {
    false,
    false,
    287,
    descriptor_table_protodef_aclk_2fv1_2flib_2eproto,
    "aclk/v1/lib.proto",
    &descriptor_table_aclk_2fv1_2flib_2eproto_once,
    descriptor_table_aclk_2fv1_2flib_2eproto_deps,
    1,
    2,
    schemas,
    file_message_globals,
    TableStruct_aclk_2fv1_2flib_2eproto::offsets,
    file_level_enum_descriptors_aclk_2fv1_2flib_2eproto,
    file_level_service_descriptors_aclk_2fv1_2flib_2eproto,
};
namespace aclk_lib {
namespace v1 {
// ===================================================================

void ACLKMessagePosition::clear_seq_id_created_at() {
  ::google::protobuf::internal::TSanWrite(&_impl_);
  if (_impl_.seq_id_created_at_ != nullptr) _impl_.seq_id_created_at_->Clear();
  ClearHasBit(_impl_._has_bits_[0], 0x00000001U);
}
ACLKMessagePosition::ACLKMessagePosition(::google::protobuf::Arena* PROTOBUF_NULLABLE arena)
#if defined(PROTOBUF_CUSTOM_VTABLE)
    : ::google::protobuf::Message(arena, ACLKMessagePosition_get_class_data()) {
#else   // PROTOBUF_CUSTOM_VTABLE
    : ::google::protobuf::Message(arena) {
#endif  // PROTOBUF_CUSTOM_VTABLE
  SharedCtor(arena);
  // @@protoc_insertion_point(arena_constructor:aclk_lib.v1.ACLKMessagePosition)
}
PROTOBUF_NDEBUG_INLINE ACLKMessagePosition::Impl_::Impl_(
    [[maybe_unused]] ::google::protobuf::internal::InternalVisibility visibility,
    [[maybe_unused]] ::google::protobuf::Arena* PROTOBUF_NULLABLE arena, const Impl_& from,
    [[maybe_unused]] const ::aclk_lib::v1::ACLKMessagePosition& from_msg)
      : _has_bits_{from._has_bits_},
        _cached_size_{0} {}

ACLKMessagePosition::ACLKMessagePosition(
    ::google::protobuf::Arena* PROTOBUF_NULLABLE arena,
    const ACLKMessagePosition& from)
#if defined(PROTOBUF_CUSTOM_VTABLE)
    : ::google::protobuf::Message(arena, ACLKMessagePosition_get_class_data()) {

#else   // PROTOBUF_CUSTOM_VTABLE
    : ::google::protobuf::Message(arena) {
#endif  // PROTOBUF_CUSTOM_VTABLE
  ACLKMessagePosition* const _this = this;
  (void)_this;
  _internal_metadata_.MergeFrom<::google::protobuf::UnknownFieldSet>(
      from._internal_metadata_);
  new (&_impl_) Impl_(internal_visibility(), arena, from._impl_, from);
  ::uint32_t cached_has_bits = _impl_._has_bits_[0];
  _impl_.seq_id_created_at_ = (CheckHasBit(cached_has_bits, 0x00000001U))
                ? ::google::protobuf::Message::CopyConstruct(arena, *from._impl_.seq_id_created_at_)
                : nullptr;
  ::memcpy(reinterpret_cast<char*>(&_impl_) +
               offsetof(Impl_, sequence_id_),
           reinterpret_cast<const char*>(&from._impl_) +
               offsetof(Impl_, sequence_id_),
           offsetof(Impl_, previous_sequence_id_) -
               offsetof(Impl_, sequence_id_) +
               sizeof(Impl_::previous_sequence_id_));

  // @@protoc_insertion_point(copy_constructor:aclk_lib.v1.ACLKMessagePosition)
}
PROTOBUF_NDEBUG_INLINE ACLKMessagePosition::Impl_::Impl_(
    [[maybe_unused]] ::google::protobuf::internal::InternalVisibility visibility,
    [[maybe_unused]] ::google::protobuf::Arena* PROTOBUF_NULLABLE arena)
      : _cached_size_{0} {}

inline void ACLKMessagePosition::SharedCtor(::_pb::Arena* PROTOBUF_NULLABLE arena) {
  new (&_impl_) Impl_(internal_visibility(), arena);
  ::memset(reinterpret_cast<char*>(&_impl_) +
               offsetof(Impl_, seq_id_created_at_),
           0,
           offsetof(Impl_, previous_sequence_id_) -
               offsetof(Impl_, seq_id_created_at_) +
               sizeof(Impl_::previous_sequence_id_));
}
ACLKMessagePosition::~ACLKMessagePosition() {
  // @@protoc_insertion_point(destructor:aclk_lib.v1.ACLKMessagePosition)
  SharedDtor(*this);
}
inline void ACLKMessagePosition::SharedDtor(MessageLite& self) {
  ACLKMessagePosition& this_ = static_cast<ACLKMessagePosition&>(self);
  if constexpr (::_pbi::DebugHardenCheckHasBitConsistency()) {
    this_.CheckHasBitConsistency();
  }
  this_._internal_metadata_.Delete<::google::protobuf::UnknownFieldSet>();
  ABSL_DCHECK(this_.GetArena() == nullptr);
  delete this_._impl_.seq_id_created_at_;
  this_._impl_.~Impl_();
}

#ifndef PROTOBUF_MESSAGE_GLOBALS
PROTOBUF_CONSTINIT PROTOBUF_ATTRIBUTE_INIT_PRIORITY1 const
    ::google::protobuf::internal::ClassDataFull ACLKMessagePosition_class_data_ =
        ACLKMessagePosition::InternalGenerateClassData_(ACLKMessagePosition_globals_._default);

PROTOBUF_ATTRIBUTE_WEAK const ::google::protobuf::internal::ClassData* PROTOBUF_NONNULL
ACLKMessagePosition::GetClassData() const {
  ::google::protobuf::internal::PrefetchToLocalCache(&ACLKMessagePosition_class_data_);
  ::google::protobuf::internal::PrefetchToLocalCache(ACLKMessagePosition_class_data_.tc_table);
  return ACLKMessagePosition_class_data_.base();
}
#else
PROTOBUF_ATTRIBUTE_WEAK const ::google::protobuf::internal::ClassData* PROTOBUF_NONNULL
ACLKMessagePosition::GetClassData() const {
  ::google::protobuf::internal::PrefetchToLocalCache(&ACLKMessagePosition_globals_);
  ::google::protobuf::internal::PrefetchToLocalCache(
      ::google::protobuf::internal::MessageGlobalsBase::ToParseTableBase(&ACLKMessagePosition_globals_));
  return ACLKMessagePosition_globals_.GetClassData();
}
#endif  // !PROTOBUF_MESSAGE_GLOBALS
#ifndef PROTOBUF_MESSAGE_GLOBALS
PROTOBUF_CONSTINIT
PROTOBUF_ATTRIBUTE_INIT_PRIORITY1 const ACLKMessagePosition::ParseTableT_
    ACLKMessagePosition::_table_ =
        ACLKMessagePosition::InternalGenerateParseTable_(ACLKMessagePosition_class_data_.base());
#endif  // !PROTOBUF_MESSAGE_GLOBALS
PROTOBUF_NOINLINE void ACLKMessagePosition::Clear() {
// @@protoc_insertion_point(message_clear_start:aclk_lib.v1.ACLKMessagePosition)
  ::google::protobuf::internal::TSanWrite(&_impl_);
  ::uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void) cached_has_bits;

  cached_has_bits = _impl_._has_bits_[0];
  if (CheckHasBit(cached_has_bits, 0x00000001U)) {
    ABSL_DCHECK(_impl_.seq_id_created_at_ != nullptr);
    _impl_.seq_id_created_at_->Clear();
  }
  if (BatchCheckHasBit(cached_has_bits, 0x00000006U)) {
    ::memset(&_impl_.sequence_id_, 0, static_cast<::size_t>(
        reinterpret_cast<char*>(&_impl_.previous_sequence_id_) -
        reinterpret_cast<char*>(&_impl_.sequence_id_)) + sizeof(_impl_.previous_sequence_id_));
  }
  _impl_._has_bits_.Clear();
  _internal_metadata_.Clear<::google::protobuf::UnknownFieldSet>();
}

#if defined(PROTOBUF_CUSTOM_VTABLE)
::uint8_t* PROTOBUF_NONNULL ACLKMessagePosition::_InternalSerialize(
    const ::google::protobuf::MessageLite& base, ::uint8_t* PROTOBUF_NONNULL target,
    ::google::protobuf::io::EpsCopyOutputStream* PROTOBUF_NONNULL stream) {
  const ACLKMessagePosition& this_ = static_cast<const ACLKMessagePosition&>(base);
#else   // PROTOBUF_CUSTOM_VTABLE
::uint8_t* PROTOBUF_NONNULL ACLKMessagePosition::_InternalSerialize(
    ::uint8_t* PROTOBUF_NONNULL target,
    ::google::protobuf::io::EpsCopyOutputStream* PROTOBUF_NONNULL stream) const {
  const ACLKMessagePosition& this_ = *this;
#endif  // PROTOBUF_CUSTOM_VTABLE
  if constexpr (::_pbi::DebugHardenCheckHasBitConsistency()) {
    this_.CheckHasBitConsistency();
  }
  // @@protoc_insertion_point(serialize_to_array_start:aclk_lib.v1.ACLKMessagePosition)
  ::uint32_t cached_has_bits = 0;
  (void)cached_has_bits;

  cached_has_bits = this_._impl_._has_bits_[0];
  // uint64 sequence_id = 1;
  if (CheckHasBit(cached_has_bits, 0x00000002U)) {
    if (this_._internal_sequence_id() != 0) {
      target = stream->EnsureSpace(target);
      target = ::_pbi::WireFormatLite::WriteUInt64ToArray(
          1, this_._internal_sequence_id(), target);
    }
  }

  // .google.protobuf.Timestamp seq_id_created_at = 2;
  if (CheckHasBit(cached_has_bits, 0x00000001U)) {
    target = ::google::protobuf::internal::WireFormatLite::InternalWriteMessage(
        2, *this_._impl_.seq_id_created_at_, this_._impl_.seq_id_created_at_->GetCachedSize(), target,
        stream);
  }

  // uint64 previous_sequence_id = 3;
  if (CheckHasBit(cached_has_bits, 0x00000004U)) {
    if (this_._internal_previous_sequence_id() != 0) {
      target = stream->EnsureSpace(target);
      target = ::_pbi::WireFormatLite::WriteUInt64ToArray(
          3, this_._internal_previous_sequence_id(), target);
    }
  }

  if (ABSL_PREDICT_FALSE(this_._internal_metadata_.have_unknown_fields())) {
    target =
        ::_pbi::WireFormat::InternalSerializeUnknownFieldsToArray(
            this_._internal_metadata_.unknown_fields<::google::protobuf::UnknownFieldSet>(::google::protobuf::UnknownFieldSet::default_instance), target, stream);
  }
  // @@protoc_insertion_point(serialize_to_array_end:aclk_lib.v1.ACLKMessagePosition)
  return target;
}

#if defined(PROTOBUF_CUSTOM_VTABLE)
::size_t ACLKMessagePosition::ByteSizeLong(const MessageLite& base) {
  const ACLKMessagePosition& this_ = static_cast<const ACLKMessagePosition&>(base);
#else   // PROTOBUF_CUSTOM_VTABLE
::size_t ACLKMessagePosition::ByteSizeLong() const {
  const ACLKMessagePosition& this_ = *this;
#endif  // PROTOBUF_CUSTOM_VTABLE
  // @@protoc_insertion_point(message_byte_size_start:aclk_lib.v1.ACLKMessagePosition)
  ::size_t total_size = 0;

  ::uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void)cached_has_bits;

  ::_pbi::Prefetch5LinesFrom7Lines(&this_);
  cached_has_bits = this_._impl_._has_bits_[0];
  if (BatchCheckHasBit(cached_has_bits, 0x00000007U)) {
    // .google.protobuf.Timestamp seq_id_created_at = 2;
    if (CheckHasBit(cached_has_bits, 0x00000001U)) {
      total_size += 1 +
                    ::google::protobuf::internal::WireFormatLite::MessageSize(*this_._impl_.seq_id_created_at_);
    }
    // uint64 sequence_id = 1;
    if (CheckHasBit(cached_has_bits, 0x00000002U)) {
      if (this_._internal_sequence_id() != 0) {
        total_size += ::_pbi::WireFormatLite::UInt64SizePlusOne(
            this_._internal_sequence_id());
      }
    }
    // uint64 previous_sequence_id = 3;
    if (CheckHasBit(cached_has_bits, 0x00000004U)) {
      if (this_._internal_previous_sequence_id() != 0) {
        total_size += ::_pbi::WireFormatLite::UInt64SizePlusOne(
            this_._internal_previous_sequence_id());
      }
    }
  }
  return this_.MaybeComputeUnknownFieldsSize(total_size,
                                             &this_._impl_._cached_size_);
}

void ACLKMessagePosition::MergeImpl(::google::protobuf::MessageLite& to_msg,
                      const ::google::protobuf::MessageLite& from_msg) {
   auto* const _this = static_cast<ACLKMessagePosition*>(&to_msg);
  auto& from = static_cast<const ACLKMessagePosition&>(from_msg);
  if constexpr (::_pbi::DebugHardenCheckHasBitConsistency()) {
    from.CheckHasBitConsistency();
  }
  ::google::protobuf::Arena* arena = _this->GetArena();
  // @@protoc_insertion_point(class_specific_merge_from_start:aclk_lib.v1.ACLKMessagePosition)
  ABSL_DCHECK_NE(&from, _this);
  ::uint32_t cached_has_bits = 0;
  (void)cached_has_bits;

  cached_has_bits = from._impl_._has_bits_[0];
  if (BatchCheckHasBit(cached_has_bits, 0x00000007U)) {
    if (CheckHasBit(cached_has_bits, 0x00000001U)) {
      ABSL_DCHECK(from._impl_.seq_id_created_at_ != nullptr);
      if (_this->_impl_.seq_id_created_at_ == nullptr) {
        _this->_impl_.seq_id_created_at_ = ::google::protobuf::Message::CopyConstruct(arena, *from._impl_.seq_id_created_at_);
      } else {
        _this->_impl_.seq_id_created_at_->MergeFrom(*from._impl_.seq_id_created_at_);
      }
    }
    if (CheckHasBit(cached_has_bits, 0x00000002U)) {
      if (from._internal_sequence_id() != 0) {
        _this->_impl_.sequence_id_ = from._impl_.sequence_id_;
      }
    }
    if (CheckHasBit(cached_has_bits, 0x00000004U)) {
      if (from._internal_previous_sequence_id() != 0) {
        _this->_impl_.previous_sequence_id_ = from._impl_.previous_sequence_id_;
      }
    }
  }
  _this->_impl_._has_bits_[0] |= cached_has_bits;
  _this->_internal_metadata_.MergeFrom<::google::protobuf::UnknownFieldSet>(
      from._internal_metadata_);
}

void ACLKMessagePosition::CopyFrom(const ACLKMessagePosition& from) {
  // @@protoc_insertion_point(class_specific_copy_from_start:aclk_lib.v1.ACLKMessagePosition)
  if (&from == this) return;
  Clear();
  MergeFrom(from);
}


void ACLKMessagePosition::InternalSwap(ACLKMessagePosition* PROTOBUF_RESTRICT PROTOBUF_NONNULL other) {
  using ::std::swap;
  _internal_metadata_.InternalSwap(&other->_internal_metadata_);
  swap(_impl_._has_bits_[0], other->_impl_._has_bits_[0]);
  ::google::protobuf::internal::memswap<
      PROTOBUF_FIELD_OFFSET(ACLKMessagePosition, _impl_.previous_sequence_id_)
      + sizeof(ACLKMessagePosition::_impl_.previous_sequence_id_)
      - PROTOBUF_FIELD_OFFSET(ACLKMessagePosition, _impl_.seq_id_created_at_)>(
          reinterpret_cast<char*>(&_impl_.seq_id_created_at_),
          reinterpret_cast<char*>(&other->_impl_.seq_id_created_at_));
}

::google::protobuf::Metadata ACLKMessagePosition::GetMetadata() const {
  return ::google::protobuf::Message::GetMetadataImpl(GetClassData()->full());
}
// ===================================================================

Capability::Capability(::google::protobuf::Arena* PROTOBUF_NULLABLE arena)
#if defined(PROTOBUF_CUSTOM_VTABLE)
    : ::google::protobuf::Message(arena, Capability_get_class_data()) {
#else   // PROTOBUF_CUSTOM_VTABLE
    : ::google::protobuf::Message(arena) {
#endif  // PROTOBUF_CUSTOM_VTABLE
  SharedCtor(arena);
  // @@protoc_insertion_point(arena_constructor:aclk_lib.v1.Capability)
}
PROTOBUF_NDEBUG_INLINE Capability::Impl_::Impl_(
    [[maybe_unused]] ::google::protobuf::internal::InternalVisibility visibility,
    [[maybe_unused]] ::google::protobuf::Arena* PROTOBUF_NULLABLE arena, const Impl_& from,
    [[maybe_unused]] const ::aclk_lib::v1::Capability& from_msg)
      : _has_bits_{from._has_bits_},
        _cached_size_{0},
        name_(arena, from.name_) {}

Capability::Capability(
    ::google::protobuf::Arena* PROTOBUF_NULLABLE arena,
    const Capability& from)
#if defined(PROTOBUF_CUSTOM_VTABLE)
    : ::google::protobuf::Message(arena, Capability_get_class_data()) {

#else   // PROTOBUF_CUSTOM_VTABLE
    : ::google::protobuf::Message(arena) {
#endif  // PROTOBUF_CUSTOM_VTABLE
  Capability* const _this = this;
  (void)_this;
  _internal_metadata_.MergeFrom<::google::protobuf::UnknownFieldSet>(
      from._internal_metadata_);
  new (&_impl_) Impl_(internal_visibility(), arena, from._impl_, from);
  ::memcpy(reinterpret_cast<char*>(&_impl_) +
               offsetof(Impl_, version_),
           reinterpret_cast<const char*>(&from._impl_) +
               offsetof(Impl_, version_),
           offsetof(Impl_, enabled_) -
               offsetof(Impl_, version_) +
               sizeof(Impl_::enabled_));

  // @@protoc_insertion_point(copy_constructor:aclk_lib.v1.Capability)
}
PROTOBUF_NDEBUG_INLINE Capability::Impl_::Impl_(
    [[maybe_unused]] ::google::protobuf::internal::InternalVisibility visibility,
    [[maybe_unused]] ::google::protobuf::Arena* PROTOBUF_NULLABLE arena)
      : _cached_size_{0},
        name_(arena) {}

inline void Capability::SharedCtor(::_pb::Arena* PROTOBUF_NULLABLE arena) {
  new (&_impl_) Impl_(internal_visibility(), arena);
  ::memset(reinterpret_cast<char*>(&_impl_) +
               offsetof(Impl_, version_),
           0,
           offsetof(Impl_, enabled_) -
               offsetof(Impl_, version_) +
               sizeof(Impl_::enabled_));
}
Capability::~Capability() {
  // @@protoc_insertion_point(destructor:aclk_lib.v1.Capability)
  SharedDtor(*this);
}
inline void Capability::SharedDtor(MessageLite& self) {
  Capability& this_ = static_cast<Capability&>(self);
  if constexpr (::_pbi::DebugHardenCheckHasBitConsistency()) {
    this_.CheckHasBitConsistency();
  }
  this_._internal_metadata_.Delete<::google::protobuf::UnknownFieldSet>();
  ABSL_DCHECK(this_.GetArena() == nullptr);
  this_._impl_.name_.Destroy();
  this_._impl_.~Impl_();
}

#ifndef PROTOBUF_MESSAGE_GLOBALS
PROTOBUF_CONSTINIT PROTOBUF_ATTRIBUTE_INIT_PRIORITY1 const
    ::google::protobuf::internal::ClassDataFull Capability_class_data_ =
        Capability::InternalGenerateClassData_(Capability_globals_._default);

PROTOBUF_ATTRIBUTE_WEAK const ::google::protobuf::internal::ClassData* PROTOBUF_NONNULL
Capability::GetClassData() const {
  ::google::protobuf::internal::PrefetchToLocalCache(&Capability_class_data_);
  ::google::protobuf::internal::PrefetchToLocalCache(Capability_class_data_.tc_table);
  return Capability_class_data_.base();
}
#else
PROTOBUF_ATTRIBUTE_WEAK const ::google::protobuf::internal::ClassData* PROTOBUF_NONNULL
Capability::GetClassData() const {
  ::google::protobuf::internal::PrefetchToLocalCache(&Capability_globals_);
  ::google::protobuf::internal::PrefetchToLocalCache(
      ::google::protobuf::internal::MessageGlobalsBase::ToParseTableBase(&Capability_globals_));
  return Capability_globals_.GetClassData();
}
#endif  // !PROTOBUF_MESSAGE_GLOBALS
#ifndef PROTOBUF_MESSAGE_GLOBALS
PROTOBUF_CONSTINIT
PROTOBUF_ATTRIBUTE_INIT_PRIORITY1 const Capability::ParseTableT_
    Capability::_table_ =
        Capability::InternalGenerateParseTable_(Capability_class_data_.base());
#endif  // !PROTOBUF_MESSAGE_GLOBALS
PROTOBUF_NOINLINE void Capability::Clear() {
// @@protoc_insertion_point(message_clear_start:aclk_lib.v1.Capability)
  ::google::protobuf::internal::TSanWrite(&_impl_);
  ::uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void) cached_has_bits;

  cached_has_bits = _impl_._has_bits_[0];
  if (CheckHasBit(cached_has_bits, 0x00000001U)) {
    _impl_.name_.ClearNonDefaultToEmpty();
  }
  if (BatchCheckHasBit(cached_has_bits, 0x00000006U)) {
    ::memset(&_impl_.version_, 0, static_cast<::size_t>(
        reinterpret_cast<char*>(&_impl_.enabled_) -
        reinterpret_cast<char*>(&_impl_.version_)) + sizeof(_impl_.enabled_));
  }
  _impl_._has_bits_.Clear();
  _internal_metadata_.Clear<::google::protobuf::UnknownFieldSet>();
}

#if defined(PROTOBUF_CUSTOM_VTABLE)
::uint8_t* PROTOBUF_NONNULL Capability::_InternalSerialize(
    const ::google::protobuf::MessageLite& base, ::uint8_t* PROTOBUF_NONNULL target,
    ::google::protobuf::io::EpsCopyOutputStream* PROTOBUF_NONNULL stream) {
  const Capability& this_ = static_cast<const Capability&>(base);
#else   // PROTOBUF_CUSTOM_VTABLE
::uint8_t* PROTOBUF_NONNULL Capability::_InternalSerialize(
    ::uint8_t* PROTOBUF_NONNULL target,
    ::google::protobuf::io::EpsCopyOutputStream* PROTOBUF_NONNULL stream) const {
  const Capability& this_ = *this;
#endif  // PROTOBUF_CUSTOM_VTABLE
  if constexpr (::_pbi::DebugHardenCheckHasBitConsistency()) {
    this_.CheckHasBitConsistency();
  }
  // @@protoc_insertion_point(serialize_to_array_start:aclk_lib.v1.Capability)
  ::uint32_t cached_has_bits = 0;
  (void)cached_has_bits;

  cached_has_bits = this_._impl_._has_bits_[0];
  // string name = 1;
  if (CheckHasBit(cached_has_bits, 0x00000001U)) {
    if (!this_._internal_name().empty()) {
      const ::std::string& _s = this_._internal_name();
      ::google::protobuf::internal::WireFormatLite::VerifyUtf8String(
          _s.data(), static_cast<int>(_s.length()), ::google::protobuf::internal::WireFormatLite::SERIALIZE, "aclk_lib.v1.Capability.name");
      target = stream->WriteStringMaybeAliased(1, _s, target);
    }
  }

  // uint32 version = 2;
  if (CheckHasBit(cached_has_bits, 0x00000002U)) {
    if (this_._internal_version() != 0) {
      target = stream->EnsureSpace(target);
      target = ::_pbi::WireFormatLite::WriteUInt32ToArray(
          2, this_._internal_version(), target);
    }
  }

  // bool enabled = 3;
  if (CheckHasBit(cached_has_bits, 0x00000004U)) {
    if (this_._internal_enabled() != 0) {
      target = stream->EnsureSpace(target);
      target = ::_pbi::WireFormatLite::WriteBoolToArray(
          3, this_._internal_enabled(), target);
    }
  }

  if (ABSL_PREDICT_FALSE(this_._internal_metadata_.have_unknown_fields())) {
    target =
        ::_pbi::WireFormat::InternalSerializeUnknownFieldsToArray(
            this_._internal_metadata_.unknown_fields<::google::protobuf::UnknownFieldSet>(::google::protobuf::UnknownFieldSet::default_instance), target, stream);
  }
  // @@protoc_insertion_point(serialize_to_array_end:aclk_lib.v1.Capability)
  return target;
}

#if defined(PROTOBUF_CUSTOM_VTABLE)
::size_t Capability::ByteSizeLong(const MessageLite& base) {
  const Capability& this_ = static_cast<const Capability&>(base);
#else   // PROTOBUF_CUSTOM_VTABLE
::size_t Capability::ByteSizeLong() const {
  const Capability& this_ = *this;
#endif  // PROTOBUF_CUSTOM_VTABLE
  // @@protoc_insertion_point(message_byte_size_start:aclk_lib.v1.Capability)
  ::size_t total_size = 0;

  ::uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void)cached_has_bits;

  ::_pbi::Prefetch5LinesFrom7Lines(&this_);
  cached_has_bits = this_._impl_._has_bits_[0];
  if (BatchCheckHasBit(cached_has_bits, 0x00000007U)) {
    // string name = 1;
    if (CheckHasBit(cached_has_bits, 0x00000001U)) {
      if (!this_._internal_name().empty()) {
        total_size += 1 + ::google::protobuf::internal::WireFormatLite::StringSize(
                                        this_._internal_name());
      }
    }
    // uint32 version = 2;
    if (CheckHasBit(cached_has_bits, 0x00000002U)) {
      if (this_._internal_version() != 0) {
        total_size += ::_pbi::WireFormatLite::UInt32SizePlusOne(
            this_._internal_version());
      }
    }
    // bool enabled = 3;
    if (CheckHasBit(cached_has_bits, 0x00000004U)) {
      if (this_._internal_enabled() != 0) {
        total_size += 2;
      }
    }
  }
  return this_.MaybeComputeUnknownFieldsSize(total_size,
                                             &this_._impl_._cached_size_);
}

void Capability::MergeImpl(::google::protobuf::MessageLite& to_msg,
                      const ::google::protobuf::MessageLite& from_msg) {
   auto* const _this = static_cast<Capability*>(&to_msg);
  auto& from = static_cast<const Capability&>(from_msg);
  if constexpr (::_pbi::DebugHardenCheckHasBitConsistency()) {
    from.CheckHasBitConsistency();
  }
  // @@protoc_insertion_point(class_specific_merge_from_start:aclk_lib.v1.Capability)
  ABSL_DCHECK_NE(&from, _this);
  ::uint32_t cached_has_bits = 0;
  (void)cached_has_bits;

  cached_has_bits = from._impl_._has_bits_[0];
  if (BatchCheckHasBit(cached_has_bits, 0x00000007U)) {
    if (CheckHasBit(cached_has_bits, 0x00000001U)) {
      if (!from._internal_name().empty()) {
        _this->_internal_set_name(from._internal_name());
      } else {
        if (_this->_impl_.name_.IsDefault()) {
          _this->_internal_set_name("");
        }
      }
    }
    if (CheckHasBit(cached_has_bits, 0x00000002U)) {
      if (from._internal_version() != 0) {
        _this->_impl_.version_ = from._impl_.version_;
      }
    }
    if (CheckHasBit(cached_has_bits, 0x00000004U)) {
      if (from._internal_enabled() != 0) {
        _this->_impl_.enabled_ = from._impl_.enabled_;
      }
    }
  }
  _this->_impl_._has_bits_[0] |= cached_has_bits;
  _this->_internal_metadata_.MergeFrom<::google::protobuf::UnknownFieldSet>(
      from._internal_metadata_);
}

void Capability::CopyFrom(const Capability& from) {
  // @@protoc_insertion_point(class_specific_copy_from_start:aclk_lib.v1.Capability)
  if (&from == this) return;
  Clear();
  MergeFrom(from);
}


void Capability::InternalSwap(Capability* PROTOBUF_RESTRICT PROTOBUF_NONNULL other) {
  using ::std::swap;
  auto* arena = GetArena();
  ABSL_DCHECK_EQ(arena, other->GetArena());
  _internal_metadata_.InternalSwap(&other->_internal_metadata_);
  swap(_impl_._has_bits_[0], other->_impl_._has_bits_[0]);
  ::_pbi::ArenaStringPtr::InternalSwap(&_impl_.name_, &other->_impl_.name_, arena);
  ::google::protobuf::internal::memswap<
      PROTOBUF_FIELD_OFFSET(Capability, _impl_.enabled_)
      + sizeof(Capability::_impl_.enabled_)
      - PROTOBUF_FIELD_OFFSET(Capability, _impl_.version_)>(
          reinterpret_cast<char*>(&_impl_.version_),
          reinterpret_cast<char*>(&other->_impl_.version_));
}

::google::protobuf::Metadata Capability::GetMetadata() const {
  return ::google::protobuf::Message::GetMetadataImpl(GetClassData()->full());
}
// @@protoc_insertion_point(namespace_scope)
}  // namespace v1
}  // namespace aclk_lib
namespace google {
namespace protobuf {
}  // namespace protobuf
}  // namespace google
// @@protoc_insertion_point(global_scope)
PROTOBUF_ATTRIBUTE_INIT_PRIORITY2 static ::std::false_type
    _static_init2_ [[maybe_unused]] =
        (::_pbi::AddDescriptors(&descriptor_table_aclk_2fv1_2flib_2eproto),
         ::std::false_type{});
#include "google/protobuf/port_undef.inc"
