// SPDX-License-Identifier: GPL-3.0-or-later

#ifndef NETDATA_SPAWN_LIBRARY_H
#define NETDATA_SPAWN_LIBRARY_H

#include "../exit/exit_initiated.h"

BUFFER *argv_to_cmdline_buffer(const char **argv);

#endif //NETDATA_SPAWN_LIBRARY_H
