// Generated by the protocol buffer compiler.  DO NOT EDIT!
// NO CHECKED-IN PROTOBUF GENCODE
// source: agent/v1/cmds.proto
// Protobuf C++ Version: 7.35.1

#include "agent/v1/cmds.pb.h"

#include <algorithm>
#include <type_traits>
#include "google/protobuf/io/coded_stream.h"
#include "google/protobuf/generated_message_tctable_impl.h"
#include "google/protobuf/internal_visibility.h"
#include "google/protobuf/extension_set.h"
#include "google/protobuf/generated_message_util.h"
#include "google/protobuf/wire_format_lite.h"
#include "google/protobuf/descriptor.h"
#include "google/protobuf/generated_message_reflection.h"
#include "google/protobuf/reflection_ops.h"
#include "google/protobuf/wire_format.h"
// @@protoc_insertion_point(includes)

// Must be included last.
#include "google/protobuf/port_def.inc"
PROTOBUF_PRAGMA_INIT_SEG
namespace _pb = ::google::protobuf;
namespace _pbi = ::google::protobuf::internal;
namespace _fl = ::google::protobuf::internal::field_layout;
#ifdef PROTOBUF_MESSAGE_GLOBALS
namespace {
PROTOBUF_CONSTINIT ::google::protobuf::internal::ReflectionData
    file_reflection_data[] = {
        // ::agent::v1::CancelPendingRequest
        {&::_pbi::kDescriptorMethods, &::descriptor_table_agent_2fv1_2fcmds_2eproto, /* tracker*/ nullptr,},
        // ::agent::v1::AgentCommand
        {&::_pbi::kDescriptorMethods, &::descriptor_table_agent_2fv1_2fcmds_2eproto, /* tracker*/ nullptr,},
        // ::agent::v1::AgentCommandAck
        {&::_pbi::kDescriptorMethods, &::descriptor_table_agent_2fv1_2fcmds_2eproto, /* tracker*/ nullptr,},
        // ::agent::v1::AgentCommandResponse
        {&::_pbi::kDescriptorMethods, &::descriptor_table_agent_2fv1_2fcmds_2eproto, /* tracker*/ nullptr,},
};
}  // namespace
#endif
namespace agent {
namespace v1 {
class AgentCommand::_Internal {
 public:
  using HasBits = decltype(::std::declval<AgentCommand>()._impl_._has_bits_);
  static constexpr ::int32_t kHasBitsOffset =
      8 * PROTOBUF_FIELD_OFFSET(AgentCommand, _impl_._has_bits_);
};

constexpr AgentCommand::ParseTableT_ AgentCommand::InternalGenerateParseTable_(const ::_pbi::ClassData* class_data) {
  return ParseTableT_{
    {
      PROTOBUF_FIELD_OFFSET(AgentCommand, _impl_._has_bits_),
      0, // no _extensions_
      6, 56,  // max_field_number, fast_idx_mask
      offsetof(ParseTableT_, field_lookup_table),
      4294967232,  // skipmap
      offsetof(ParseTableT_, field_entries),
      6,  // num_field_entries
      0,  // num_aux_entries
      offsetof(ParseTableT_, field_names),  // no aux_entries
      class_data,
      nullptr,  // post_loop_handler
      ::_pbi::TcParser::GenericFallback,  // fallback
      #ifdef PROTOBUF_PREFETCH_PARSE_TABLE
      ::_pbi::TcParser::GetTable<::agent::v1::AgentCommand>(),  // to_prefetch
      #endif  // PROTOBUF_PREFETCH_PARSE_TABLE
    }, {{
      {::_pbi::TcParser::MiniParse, {}},
      // string callback_topic = 1;
      {::_pbi::TcParser::FastUS1,
       {10, 0, 0,
        PROTOBUF_FIELD_OFFSET(AgentCommand, _impl_.callback_topic_)}},
      // string ack_topic = 2;
      {::_pbi::TcParser::FastUS1,
       {18, 1, 0,
        PROTOBUF_FIELD_OFFSET(AgentCommand, _impl_.ack_topic_)}},
      // string message_id = 3;
      {::_pbi::TcParser::FastUS1,
       {26, 2, 0,
        PROTOBUF_FIELD_OFFSET(AgentCommand, _impl_.message_id_)}},
      // uint64 timeout = 4;
      {::_pbi::TcParser::SingularVarintNoZag1<::uint64_t, offsetof(AgentCommand, _impl_.timeout_), 4>(),
       {32, 4, 0,
        PROTOBUF_FIELD_OFFSET(AgentCommand, _impl_.timeout_)}},
      // uint64 ack_timeout = 5;
      {::_pbi::TcParser::SingularVarintNoZag1<::uint64_t, offsetof(AgentCommand, _impl_.ack_timeout_), 5>(),
       {40, 5, 0,
        PROTOBUF_FIELD_OFFSET(AgentCommand, _impl_.ack_timeout_)}},
      // string resource = 6;
      {::_pbi::TcParser::FastUS1,
       {50, 3, 0,
        PROTOBUF_FIELD_OFFSET(AgentCommand, _impl_.resource_)}},
      {::_pbi::TcParser::MiniParse, {}},
    }}, {{
      65535, 65535
    }}, {{
      // string callback_topic = 1;
      {PROTOBUF_FIELD_OFFSET(AgentCommand, _impl_.callback_topic_), _Internal::kHasBitsOffset + 0, 0, (0 | ::_fl::kFcOptional | ::_fl::kUtf8String | ::_fl::kRepAString)},
      // string ack_topic = 2;
      {PROTOBUF_FIELD_OFFSET(AgentCommand, _impl_.ack_topic_), _Internal::kHasBitsOffset + 1, 0, (0 | ::_fl::kFcOptional | ::_fl::kUtf8String | ::_fl::kRepAString)},
      // string message_id = 3;
      {PROTOBUF_FIELD_OFFSET(AgentCommand, _impl_.message_id_), _Internal::kHasBitsOffset + 2, 0, (0 | ::_fl::kFcOptional | ::_fl::kUtf8String | ::_fl::kRepAString)},
      // uint64 timeout = 4;
      {PROTOBUF_FIELD_OFFSET(AgentCommand, _impl_.timeout_), _Internal::kHasBitsOffset + 4, 0, (0 | ::_fl::kFcOptional | ::_fl::kUInt64)},
      // uint64 ack_timeout = 5;
      {PROTOBUF_FIELD_OFFSET(AgentCommand, _impl_.ack_timeout_), _Internal::kHasBitsOffset + 5, 0, (0 | ::_fl::kFcOptional | ::_fl::kUInt64)},
      // string resource = 6;
      {PROTOBUF_FIELD_OFFSET(AgentCommand, _impl_.resource_), _Internal::kHasBitsOffset + 3, 0, (0 | ::_fl::kFcOptional | ::_fl::kUtf8String | ::_fl::kRepAString)},
    }},
    // no aux_entries
    {{
      "\25\16\11\12\0\0\10\0"
      "agent.v1.AgentCommand"
      "callback_topic"
      "ack_topic"
      "message_id"
      "resource"
    }},
  };
}


inline constexpr AgentCommand::Impl_::Impl_(
    [[maybe_unused]] ::google::protobuf::internal::InternalVisibility visibility,
    ::_pbi::ConstantInitialized) noexcept
      : _cached_size_{0},
        callback_topic_(
            &::google::protobuf::internal::fixed_address_empty_string,
            ::_pbi::ConstantInitialized()),
        ack_topic_(
            &::google::protobuf::internal::fixed_address_empty_string,
            ::_pbi::ConstantInitialized()),
        message_id_(
            &::google::protobuf::internal::fixed_address_empty_string,
            ::_pbi::ConstantInitialized()),
        resource_(
            &::google::protobuf::internal::fixed_address_empty_string,
            ::_pbi::ConstantInitialized()),
        timeout_{::uint64_t{0u}},
        ack_timeout_{::uint64_t{0u}} {}

template <typename>
constexpr AgentCommand::AgentCommand(::_pbi::ConstantInitialized,
                       const ::_pbi::ClassData* PROTOBUF_NONNULL class_data)
    : ::google::protobuf::Message(
#if defined(PROTOBUF_CUSTOM_VTABLE)
          class_data
#endif  // PROTOBUF_CUSTOM_VTABLE
          ),
      _impl_(internal_visibility(), ::_pbi::ConstantInitialized()) {
}
inline void* PROTOBUF_NONNULL AgentCommand::PlacementNew_(
    const void* PROTOBUF_NONNULL, void* PROTOBUF_NONNULL mem,
    ::google::protobuf::Arena* PROTOBUF_NULLABLE arena) {
  return ::new (mem) AgentCommand(arena);
}
constexpr auto AgentCommand::InternalNewImpl_() {
  return ::google::protobuf::internal::MessageCreator::CopyInit(sizeof(AgentCommand), alignof(AgentCommand));
}
constexpr auto AgentCommand::InternalGenerateClassData_(
    const MessageLite& prototype,
    const ::google::protobuf::internal::TcParseTableBase* tc_table) {
  return ::google::protobuf::internal::ClassDataFull{
      ::google::protobuf::internal::ClassData{
          &prototype,
#ifndef PROTOBUF_MESSAGE_GLOBALS
          &_table_.header,
#else
          tc_table,
#endif
          nullptr,  // IsInitialized
          &AgentCommand::MergeImpl,
          ::google::protobuf::Message::GetNewImpl<AgentCommand>(),
#if defined(PROTOBUF_CUSTOM_VTABLE)
          &AgentCommand::SharedDtor,
          ::google::protobuf::Message::GetClearImpl<AgentCommand>(), &AgentCommand::ByteSizeLong,
              &AgentCommand::_InternalSerialize,
#endif  // PROTOBUF_CUSTOM_VTABLE
          PROTOBUF_FIELD_OFFSET(AgentCommand, _impl_._cached_size_),
          false,
      },
#ifdef PROTOBUF_MESSAGE_GLOBALS
      &file_reflection_data[1],
#else   // !PROTOBUF_MESSAGE_GLOBALS
      &::_pbi::kDescriptorMethods,
      &descriptor_table_agent_2fv1_2fcmds_2eproto,
      nullptr,  // tracker
#endif  // PROTOBUF_MESSAGE_GLOBALS
  };
}
struct AgentCommandGlobalsTypeInternal : ::_pbi::MessageGlobalsBase {
  constexpr AgentCommandGlobalsTypeInternal()
      :
#ifndef PROTOBUF_MESSAGE_GLOBALS
        _default(::_pbi::ConstantInitialized{},
                 AgentCommand_class_data_.base())
#else   // !PROTOBUF_MESSAGE_GLOBALS
        MessageGlobalsBase(AgentCommand::InternalGenerateClassData_(
            _default, &AgentCommand_globals_._table.header)),
        _default(::_pbi::ConstantInitialized{}, GetClassData()),
        _table(::_pbi::PrivateAccess::GenerateParseTable<AgentCommand>(
            GetClassData()))
#endif  // PROTOBUF_MESSAGE_GLOBALS
  {
  }
  ~AgentCommandGlobalsTypeInternal() {}
  union {
    alignas(::_pbi::kMaxMessageAlignment) AgentCommand _default;
  };
#ifdef PROTOBUF_MESSAGE_GLOBALS
  decltype(::_pbi::PrivateAccess::GenerateParseTable<AgentCommand>(
      ::std::declval<const ::_pbi::ClassData*>())) _table;
#endif
};
#ifdef PROTOBUF_MESSAGE_GLOBALS
static_assert(PROTOBUF_FIELD_OFFSET(AgentCommandGlobalsTypeInternal, _default) ==
              ::_pbi::MessageGlobalsBase::OffsetToDefault());
#endif  // PROTOBUF_MESSAGE_GLOBALS

PROTOBUF_ATTRIBUTE_NO_DESTROY PROTOBUF_CONSTINIT
    PROTOBUF_ATTRIBUTE_INIT_PRIORITY1 PROTOBUF_MESSAGE_GLOBALS_CONST AgentCommandGlobalsTypeInternal AgentCommand_globals_
        PROTOBUF_MESSAGE_GLOBALS_SECTION(.data.rel.ro);
#if defined(PROTOBUF_CUSTOM_VTABLE)
namespace {
const ::_pbi::ClassData* AgentCommand_get_class_data() {
#ifdef PROTOBUF_MESSAGE_GLOBALS
  return AgentCommand_globals_.GetClassData();
#else
  return AgentCommand_class_data_.base();
#endif  // PROTOBUF_MESSAGE_GLOBALS
}
}  // namespace
#endif  // PROTOBUF_CUSTOM_VTABLE
class CancelPendingRequest::_Internal {
 public:
  using HasBits = decltype(::std::declval<CancelPendingRequest>()._impl_._has_bits_);
  static constexpr ::int32_t kHasBitsOffset =
      8 * PROTOBUF_FIELD_OFFSET(CancelPendingRequest, _impl_._has_bits_);
};

constexpr CancelPendingRequest::ParseTableT_ CancelPendingRequest::InternalGenerateParseTable_(const ::_pbi::ClassData* class_data) {
  return ParseTableT_{
    {
      PROTOBUF_FIELD_OFFSET(CancelPendingRequest, _impl_._has_bits_),
      0, // no _extensions_
      3, 24,  // max_field_number, fast_idx_mask
      offsetof(ParseTableT_, field_lookup_table),
      4294967288,  // skipmap
      offsetof(ParseTableT_, field_entries),
      3,  // num_field_entries
      1,  // num_aux_entries
      offsetof(ParseTableT_, aux_entries),
      class_data,
      nullptr,  // post_loop_handler
      ::_pbi::TcParser::GenericFallback,  // fallback
      #ifdef PROTOBUF_PREFETCH_PARSE_TABLE
      ::_pbi::TcParser::GetTable<::agent::v1::CancelPendingRequest>(),  // to_prefetch
      #endif  // PROTOBUF_PREFETCH_PARSE_TABLE
    }, {{
      {::_pbi::TcParser::MiniParse, {}},
      // string request_id = 1;
      {::_pbi::TcParser::FastUS1,
       {10, 0, 0,
        PROTOBUF_FIELD_OFFSET(CancelPendingRequest, _impl_.request_id_)}},
      // .google.protobuf.Timestamp timestamp = 2;
      {::_pbi::TcParser::FastMtS1,
       {18, 2, 0,
        PROTOBUF_FIELD_OFFSET(CancelPendingRequest, _impl_.timestamp_)}},
      // string trace_id = 3;
      {::_pbi::TcParser::FastUS1,
       {26, 1, 0,
        PROTOBUF_FIELD_OFFSET(CancelPendingRequest, _impl_.trace_id_)}},
    }}, {{
      65535, 65535
    }}, {{
      // string request_id = 1;
      {PROTOBUF_FIELD_OFFSET(CancelPendingRequest, _impl_.request_id_), _Internal::kHasBitsOffset + 0, 0, (0 | ::_fl::kFcOptional | ::_fl::kUtf8String | ::_fl::kRepAString)},
      // .google.protobuf.Timestamp timestamp = 2;
      {PROTOBUF_FIELD_OFFSET(CancelPendingRequest, _impl_.timestamp_), _Internal::kHasBitsOffset + 2, 0, (0 | ::_fl::kFcOptional | ::_fl::kMessage | ::_fl::kTvTable)},
      // string trace_id = 3;
      {PROTOBUF_FIELD_OFFSET(CancelPendingRequest, _impl_.trace_id_), _Internal::kHasBitsOffset + 1, 0, (0 | ::_fl::kFcOptional | ::_fl::kUtf8String | ::_fl::kRepAString)},
    }},
    {{
        #ifndef PROTOBUF_MESSAGE_GLOBALS
        {::_pbi::TcParser::GetTable<::google::protobuf::Timestamp>()},
        #else
        {::_pbi::FieldAuxMessageGlobals(), &::google::protobuf::Timestamp_globals_},
        #endif
    }},
    {{
      "\35\12\0\10\0\0\0\0"
      "agent.v1.CancelPendingRequest"
      "request_id"
      "trace_id"
    }},
  };
}


inline constexpr CancelPendingRequest::Impl_::Impl_(
    [[maybe_unused]] ::google::protobuf::internal::InternalVisibility visibility,
    ::_pbi::ConstantInitialized) noexcept
      : _cached_size_{0},
        request_id_(
            &::google::protobuf::internal::fixed_address_empty_string,
            ::_pbi::ConstantInitialized()),
        trace_id_(
            &::google::protobuf::internal::fixed_address_empty_string,
            ::_pbi::ConstantInitialized()),
        timestamp_{nullptr} {}

template <typename>
constexpr CancelPendingRequest::CancelPendingRequest(::_pbi::ConstantInitialized,
                       const ::_pbi::ClassData* PROTOBUF_NONNULL class_data)
    : ::google::protobuf::Message(
#if defined(PROTOBUF_CUSTOM_VTABLE)
          class_data
#endif  // PROTOBUF_CUSTOM_VTABLE
          ),
      _impl_(internal_visibility(), ::_pbi::ConstantInitialized()) {
}
inline void* PROTOBUF_NONNULL CancelPendingRequest::PlacementNew_(
    const void* PROTOBUF_NONNULL, void* PROTOBUF_NONNULL mem,
    ::google::protobuf::Arena* PROTOBUF_NULLABLE arena) {
  return ::new (mem) CancelPendingRequest(arena);
}
constexpr auto CancelPendingRequest::InternalNewImpl_() {
  return ::google::protobuf::internal::MessageCreator::CopyInit(sizeof(CancelPendingRequest), alignof(CancelPendingRequest));
}
constexpr auto CancelPendingRequest::InternalGenerateClassData_(
    const MessageLite& prototype,
    const ::google::protobuf::internal::TcParseTableBase* tc_table) {
  return ::google::protobuf::internal::ClassDataFull{
      ::google::protobuf::internal::ClassData{
          &prototype,
#ifndef PROTOBUF_MESSAGE_GLOBALS
          &_table_.header,
#else
          tc_table,
#endif
          nullptr,  // IsInitialized
          &CancelPendingRequest::MergeImpl,
          ::google::protobuf::Message::GetNewImpl<CancelPendingRequest>(),
#if defined(PROTOBUF_CUSTOM_VTABLE)
          &CancelPendingRequest::SharedDtor,
          ::google::protobuf::Message::GetClearImpl<CancelPendingRequest>(), &CancelPendingRequest::ByteSizeLong,
              &CancelPendingRequest::_InternalSerialize,
#endif  // PROTOBUF_CUSTOM_VTABLE
          PROTOBUF_FIELD_OFFSET(CancelPendingRequest, _impl_._cached_size_),
          false,
      },
#ifdef PROTOBUF_MESSAGE_GLOBALS
      &file_reflection_data[0],
#else   // !PROTOBUF_MESSAGE_GLOBALS
      &::_pbi::kDescriptorMethods,
      &descriptor_table_agent_2fv1_2fcmds_2eproto,
      nullptr,  // tracker
#endif  // PROTOBUF_MESSAGE_GLOBALS
  };
}
struct CancelPendingRequestGlobalsTypeInternal : ::_pbi::MessageGlobalsBase {
  constexpr CancelPendingRequestGlobalsTypeInternal()
      :
#ifndef PROTOBUF_MESSAGE_GLOBALS
        _default(::_pbi::ConstantInitialized{},
                 CancelPendingRequest_class_data_.base())
#else   // !PROTOBUF_MESSAGE_GLOBALS
        MessageGlobalsBase(CancelPendingRequest::InternalGenerateClassData_(
            _default, &CancelPendingRequest_globals_._table.header)),
        _default(::_pbi::ConstantInitialized{}, GetClassData()),
        _table(::_pbi::PrivateAccess::GenerateParseTable<CancelPendingRequest>(
            GetClassData()))
#endif  // PROTOBUF_MESSAGE_GLOBALS
  {
  }
  ~CancelPendingRequestGlobalsTypeInternal() {}
  union {
    alignas(::_pbi::kMaxMessageAlignment) CancelPendingRequest _default;
  };
#ifdef PROTOBUF_MESSAGE_GLOBALS
  decltype(::_pbi::PrivateAccess::GenerateParseTable<CancelPendingRequest>(
      ::std::declval<const ::_pbi::ClassData*>())) _table;
#endif
};
#ifdef PROTOBUF_MESSAGE_GLOBALS
static_assert(PROTOBUF_FIELD_OFFSET(CancelPendingRequestGlobalsTypeInternal, _default) ==
              ::_pbi::MessageGlobalsBase::OffsetToDefault());
#endif  // PROTOBUF_MESSAGE_GLOBALS

PROTOBUF_ATTRIBUTE_NO_DESTROY PROTOBUF_CONSTINIT
    PROTOBUF_ATTRIBUTE_INIT_PRIORITY1 PROTOBUF_MESSAGE_GLOBALS_CONST CancelPendingRequestGlobalsTypeInternal CancelPendingRequest_globals_
        PROTOBUF_MESSAGE_GLOBALS_SECTION(.data.rel.ro);
#if defined(PROTOBUF_CUSTOM_VTABLE)
namespace {
const ::_pbi::ClassData* CancelPendingRequest_get_class_data() {
#ifdef PROTOBUF_MESSAGE_GLOBALS
  return CancelPendingRequest_globals_.GetClassData();
#else
  return CancelPendingRequest_class_data_.base();
#endif  // PROTOBUF_MESSAGE_GLOBALS
}
}  // namespace
#endif  // PROTOBUF_CUSTOM_VTABLE
class AgentCommandResponse::_Internal {
 public:
  using HasBits = decltype(::std::declval<AgentCommandResponse>()._impl_._has_bits_);
  static constexpr ::int32_t kHasBitsOffset =
      8 * PROTOBUF_FIELD_OFFSET(AgentCommandResponse, _impl_._has_bits_);
};

constexpr AgentCommandResponse::ParseTableT_ AgentCommandResponse::InternalGenerateParseTable_(const ::_pbi::ClassData* class_data) {
  return ParseTableT_{
    {
      PROTOBUF_FIELD_OFFSET(AgentCommandResponse, _impl_._has_bits_),
      0, // no _extensions_
      6, 56,  // max_field_number, fast_idx_mask
      offsetof(ParseTableT_, field_lookup_table),
      4294967232,  // skipmap
      offsetof(ParseTableT_, field_entries),
      6,  // num_field_entries
      2,  // num_aux_entries
      offsetof(ParseTableT_, aux_entries),
      class_data,
      nullptr,  // post_loop_handler
      ::_pbi::TcParser::GenericFallback,  // fallback
      #ifdef PROTOBUF_PREFETCH_PARSE_TABLE
      ::_pbi::TcParser::GetTable<::agent::v1::AgentCommandResponse>(),  // to_prefetch
      #endif  // PROTOBUF_PREFETCH_PARSE_TABLE
    }, {{
      {::_pbi::TcParser::MiniParse, {}},
      // string message_id = 1;
      {::_pbi::TcParser::FastUS1,
       {10, 0, 0,
        PROTOBUF_FIELD_OFFSET(AgentCommandResponse, _impl_.message_id_)}},
      // uint32 status_code = 2;
      {::_pbi::TcParser::SingularVarintNoZag1<::uint32_t, offsetof(AgentCommandResponse, _impl_.status_code_), 5>(),
       {16, 5, 0,
        PROTOBUF_FIELD_OFFSET(AgentCommandResponse, _impl_.status_code_)}},
      // bytes response = 3;
      {::_pbi::TcParser::FastBS1,
       {26, 1, 0,
        PROTOBUF_FIELD_OFFSET(AgentCommandResponse, _impl_.response_)}},
      // .google.protobuf.Timestamp timestamp = 4;
      {::_pbi::TcParser::FastMtS1,
       {34, 2, 0,
        PROTOBUF_FIELD_OFFSET(AgentCommandResponse, _impl_.timestamp_)}},
      // .google.protobuf.Timestamp received_at = 5;
      {::_pbi::TcParser::FastMtS1,
       {42, 3, 1,
        PROTOBUF_FIELD_OFFSET(AgentCommandResponse, _impl_.received_at_)}},
      // uint64 exec_time = 6;
      {::_pbi::TcParser::SingularVarintNoZag1<::uint64_t, offsetof(AgentCommandResponse, _impl_.exec_time_), 4>(),
       {48, 4, 0,
        PROTOBUF_FIELD_OFFSET(AgentCommandResponse, _impl_.exec_time_)}},
      {::_pbi::TcParser::MiniParse, {}},
    }}, {{
      65535, 65535
    }}, {{
      // string message_id = 1;
      {PROTOBUF_FIELD_OFFSET(AgentCommandResponse, _impl_.message_id_), _Internal::kHasBitsOffset + 0, 0, (0 | ::_fl::kFcOptional | ::_fl::kUtf8String | ::_fl::kRepAString)},
      // uint32 status_code = 2;
      {PROTOBUF_FIELD_OFFSET(AgentCommandResponse, _impl_.status_code_), _Internal::kHasBitsOffset + 5, 0, (0 | ::_fl::kFcOptional | ::_fl::kUInt32)},
      // bytes response = 3;
      {PROTOBUF_FIELD_OFFSET(AgentCommandResponse, _impl_.response_), _Internal::kHasBitsOffset + 1, 0, (0 | ::_fl::kFcOptional | ::_fl::kBytes | ::_fl::kRepAString)},
      // .google.protobuf.Timestamp timestamp = 4;
      {PROTOBUF_FIELD_OFFSET(AgentCommandResponse, _impl_.timestamp_), _Internal::kHasBitsOffset + 2, 0, (0 | ::_fl::kFcOptional | ::_fl::kMessage | ::_fl::kTvTable)},
      // .google.protobuf.Timestamp received_at = 5;
      {PROTOBUF_FIELD_OFFSET(AgentCommandResponse, _impl_.received_at_), _Internal::kHasBitsOffset + 3, 1, (0 | ::_fl::kFcOptional | ::_fl::kMessage | ::_fl::kTvTable)},
      // uint64 exec_time = 6;
      {PROTOBUF_FIELD_OFFSET(AgentCommandResponse, _impl_.exec_time_), _Internal::kHasBitsOffset + 4, 0, (0 | ::_fl::kFcOptional | ::_fl::kUInt64)},
    }},
    {{
        #ifndef PROTOBUF_MESSAGE_GLOBALS
        {::_pbi::TcParser::GetTable<::google::protobuf::Timestamp>()},
        #else
        {::_pbi::FieldAuxMessageGlobals(), &::google::protobuf::Timestamp_globals_},
        #endif
        #ifndef PROTOBUF_MESSAGE_GLOBALS
        {::_pbi::TcParser::GetTable<::google::protobuf::Timestamp>()},
        #else
        {::_pbi::FieldAuxMessageGlobals(), &::google::protobuf::Timestamp_globals_},
        #endif
    }},
    {{
      "\35\12\0\0\0\0\0\0"
      "agent.v1.AgentCommandResponse"
      "message_id"
    }},
  };
}


inline constexpr AgentCommandResponse::Impl_::Impl_(
    [[maybe_unused]] ::google::protobuf::internal::InternalVisibility visibility,
    ::_pbi::ConstantInitialized) noexcept
      : _cached_size_{0},
        message_id_(
            &::google::protobuf::internal::fixed_address_empty_string,
            ::_pbi::ConstantInitialized()),
        response_(
            &::google::protobuf::internal::fixed_address_empty_string,
            ::_pbi::ConstantInitialized()),
        timestamp_{nullptr},
        received_at_{nullptr},
        exec_time_{::uint64_t{0u}},
        status_code_{0u} {}

template <typename>
constexpr AgentCommandResponse::AgentCommandResponse(::_pbi::ConstantInitialized,
                       const ::_pbi::ClassData* PROTOBUF_NONNULL class_data)
    : ::google::protobuf::Message(
#if defined(PROTOBUF_CUSTOM_VTABLE)
          class_data
#endif  // PROTOBUF_CUSTOM_VTABLE
          ),
      _impl_(internal_visibility(), ::_pbi::ConstantInitialized()) {
}
inline void* PROTOBUF_NONNULL AgentCommandResponse::PlacementNew_(
    const void* PROTOBUF_NONNULL, void* PROTOBUF_NONNULL mem,
    ::google::protobuf::Arena* PROTOBUF_NULLABLE arena) {
  return ::new (mem) AgentCommandResponse(arena);
}
constexpr auto AgentCommandResponse::InternalNewImpl_() {
  return ::google::protobuf::internal::MessageCreator::CopyInit(sizeof(AgentCommandResponse), alignof(AgentCommandResponse));
}
constexpr auto AgentCommandResponse::InternalGenerateClassData_(
    const MessageLite& prototype,
    const ::google::protobuf::internal::TcParseTableBase* tc_table) {
  return ::google::protobuf::internal::ClassDataFull{
      ::google::protobuf::internal::ClassData{
          &prototype,
#ifndef PROTOBUF_MESSAGE_GLOBALS
          &_table_.header,
#else
          tc_table,
#endif
          nullptr,  // IsInitialized
          &AgentCommandResponse::MergeImpl,
          ::google::protobuf::Message::GetNewImpl<AgentCommandResponse>(),
#if defined(PROTOBUF_CUSTOM_VTABLE)
          &AgentCommandResponse::SharedDtor,
          ::google::protobuf::Message::GetClearImpl<AgentCommandResponse>(), &AgentCommandResponse::ByteSizeLong,
              &AgentCommandResponse::_InternalSerialize,
#endif  // PROTOBUF_CUSTOM_VTABLE
          PROTOBUF_FIELD_OFFSET(AgentCommandResponse, _impl_._cached_size_),
          false,
      },
#ifdef PROTOBUF_MESSAGE_GLOBALS
      &file_reflection_data[3],
#else   // !PROTOBUF_MESSAGE_GLOBALS
      &::_pbi::kDescriptorMethods,
      &descriptor_table_agent_2fv1_2fcmds_2eproto,
      nullptr,  // tracker
#endif  // PROTOBUF_MESSAGE_GLOBALS
  };
}
struct AgentCommandResponseGlobalsTypeInternal : ::_pbi::MessageGlobalsBase {
  constexpr AgentCommandResponseGlobalsTypeInternal()
      :
#ifndef PROTOBUF_MESSAGE_GLOBALS
        _default(::_pbi::ConstantInitialized{},
                 AgentCommandResponse_class_data_.base())
#else   // !PROTOBUF_MESSAGE_GLOBALS
        MessageGlobalsBase(AgentCommandResponse::InternalGenerateClassData_(
            _default, &AgentCommandResponse_globals_._table.header)),
        _default(::_pbi::ConstantInitialized{}, GetClassData()),
        _table(::_pbi::PrivateAccess::GenerateParseTable<AgentCommandResponse>(
            GetClassData()))
#endif  // PROTOBUF_MESSAGE_GLOBALS
  {
  }
  ~AgentCommandResponseGlobalsTypeInternal() {}
  union {
    alignas(::_pbi::kMaxMessageAlignment) AgentCommandResponse _default;
  };
#ifdef PROTOBUF_MESSAGE_GLOBALS
  decltype(::_pbi::PrivateAccess::GenerateParseTable<AgentCommandResponse>(
      ::std::declval<const ::_pbi::ClassData*>())) _table;
#endif
};
#ifdef PROTOBUF_MESSAGE_GLOBALS
static_assert(PROTOBUF_FIELD_OFFSET(AgentCommandResponseGlobalsTypeInternal, _default) ==
              ::_pbi::MessageGlobalsBase::OffsetToDefault());
#endif  // PROTOBUF_MESSAGE_GLOBALS

PROTOBUF_ATTRIBUTE_NO_DESTROY PROTOBUF_CONSTINIT
    PROTOBUF_ATTRIBUTE_INIT_PRIORITY1 PROTOBUF_MESSAGE_GLOBALS_CONST AgentCommandResponseGlobalsTypeInternal AgentCommandResponse_globals_
        PROTOBUF_MESSAGE_GLOBALS_SECTION(.data.rel.ro);
#if defined(PROTOBUF_CUSTOM_VTABLE)
namespace {
const ::_pbi::ClassData* AgentCommandResponse_get_class_data() {
#ifdef PROTOBUF_MESSAGE_GLOBALS
  return AgentCommandResponse_globals_.GetClassData();
#else
  return AgentCommandResponse_class_data_.base();
#endif  // PROTOBUF_MESSAGE_GLOBALS
}
}  // namespace
#endif  // PROTOBUF_CUSTOM_VTABLE
class AgentCommandAck::_Internal {
 public:
  using HasBits = decltype(::std::declval<AgentCommandAck>()._impl_._has_bits_);
  static constexpr ::int32_t kHasBitsOffset =
      8 * PROTOBUF_FIELD_OFFSET(AgentCommandAck, _impl_._has_bits_);
};

constexpr AgentCommandAck::ParseTableT_ AgentCommandAck::InternalGenerateParseTable_(const ::_pbi::ClassData* class_data) {
  return ParseTableT_{
    {
      PROTOBUF_FIELD_OFFSET(AgentCommandAck, _impl_._has_bits_),
      0, // no _extensions_
      3, 24,  // max_field_number, fast_idx_mask
      offsetof(ParseTableT_, field_lookup_table),
      4294967288,  // skipmap
      offsetof(ParseTableT_, field_entries),
      3,  // num_field_entries
      1,  // num_aux_entries
      offsetof(ParseTableT_, aux_entries),
      class_data,
      nullptr,  // post_loop_handler
      ::_pbi::TcParser::GenericFallback,  // fallback
      #ifdef PROTOBUF_PREFETCH_PARSE_TABLE
      ::_pbi::TcParser::GetTable<::agent::v1::AgentCommandAck>(),  // to_prefetch
      #endif  // PROTOBUF_PREFETCH_PARSE_TABLE
    }, {{
      {::_pbi::TcParser::MiniParse, {}},
      // string message_id = 1;
      {::_pbi::TcParser::FastUS1,
       {10, 0, 0,
        PROTOBUF_FIELD_OFFSET(AgentCommandAck, _impl_.message_id_)}},
      // .google.protobuf.Timestamp created_at = 2;
      {::_pbi::TcParser::FastMtS1,
       {18, 1, 0,
        PROTOBUF_FIELD_OFFSET(AgentCommandAck, _impl_.created_at_)}},
      // uint32 progress_percent = 3;
      {::_pbi::TcParser::SingularVarintNoZag1<::uint32_t, offsetof(AgentCommandAck, _impl_.progress_percent_), 2>(),
       {24, 2, 0,
        PROTOBUF_FIELD_OFFSET(AgentCommandAck, _impl_.progress_percent_)}},
    }}, {{
      65535, 65535
    }}, {{
      // string message_id = 1;
      {PROTOBUF_FIELD_OFFSET(AgentCommandAck, _impl_.message_id_), _Internal::kHasBitsOffset + 0, 0, (0 | ::_fl::kFcOptional | ::_fl::kUtf8String | ::_fl::kRepAString)},
      // .google.protobuf.Timestamp created_at = 2;
      {PROTOBUF_FIELD_OFFSET(AgentCommandAck, _impl_.created_at_), _Internal::kHasBitsOffset + 1, 0, (0 | ::_fl::kFcOptional | ::_fl::kMessage | ::_fl::kTvTable)},
      // uint32 progress_percent = 3;
      {PROTOBUF_FIELD_OFFSET(AgentCommandAck, _impl_.progress_percent_), _Internal::kHasBitsOffset + 2, 0, (0 | ::_fl::kFcOptional | ::_fl::kUInt32)},
    }},
    {{
        #ifndef PROTOBUF_MESSAGE_GLOBALS
        {::_pbi::TcParser::GetTable<::google::protobuf::Timestamp>()},
        #else
        {::_pbi::FieldAuxMessageGlobals(), &::google::protobuf::Timestamp_globals_},
        #endif
    }},
    {{
      "\30\12\0\0\0\0\0\0"
      "agent.v1.AgentCommandAck"
      "message_id"
    }},
  };
}


inline constexpr AgentCommandAck::Impl_::Impl_(
    [[maybe_unused]] ::google::protobuf::internal::InternalVisibility visibility,
    ::_pbi::ConstantInitialized) noexcept
      : _cached_size_{0},
        message_id_(
            &::google::protobuf::internal::fixed_address_empty_string,
            ::_pbi::ConstantInitialized()),
        created_at_{nullptr},
        progress_percent_{0u} {}

template <typename>
constexpr AgentCommandAck::AgentCommandAck(::_pbi::ConstantInitialized,
                       const ::_pbi::ClassData* PROTOBUF_NONNULL class_data)
    : ::google::protobuf::Message(
#if defined(PROTOBUF_CUSTOM_VTABLE)
          class_data
#endif  // PROTOBUF_CUSTOM_VTABLE
          ),
      _impl_(internal_visibility(), ::_pbi::ConstantInitialized()) {
}
inline void* PROTOBUF_NONNULL AgentCommandAck::PlacementNew_(
    const void* PROTOBUF_NONNULL, void* PROTOBUF_NONNULL mem,
    ::google::protobuf::Arena* PROTOBUF_NULLABLE arena) {
  return ::new (mem) AgentCommandAck(arena);
}
constexpr auto AgentCommandAck::InternalNewImpl_() {
  return ::google::protobuf::internal::MessageCreator::CopyInit(sizeof(AgentCommandAck), alignof(AgentCommandAck));
}
constexpr auto AgentCommandAck::InternalGenerateClassData_(
    const MessageLite& prototype,
    const ::google::protobuf::internal::TcParseTableBase* tc_table) {
  return ::google::protobuf::internal::ClassDataFull{
      ::google::protobuf::internal::ClassData{
          &prototype,
#ifndef PROTOBUF_MESSAGE_GLOBALS
          &_table_.header,
#else
          tc_table,
#endif
          nullptr,  // IsInitialized
          &AgentCommandAck::MergeImpl,
          ::google::protobuf::Message::GetNewImpl<AgentCommandAck>(),
#if defined(PROTOBUF_CUSTOM_VTABLE)
          &AgentCommandAck::SharedDtor,
          ::google::protobuf::Message::GetClearImpl<AgentCommandAck>(), &AgentCommandAck::ByteSizeLong,
              &AgentCommandAck::_InternalSerialize,
#endif  // PROTOBUF_CUSTOM_VTABLE
          PROTOBUF_FIELD_OFFSET(AgentCommandAck, _impl_._cached_size_),
          false,
      },
#ifdef PROTOBUF_MESSAGE_GLOBALS
      &file_reflection_data[2],
#else   // !PROTOBUF_MESSAGE_GLOBALS
      &::_pbi::kDescriptorMethods,
      &descriptor_table_agent_2fv1_2fcmds_2eproto,
      nullptr,  // tracker
#endif  // PROTOBUF_MESSAGE_GLOBALS
  };
}
struct AgentCommandAckGlobalsTypeInternal : ::_pbi::MessageGlobalsBase {
  constexpr AgentCommandAckGlobalsTypeInternal()
      :
#ifndef PROTOBUF_MESSAGE_GLOBALS
        _default(::_pbi::ConstantInitialized{},
                 AgentCommandAck_class_data_.base())
#else   // !PROTOBUF_MESSAGE_GLOBALS
        MessageGlobalsBase(AgentCommandAck::InternalGenerateClassData_(
            _default, &AgentCommandAck_globals_._table.header)),
        _default(::_pbi::ConstantInitialized{}, GetClassData()),
        _table(::_pbi::PrivateAccess::GenerateParseTable<AgentCommandAck>(
            GetClassData()))
#endif  // PROTOBUF_MESSAGE_GLOBALS
  {
  }
  ~AgentCommandAckGlobalsTypeInternal() {}
  union {
    alignas(::_pbi::kMaxMessageAlignment) AgentCommandAck _default;
  };
#ifdef PROTOBUF_MESSAGE_GLOBALS
  decltype(::_pbi::PrivateAccess::GenerateParseTable<AgentCommandAck>(
      ::std::declval<const ::_pbi::ClassData*>())) _table;
#endif
};
#ifdef PROTOBUF_MESSAGE_GLOBALS
static_assert(PROTOBUF_FIELD_OFFSET(AgentCommandAckGlobalsTypeInternal, _default) ==
              ::_pbi::MessageGlobalsBase::OffsetToDefault());
#endif  // PROTOBUF_MESSAGE_GLOBALS

PROTOBUF_ATTRIBUTE_NO_DESTROY PROTOBUF_CONSTINIT
    PROTOBUF_ATTRIBUTE_INIT_PRIORITY1 PROTOBUF_MESSAGE_GLOBALS_CONST AgentCommandAckGlobalsTypeInternal AgentCommandAck_globals_
        PROTOBUF_MESSAGE_GLOBALS_SECTION(.data.rel.ro);
#if defined(PROTOBUF_CUSTOM_VTABLE)
namespace {
const ::_pbi::ClassData* AgentCommandAck_get_class_data() {
#ifdef PROTOBUF_MESSAGE_GLOBALS
  return AgentCommandAck_globals_.GetClassData();
#else
  return AgentCommandAck_class_data_.base();
#endif  // PROTOBUF_MESSAGE_GLOBALS
}
}  // namespace
#endif  // PROTOBUF_CUSTOM_VTABLE
}  // namespace v1
}  // namespace agent
static constexpr const ::_pb::EnumDescriptor* PROTOBUF_NONNULL* PROTOBUF_NULLABLE
    file_level_enum_descriptors_agent_2fv1_2fcmds_2eproto = nullptr;
static constexpr const ::_pb::ServiceDescriptor* PROTOBUF_NONNULL* PROTOBUF_NULLABLE
    file_level_service_descriptors_agent_2fv1_2fcmds_2eproto = nullptr;
const ::uint32_t
    TableStruct_agent_2fv1_2fcmds_2eproto::offsets[] ABSL_ATTRIBUTE_SECTION_VARIABLE(
        protodesc_cold) = {
        0x081, // bitmap
        PROTOBUF_FIELD_OFFSET(::agent::v1::CancelPendingRequest, _impl_._has_bits_),
        6, // hasbit index offset
        PROTOBUF_FIELD_OFFSET(::agent::v1::CancelPendingRequest, _impl_.request_id_),
        PROTOBUF_FIELD_OFFSET(::agent::v1::CancelPendingRequest, _impl_.timestamp_),
        PROTOBUF_FIELD_OFFSET(::agent::v1::CancelPendingRequest, _impl_.trace_id_),
        0,
        2,
        1,
        0x081, // bitmap
        PROTOBUF_FIELD_OFFSET(::agent::v1::AgentCommand, _impl_._has_bits_),
        9, // hasbit index offset
        PROTOBUF_FIELD_OFFSET(::agent::v1::AgentCommand, _impl_.callback_topic_),
        PROTOBUF_FIELD_OFFSET(::agent::v1::AgentCommand, _impl_.ack_topic_),
        PROTOBUF_FIELD_OFFSET(::agent::v1::AgentCommand, _impl_.message_id_),
        PROTOBUF_FIELD_OFFSET(::agent::v1::AgentCommand, _impl_.timeout_),
        PROTOBUF_FIELD_OFFSET(::agent::v1::AgentCommand, _impl_.ack_timeout_),
        PROTOBUF_FIELD_OFFSET(::agent::v1::AgentCommand, _impl_.resource_),
        0,
        1,
        2,
        4,
        5,
        3,
        0x081, // bitmap
        PROTOBUF_FIELD_OFFSET(::agent::v1::AgentCommandAck, _impl_._has_bits_),
        6, // hasbit index offset
        PROTOBUF_FIELD_OFFSET(::agent::v1::AgentCommandAck, _impl_.message_id_),
        PROTOBUF_FIELD_OFFSET(::agent::v1::AgentCommandAck, _impl_.created_at_),
        PROTOBUF_FIELD_OFFSET(::agent::v1::AgentCommandAck, _impl_.progress_percent_),
        0,
        1,
        2,
        0x081, // bitmap
        PROTOBUF_FIELD_OFFSET(::agent::v1::AgentCommandResponse, _impl_._has_bits_),
        9, // hasbit index offset
        PROTOBUF_FIELD_OFFSET(::agent::v1::AgentCommandResponse, _impl_.message_id_),
        PROTOBUF_FIELD_OFFSET(::agent::v1::AgentCommandResponse, _impl_.status_code_),
        PROTOBUF_FIELD_OFFSET(::agent::v1::AgentCommandResponse, _impl_.response_),
        PROTOBUF_FIELD_OFFSET(::agent::v1::AgentCommandResponse, _impl_.timestamp_),
        PROTOBUF_FIELD_OFFSET(::agent::v1::AgentCommandResponse, _impl_.received_at_),
        PROTOBUF_FIELD_OFFSET(::agent::v1::AgentCommandResponse, _impl_.exec_time_),
        0,
        5,
        1,
        2,
        3,
        4,
};

static const ::_pbi::MigrationSchema
    schemas[] ABSL_ATTRIBUTE_SECTION_VARIABLE(protodesc_cold) = {
        {0, sizeof(::agent::v1::CancelPendingRequest)},
        {9, sizeof(::agent::v1::AgentCommand)},
        {24, sizeof(::agent::v1::AgentCommandAck)},
        {33, sizeof(::agent::v1::AgentCommandResponse)},
};
static const ::_pbi::MessageGlobalsBase* PROTOBUF_NONNULL const
    file_message_globals[] = {
        &::agent::v1::CancelPendingRequest_globals_,
        &::agent::v1::AgentCommand_globals_,
        &::agent::v1::AgentCommandAck_globals_,
        &::agent::v1::AgentCommandResponse_globals_,
};
const char descriptor_table_protodef_agent_2fv1_2fcmds_2eproto[] ABSL_ATTRIBUTE_SECTION_VARIABLE(
    protodesc_cold) = {
    "\n\023agent/v1/cmds.proto\022\010agent.v1\032\037google/"
    "protobuf/timestamp.proto\"k\n\024CancelPendin"
    "gRequest\022\022\n\nrequest_id\030\001 \001(\t\022-\n\ttimestam"
    "p\030\002 \001(\0132\032.google.protobuf.Timestamp\022\020\n\010t"
    "race_id\030\003 \001(\t\"\205\001\n\014AgentCommand\022\026\n\016callba"
    "ck_topic\030\001 \001(\t\022\021\n\tack_topic\030\002 \001(\t\022\022\n\nmes"
    "sage_id\030\003 \001(\t\022\017\n\007timeout\030\004 \001(\004\022\023\n\013ack_ti"
    "meout\030\005 \001(\004\022\020\n\010resource\030\006 \001(\t\"o\n\017AgentCo"
    "mmandAck\022\022\n\nmessage_id\030\001 \001(\t\022.\n\ncreated_"
    "at\030\002 \001(\0132\032.google.protobuf.Timestamp\022\030\n\020"
    "progress_percent\030\003 \001(\r\"\304\001\n\024AgentCommandR"
    "esponse\022\022\n\nmessage_id\030\001 \001(\t\022\023\n\013status_co"
    "de\030\002 \001(\r\022\020\n\010response\030\003 \001(\014\022-\n\ttimestamp\030"
    "\004 \001(\0132\032.google.protobuf.Timestamp\022/\n\013rec"
    "eived_at\030\005 \001(\0132\032.google.protobuf.Timesta"
    "mp\022\021\n\texec_time\030\006 \001(\004B\020Z\016agent/v1;agentb"
    "\006proto3"
};
static const ::_pbi::DescriptorTable* PROTOBUF_NONNULL const
    descriptor_table_agent_2fv1_2fcmds_2eproto_deps[1] = {
        &::descriptor_table_google_2fprotobuf_2ftimestamp_2eproto,
};
static ::absl::once_flag descriptor_table_agent_2fv1_2fcmds_2eproto_once;
PROTOBUF_CONSTINIT const ::_pbi::DescriptorTable descriptor_table_agent_2fv1_2fcmds_2eproto = {
    false,
    false,
    647,
    descriptor_table_protodef_agent_2fv1_2fcmds_2eproto,
    "agent/v1/cmds.proto",
    &descriptor_table_agent_2fv1_2fcmds_2eproto_once,
    descriptor_table_agent_2fv1_2fcmds_2eproto_deps,
    1,
    4,
    schemas,
    file_message_globals,
    TableStruct_agent_2fv1_2fcmds_2eproto::offsets,
    file_level_enum_descriptors_agent_2fv1_2fcmds_2eproto,
    file_level_service_descriptors_agent_2fv1_2fcmds_2eproto,
};
namespace agent {
namespace v1 {
// ===================================================================

void CancelPendingRequest::clear_timestamp() {
  ::google::protobuf::internal::TSanWrite(&_impl_);
  if (_impl_.timestamp_ != nullptr) _impl_.timestamp_->Clear();
  ClearHasBit(_impl_._has_bits_[0], 0x00000004U);
}
CancelPendingRequest::CancelPendingRequest(::google::protobuf::Arena* PROTOBUF_NULLABLE arena)
#if defined(PROTOBUF_CUSTOM_VTABLE)
    : ::google::protobuf::Message(arena, CancelPendingRequest_get_class_data()) {
#else   // PROTOBUF_CUSTOM_VTABLE
    : ::google::protobuf::Message(arena) {
#endif  // PROTOBUF_CUSTOM_VTABLE
  SharedCtor(arena);
  // @@protoc_insertion_point(arena_constructor:agent.v1.CancelPendingRequest)
}
PROTOBUF_NDEBUG_INLINE CancelPendingRequest::Impl_::Impl_(
    [[maybe_unused]] ::google::protobuf::internal::InternalVisibility visibility,
    [[maybe_unused]] ::google::protobuf::Arena* PROTOBUF_NULLABLE arena, const Impl_& from,
    [[maybe_unused]] const ::agent::v1::CancelPendingRequest& from_msg)
      : _has_bits_{from._has_bits_},
        _cached_size_{0},
        request_id_(arena, from.request_id_),
        trace_id_(arena, from.trace_id_) {}

CancelPendingRequest::CancelPendingRequest(
    ::google::protobuf::Arena* PROTOBUF_NULLABLE arena,
    const CancelPendingRequest& from)
#if defined(PROTOBUF_CUSTOM_VTABLE)
    : ::google::protobuf::Message(arena, CancelPendingRequest_get_class_data()) {

#else   // PROTOBUF_CUSTOM_VTABLE
    : ::google::protobuf::Message(arena) {
#endif  // PROTOBUF_CUSTOM_VTABLE
  CancelPendingRequest* const _this = this;
  (void)_this;
  _internal_metadata_.MergeFrom<::google::protobuf::UnknownFieldSet>(
      from._internal_metadata_);
  new (&_impl_) Impl_(internal_visibility(), arena, from._impl_, from);
  ::uint32_t cached_has_bits = _impl_._has_bits_[0];
  _impl_.timestamp_ = (CheckHasBit(cached_has_bits, 0x00000004U))
                ? ::google::protobuf::Message::CopyConstruct(arena, *from._impl_.timestamp_)
                : nullptr;

  // @@protoc_insertion_point(copy_constructor:agent.v1.CancelPendingRequest)
}
PROTOBUF_NDEBUG_INLINE CancelPendingRequest::Impl_::Impl_(
    [[maybe_unused]] ::google::protobuf::internal::InternalVisibility visibility,
    [[maybe_unused]] ::google::protobuf::Arena* PROTOBUF_NULLABLE arena)
      : _cached_size_{0},
        request_id_(arena),
        trace_id_(arena) {}

inline void CancelPendingRequest::SharedCtor(::_pb::Arena* PROTOBUF_NULLABLE arena) {
  new (&_impl_) Impl_(internal_visibility(), arena);
  _impl_.timestamp_ = {};
}
CancelPendingRequest::~CancelPendingRequest() {
  // @@protoc_insertion_point(destructor:agent.v1.CancelPendingRequest)
  SharedDtor(*this);
}
inline void CancelPendingRequest::SharedDtor(MessageLite& self) {
  CancelPendingRequest& this_ = static_cast<CancelPendingRequest&>(self);
  if constexpr (::_pbi::DebugHardenCheckHasBitConsistency()) {
    this_.CheckHasBitConsistency();
  }
  this_._internal_metadata_.Delete<::google::protobuf::UnknownFieldSet>();
  ABSL_DCHECK(this_.GetArena() == nullptr);
  this_._impl_.request_id_.Destroy();
  this_._impl_.trace_id_.Destroy();
  delete this_._impl_.timestamp_;
  this_._impl_.~Impl_();
}

#ifndef PROTOBUF_MESSAGE_GLOBALS
PROTOBUF_CONSTINIT PROTOBUF_ATTRIBUTE_INIT_PRIORITY1 const
    ::google::protobuf::internal::ClassDataFull CancelPendingRequest_class_data_ =
        CancelPendingRequest::InternalGenerateClassData_(CancelPendingRequest_globals_._default);

PROTOBUF_ATTRIBUTE_WEAK const ::google::protobuf::internal::ClassData* PROTOBUF_NONNULL
CancelPendingRequest::GetClassData() const {
  ::google::protobuf::internal::PrefetchToLocalCache(&CancelPendingRequest_class_data_);
  ::google::protobuf::internal::PrefetchToLocalCache(CancelPendingRequest_class_data_.tc_table);
  return CancelPendingRequest_class_data_.base();
}
#else
PROTOBUF_ATTRIBUTE_WEAK const ::google::protobuf::internal::ClassData* PROTOBUF_NONNULL
CancelPendingRequest::GetClassData() const {
  ::google::protobuf::internal::PrefetchToLocalCache(&CancelPendingRequest_globals_);
  ::google::protobuf::internal::PrefetchToLocalCache(
      ::google::protobuf::internal::MessageGlobalsBase::ToParseTableBase(&CancelPendingRequest_globals_));
  return CancelPendingRequest_globals_.GetClassData();
}
#endif  // !PROTOBUF_MESSAGE_GLOBALS
#ifndef PROTOBUF_MESSAGE_GLOBALS
PROTOBUF_CONSTINIT
PROTOBUF_ATTRIBUTE_INIT_PRIORITY1 const CancelPendingRequest::ParseTableT_
    CancelPendingRequest::_table_ =
        CancelPendingRequest::InternalGenerateParseTable_(CancelPendingRequest_class_data_.base());
#endif  // !PROTOBUF_MESSAGE_GLOBALS
PROTOBUF_NOINLINE void CancelPendingRequest::Clear() {
// @@protoc_insertion_point(message_clear_start:agent.v1.CancelPendingRequest)
  ::google::protobuf::internal::TSanWrite(&_impl_);
  ::uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void) cached_has_bits;

  cached_has_bits = _impl_._has_bits_[0];
  if (BatchCheckHasBit(cached_has_bits, 0x00000007U)) {
    if (CheckHasBit(cached_has_bits, 0x00000001U)) {
      _impl_.request_id_.ClearNonDefaultToEmpty();
    }
    if (CheckHasBit(cached_has_bits, 0x00000002U)) {
      _impl_.trace_id_.ClearNonDefaultToEmpty();
    }
    if (CheckHasBit(cached_has_bits, 0x00000004U)) {
      ABSL_DCHECK(_impl_.timestamp_ != nullptr);
      _impl_.timestamp_->Clear();
    }
  }
  _impl_._has_bits_.Clear();
  _internal_metadata_.Clear<::google::protobuf::UnknownFieldSet>();
}

#if defined(PROTOBUF_CUSTOM_VTABLE)
::uint8_t* PROTOBUF_NONNULL CancelPendingRequest::_InternalSerialize(
    const ::google::protobuf::MessageLite& base, ::uint8_t* PROTOBUF_NONNULL target,
    ::google::protobuf::io::EpsCopyOutputStream* PROTOBUF_NONNULL stream) {
  const CancelPendingRequest& this_ = static_cast<const CancelPendingRequest&>(base);
#else   // PROTOBUF_CUSTOM_VTABLE
::uint8_t* PROTOBUF_NONNULL CancelPendingRequest::_InternalSerialize(
    ::uint8_t* PROTOBUF_NONNULL target,
    ::google::protobuf::io::EpsCopyOutputStream* PROTOBUF_NONNULL stream) const {
  const CancelPendingRequest& this_ = *this;
#endif  // PROTOBUF_CUSTOM_VTABLE
  if constexpr (::_pbi::DebugHardenCheckHasBitConsistency()) {
    this_.CheckHasBitConsistency();
  }
  // @@protoc_insertion_point(serialize_to_array_start:agent.v1.CancelPendingRequest)
  ::uint32_t cached_has_bits = 0;
  (void)cached_has_bits;

  cached_has_bits = this_._impl_._has_bits_[0];
  // string request_id = 1;
  if (CheckHasBit(cached_has_bits, 0x00000001U)) {
    if (!this_._internal_request_id().empty()) {
      const ::std::string& _s = this_._internal_request_id();
      ::google::protobuf::internal::WireFormatLite::VerifyUtf8String(
          _s.data(), static_cast<int>(_s.length()), ::google::protobuf::internal::WireFormatLite::SERIALIZE, "agent.v1.CancelPendingRequest.request_id");
      target = stream->WriteStringMaybeAliased(1, _s, target);
    }
  }

  // .google.protobuf.Timestamp timestamp = 2;
  if (CheckHasBit(cached_has_bits, 0x00000004U)) {
    target = ::google::protobuf::internal::WireFormatLite::InternalWriteMessage(
        2, *this_._impl_.timestamp_, this_._impl_.timestamp_->GetCachedSize(), target,
        stream);
  }

  // string trace_id = 3;
  if (CheckHasBit(cached_has_bits, 0x00000002U)) {
    if (!this_._internal_trace_id().empty()) {
      const ::std::string& _s = this_._internal_trace_id();
      ::google::protobuf::internal::WireFormatLite::VerifyUtf8String(
          _s.data(), static_cast<int>(_s.length()), ::google::protobuf::internal::WireFormatLite::SERIALIZE, "agent.v1.CancelPendingRequest.trace_id");
      target = stream->WriteStringMaybeAliased(3, _s, target);
    }
  }

  if (ABSL_PREDICT_FALSE(this_._internal_metadata_.have_unknown_fields())) {
    target =
        ::_pbi::WireFormat::InternalSerializeUnknownFieldsToArray(
            this_._internal_metadata_.unknown_fields<::google::protobuf::UnknownFieldSet>(::google::protobuf::UnknownFieldSet::default_instance), target, stream);
  }
  // @@protoc_insertion_point(serialize_to_array_end:agent.v1.CancelPendingRequest)
  return target;
}

#if defined(PROTOBUF_CUSTOM_VTABLE)
::size_t CancelPendingRequest::ByteSizeLong(const MessageLite& base) {
  const CancelPendingRequest& this_ = static_cast<const CancelPendingRequest&>(base);
#else   // PROTOBUF_CUSTOM_VTABLE
::size_t CancelPendingRequest::ByteSizeLong() const {
  const CancelPendingRequest& this_ = *this;
#endif  // PROTOBUF_CUSTOM_VTABLE
  // @@protoc_insertion_point(message_byte_size_start:agent.v1.CancelPendingRequest)
  ::size_t total_size = 0;

  ::uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void)cached_has_bits;

  ::_pbi::Prefetch5LinesFrom7Lines(&this_);
  cached_has_bits = this_._impl_._has_bits_[0];
  if (BatchCheckHasBit(cached_has_bits, 0x00000007U)) {
    // string request_id = 1;
    if (CheckHasBit(cached_has_bits, 0x00000001U)) {
      if (!this_._internal_request_id().empty()) {
        total_size += 1 + ::google::protobuf::internal::WireFormatLite::StringSize(
                                        this_._internal_request_id());
      }
    }
    // string trace_id = 3;
    if (CheckHasBit(cached_has_bits, 0x00000002U)) {
      if (!this_._internal_trace_id().empty()) {
        total_size += 1 + ::google::protobuf::internal::WireFormatLite::StringSize(
                                        this_._internal_trace_id());
      }
    }
    // .google.protobuf.Timestamp timestamp = 2;
    if (CheckHasBit(cached_has_bits, 0x00000004U)) {
      total_size += 1 +
                    ::google::protobuf::internal::WireFormatLite::MessageSize(*this_._impl_.timestamp_);
    }
  }
  return this_.MaybeComputeUnknownFieldsSize(total_size,
                                             &this_._impl_._cached_size_);
}

void CancelPendingRequest::MergeImpl(::google::protobuf::MessageLite& to_msg,
                      const ::google::protobuf::MessageLite& from_msg) {
   auto* const _this = static_cast<CancelPendingRequest*>(&to_msg);
  auto& from = static_cast<const CancelPendingRequest&>(from_msg);
  if constexpr (::_pbi::DebugHardenCheckHasBitConsistency()) {
    from.CheckHasBitConsistency();
  }
  ::google::protobuf::Arena* arena = _this->GetArena();
  // @@protoc_insertion_point(class_specific_merge_from_start:agent.v1.CancelPendingRequest)
  ABSL_DCHECK_NE(&from, _this);
  ::uint32_t cached_has_bits = 0;
  (void)cached_has_bits;

  cached_has_bits = from._impl_._has_bits_[0];
  if (BatchCheckHasBit(cached_has_bits, 0x00000007U)) {
    if (CheckHasBit(cached_has_bits, 0x00000001U)) {
      if (!from._internal_request_id().empty()) {
        _this->_internal_set_request_id(from._internal_request_id());
      } else {
        if (_this->_impl_.request_id_.IsDefault()) {
          _this->_internal_set_request_id("");
        }
      }
    }
    if (CheckHasBit(cached_has_bits, 0x00000002U)) {
      if (!from._internal_trace_id().empty()) {
        _this->_internal_set_trace_id(from._internal_trace_id());
      } else {
        if (_this->_impl_.trace_id_.IsDefault()) {
          _this->_internal_set_trace_id("");
        }
      }
    }
    if (CheckHasBit(cached_has_bits, 0x00000004U)) {
      ABSL_DCHECK(from._impl_.timestamp_ != nullptr);
      if (_this->_impl_.timestamp_ == nullptr) {
        _this->_impl_.timestamp_ = ::google::protobuf::Message::CopyConstruct(arena, *from._impl_.timestamp_);
      } else {
        _this->_impl_.timestamp_->MergeFrom(*from._impl_.timestamp_);
      }
    }
  }
  _this->_impl_._has_bits_[0] |= cached_has_bits;
  _this->_internal_metadata_.MergeFrom<::google::protobuf::UnknownFieldSet>(
      from._internal_metadata_);
}

void CancelPendingRequest::CopyFrom(const CancelPendingRequest& from) {
  // @@protoc_insertion_point(class_specific_copy_from_start:agent.v1.CancelPendingRequest)
  if (&from == this) return;
  Clear();
  MergeFrom(from);
}


void CancelPendingRequest::InternalSwap(CancelPendingRequest* PROTOBUF_RESTRICT PROTOBUF_NONNULL other) {
  using ::std::swap;
  auto* arena = GetArena();
  ABSL_DCHECK_EQ(arena, other->GetArena());
  _internal_metadata_.InternalSwap(&other->_internal_metadata_);
  swap(_impl_._has_bits_[0], other->_impl_._has_bits_[0]);
  ::_pbi::ArenaStringPtr::InternalSwap(&_impl_.request_id_, &other->_impl_.request_id_, arena);
  ::_pbi::ArenaStringPtr::InternalSwap(&_impl_.trace_id_, &other->_impl_.trace_id_, arena);
  swap(_impl_.timestamp_, other->_impl_.timestamp_);
}

::google::protobuf::Metadata CancelPendingRequest::GetMetadata() const {
  return ::google::protobuf::Message::GetMetadataImpl(GetClassData()->full());
}
// ===================================================================

AgentCommand::AgentCommand(::google::protobuf::Arena* PROTOBUF_NULLABLE arena)
#if defined(PROTOBUF_CUSTOM_VTABLE)
    : ::google::protobuf::Message(arena, AgentCommand_get_class_data()) {
#else   // PROTOBUF_CUSTOM_VTABLE
    : ::google::protobuf::Message(arena) {
#endif  // PROTOBUF_CUSTOM_VTABLE
  SharedCtor(arena);
  // @@protoc_insertion_point(arena_constructor:agent.v1.AgentCommand)
}
PROTOBUF_NDEBUG_INLINE AgentCommand::Impl_::Impl_(
    [[maybe_unused]] ::google::protobuf::internal::InternalVisibility visibility,
    [[maybe_unused]] ::google::protobuf::Arena* PROTOBUF_NULLABLE arena, const Impl_& from,
    [[maybe_unused]] const ::agent::v1::AgentCommand& from_msg)
      : _has_bits_{from._has_bits_},
        _cached_size_{0},
        callback_topic_(arena, from.callback_topic_),
        ack_topic_(arena, from.ack_topic_),
        message_id_(arena, from.message_id_),
        resource_(arena, from.resource_) {}

AgentCommand::AgentCommand(
    ::google::protobuf::Arena* PROTOBUF_NULLABLE arena,
    const AgentCommand& from)
#if defined(PROTOBUF_CUSTOM_VTABLE)
    : ::google::protobuf::Message(arena, AgentCommand_get_class_data()) {

#else   // PROTOBUF_CUSTOM_VTABLE
    : ::google::protobuf::Message(arena) {
#endif  // PROTOBUF_CUSTOM_VTABLE
  AgentCommand* const _this = this;
  (void)_this;
  _internal_metadata_.MergeFrom<::google::protobuf::UnknownFieldSet>(
      from._internal_metadata_);
  new (&_impl_) Impl_(internal_visibility(), arena, from._impl_, from);
  ::memcpy(reinterpret_cast<char*>(&_impl_) +
               offsetof(Impl_, timeout_),
           reinterpret_cast<const char*>(&from._impl_) +
               offsetof(Impl_, timeout_),
           offsetof(Impl_, ack_timeout_) -
               offsetof(Impl_, timeout_) +
               sizeof(Impl_::ack_timeout_));

  // @@protoc_insertion_point(copy_constructor:agent.v1.AgentCommand)
}
PROTOBUF_NDEBUG_INLINE AgentCommand::Impl_::Impl_(
    [[maybe_unused]] ::google::protobuf::internal::InternalVisibility visibility,
    [[maybe_unused]] ::google::protobuf::Arena* PROTOBUF_NULLABLE arena)
      : _cached_size_{0},
        callback_topic_(arena),
        ack_topic_(arena),
        message_id_(arena),
        resource_(arena) {}

inline void AgentCommand::SharedCtor(::_pb::Arena* PROTOBUF_NULLABLE arena) {
  new (&_impl_) Impl_(internal_visibility(), arena);
  ::memset(reinterpret_cast<char*>(&_impl_) +
               offsetof(Impl_, timeout_),
           0,
           offsetof(Impl_, ack_timeout_) -
               offsetof(Impl_, timeout_) +
               sizeof(Impl_::ack_timeout_));
}
AgentCommand::~AgentCommand() {
  // @@protoc_insertion_point(destructor:agent.v1.AgentCommand)
  SharedDtor(*this);
}
inline void AgentCommand::SharedDtor(MessageLite& self) {
  AgentCommand& this_ = static_cast<AgentCommand&>(self);
  if constexpr (::_pbi::DebugHardenCheckHasBitConsistency()) {
    this_.CheckHasBitConsistency();
  }
  this_._internal_metadata_.Delete<::google::protobuf::UnknownFieldSet>();
  ABSL_DCHECK(this_.GetArena() == nullptr);
  this_._impl_.callback_topic_.Destroy();
  this_._impl_.ack_topic_.Destroy();
  this_._impl_.message_id_.Destroy();
  this_._impl_.resource_.Destroy();
  this_._impl_.~Impl_();
}

#ifndef PROTOBUF_MESSAGE_GLOBALS
PROTOBUF_CONSTINIT PROTOBUF_ATTRIBUTE_INIT_PRIORITY1 const
    ::google::protobuf::internal::ClassDataFull AgentCommand_class_data_ =
        AgentCommand::InternalGenerateClassData_(AgentCommand_globals_._default);

PROTOBUF_ATTRIBUTE_WEAK const ::google::protobuf::internal::ClassData* PROTOBUF_NONNULL
AgentCommand::GetClassData() const {
  ::google::protobuf::internal::PrefetchToLocalCache(&AgentCommand_class_data_);
  ::google::protobuf::internal::PrefetchToLocalCache(AgentCommand_class_data_.tc_table);
  return AgentCommand_class_data_.base();
}
#else
PROTOBUF_ATTRIBUTE_WEAK const ::google::protobuf::internal::ClassData* PROTOBUF_NONNULL
AgentCommand::GetClassData() const {
  ::google::protobuf::internal::PrefetchToLocalCache(&AgentCommand_globals_);
  ::google::protobuf::internal::PrefetchToLocalCache(
      ::google::protobuf::internal::MessageGlobalsBase::ToParseTableBase(&AgentCommand_globals_));
  return AgentCommand_globals_.GetClassData();
}
#endif  // !PROTOBUF_MESSAGE_GLOBALS
#ifndef PROTOBUF_MESSAGE_GLOBALS
PROTOBUF_CONSTINIT
PROTOBUF_ATTRIBUTE_INIT_PRIORITY1 const AgentCommand::ParseTableT_
    AgentCommand::_table_ =
        AgentCommand::InternalGenerateParseTable_(AgentCommand_class_data_.base());
#endif  // !PROTOBUF_MESSAGE_GLOBALS
PROTOBUF_NOINLINE void AgentCommand::Clear() {
// @@protoc_insertion_point(message_clear_start:agent.v1.AgentCommand)
  ::google::protobuf::internal::TSanWrite(&_impl_);
  ::uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void) cached_has_bits;

  cached_has_bits = _impl_._has_bits_[0];
  if (BatchCheckHasBit(cached_has_bits, 0x0000000fU)) {
    if (CheckHasBit(cached_has_bits, 0x00000001U)) {
      _impl_.callback_topic_.ClearNonDefaultToEmpty();
    }
    if (CheckHasBit(cached_has_bits, 0x00000002U)) {
      _impl_.ack_topic_.ClearNonDefaultToEmpty();
    }
    if (CheckHasBit(cached_has_bits, 0x00000004U)) {
      _impl_.message_id_.ClearNonDefaultToEmpty();
    }
    if (CheckHasBit(cached_has_bits, 0x00000008U)) {
      _impl_.resource_.ClearNonDefaultToEmpty();
    }
  }
  if (BatchCheckHasBit(cached_has_bits, 0x00000030U)) {
    ::memset(&_impl_.timeout_, 0, static_cast<::size_t>(
        reinterpret_cast<char*>(&_impl_.ack_timeout_) -
        reinterpret_cast<char*>(&_impl_.timeout_)) + sizeof(_impl_.ack_timeout_));
  }
  _impl_._has_bits_.Clear();
  _internal_metadata_.Clear<::google::protobuf::UnknownFieldSet>();
}

#if defined(PROTOBUF_CUSTOM_VTABLE)
::uint8_t* PROTOBUF_NONNULL AgentCommand::_InternalSerialize(
    const ::google::protobuf::MessageLite& base, ::uint8_t* PROTOBUF_NONNULL target,
    ::google::protobuf::io::EpsCopyOutputStream* PROTOBUF_NONNULL stream) {
  const AgentCommand& this_ = static_cast<const AgentCommand&>(base);
#else   // PROTOBUF_CUSTOM_VTABLE
::uint8_t* PROTOBUF_NONNULL AgentCommand::_InternalSerialize(
    ::uint8_t* PROTOBUF_NONNULL target,
    ::google::protobuf::io::EpsCopyOutputStream* PROTOBUF_NONNULL stream) const {
  const AgentCommand& this_ = *this;
#endif  // PROTOBUF_CUSTOM_VTABLE
  if constexpr (::_pbi::DebugHardenCheckHasBitConsistency()) {
    this_.CheckHasBitConsistency();
  }
  // @@protoc_insertion_point(serialize_to_array_start:agent.v1.AgentCommand)
  ::uint32_t cached_has_bits = 0;
  (void)cached_has_bits;

  cached_has_bits = this_._impl_._has_bits_[0];
  // string callback_topic = 1;
  if (CheckHasBit(cached_has_bits, 0x00000001U)) {
    if (!this_._internal_callback_topic().empty()) {
      const ::std::string& _s = this_._internal_callback_topic();
      ::google::protobuf::internal::WireFormatLite::VerifyUtf8String(
          _s.data(), static_cast<int>(_s.length()), ::google::protobuf::internal::WireFormatLite::SERIALIZE, "agent.v1.AgentCommand.callback_topic");
      target = stream->WriteStringMaybeAliased(1, _s, target);
    }
  }

  // string ack_topic = 2;
  if (CheckHasBit(cached_has_bits, 0x00000002U)) {
    if (!this_._internal_ack_topic().empty()) {
      const ::std::string& _s = this_._internal_ack_topic();
      ::google::protobuf::internal::WireFormatLite::VerifyUtf8String(
          _s.data(), static_cast<int>(_s.length()), ::google::protobuf::internal::WireFormatLite::SERIALIZE, "agent.v1.AgentCommand.ack_topic");
      target = stream->WriteStringMaybeAliased(2, _s, target);
    }
  }

  // string message_id = 3;
  if (CheckHasBit(cached_has_bits, 0x00000004U)) {
    if (!this_._internal_message_id().empty()) {
      const ::std::string& _s = this_._internal_message_id();
      ::google::protobuf::internal::WireFormatLite::VerifyUtf8String(
          _s.data(), static_cast<int>(_s.length()), ::google::protobuf::internal::WireFormatLite::SERIALIZE, "agent.v1.AgentCommand.message_id");
      target = stream->WriteStringMaybeAliased(3, _s, target);
    }
  }

  // uint64 timeout = 4;
  if (CheckHasBit(cached_has_bits, 0x00000010U)) {
    if (this_._internal_timeout() != 0) {
      target = stream->EnsureSpace(target);
      target = ::_pbi::WireFormatLite::WriteUInt64ToArray(
          4, this_._internal_timeout(), target);
    }
  }

  // uint64 ack_timeout = 5;
  if (CheckHasBit(cached_has_bits, 0x00000020U)) {
    if (this_._internal_ack_timeout() != 0) {
      target = stream->EnsureSpace(target);
      target = ::_pbi::WireFormatLite::WriteUInt64ToArray(
          5, this_._internal_ack_timeout(), target);
    }
  }

  // string resource = 6;
  if (CheckHasBit(cached_has_bits, 0x00000008U)) {
    if (!this_._internal_resource().empty()) {
      const ::std::string& _s = this_._internal_resource();
      ::google::protobuf::internal::WireFormatLite::VerifyUtf8String(
          _s.data(), static_cast<int>(_s.length()), ::google::protobuf::internal::WireFormatLite::SERIALIZE, "agent.v1.AgentCommand.resource");
      target = stream->WriteStringMaybeAliased(6, _s, target);
    }
  }

  if (ABSL_PREDICT_FALSE(this_._internal_metadata_.have_unknown_fields())) {
    target =
        ::_pbi::WireFormat::InternalSerializeUnknownFieldsToArray(
            this_._internal_metadata_.unknown_fields<::google::protobuf::UnknownFieldSet>(::google::protobuf::UnknownFieldSet::default_instance), target, stream);
  }
  // @@protoc_insertion_point(serialize_to_array_end:agent.v1.AgentCommand)
  return target;
}

#if defined(PROTOBUF_CUSTOM_VTABLE)
::size_t AgentCommand::ByteSizeLong(const MessageLite& base) {
  const AgentCommand& this_ = static_cast<const AgentCommand&>(base);
#else   // PROTOBUF_CUSTOM_VTABLE
::size_t AgentCommand::ByteSizeLong() const {
  const AgentCommand& this_ = *this;
#endif  // PROTOBUF_CUSTOM_VTABLE
  // @@protoc_insertion_point(message_byte_size_start:agent.v1.AgentCommand)
  ::size_t total_size = 0;

  ::uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void)cached_has_bits;

  ::_pbi::Prefetch5LinesFrom7Lines(&this_);
  cached_has_bits = this_._impl_._has_bits_[0];
  if (BatchCheckHasBit(cached_has_bits, 0x0000003fU)) {
    // string callback_topic = 1;
    if (CheckHasBit(cached_has_bits, 0x00000001U)) {
      if (!this_._internal_callback_topic().empty()) {
        total_size += 1 + ::google::protobuf::internal::WireFormatLite::StringSize(
                                        this_._internal_callback_topic());
      }
    }
    // string ack_topic = 2;
    if (CheckHasBit(cached_has_bits, 0x00000002U)) {
      if (!this_._internal_ack_topic().empty()) {
        total_size += 1 + ::google::protobuf::internal::WireFormatLite::StringSize(
                                        this_._internal_ack_topic());
      }
    }
    // string message_id = 3;
    if (CheckHasBit(cached_has_bits, 0x00000004U)) {
      if (!this_._internal_message_id().empty()) {
        total_size += 1 + ::google::protobuf::internal::WireFormatLite::StringSize(
                                        this_._internal_message_id());
      }
    }
    // string resource = 6;
    if (CheckHasBit(cached_has_bits, 0x00000008U)) {
      if (!this_._internal_resource().empty()) {
        total_size += 1 + ::google::protobuf::internal::WireFormatLite::StringSize(
                                        this_._internal_resource());
      }
    }
    // uint64 timeout = 4;
    if (CheckHasBit(cached_has_bits, 0x00000010U)) {
      if (this_._internal_timeout() != 0) {
        total_size += ::_pbi::WireFormatLite::UInt64SizePlusOne(
            this_._internal_timeout());
      }
    }
    // uint64 ack_timeout = 5;
    if (CheckHasBit(cached_has_bits, 0x00000020U)) {
      if (this_._internal_ack_timeout() != 0) {
        total_size += ::_pbi::WireFormatLite::UInt64SizePlusOne(
            this_._internal_ack_timeout());
      }
    }
  }
  return this_.MaybeComputeUnknownFieldsSize(total_size,
                                             &this_._impl_._cached_size_);
}

void AgentCommand::MergeImpl(::google::protobuf::MessageLite& to_msg,
                      const ::google::protobuf::MessageLite& from_msg) {
   auto* const _this = static_cast<AgentCommand*>(&to_msg);
  auto& from = static_cast<const AgentCommand&>(from_msg);
  if constexpr (::_pbi::DebugHardenCheckHasBitConsistency()) {
    from.CheckHasBitConsistency();
  }
  // @@protoc_insertion_point(class_specific_merge_from_start:agent.v1.AgentCommand)
  ABSL_DCHECK_NE(&from, _this);
  ::uint32_t cached_has_bits = 0;
  (void)cached_has_bits;

  cached_has_bits = from._impl_._has_bits_[0];
  if (BatchCheckHasBit(cached_has_bits, 0x0000003fU)) {
    if (CheckHasBit(cached_has_bits, 0x00000001U)) {
      if (!from._internal_callback_topic().empty()) {
        _this->_internal_set_callback_topic(from._internal_callback_topic());
      } else {
        if (_this->_impl_.callback_topic_.IsDefault()) {
          _this->_internal_set_callback_topic("");
        }
      }
    }
    if (CheckHasBit(cached_has_bits, 0x00000002U)) {
      if (!from._internal_ack_topic().empty()) {
        _this->_internal_set_ack_topic(from._internal_ack_topic());
      } else {
        if (_this->_impl_.ack_topic_.IsDefault()) {
          _this->_internal_set_ack_topic("");
        }
      }
    }
    if (CheckHasBit(cached_has_bits, 0x00000004U)) {
      if (!from._internal_message_id().empty()) {
        _this->_internal_set_message_id(from._internal_message_id());
      } else {
        if (_this->_impl_.message_id_.IsDefault()) {
          _this->_internal_set_message_id("");
        }
      }
    }
    if (CheckHasBit(cached_has_bits, 0x00000008U)) {
      if (!from._internal_resource().empty()) {
        _this->_internal_set_resource(from._internal_resource());
      } else {
        if (_this->_impl_.resource_.IsDefault()) {
          _this->_internal_set_resource("");
        }
      }
    }
    if (CheckHasBit(cached_has_bits, 0x00000010U)) {
      if (from._internal_timeout() != 0) {
        _this->_impl_.timeout_ = from._impl_.timeout_;
      }
    }
    if (CheckHasBit(cached_has_bits, 0x00000020U)) {
      if (from._internal_ack_timeout() != 0) {
        _this->_impl_.ack_timeout_ = from._impl_.ack_timeout_;
      }
    }
  }
  _this->_impl_._has_bits_[0] |= cached_has_bits;
  _this->_internal_metadata_.MergeFrom<::google::protobuf::UnknownFieldSet>(
      from._internal_metadata_);
}

void AgentCommand::CopyFrom(const AgentCommand& from) {
  // @@protoc_insertion_point(class_specific_copy_from_start:agent.v1.AgentCommand)
  if (&from == this) return;
  Clear();
  MergeFrom(from);
}


void AgentCommand::InternalSwap(AgentCommand* PROTOBUF_RESTRICT PROTOBUF_NONNULL other) {
  using ::std::swap;
  auto* arena = GetArena();
  ABSL_DCHECK_EQ(arena, other->GetArena());
  _internal_metadata_.InternalSwap(&other->_internal_metadata_);
  swap(_impl_._has_bits_[0], other->_impl_._has_bits_[0]);
  ::_pbi::ArenaStringPtr::InternalSwap(&_impl_.callback_topic_, &other->_impl_.callback_topic_, arena);
  ::_pbi::ArenaStringPtr::InternalSwap(&_impl_.ack_topic_, &other->_impl_.ack_topic_, arena);
  ::_pbi::ArenaStringPtr::InternalSwap(&_impl_.message_id_, &other->_impl_.message_id_, arena);
  ::_pbi::ArenaStringPtr::InternalSwap(&_impl_.resource_, &other->_impl_.resource_, arena);
  ::google::protobuf::internal::memswap<
      PROTOBUF_FIELD_OFFSET(AgentCommand, _impl_.ack_timeout_)
      + sizeof(AgentCommand::_impl_.ack_timeout_)
      - PROTOBUF_FIELD_OFFSET(AgentCommand, _impl_.timeout_)>(
          reinterpret_cast<char*>(&_impl_.timeout_),
          reinterpret_cast<char*>(&other->_impl_.timeout_));
}

::google::protobuf::Metadata AgentCommand::GetMetadata() const {
  return ::google::protobuf::Message::GetMetadataImpl(GetClassData()->full());
}
// ===================================================================

void AgentCommandAck::clear_created_at() {
  ::google::protobuf::internal::TSanWrite(&_impl_);
  if (_impl_.created_at_ != nullptr) _impl_.created_at_->Clear();
  ClearHasBit(_impl_._has_bits_[0], 0x00000002U);
}
AgentCommandAck::AgentCommandAck(::google::protobuf::Arena* PROTOBUF_NULLABLE arena)
#if defined(PROTOBUF_CUSTOM_VTABLE)
    : ::google::protobuf::Message(arena, AgentCommandAck_get_class_data()) {
#else   // PROTOBUF_CUSTOM_VTABLE
    : ::google::protobuf::Message(arena) {
#endif  // PROTOBUF_CUSTOM_VTABLE
  SharedCtor(arena);
  // @@protoc_insertion_point(arena_constructor:agent.v1.AgentCommandAck)
}
PROTOBUF_NDEBUG_INLINE AgentCommandAck::Impl_::Impl_(
    [[maybe_unused]] ::google::protobuf::internal::InternalVisibility visibility,
    [[maybe_unused]] ::google::protobuf::Arena* PROTOBUF_NULLABLE arena, const Impl_& from,
    [[maybe_unused]] const ::agent::v1::AgentCommandAck& from_msg)
      : _has_bits_{from._has_bits_},
        _cached_size_{0},
        message_id_(arena, from.message_id_) {}

AgentCommandAck::AgentCommandAck(
    ::google::protobuf::Arena* PROTOBUF_NULLABLE arena,
    const AgentCommandAck& from)
#if defined(PROTOBUF_CUSTOM_VTABLE)
    : ::google::protobuf::Message(arena, AgentCommandAck_get_class_data()) {

#else   // PROTOBUF_CUSTOM_VTABLE
    : ::google::protobuf::Message(arena) {
#endif  // PROTOBUF_CUSTOM_VTABLE
  AgentCommandAck* const _this = this;
  (void)_this;
  _internal_metadata_.MergeFrom<::google::protobuf::UnknownFieldSet>(
      from._internal_metadata_);
  new (&_impl_) Impl_(internal_visibility(), arena, from._impl_, from);
  ::uint32_t cached_has_bits = _impl_._has_bits_[0];
  _impl_.created_at_ = (CheckHasBit(cached_has_bits, 0x00000002U))
                ? ::google::protobuf::Message::CopyConstruct(arena, *from._impl_.created_at_)
                : nullptr;
  _impl_.progress_percent_ = from._impl_.progress_percent_;

  // @@protoc_insertion_point(copy_constructor:agent.v1.AgentCommandAck)
}
PROTOBUF_NDEBUG_INLINE AgentCommandAck::Impl_::Impl_(
    [[maybe_unused]] ::google::protobuf::internal::InternalVisibility visibility,
    [[maybe_unused]] ::google::protobuf::Arena* PROTOBUF_NULLABLE arena)
      : _cached_size_{0},
        message_id_(arena) {}

inline void AgentCommandAck::SharedCtor(::_pb::Arena* PROTOBUF_NULLABLE arena) {
  new (&_impl_) Impl_(internal_visibility(), arena);
  ::memset(reinterpret_cast<char*>(&_impl_) +
               offsetof(Impl_, created_at_),
           0,
           offsetof(Impl_, progress_percent_) -
               offsetof(Impl_, created_at_) +
               sizeof(Impl_::progress_percent_));
}
AgentCommandAck::~AgentCommandAck() {
  // @@protoc_insertion_point(destructor:agent.v1.AgentCommandAck)
  SharedDtor(*this);
}
inline void AgentCommandAck::SharedDtor(MessageLite& self) {
  AgentCommandAck& this_ = static_cast<AgentCommandAck&>(self);
  if constexpr (::_pbi::DebugHardenCheckHasBitConsistency()) {
    this_.CheckHasBitConsistency();
  }
  this_._internal_metadata_.Delete<::google::protobuf::UnknownFieldSet>();
  ABSL_DCHECK(this_.GetArena() == nullptr);
  this_._impl_.message_id_.Destroy();
  delete this_._impl_.created_at_;
  this_._impl_.~Impl_();
}

#ifndef PROTOBUF_MESSAGE_GLOBALS
PROTOBUF_CONSTINIT PROTOBUF_ATTRIBUTE_INIT_PRIORITY1 const
    ::google::protobuf::internal::ClassDataFull AgentCommandAck_class_data_ =
        AgentCommandAck::InternalGenerateClassData_(AgentCommandAck_globals_._default);

PROTOBUF_ATTRIBUTE_WEAK const ::google::protobuf::internal::ClassData* PROTOBUF_NONNULL
AgentCommandAck::GetClassData() const {
  ::google::protobuf::internal::PrefetchToLocalCache(&AgentCommandAck_class_data_);
  ::google::protobuf::internal::PrefetchToLocalCache(AgentCommandAck_class_data_.tc_table);
  return AgentCommandAck_class_data_.base();
}
#else
PROTOBUF_ATTRIBUTE_WEAK const ::google::protobuf::internal::ClassData* PROTOBUF_NONNULL
AgentCommandAck::GetClassData() const {
  ::google::protobuf::internal::PrefetchToLocalCache(&AgentCommandAck_globals_);
  ::google::protobuf::internal::PrefetchToLocalCache(
      ::google::protobuf::internal::MessageGlobalsBase::ToParseTableBase(&AgentCommandAck_globals_));
  return AgentCommandAck_globals_.GetClassData();
}
#endif  // !PROTOBUF_MESSAGE_GLOBALS
#ifndef PROTOBUF_MESSAGE_GLOBALS
PROTOBUF_CONSTINIT
PROTOBUF_ATTRIBUTE_INIT_PRIORITY1 const AgentCommandAck::ParseTableT_
    AgentCommandAck::_table_ =
        AgentCommandAck::InternalGenerateParseTable_(AgentCommandAck_class_data_.base());
#endif  // !PROTOBUF_MESSAGE_GLOBALS
PROTOBUF_NOINLINE void AgentCommandAck::Clear() {
// @@protoc_insertion_point(message_clear_start:agent.v1.AgentCommandAck)
  ::google::protobuf::internal::TSanWrite(&_impl_);
  ::uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void) cached_has_bits;

  cached_has_bits = _impl_._has_bits_[0];
  if (BatchCheckHasBit(cached_has_bits, 0x00000003U)) {
    if (CheckHasBit(cached_has_bits, 0x00000001U)) {
      _impl_.message_id_.ClearNonDefaultToEmpty();
    }
    if (CheckHasBit(cached_has_bits, 0x00000002U)) {
      ABSL_DCHECK(_impl_.created_at_ != nullptr);
      _impl_.created_at_->Clear();
    }
  }
  _impl_.progress_percent_ = 0u;
  _impl_._has_bits_.Clear();
  _internal_metadata_.Clear<::google::protobuf::UnknownFieldSet>();
}

#if defined(PROTOBUF_CUSTOM_VTABLE)
::uint8_t* PROTOBUF_NONNULL AgentCommandAck::_InternalSerialize(
    const ::google::protobuf::MessageLite& base, ::uint8_t* PROTOBUF_NONNULL target,
    ::google::protobuf::io::EpsCopyOutputStream* PROTOBUF_NONNULL stream) {
  const AgentCommandAck& this_ = static_cast<const AgentCommandAck&>(base);
#else   // PROTOBUF_CUSTOM_VTABLE
::uint8_t* PROTOBUF_NONNULL AgentCommandAck::_InternalSerialize(
    ::uint8_t* PROTOBUF_NONNULL target,
    ::google::protobuf::io::EpsCopyOutputStream* PROTOBUF_NONNULL stream) const {
  const AgentCommandAck& this_ = *this;
#endif  // PROTOBUF_CUSTOM_VTABLE
  if constexpr (::_pbi::DebugHardenCheckHasBitConsistency()) {
    this_.CheckHasBitConsistency();
  }
  // @@protoc_insertion_point(serialize_to_array_start:agent.v1.AgentCommandAck)
  ::uint32_t cached_has_bits = 0;
  (void)cached_has_bits;

  cached_has_bits = this_._impl_._has_bits_[0];
  // string message_id = 1;
  if (CheckHasBit(cached_has_bits, 0x00000001U)) {
    if (!this_._internal_message_id().empty()) {
      const ::std::string& _s = this_._internal_message_id();
      ::google::protobuf::internal::WireFormatLite::VerifyUtf8String(
          _s.data(), static_cast<int>(_s.length()), ::google::protobuf::internal::WireFormatLite::SERIALIZE, "agent.v1.AgentCommandAck.message_id");
      target = stream->WriteStringMaybeAliased(1, _s, target);
    }
  }

  // .google.protobuf.Timestamp created_at = 2;
  if (CheckHasBit(cached_has_bits, 0x00000002U)) {
    target = ::google::protobuf::internal::WireFormatLite::InternalWriteMessage(
        2, *this_._impl_.created_at_, this_._impl_.created_at_->GetCachedSize(), target,
        stream);
  }

  // uint32 progress_percent = 3;
  if (CheckHasBit(cached_has_bits, 0x00000004U)) {
    if (this_._internal_progress_percent() != 0) {
      target = stream->EnsureSpace(target);
      target = ::_pbi::WireFormatLite::WriteUInt32ToArray(
          3, this_._internal_progress_percent(), target);
    }
  }

  if (ABSL_PREDICT_FALSE(this_._internal_metadata_.have_unknown_fields())) {
    target =
        ::_pbi::WireFormat::InternalSerializeUnknownFieldsToArray(
            this_._internal_metadata_.unknown_fields<::google::protobuf::UnknownFieldSet>(::google::protobuf::UnknownFieldSet::default_instance), target, stream);
  }
  // @@protoc_insertion_point(serialize_to_array_end:agent.v1.AgentCommandAck)
  return target;
}

#if defined(PROTOBUF_CUSTOM_VTABLE)
::size_t AgentCommandAck::ByteSizeLong(const MessageLite& base) {
  const AgentCommandAck& this_ = static_cast<const AgentCommandAck&>(base);
#else   // PROTOBUF_CUSTOM_VTABLE
::size_t AgentCommandAck::ByteSizeLong() const {
  const AgentCommandAck& this_ = *this;
#endif  // PROTOBUF_CUSTOM_VTABLE
  // @@protoc_insertion_point(message_byte_size_start:agent.v1.AgentCommandAck)
  ::size_t total_size = 0;

  ::uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void)cached_has_bits;

  ::_pbi::Prefetch5LinesFrom7Lines(&this_);
  cached_has_bits = this_._impl_._has_bits_[0];
  if (BatchCheckHasBit(cached_has_bits, 0x00000007U)) {
    // string message_id = 1;
    if (CheckHasBit(cached_has_bits, 0x00000001U)) {
      if (!this_._internal_message_id().empty()) {
        total_size += 1 + ::google::protobuf::internal::WireFormatLite::StringSize(
                                        this_._internal_message_id());
      }
    }
    // .google.protobuf.Timestamp created_at = 2;
    if (CheckHasBit(cached_has_bits, 0x00000002U)) {
      total_size += 1 +
                    ::google::protobuf::internal::WireFormatLite::MessageSize(*this_._impl_.created_at_);
    }
    // uint32 progress_percent = 3;
    if (CheckHasBit(cached_has_bits, 0x00000004U)) {
      if (this_._internal_progress_percent() != 0) {
        total_size += ::_pbi::WireFormatLite::UInt32SizePlusOne(
            this_._internal_progress_percent());
      }
    }
  }
  return this_.MaybeComputeUnknownFieldsSize(total_size,
                                             &this_._impl_._cached_size_);
}

void AgentCommandAck::MergeImpl(::google::protobuf::MessageLite& to_msg,
                      const ::google::protobuf::MessageLite& from_msg) {
   auto* const _this = static_cast<AgentCommandAck*>(&to_msg);
  auto& from = static_cast<const AgentCommandAck&>(from_msg);
  if constexpr (::_pbi::DebugHardenCheckHasBitConsistency()) {
    from.CheckHasBitConsistency();
  }
  ::google::protobuf::Arena* arena = _this->GetArena();
  // @@protoc_insertion_point(class_specific_merge_from_start:agent.v1.AgentCommandAck)
  ABSL_DCHECK_NE(&from, _this);
  ::uint32_t cached_has_bits = 0;
  (void)cached_has_bits;

  cached_has_bits = from._impl_._has_bits_[0];
  if (BatchCheckHasBit(cached_has_bits, 0x00000007U)) {
    if (CheckHasBit(cached_has_bits, 0x00000001U)) {
      if (!from._internal_message_id().empty()) {
        _this->_internal_set_message_id(from._internal_message_id());
      } else {
        if (_this->_impl_.message_id_.IsDefault()) {
          _this->_internal_set_message_id("");
        }
      }
    }
    if (CheckHasBit(cached_has_bits, 0x00000002U)) {
      ABSL_DCHECK(from._impl_.created_at_ != nullptr);
      if (_this->_impl_.created_at_ == nullptr) {
        _this->_impl_.created_at_ = ::google::protobuf::Message::CopyConstruct(arena, *from._impl_.created_at_);
      } else {
        _this->_impl_.created_at_->MergeFrom(*from._impl_.created_at_);
      }
    }
    if (CheckHasBit(cached_has_bits, 0x00000004U)) {
      if (from._internal_progress_percent() != 0) {
        _this->_impl_.progress_percent_ = from._impl_.progress_percent_;
      }
    }
  }
  _this->_impl_._has_bits_[0] |= cached_has_bits;
  _this->_internal_metadata_.MergeFrom<::google::protobuf::UnknownFieldSet>(
      from._internal_metadata_);
}

void AgentCommandAck::CopyFrom(const AgentCommandAck& from) {
  // @@protoc_insertion_point(class_specific_copy_from_start:agent.v1.AgentCommandAck)
  if (&from == this) return;
  Clear();
  MergeFrom(from);
}


void AgentCommandAck::InternalSwap(AgentCommandAck* PROTOBUF_RESTRICT PROTOBUF_NONNULL other) {
  using ::std::swap;
  auto* arena = GetArena();
  ABSL_DCHECK_EQ(arena, other->GetArena());
  _internal_metadata_.InternalSwap(&other->_internal_metadata_);
  swap(_impl_._has_bits_[0], other->_impl_._has_bits_[0]);
  ::_pbi::ArenaStringPtr::InternalSwap(&_impl_.message_id_, &other->_impl_.message_id_, arena);
  ::google::protobuf::internal::memswap<
      PROTOBUF_FIELD_OFFSET(AgentCommandAck, _impl_.progress_percent_)
      + sizeof(AgentCommandAck::_impl_.progress_percent_)
      - PROTOBUF_FIELD_OFFSET(AgentCommandAck, _impl_.created_at_)>(
          reinterpret_cast<char*>(&_impl_.created_at_),
          reinterpret_cast<char*>(&other->_impl_.created_at_));
}

::google::protobuf::Metadata AgentCommandAck::GetMetadata() const {
  return ::google::protobuf::Message::GetMetadataImpl(GetClassData()->full());
}
// ===================================================================

void AgentCommandResponse::clear_timestamp() {
  ::google::protobuf::internal::TSanWrite(&_impl_);
  if (_impl_.timestamp_ != nullptr) _impl_.timestamp_->Clear();
  ClearHasBit(_impl_._has_bits_[0], 0x00000004U);
}
void AgentCommandResponse::clear_received_at() {
  ::google::protobuf::internal::TSanWrite(&_impl_);
  if (_impl_.received_at_ != nullptr) _impl_.received_at_->Clear();
  ClearHasBit(_impl_._has_bits_[0], 0x00000008U);
}
AgentCommandResponse::AgentCommandResponse(::google::protobuf::Arena* PROTOBUF_NULLABLE arena)
#if defined(PROTOBUF_CUSTOM_VTABLE)
    : ::google::protobuf::Message(arena, AgentCommandResponse_get_class_data()) {
#else   // PROTOBUF_CUSTOM_VTABLE
    : ::google::protobuf::Message(arena) {
#endif  // PROTOBUF_CUSTOM_VTABLE
  SharedCtor(arena);
  // @@protoc_insertion_point(arena_constructor:agent.v1.AgentCommandResponse)
}
PROTOBUF_NDEBUG_INLINE AgentCommandResponse::Impl_::Impl_(
    [[maybe_unused]] ::google::protobuf::internal::InternalVisibility visibility,
    [[maybe_unused]] ::google::protobuf::Arena* PROTOBUF_NULLABLE arena, const Impl_& from,
    [[maybe_unused]] const ::agent::v1::AgentCommandResponse& from_msg)
      : _has_bits_{from._has_bits_},
        _cached_size_{0},
        message_id_(arena, from.message_id_),
        response_(arena, from.response_) {}

AgentCommandResponse::AgentCommandResponse(
    ::google::protobuf::Arena* PROTOBUF_NULLABLE arena,
    const AgentCommandResponse& from)
#if defined(PROTOBUF_CUSTOM_VTABLE)
    : ::google::protobuf::Message(arena, AgentCommandResponse_get_class_data()) {

#else   // PROTOBUF_CUSTOM_VTABLE
    : ::google::protobuf::Message(arena) {
#endif  // PROTOBUF_CUSTOM_VTABLE
  AgentCommandResponse* const _this = this;
  (void)_this;
  _internal_metadata_.MergeFrom<::google::protobuf::UnknownFieldSet>(
      from._internal_metadata_);
  new (&_impl_) Impl_(internal_visibility(), arena, from._impl_, from);
  ::uint32_t cached_has_bits = _impl_._has_bits_[0];
  _impl_.timestamp_ = (CheckHasBit(cached_has_bits, 0x00000004U))
                ? ::google::protobuf::Message::CopyConstruct(arena, *from._impl_.timestamp_)
                : nullptr;
  _impl_.received_at_ = (CheckHasBit(cached_has_bits, 0x00000008U))
                ? ::google::protobuf::Message::CopyConstruct(arena, *from._impl_.received_at_)
                : nullptr;
  ::memcpy(reinterpret_cast<char*>(&_impl_) +
               offsetof(Impl_, exec_time_),
           reinterpret_cast<const char*>(&from._impl_) +
               offsetof(Impl_, exec_time_),
           offsetof(Impl_, status_code_) -
               offsetof(Impl_, exec_time_) +
               sizeof(Impl_::status_code_));

  // @@protoc_insertion_point(copy_constructor:agent.v1.AgentCommandResponse)
}
PROTOBUF_NDEBUG_INLINE AgentCommandResponse::Impl_::Impl_(
    [[maybe_unused]] ::google::protobuf::internal::InternalVisibility visibility,
    [[maybe_unused]] ::google::protobuf::Arena* PROTOBUF_NULLABLE arena)
      : _cached_size_{0},
        message_id_(arena),
        response_(arena) {}

inline void AgentCommandResponse::SharedCtor(::_pb::Arena* PROTOBUF_NULLABLE arena) {
  new (&_impl_) Impl_(internal_visibility(), arena);
  ::memset(reinterpret_cast<char*>(&_impl_) +
               offsetof(Impl_, timestamp_),
           0,
           offsetof(Impl_, status_code_) -
               offsetof(Impl_, timestamp_) +
               sizeof(Impl_::status_code_));
}
AgentCommandResponse::~AgentCommandResponse() {
  // @@protoc_insertion_point(destructor:agent.v1.AgentCommandResponse)
  SharedDtor(*this);
}
inline void AgentCommandResponse::SharedDtor(MessageLite& self) {
  AgentCommandResponse& this_ = static_cast<AgentCommandResponse&>(self);
  if constexpr (::_pbi::DebugHardenCheckHasBitConsistency()) {
    this_.CheckHasBitConsistency();
  }
  this_._internal_metadata_.Delete<::google::protobuf::UnknownFieldSet>();
  ABSL_DCHECK(this_.GetArena() == nullptr);
  this_._impl_.message_id_.Destroy();
  this_._impl_.response_.Destroy();
  delete this_._impl_.timestamp_;
  delete this_._impl_.received_at_;
  this_._impl_.~Impl_();
}

#ifndef PROTOBUF_MESSAGE_GLOBALS
PROTOBUF_CONSTINIT PROTOBUF_ATTRIBUTE_INIT_PRIORITY1 const
    ::google::protobuf::internal::ClassDataFull AgentCommandResponse_class_data_ =
        AgentCommandResponse::InternalGenerateClassData_(AgentCommandResponse_globals_._default);

PROTOBUF_ATTRIBUTE_WEAK const ::google::protobuf::internal::ClassData* PROTOBUF_NONNULL
AgentCommandResponse::GetClassData() const {
  ::google::protobuf::internal::PrefetchToLocalCache(&AgentCommandResponse_class_data_);
  ::google::protobuf::internal::PrefetchToLocalCache(AgentCommandResponse_class_data_.tc_table);
  return AgentCommandResponse_class_data_.base();
}
#else
PROTOBUF_ATTRIBUTE_WEAK const ::google::protobuf::internal::ClassData* PROTOBUF_NONNULL
AgentCommandResponse::GetClassData() const {
  ::google::protobuf::internal::PrefetchToLocalCache(&AgentCommandResponse_globals_);
  ::google::protobuf::internal::PrefetchToLocalCache(
      ::google::protobuf::internal::MessageGlobalsBase::ToParseTableBase(&AgentCommandResponse_globals_));
  return AgentCommandResponse_globals_.GetClassData();
}
#endif  // !PROTOBUF_MESSAGE_GLOBALS
#ifndef PROTOBUF_MESSAGE_GLOBALS
PROTOBUF_CONSTINIT
PROTOBUF_ATTRIBUTE_INIT_PRIORITY1 const AgentCommandResponse::ParseTableT_
    AgentCommandResponse::_table_ =
        AgentCommandResponse::InternalGenerateParseTable_(AgentCommandResponse_class_data_.base());
#endif  // !PROTOBUF_MESSAGE_GLOBALS
PROTOBUF_NOINLINE void AgentCommandResponse::Clear() {
// @@protoc_insertion_point(message_clear_start:agent.v1.AgentCommandResponse)
  ::google::protobuf::internal::TSanWrite(&_impl_);
  ::uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void) cached_has_bits;

  cached_has_bits = _impl_._has_bits_[0];
  if (BatchCheckHasBit(cached_has_bits, 0x0000000fU)) {
    if (CheckHasBit(cached_has_bits, 0x00000001U)) {
      _impl_.message_id_.ClearNonDefaultToEmpty();
    }
    if (CheckHasBit(cached_has_bits, 0x00000002U)) {
      _impl_.response_.ClearNonDefaultToEmpty();
    }
    if (CheckHasBit(cached_has_bits, 0x00000004U)) {
      ABSL_DCHECK(_impl_.timestamp_ != nullptr);
      _impl_.timestamp_->Clear();
    }
    if (CheckHasBit(cached_has_bits, 0x00000008U)) {
      ABSL_DCHECK(_impl_.received_at_ != nullptr);
      _impl_.received_at_->Clear();
    }
  }
  if (BatchCheckHasBit(cached_has_bits, 0x00000030U)) {
    ::memset(&_impl_.exec_time_, 0, static_cast<::size_t>(
        reinterpret_cast<char*>(&_impl_.status_code_) -
        reinterpret_cast<char*>(&_impl_.exec_time_)) + sizeof(_impl_.status_code_));
  }
  _impl_._has_bits_.Clear();
  _internal_metadata_.Clear<::google::protobuf::UnknownFieldSet>();
}

#if defined(PROTOBUF_CUSTOM_VTABLE)
::uint8_t* PROTOBUF_NONNULL AgentCommandResponse::_InternalSerialize(
    const ::google::protobuf::MessageLite& base, ::uint8_t* PROTOBUF_NONNULL target,
    ::google::protobuf::io::EpsCopyOutputStream* PROTOBUF_NONNULL stream) {
  const AgentCommandResponse& this_ = static_cast<const AgentCommandResponse&>(base);
#else   // PROTOBUF_CUSTOM_VTABLE
::uint8_t* PROTOBUF_NONNULL AgentCommandResponse::_InternalSerialize(
    ::uint8_t* PROTOBUF_NONNULL target,
    ::google::protobuf::io::EpsCopyOutputStream* PROTOBUF_NONNULL stream) const {
  const AgentCommandResponse& this_ = *this;
#endif  // PROTOBUF_CUSTOM_VTABLE
  if constexpr (::_pbi::DebugHardenCheckHasBitConsistency()) {
    this_.CheckHasBitConsistency();
  }
  // @@protoc_insertion_point(serialize_to_array_start:agent.v1.AgentCommandResponse)
  ::uint32_t cached_has_bits = 0;
  (void)cached_has_bits;

  cached_has_bits = this_._impl_._has_bits_[0];
  // string message_id = 1;
  if (CheckHasBit(cached_has_bits, 0x00000001U)) {
    if (!this_._internal_message_id().empty()) {
      const ::std::string& _s = this_._internal_message_id();
      ::google::protobuf::internal::WireFormatLite::VerifyUtf8String(
          _s.data(), static_cast<int>(_s.length()), ::google::protobuf::internal::WireFormatLite::SERIALIZE, "agent.v1.AgentCommandResponse.message_id");
      target = stream->WriteStringMaybeAliased(1, _s, target);
    }
  }

  // uint32 status_code = 2;
  if (CheckHasBit(cached_has_bits, 0x00000020U)) {
    if (this_._internal_status_code() != 0) {
      target = stream->EnsureSpace(target);
      target = ::_pbi::WireFormatLite::WriteUInt32ToArray(
          2, this_._internal_status_code(), target);
    }
  }

  // bytes response = 3;
  if (CheckHasBit(cached_has_bits, 0x00000002U)) {
    if (!this_._internal_response().empty()) {
      const ::std::string& _s = this_._internal_response();
      target = stream->WriteBytesMaybeAliased(3, _s, target);
    }
  }

  // .google.protobuf.Timestamp timestamp = 4;
  if (CheckHasBit(cached_has_bits, 0x00000004U)) {
    target = ::google::protobuf::internal::WireFormatLite::InternalWriteMessage(
        4, *this_._impl_.timestamp_, this_._impl_.timestamp_->GetCachedSize(), target,
        stream);
  }

  // .google.protobuf.Timestamp received_at = 5;
  if (CheckHasBit(cached_has_bits, 0x00000008U)) {
    target = ::google::protobuf::internal::WireFormatLite::InternalWriteMessage(
        5, *this_._impl_.received_at_, this_._impl_.received_at_->GetCachedSize(), target,
        stream);
  }

  // uint64 exec_time = 6;
  if (CheckHasBit(cached_has_bits, 0x00000010U)) {
    if (this_._internal_exec_time() != 0) {
      target = stream->EnsureSpace(target);
      target = ::_pbi::WireFormatLite::WriteUInt64ToArray(
          6, this_._internal_exec_time(), target);
    }
  }

  if (ABSL_PREDICT_FALSE(this_._internal_metadata_.have_unknown_fields())) {
    target =
        ::_pbi::WireFormat::InternalSerializeUnknownFieldsToArray(
            this_._internal_metadata_.unknown_fields<::google::protobuf::UnknownFieldSet>(::google::protobuf::UnknownFieldSet::default_instance), target, stream);
  }
  // @@protoc_insertion_point(serialize_to_array_end:agent.v1.AgentCommandResponse)
  return target;
}

#if defined(PROTOBUF_CUSTOM_VTABLE)
::size_t AgentCommandResponse::ByteSizeLong(const MessageLite& base) {
  const AgentCommandResponse& this_ = static_cast<const AgentCommandResponse&>(base);
#else   // PROTOBUF_CUSTOM_VTABLE
::size_t AgentCommandResponse::ByteSizeLong() const {
  const AgentCommandResponse& this_ = *this;
#endif  // PROTOBUF_CUSTOM_VTABLE
  // @@protoc_insertion_point(message_byte_size_start:agent.v1.AgentCommandResponse)
  ::size_t total_size = 0;

  ::uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void)cached_has_bits;

  ::_pbi::Prefetch5LinesFrom7Lines(&this_);
  cached_has_bits = this_._impl_._has_bits_[0];
  if (BatchCheckHasBit(cached_has_bits, 0x0000003fU)) {
    // string message_id = 1;
    if (CheckHasBit(cached_has_bits, 0x00000001U)) {
      if (!this_._internal_message_id().empty()) {
        total_size += 1 + ::google::protobuf::internal::WireFormatLite::StringSize(
                                        this_._internal_message_id());
      }
    }
    // bytes response = 3;
    if (CheckHasBit(cached_has_bits, 0x00000002U)) {
      if (!this_._internal_response().empty()) {
        total_size += 1 + ::google::protobuf::internal::WireFormatLite::BytesSize(
                                        this_._internal_response());
      }
    }
    // .google.protobuf.Timestamp timestamp = 4;
    if (CheckHasBit(cached_has_bits, 0x00000004U)) {
      total_size += 1 +
                    ::google::protobuf::internal::WireFormatLite::MessageSize(*this_._impl_.timestamp_);
    }
    // .google.protobuf.Timestamp received_at = 5;
    if (CheckHasBit(cached_has_bits, 0x00000008U)) {
      total_size += 1 +
                    ::google::protobuf::internal::WireFormatLite::MessageSize(*this_._impl_.received_at_);
    }
    // uint64 exec_time = 6;
    if (CheckHasBit(cached_has_bits, 0x00000010U)) {
      if (this_._internal_exec_time() != 0) {
        total_size += ::_pbi::WireFormatLite::UInt64SizePlusOne(
            this_._internal_exec_time());
      }
    }
    // uint32 status_code = 2;
    if (CheckHasBit(cached_has_bits, 0x00000020U)) {
      if (this_._internal_status_code() != 0) {
        total_size += ::_pbi::WireFormatLite::UInt32SizePlusOne(
            this_._internal_status_code());
      }
    }
  }
  return this_.MaybeComputeUnknownFieldsSize(total_size,
                                             &this_._impl_._cached_size_);
}

void AgentCommandResponse::MergeImpl(::google::protobuf::MessageLite& to_msg,
                      const ::google::protobuf::MessageLite& from_msg) {
   auto* const _this = static_cast<AgentCommandResponse*>(&to_msg);
  auto& from = static_cast<const AgentCommandResponse&>(from_msg);
  if constexpr (::_pbi::DebugHardenCheckHasBitConsistency()) {
    from.CheckHasBitConsistency();
  }
  ::google::protobuf::Arena* arena = _this->GetArena();
  // @@protoc_insertion_point(class_specific_merge_from_start:agent.v1.AgentCommandResponse)
  ABSL_DCHECK_NE(&from, _this);
  ::uint32_t cached_has_bits = 0;
  (void)cached_has_bits;

  cached_has_bits = from._impl_._has_bits_[0];
  if (BatchCheckHasBit(cached_has_bits, 0x0000003fU)) {
    if (CheckHasBit(cached_has_bits, 0x00000001U)) {
      if (!from._internal_message_id().empty()) {
        _this->_internal_set_message_id(from._internal_message_id());
      } else {
        if (_this->_impl_.message_id_.IsDefault()) {
          _this->_internal_set_message_id("");
        }
      }
    }
    if (CheckHasBit(cached_has_bits, 0x00000002U)) {
      if (!from._internal_response().empty()) {
        _this->_internal_set_response(from._internal_response());
      } else {
        if (_this->_impl_.response_.IsDefault()) {
          _this->_internal_set_response("");
        }
      }
    }
    if (CheckHasBit(cached_has_bits, 0x00000004U)) {
      ABSL_DCHECK(from._impl_.timestamp_ != nullptr);
      if (_this->_impl_.timestamp_ == nullptr) {
        _this->_impl_.timestamp_ = ::google::protobuf::Message::CopyConstruct(arena, *from._impl_.timestamp_);
      } else {
        _this->_impl_.timestamp_->MergeFrom(*from._impl_.timestamp_);
      }
    }
    if (CheckHasBit(cached_has_bits, 0x00000008U)) {
      ABSL_DCHECK(from._impl_.received_at_ != nullptr);
      if (_this->_impl_.received_at_ == nullptr) {
        _this->_impl_.received_at_ = ::google::protobuf::Message::CopyConstruct(arena, *from._impl_.received_at_);
      } else {
        _this->_impl_.received_at_->MergeFrom(*from._impl_.received_at_);
      }
    }
    if (CheckHasBit(cached_has_bits, 0x00000010U)) {
      if (from._internal_exec_time() != 0) {
        _this->_impl_.exec_time_ = from._impl_.exec_time_;
      }
    }
    if (CheckHasBit(cached_has_bits, 0x00000020U)) {
      if (from._internal_status_code() != 0) {
        _this->_impl_.status_code_ = from._impl_.status_code_;
      }
    }
  }
  _this->_impl_._has_bits_[0] |= cached_has_bits;
  _this->_internal_metadata_.MergeFrom<::google::protobuf::UnknownFieldSet>(
      from._internal_metadata_);
}

void AgentCommandResponse::CopyFrom(const AgentCommandResponse& from) {
  // @@protoc_insertion_point(class_specific_copy_from_start:agent.v1.AgentCommandResponse)
  if (&from == this) return;
  Clear();
  MergeFrom(from);
}


void AgentCommandResponse::InternalSwap(AgentCommandResponse* PROTOBUF_RESTRICT PROTOBUF_NONNULL other) {
  using ::std::swap;
  auto* arena = GetArena();
  ABSL_DCHECK_EQ(arena, other->GetArena());
  _internal_metadata_.InternalSwap(&other->_internal_metadata_);
  swap(_impl_._has_bits_[0], other->_impl_._has_bits_[0]);
  ::_pbi::ArenaStringPtr::InternalSwap(&_impl_.message_id_, &other->_impl_.message_id_, arena);
  ::_pbi::ArenaStringPtr::InternalSwap(&_impl_.response_, &other->_impl_.response_, arena);
  ::google::protobuf::internal::memswap<
      PROTOBUF_FIELD_OFFSET(AgentCommandResponse, _impl_.status_code_)
      + sizeof(AgentCommandResponse::_impl_.status_code_)
      - PROTOBUF_FIELD_OFFSET(AgentCommandResponse, _impl_.timestamp_)>(
          reinterpret_cast<char*>(&_impl_.timestamp_),
          reinterpret_cast<char*>(&other->_impl_.timestamp_));
}

::google::protobuf::Metadata AgentCommandResponse::GetMetadata() const {
  return ::google::protobuf::Message::GetMetadataImpl(GetClassData()->full());
}
// @@protoc_insertion_point(namespace_scope)
}  // namespace v1
}  // namespace agent
namespace google {
namespace protobuf {
}  // namespace protobuf
}  // namespace google
// @@protoc_insertion_point(global_scope)
PROTOBUF_ATTRIBUTE_INIT_PRIORITY2 static ::std::false_type
    _static_init2_ [[maybe_unused]] =
        (::_pbi::AddDescriptors(&descriptor_table_agent_2fv1_2fcmds_2eproto),
         ::std::false_type{});
#include "google/protobuf/port_undef.inc"
