// SPDX-License-Identifier: GPL-3.0-or-later

#ifndef NETDATA_CACHE_HOST_USERS_AND_GROUPS_H
#define NETDATA_CACHE_HOST_USERS_AND_GROUPS_H

void update_cached_host_users(void);
void update_cached_host_groups(void);

#endif //NETDATA_CACHE_HOST_USERS_AND_GROUPS_H
