// Generated by the protocol buffer compiler.  DO NOT EDIT!
// NO CHECKED-IN PROTOBUF GENCODE
// source: nodeinstance/manifest/v1/manifest.proto
// Protobuf C++ Version: 7.35.1

#ifndef nodeinstance_2fmanifest_2fv1_2fmanifest_2eproto_2epb_2eh
#define nodeinstance_2fmanifest_2fv1_2fmanifest_2eproto_2epb_2eh

#include <limits>
#include <string>
#include <type_traits>
#include <utility>

// clang-format off
#include "google/protobuf/runtime_version.h"
#if PROTOBUF_VERSION != 7035001
#error "Protobuf C++ gencode is built with an incompatible version of"
#error "Protobuf C++ headers/runtime. See"
#error "https://protobuf.dev/support/cross-version-runtime-guarantee/#cpp"
#endif
#include "google/protobuf/io/coded_stream.h"
#include "google/protobuf/arena.h"
#include "google/protobuf/arenastring.h"
#include "google/protobuf/generated_message_tctable_decl.h"
#include "google/protobuf/generated_message_util.h"
#include "google/protobuf/metadata_lite.h"
#include "google/protobuf/generated_message_reflection.h"
#include "google/protobuf/message.h"
#include "google/protobuf/message_lite.h"
#include "google/protobuf/repeated_field.h"  // IWYU pragma: export
#include "google/protobuf/extension_set.h"  // IWYU pragma: export
#include "google/protobuf/generated_enum_reflection.h"
#include "google/protobuf/unknown_field_set.h"
#include "google/protobuf/timestamp.pb.h"
// @@protoc_insertion_point(includes)

// Must be included last.
#include "google/protobuf/port_def.inc"

#define PROTOBUF_INTERNAL_EXPORT_nodeinstance_2fmanifest_2fv1_2fmanifest_2eproto

namespace google {
namespace protobuf {
namespace internal {
template <typename T>
::absl::string_view GetAnyMessageName();
}  // namespace internal
}  // namespace protobuf
}  // namespace google

// Internal implementation detail -- do not use these members.
struct TableStruct_nodeinstance_2fmanifest_2fv1_2fmanifest_2eproto {
  static const ::uint32_t offsets[];
};
extern "C" {
extern const ::google::protobuf::internal::DescriptorTable descriptor_table_nodeinstance_2fmanifest_2fv1_2fmanifest_2eproto;
}  // extern "C"
namespace nodeinstance {
namespace manifest {
namespace v1 {
enum HttpAccess : int;
extern const uint32_t HttpAccess_internal_data_[];
class Function;
struct FunctionGlobalsTypeInternal;
#ifndef PROTOBUF_MESSAGE_GLOBALS
extern FunctionGlobalsTypeInternal Function_globals_;
extern const ::google::protobuf::internal::ClassDataFull Function_class_data_;
#else
extern const FunctionGlobalsTypeInternal Function_globals_;
#endif  // PROTOBUF_MESSAGE_GLOBALS
class Functions;
struct FunctionsGlobalsTypeInternal;
#ifndef PROTOBUF_MESSAGE_GLOBALS
extern FunctionsGlobalsTypeInternal Functions_globals_;
extern const ::google::protobuf::internal::ClassDataFull Functions_class_data_;
#else
extern const FunctionsGlobalsTypeInternal Functions_globals_;
#endif  // PROTOBUF_MESSAGE_GLOBALS
class UpdateNodeInstanceManifest;
struct UpdateNodeInstanceManifestGlobalsTypeInternal;
#ifndef PROTOBUF_MESSAGE_GLOBALS
extern UpdateNodeInstanceManifestGlobalsTypeInternal UpdateNodeInstanceManifest_globals_;
extern const ::google::protobuf::internal::ClassDataFull UpdateNodeInstanceManifest_class_data_;
#else
extern const UpdateNodeInstanceManifestGlobalsTypeInternal UpdateNodeInstanceManifest_globals_;
#endif  // PROTOBUF_MESSAGE_GLOBALS
}  // namespace v1
}  // namespace manifest
}  // namespace nodeinstance
namespace google {
namespace protobuf {
template <>
internal::EnumTraitsT<::nodeinstance::manifest::v1::HttpAccess_internal_data_>
    internal::EnumTraitsImpl::value<::nodeinstance::manifest::v1::HttpAccess>;
}  // namespace protobuf
}  // namespace google

namespace nodeinstance {
namespace manifest {
namespace v1 {
enum HttpAccess : int {
  HTTP_ACCESS_UNSPECIFIED = 0,
  HTTP_ACCESS_SIGNED_IN = 1,
  HTTP_ACCESS_SAME_SPACE = 2,
  HTTP_ACCESS_COMMERCIAL = 3,
  HTTP_ACCESS_ANONYMOUS_DATA = 4,
  HTTP_ACCESS_SENSITIVE_DATA = 5,
  HTTP_ACCESS_VIEW_CONFIG = 6,
  HTTP_ACCESS_EDIT_CONFIG = 7,
  HTTP_ACCESS_VIEW_NOTIFICATIONS_CONFIG = 8,
  HTTP_ACCESS_EDIT_NOTIFICATIONS_CONFIG = 9,
  HTTP_ACCESS_VIEW_ALERTS_SILENCING = 10,
  HTTP_ACCESS_EDIT_ALERTS_SILENCING = 11,
  HttpAccess_INT_MIN_SENTINEL_DO_NOT_USE_ =
      ::std::numeric_limits<::int32_t>::min(),
  HttpAccess_INT_MAX_SENTINEL_DO_NOT_USE_ =
      ::std::numeric_limits<::int32_t>::max(),
};

extern const uint32_t HttpAccess_internal_data_[];
inline constexpr HttpAccess HttpAccess_MIN =
    static_cast<HttpAccess>(0);
inline constexpr HttpAccess HttpAccess_MAX =
    static_cast<HttpAccess>(11);
[[nodiscard]] inline bool HttpAccess_IsValid(int value) {
  return 0 <= value && value <= 11;
}
inline constexpr int HttpAccess_ARRAYSIZE = 11 + 1;
[[nodiscard]] const ::google::protobuf::EnumDescriptor* PROTOBUF_NONNULL
HttpAccess_descriptor();
[[nodiscard]] inline auto ProtobufInternalGetEnumDescriptor(HttpAccess) {
  return HttpAccess_descriptor();
}
template <typename T>
[[nodiscard]] const ::std::string& HttpAccess_Name(T value) {
  static_assert(::std::is_same<T, HttpAccess>::value ||
                    ::std::is_integral<T>::value,
                "Incorrect type passed to HttpAccess_Name().");
  return HttpAccess_Name(static_cast<HttpAccess>(value));
}
template <>
[[nodiscard]] inline const ::std::string& HttpAccess_Name(HttpAccess value) {
  return ::google::protobuf::internal::NameOfDenseEnum<HttpAccess_descriptor, 0, 11>(
      static_cast<int>(value));
}
[[nodiscard]] inline bool HttpAccess_Parse(
    ::absl::string_view name, HttpAccess* PROTOBUF_NONNULL value) {
  return ::google::protobuf::internal::ParseNamedEnum<HttpAccess>(HttpAccess_descriptor(), name,
                                           value);
}
using ::google::protobuf::internal::generated_enum::AbslParseFlag;
using ::google::protobuf::internal::generated_enum::AbslUnparseFlag;

// ===================================================================


// -------------------------------------------------------------------

class  PROTOBUF_FUTURE_ADD_EARLY_WARN_UNUSED Function final : public ::google::protobuf::Message
/* @@protoc_insertion_point(class_definition:nodeinstance.manifest.v1.Function) */ {
 public:
  inline Function() : Function(nullptr) {}
  ~Function() PROTOBUF_FINAL;

#if defined(PROTOBUF_CUSTOM_VTABLE)
  void operator delete(Function* PROTOBUF_NONNULL msg, ::std::destroying_delete_t) {
    SharedDtor(*msg);
    ::google::protobuf::internal::SizedDelete(msg, sizeof(Function));
  }
#endif

  template <typename = void>
  explicit constexpr Function(::google::protobuf::internal::ConstantInitialized,
                           const ::google::protobuf::internal::ClassData* PROTOBUF_NONNULL
                               class_data);

  inline Function(const Function& from) : Function(nullptr, from) {}
  inline Function(Function&& from) noexcept : Function(nullptr, ::std::move(from)) {}
  inline Function& operator=(const Function& from) {
    CopyFrom(from);
    return *this;
  }
  inline Function& operator=(Function&& from) noexcept {
    if (this == &from) return *this;
    if (::google::protobuf::internal::CanMoveWithInternalSwap(GetArena(), from.GetArena())) {
      InternalSwap(&from);
    } else {
      CopyFrom(from);
    }
    return *this;
  }

  [[nodiscard]] inline const ::google::protobuf::UnknownFieldSet& unknown_fields() const
      ABSL_ATTRIBUTE_LIFETIME_BOUND {
    return _internal_metadata_.unknown_fields<::google::protobuf::UnknownFieldSet>(::google::protobuf::UnknownFieldSet::default_instance);
  }
  [[nodiscard]] inline ::google::protobuf::UnknownFieldSet* PROTOBUF_NONNULL
  mutable_unknown_fields() ABSL_ATTRIBUTE_LIFETIME_BOUND {
    return _internal_metadata_.mutable_unknown_fields<::google::protobuf::UnknownFieldSet>();
  }

  [[nodiscard]] static const ::google::protobuf::Descriptor* PROTOBUF_NONNULL descriptor() {
    return GetDescriptor();
  }
  [[nodiscard]] static const ::google::protobuf::Descriptor* PROTOBUF_NONNULL
  GetDescriptor() {
    return default_instance().GetMetadata().descriptor;
  }
  [[nodiscard]] static const ::google::protobuf::Reflection* PROTOBUF_NONNULL GetReflection() {
    return default_instance().GetMetadata().reflection;
  }
  [[nodiscard]] static const Function& default_instance() {
    return *::google::protobuf::internal::MessageGlobalsBase::ToDefaultInstance<Function>(&Function_globals_);
  }
  static constexpr int kIndexInFileMessages = 2;
  friend void swap(Function& a, Function& b) { a.Swap(&b); }
  inline void Swap(Function* PROTOBUF_NONNULL other) {
    if (other == this) return;
    if (::google::protobuf::internal::CanUseInternalSwap(GetArena(), other->GetArena())) {
      InternalSwap(other);
    } else {
      ::google::protobuf::internal::GenericSwap(this, other);
    }
  }
  void UnsafeArenaSwap(Function* PROTOBUF_NONNULL other) {
    if (other == this) return;
    ABSL_DCHECK(GetArena() == other->GetArena());
    InternalSwap(other);
  }

  // implements Message ----------------------------------------------

  [[nodiscard]] Function* PROTOBUF_NONNULL
  New(::google::protobuf::Arena* PROTOBUF_NULLABLE arena = nullptr) const {
    return ::google::protobuf::Message::DefaultConstruct<Function>(arena);
  }
  using ::google::protobuf::Message::CopyFrom;
  void CopyFrom(const Function& from);
  using ::google::protobuf::Message::MergeFrom;
  void MergeFrom(const Function& from) { Function::MergeImpl(*this, from); }

  private:
  static void MergeImpl(::google::protobuf::MessageLite& to_msg,
                        const ::google::protobuf::MessageLite& from_msg);

  public:
  [[nodiscard]] bool IsInitialized() const {
    return true;
  }
  ABSL_ATTRIBUTE_REINITIALIZES void Clear() PROTOBUF_FINAL;
  #if defined(PROTOBUF_CUSTOM_VTABLE)
  private:
  [[nodiscard]] static ::size_t ByteSizeLong(const ::google::protobuf::MessageLite& msg);
  [[nodiscard]] static ::uint8_t* PROTOBUF_NONNULL _InternalSerialize(
      const ::google::protobuf::MessageLite& msg, ::uint8_t* PROTOBUF_NONNULL target,
      ::google::protobuf::io::EpsCopyOutputStream* PROTOBUF_NONNULL stream);

  public:
  [[nodiscard]] ::size_t ByteSizeLong() const { return ByteSizeLong(*this); }
  [[nodiscard]] ::uint8_t* PROTOBUF_NONNULL _InternalSerialize(
      ::uint8_t* PROTOBUF_NONNULL target,
      ::google::protobuf::io::EpsCopyOutputStream* PROTOBUF_NONNULL stream) const {
    return _InternalSerialize(*this, target, stream);
  }
  #else   // PROTOBUF_CUSTOM_VTABLE
  [[nodiscard]] ::size_t ByteSizeLong() const final;
  [[nodiscard]] ::uint8_t* PROTOBUF_NONNULL _InternalSerialize(
      ::uint8_t* PROTOBUF_NONNULL target,
      ::google::protobuf::io::EpsCopyOutputStream* PROTOBUF_NONNULL stream) const final;
  #endif  // PROTOBUF_CUSTOM_VTABLE
  [[nodiscard]] int GetCachedSize() const {
    return _impl_._cached_size_.Get();
  }

  private:
  void SharedCtor(::google::protobuf::Arena* PROTOBUF_NULLABLE arena);
  static void SharedDtor(MessageLite& self);
  void InternalSwap(Function* PROTOBUF_NONNULL other);
 private:
  template <typename T>
  friend ::absl::string_view(::google::protobuf::internal::GetAnyMessageName)();
  static ::absl::string_view FullMessageName() { return "nodeinstance.manifest.v1.Function"; }

  explicit Function(::google::protobuf::Arena* PROTOBUF_NULLABLE arena);
  Function(::google::protobuf::Arena* PROTOBUF_NULLABLE arena, const Function& from);
  Function(
      ::google::protobuf::Arena* PROTOBUF_NULLABLE arena, Function&& from) noexcept
      : Function(arena) {
    *this = ::std::move(from);
  }
  const ::google::protobuf::internal::ClassData* PROTOBUF_NONNULL GetClassData() const PROTOBUF_FINAL;
  static void* PROTOBUF_NONNULL PlacementNew_(
      const void* PROTOBUF_NONNULL, void* PROTOBUF_NONNULL mem,
      ::google::protobuf::Arena* PROTOBUF_NULLABLE arena);
  static constexpr auto InternalNewImpl_();

 public:
  static constexpr auto InternalGenerateClassData_(
      const MessageLite& prototype,
      const ::google::protobuf::internal::TcParseTableBase* PROTOBUF_NULLABLE tc_table = nullptr);

  [[nodiscard]] ::google::protobuf::Metadata GetMetadata() const;
  // nested types ----------------------------------------------------

  // accessors -------------------------------------------------------
  enum : int {
    kTagsFieldNumber = 3,
    kAccessFieldNumber = 6,
    kNameFieldNumber = 1,
    kHelpFieldNumber = 2,
    kPriorityFieldNumber = 4,
    kVersionFieldNumber = 5,
  };
  // repeated string tags = 3;
  [[nodiscard]] int tags_size()
      const;
  private:
  int _internal_tags_size() const;

  public:
  void clear_tags() ;
  [[nodiscard]] const ::std::string& tags(int index) const;
  ::std::string* PROTOBUF_NONNULL mutable_tags(int index);
  template <typename Arg_ = const ::std::string&, typename... Args_>
  void set_tags(int index, Arg_&& value, Args_... args);
  ::std::string* PROTOBUF_NONNULL add_tags();
  template <typename Arg_ = const ::std::string&, typename... Args_>
  void add_tags(Arg_&& value, Args_... args);
  [[nodiscard]] const
      ::google::protobuf::RepeatedPtrField<::std::string>&
      tags() const;
  [[nodiscard]] ::google::protobuf::RepeatedPtrField<::std::string>*
      PROTOBUF_NONNULL
      mutable_tags();

  private:
  const ::google::protobuf::RepeatedPtrField<::std::string>& _internal_tags() const;
  ::google::protobuf::RepeatedPtrField<::std::string>* PROTOBUF_NONNULL _internal_mutable_tags();

  public:
  // repeated .nodeinstance.manifest.v1.HttpAccess access = 6;
  [[nodiscard]] int access_size()
      const;
  private:
  int _internal_access_size() const;

  public:
  void clear_access() ;
  public:
  [[nodiscard]] ::nodeinstance::manifest::v1::HttpAccess access(int index) const;
  void set_access(int index, ::nodeinstance::manifest::v1::HttpAccess value);
  void add_access(::nodeinstance::manifest::v1::HttpAccess value);
  [[nodiscard]] const ::google::protobuf::RepeatedField<int>& access()
      const;
  [[nodiscard]] ::google::protobuf::RepeatedField<int>* PROTOBUF_NONNULL mutable_access();

  private:
  const ::google::protobuf::RepeatedField<int>& _internal_access() const;
  ::google::protobuf::RepeatedField<int>* PROTOBUF_NONNULL _internal_mutable_access();

  public:
  // string name = 1;
  void clear_name() ;
  [[nodiscard]] const ::std::string& name() const;
  template <typename Arg_ = const ::std::string&, typename... Args_>
  void set_name(Arg_&& arg, Args_... args);
  ::std::string* PROTOBUF_NONNULL mutable_name();
  [[nodiscard]] ::std::string* PROTOBUF_NULLABLE release_name();
  void set_allocated_name(::std::string* PROTOBUF_NULLABLE value);

  private:
  const ::std::string& _internal_name() const;
  PROTOBUF_ALWAYS_INLINE void _internal_set_name(const ::std::string& value);
  ::std::string* PROTOBUF_NONNULL _internal_mutable_name();

  public:
  // string help = 2;
  void clear_help() ;
  [[nodiscard]] const ::std::string& help() const;
  template <typename Arg_ = const ::std::string&, typename... Args_>
  void set_help(Arg_&& arg, Args_... args);
  ::std::string* PROTOBUF_NONNULL mutable_help();
  [[nodiscard]] ::std::string* PROTOBUF_NULLABLE release_help();
  void set_allocated_help(::std::string* PROTOBUF_NULLABLE value);

  private:
  const ::std::string& _internal_help() const;
  PROTOBUF_ALWAYS_INLINE void _internal_set_help(const ::std::string& value);
  ::std::string* PROTOBUF_NONNULL _internal_mutable_help();

  public:
  // uint32 priority = 4;
  void clear_priority() ;
  [[nodiscard]] ::uint32_t priority() const;
  void set_priority(::uint32_t value);

  private:
  ::uint32_t _internal_priority() const;
  void _internal_set_priority(::uint32_t value);

  public:
  // uint32 version = 5;
  void clear_version() ;
  [[nodiscard]] ::uint32_t version() const;
  void set_version(::uint32_t value);

  private:
  ::uint32_t _internal_version() const;
  void _internal_set_version(::uint32_t value);

  public:
  // @@protoc_insertion_point(class_scope:nodeinstance.manifest.v1.Function)
 private:
  class _Internal;
  using ParseTableT_ =
      ::google::protobuf::internal::TcParseTable<3, 6,
                          0, 54,
                          2>;
  static constexpr ParseTableT_ InternalGenerateParseTable_(
      const ::google::protobuf::internal::ClassData* PROTOBUF_NONNULL class_data);
  friend class ::google::protobuf::internal::TcParser;
  #ifndef PROTOBUF_MESSAGE_GLOBALS
  static const ParseTableT_ _table_;
  #endif

  friend class ::google::protobuf::MessageLite;
  friend class ::google::protobuf::Arena;
  friend ::google::protobuf::internal::PrivateAccess;
  template <typename T>
  friend class ::google::protobuf::Arena::InternalHelper;
  using InternalArenaConstructable_ = void;
  using DestructorSkippable_ = void;
  struct Impl_ {
    inline explicit constexpr Impl_(::google::protobuf::internal::InternalVisibility visibility,
                                    ::google::protobuf::internal::ConstantInitialized) noexcept;
    inline explicit Impl_(
        ::google::protobuf::internal::InternalVisibility visibility,
        ::google::protobuf::Arena* PROTOBUF_NULLABLE arena);
    inline explicit Impl_(
        ::google::protobuf::internal::InternalVisibility visibility,
        ::google::protobuf::Arena* PROTOBUF_NULLABLE arena, const Impl_& from,
        const Function& from_msg);
    ::google::protobuf::internal::HasBits<1> _has_bits_;
    ::google::protobuf::internal::CachedSize _cached_size_;
    ::google::protobuf::RepeatedPtrField<::std::string> tags_;
    ::google::protobuf::RepeatedField<int> access_;
    ::google::protobuf::internal::CachedSize _access_cached_byte_size_;
    ::google::protobuf::internal::ArenaStringPtr name_;
    ::google::protobuf::internal::ArenaStringPtr help_;
    ::uint32_t priority_;
    ::uint32_t version_;
    PROTOBUF_TSAN_DECLARE_MEMBER
  };
  union { Impl_ _impl_; };
  friend struct ::TableStruct_nodeinstance_2fmanifest_2fv1_2fmanifest_2eproto;
};
// -------------------------------------------------------------------

class  PROTOBUF_FUTURE_ADD_EARLY_WARN_UNUSED Functions final : public ::google::protobuf::Message
/* @@protoc_insertion_point(class_definition:nodeinstance.manifest.v1.Functions) */ {
 public:
  inline Functions() : Functions(nullptr) {}
  ~Functions() PROTOBUF_FINAL;

#if defined(PROTOBUF_CUSTOM_VTABLE)
  void operator delete(Functions* PROTOBUF_NONNULL msg, ::std::destroying_delete_t) {
    SharedDtor(*msg);
    ::google::protobuf::internal::SizedDelete(msg, sizeof(Functions));
  }
#endif

  template <typename = void>
  explicit constexpr Functions(::google::protobuf::internal::ConstantInitialized,
                           const ::google::protobuf::internal::ClassData* PROTOBUF_NONNULL
                               class_data);

  inline Functions(const Functions& from) : Functions(nullptr, from) {}
  inline Functions(Functions&& from) noexcept : Functions(nullptr, ::std::move(from)) {}
  inline Functions& operator=(const Functions& from) {
    CopyFrom(from);
    return *this;
  }
  inline Functions& operator=(Functions&& from) noexcept {
    if (this == &from) return *this;
    if (::google::protobuf::internal::CanMoveWithInternalSwap(GetArena(), from.GetArena())) {
      InternalSwap(&from);
    } else {
      CopyFrom(from);
    }
    return *this;
  }

  [[nodiscard]] inline const ::google::protobuf::UnknownFieldSet& unknown_fields() const
      ABSL_ATTRIBUTE_LIFETIME_BOUND {
    return _internal_metadata_.unknown_fields<::google::protobuf::UnknownFieldSet>(::google::protobuf::UnknownFieldSet::default_instance);
  }
  [[nodiscard]] inline ::google::protobuf::UnknownFieldSet* PROTOBUF_NONNULL
  mutable_unknown_fields() ABSL_ATTRIBUTE_LIFETIME_BOUND {
    return _internal_metadata_.mutable_unknown_fields<::google::protobuf::UnknownFieldSet>();
  }

  [[nodiscard]] static const ::google::protobuf::Descriptor* PROTOBUF_NONNULL descriptor() {
    return GetDescriptor();
  }
  [[nodiscard]] static const ::google::protobuf::Descriptor* PROTOBUF_NONNULL
  GetDescriptor() {
    return default_instance().GetMetadata().descriptor;
  }
  [[nodiscard]] static const ::google::protobuf::Reflection* PROTOBUF_NONNULL GetReflection() {
    return default_instance().GetMetadata().reflection;
  }
  [[nodiscard]] static const Functions& default_instance() {
    return *::google::protobuf::internal::MessageGlobalsBase::ToDefaultInstance<Functions>(&Functions_globals_);
  }
  static constexpr int kIndexInFileMessages = 1;
  friend void swap(Functions& a, Functions& b) { a.Swap(&b); }
  inline void Swap(Functions* PROTOBUF_NONNULL other) {
    if (other == this) return;
    if (::google::protobuf::internal::CanUseInternalSwap(GetArena(), other->GetArena())) {
      InternalSwap(other);
    } else {
      ::google::protobuf::internal::GenericSwap(this, other);
    }
  }
  void UnsafeArenaSwap(Functions* PROTOBUF_NONNULL other) {
    if (other == this) return;
    ABSL_DCHECK(GetArena() == other->GetArena());
    InternalSwap(other);
  }

  // implements Message ----------------------------------------------

  [[nodiscard]] Functions* PROTOBUF_NONNULL
  New(::google::protobuf::Arena* PROTOBUF_NULLABLE arena = nullptr) const {
    return ::google::protobuf::Message::DefaultConstruct<Functions>(arena);
  }
  using ::google::protobuf::Message::CopyFrom;
  void CopyFrom(const Functions& from);
  using ::google::protobuf::Message::MergeFrom;
  void MergeFrom(const Functions& from) { Functions::MergeImpl(*this, from); }

  private:
  static void MergeImpl(::google::protobuf::MessageLite& to_msg,
                        const ::google::protobuf::MessageLite& from_msg);

  public:
  [[nodiscard]] bool IsInitialized() const {
    return true;
  }
  ABSL_ATTRIBUTE_REINITIALIZES void Clear() PROTOBUF_FINAL;
  #if defined(PROTOBUF_CUSTOM_VTABLE)
  private:
  [[nodiscard]] static ::size_t ByteSizeLong(const ::google::protobuf::MessageLite& msg);
  [[nodiscard]] static ::uint8_t* PROTOBUF_NONNULL _InternalSerialize(
      const ::google::protobuf::MessageLite& msg, ::uint8_t* PROTOBUF_NONNULL target,
      ::google::protobuf::io::EpsCopyOutputStream* PROTOBUF_NONNULL stream);

  public:
  [[nodiscard]] ::size_t ByteSizeLong() const { return ByteSizeLong(*this); }
  [[nodiscard]] ::uint8_t* PROTOBUF_NONNULL _InternalSerialize(
      ::uint8_t* PROTOBUF_NONNULL target,
      ::google::protobuf::io::EpsCopyOutputStream* PROTOBUF_NONNULL stream) const {
    return _InternalSerialize(*this, target, stream);
  }
  #else   // PROTOBUF_CUSTOM_VTABLE
  [[nodiscard]] ::size_t ByteSizeLong() const final;
  [[nodiscard]] ::uint8_t* PROTOBUF_NONNULL _InternalSerialize(
      ::uint8_t* PROTOBUF_NONNULL target,
      ::google::protobuf::io::EpsCopyOutputStream* PROTOBUF_NONNULL stream) const final;
  #endif  // PROTOBUF_CUSTOM_VTABLE
  [[nodiscard]] int GetCachedSize() const {
    return _impl_._cached_size_.Get();
  }

  private:
  void SharedCtor(::google::protobuf::Arena* PROTOBUF_NULLABLE arena);
  static void SharedDtor(MessageLite& self);
  void InternalSwap(Functions* PROTOBUF_NONNULL other);
 private:
  template <typename T>
  friend ::absl::string_view(::google::protobuf::internal::GetAnyMessageName)();
  static ::absl::string_view FullMessageName() { return "nodeinstance.manifest.v1.Functions"; }

  explicit Functions(::google::protobuf::Arena* PROTOBUF_NULLABLE arena);
  Functions(::google::protobuf::Arena* PROTOBUF_NULLABLE arena, const Functions& from);
  Functions(
      ::google::protobuf::Arena* PROTOBUF_NULLABLE arena, Functions&& from) noexcept
      : Functions(arena) {
    *this = ::std::move(from);
  }
  const ::google::protobuf::internal::ClassData* PROTOBUF_NONNULL GetClassData() const PROTOBUF_FINAL;
  static void* PROTOBUF_NONNULL PlacementNew_(
      const void* PROTOBUF_NONNULL, void* PROTOBUF_NONNULL mem,
      ::google::protobuf::Arena* PROTOBUF_NULLABLE arena);
  static constexpr auto InternalNewImpl_();

 public:
  static constexpr auto InternalGenerateClassData_(
      const MessageLite& prototype,
      const ::google::protobuf::internal::TcParseTableBase* PROTOBUF_NULLABLE tc_table = nullptr);

  [[nodiscard]] ::google::protobuf::Metadata GetMetadata() const;
  // nested types ----------------------------------------------------

  // accessors -------------------------------------------------------
  enum : int {
    kItemsFieldNumber = 1,
  };
  // repeated .nodeinstance.manifest.v1.Function items = 1;
  [[nodiscard]] int items_size()
      const;
  private:
  int _internal_items_size() const;

  public:
  void clear_items() ;
  [[nodiscard]] const ::nodeinstance::manifest::v1::Function& items(int index) const;
  [[nodiscard]] ::nodeinstance::manifest::v1::Function* PROTOBUF_NONNULL mutable_items(int index);
  ::nodeinstance::manifest::v1::Function* PROTOBUF_NONNULL add_items();
  [[nodiscard]] const ::google::protobuf::RepeatedPtrField<::nodeinstance::manifest::v1::Function>&
  items() const;
  [[nodiscard]] ::google::protobuf::RepeatedPtrField<::nodeinstance::manifest::v1::Function>* PROTOBUF_NONNULL
  mutable_items();

  private:
  const ::google::protobuf::RepeatedPtrField<::nodeinstance::manifest::v1::Function>& _internal_items() const;
  ::google::protobuf::RepeatedPtrField<::nodeinstance::manifest::v1::Function>* PROTOBUF_NONNULL _internal_mutable_items();

  public:
  // @@protoc_insertion_point(class_scope:nodeinstance.manifest.v1.Functions)
 private:
  class _Internal;
  using ParseTableT_ =
      ::google::protobuf::internal::TcParseTable<0, 1,
                          1, 0,
                          2>;
  static constexpr ParseTableT_ InternalGenerateParseTable_(
      const ::google::protobuf::internal::ClassData* PROTOBUF_NONNULL class_data);
  friend class ::google::protobuf::internal::TcParser;
  #ifndef PROTOBUF_MESSAGE_GLOBALS
  static const ParseTableT_ _table_;
  #endif

  friend class ::google::protobuf::MessageLite;
  friend class ::google::protobuf::Arena;
  friend ::google::protobuf::internal::PrivateAccess;
  template <typename T>
  friend class ::google::protobuf::Arena::InternalHelper;
  using InternalArenaConstructable_ = void;
  using DestructorSkippable_ = void;
  struct Impl_ {
    inline explicit constexpr Impl_(::google::protobuf::internal::InternalVisibility visibility,
                                    ::google::protobuf::internal::ConstantInitialized) noexcept;
    inline explicit Impl_(
        ::google::protobuf::internal::InternalVisibility visibility,
        ::google::protobuf::Arena* PROTOBUF_NULLABLE arena);
    inline explicit Impl_(
        ::google::protobuf::internal::InternalVisibility visibility,
        ::google::protobuf::Arena* PROTOBUF_NULLABLE arena, const Impl_& from,
        const Functions& from_msg);
    ::google::protobuf::internal::HasBits<1> _has_bits_;
    ::google::protobuf::internal::CachedSize _cached_size_;
    ::google::protobuf::RepeatedPtrField< ::nodeinstance::manifest::v1::Function > items_;
    PROTOBUF_TSAN_DECLARE_MEMBER
  };
  union { Impl_ _impl_; };
  friend struct ::TableStruct_nodeinstance_2fmanifest_2fv1_2fmanifest_2eproto;
};
// -------------------------------------------------------------------

class  PROTOBUF_FUTURE_ADD_EARLY_WARN_UNUSED UpdateNodeInstanceManifest final : public ::google::protobuf::Message
/* @@protoc_insertion_point(class_definition:nodeinstance.manifest.v1.UpdateNodeInstanceManifest) */ {
 public:
  inline UpdateNodeInstanceManifest() : UpdateNodeInstanceManifest(nullptr) {}
  ~UpdateNodeInstanceManifest() PROTOBUF_FINAL;

#if defined(PROTOBUF_CUSTOM_VTABLE)
  void operator delete(UpdateNodeInstanceManifest* PROTOBUF_NONNULL msg, ::std::destroying_delete_t) {
    SharedDtor(*msg);
    ::google::protobuf::internal::SizedDelete(msg, sizeof(UpdateNodeInstanceManifest));
  }
#endif

  template <typename = void>
  explicit constexpr UpdateNodeInstanceManifest(::google::protobuf::internal::ConstantInitialized,
                           const ::google::protobuf::internal::ClassData* PROTOBUF_NONNULL
                               class_data);

  inline UpdateNodeInstanceManifest(const UpdateNodeInstanceManifest& from) : UpdateNodeInstanceManifest(nullptr, from) {}
  inline UpdateNodeInstanceManifest(UpdateNodeInstanceManifest&& from) noexcept : UpdateNodeInstanceManifest(nullptr, ::std::move(from)) {}
  inline UpdateNodeInstanceManifest& operator=(const UpdateNodeInstanceManifest& from) {
    CopyFrom(from);
    return *this;
  }
  inline UpdateNodeInstanceManifest& operator=(UpdateNodeInstanceManifest&& from) noexcept {
    if (this == &from) return *this;
    if (::google::protobuf::internal::CanMoveWithInternalSwap(GetArena(), from.GetArena())) {
      InternalSwap(&from);
    } else {
      CopyFrom(from);
    }
    return *this;
  }

  [[nodiscard]] inline const ::google::protobuf::UnknownFieldSet& unknown_fields() const
      ABSL_ATTRIBUTE_LIFETIME_BOUND {
    return _internal_metadata_.unknown_fields<::google::protobuf::UnknownFieldSet>(::google::protobuf::UnknownFieldSet::default_instance);
  }
  [[nodiscard]] inline ::google::protobuf::UnknownFieldSet* PROTOBUF_NONNULL
  mutable_unknown_fields() ABSL_ATTRIBUTE_LIFETIME_BOUND {
    return _internal_metadata_.mutable_unknown_fields<::google::protobuf::UnknownFieldSet>();
  }

  [[nodiscard]] static const ::google::protobuf::Descriptor* PROTOBUF_NONNULL descriptor() {
    return GetDescriptor();
  }
  [[nodiscard]] static const ::google::protobuf::Descriptor* PROTOBUF_NONNULL
  GetDescriptor() {
    return default_instance().GetMetadata().descriptor;
  }
  [[nodiscard]] static const ::google::protobuf::Reflection* PROTOBUF_NONNULL GetReflection() {
    return default_instance().GetMetadata().reflection;
  }
  [[nodiscard]] static const UpdateNodeInstanceManifest& default_instance() {
    return *::google::protobuf::internal::MessageGlobalsBase::ToDefaultInstance<UpdateNodeInstanceManifest>(&UpdateNodeInstanceManifest_globals_);
  }
  enum PayloadCase {
    kFunctions = 10,
    PAYLOAD_NOT_SET = 0,
  };
  static constexpr int kIndexInFileMessages = 0;
  friend void swap(UpdateNodeInstanceManifest& a, UpdateNodeInstanceManifest& b) { a.Swap(&b); }
  inline void Swap(UpdateNodeInstanceManifest* PROTOBUF_NONNULL other) {
    if (other == this) return;
    if (::google::protobuf::internal::CanUseInternalSwap(GetArena(), other->GetArena())) {
      InternalSwap(other);
    } else {
      ::google::protobuf::internal::GenericSwap(this, other);
    }
  }
  void UnsafeArenaSwap(UpdateNodeInstanceManifest* PROTOBUF_NONNULL other) {
    if (other == this) return;
    ABSL_DCHECK(GetArena() == other->GetArena());
    InternalSwap(other);
  }

  // implements Message ----------------------------------------------

  [[nodiscard]] UpdateNodeInstanceManifest* PROTOBUF_NONNULL
  New(::google::protobuf::Arena* PROTOBUF_NULLABLE arena = nullptr) const {
    return ::google::protobuf::Message::DefaultConstruct<UpdateNodeInstanceManifest>(arena);
  }
  using ::google::protobuf::Message::CopyFrom;
  void CopyFrom(const UpdateNodeInstanceManifest& from);
  using ::google::protobuf::Message::MergeFrom;
  void MergeFrom(const UpdateNodeInstanceManifest& from) { UpdateNodeInstanceManifest::MergeImpl(*this, from); }

  private:
  static void MergeImpl(::google::protobuf::MessageLite& to_msg,
                        const ::google::protobuf::MessageLite& from_msg);

  public:
  [[nodiscard]] bool IsInitialized() const {
    return true;
  }
  ABSL_ATTRIBUTE_REINITIALIZES void Clear() PROTOBUF_FINAL;
  #if defined(PROTOBUF_CUSTOM_VTABLE)
  private:
  [[nodiscard]] static ::size_t ByteSizeLong(const ::google::protobuf::MessageLite& msg);
  [[nodiscard]] static ::uint8_t* PROTOBUF_NONNULL _InternalSerialize(
      const ::google::protobuf::MessageLite& msg, ::uint8_t* PROTOBUF_NONNULL target,
      ::google::protobuf::io::EpsCopyOutputStream* PROTOBUF_NONNULL stream);

  public:
  [[nodiscard]] ::size_t ByteSizeLong() const { return ByteSizeLong(*this); }
  [[nodiscard]] ::uint8_t* PROTOBUF_NONNULL _InternalSerialize(
      ::uint8_t* PROTOBUF_NONNULL target,
      ::google::protobuf::io::EpsCopyOutputStream* PROTOBUF_NONNULL stream) const {
    return _InternalSerialize(*this, target, stream);
  }
  #else   // PROTOBUF_CUSTOM_VTABLE
  [[nodiscard]] ::size_t ByteSizeLong() const final;
  [[nodiscard]] ::uint8_t* PROTOBUF_NONNULL _InternalSerialize(
      ::uint8_t* PROTOBUF_NONNULL target,
      ::google::protobuf::io::EpsCopyOutputStream* PROTOBUF_NONNULL stream) const final;
  #endif  // PROTOBUF_CUSTOM_VTABLE
  [[nodiscard]] int GetCachedSize() const {
    return _impl_._cached_size_.Get();
  }

  private:
  void SharedCtor(::google::protobuf::Arena* PROTOBUF_NULLABLE arena);
  static void SharedDtor(MessageLite& self);
  void InternalSwap(UpdateNodeInstanceManifest* PROTOBUF_NONNULL other);
 private:
  template <typename T>
  friend ::absl::string_view(::google::protobuf::internal::GetAnyMessageName)();
  static ::absl::string_view FullMessageName() { return "nodeinstance.manifest.v1.UpdateNodeInstanceManifest"; }

  explicit UpdateNodeInstanceManifest(::google::protobuf::Arena* PROTOBUF_NULLABLE arena);
  UpdateNodeInstanceManifest(::google::protobuf::Arena* PROTOBUF_NULLABLE arena, const UpdateNodeInstanceManifest& from);
  UpdateNodeInstanceManifest(
      ::google::protobuf::Arena* PROTOBUF_NULLABLE arena, UpdateNodeInstanceManifest&& from) noexcept
      : UpdateNodeInstanceManifest(arena) {
    *this = ::std::move(from);
  }
  const ::google::protobuf::internal::ClassData* PROTOBUF_NONNULL GetClassData() const PROTOBUF_FINAL;
  static void* PROTOBUF_NONNULL PlacementNew_(
      const void* PROTOBUF_NONNULL, void* PROTOBUF_NONNULL mem,
      ::google::protobuf::Arena* PROTOBUF_NULLABLE arena);
  static constexpr auto InternalNewImpl_();

 public:
  static constexpr auto InternalGenerateClassData_(
      const MessageLite& prototype,
      const ::google::protobuf::internal::TcParseTableBase* PROTOBUF_NULLABLE tc_table = nullptr);

  [[nodiscard]] ::google::protobuf::Metadata GetMetadata() const;
  // nested types ----------------------------------------------------

  // accessors -------------------------------------------------------
  enum : int {
    kClaimIdFieldNumber = 1,
    kNodeIdFieldNumber = 2,
    kUpdatedAtFieldNumber = 3,
    kFunctionsFieldNumber = 10,
  };
  // string claim_id = 1;
  void clear_claim_id() ;
  [[nodiscard]] const ::std::string& claim_id() const;
  template <typename Arg_ = const ::std::string&, typename... Args_>
  void set_claim_id(Arg_&& arg, Args_... args);
  ::std::string* PROTOBUF_NONNULL mutable_claim_id();
  [[nodiscard]] ::std::string* PROTOBUF_NULLABLE release_claim_id();
  void set_allocated_claim_id(::std::string* PROTOBUF_NULLABLE value);

  private:
  const ::std::string& _internal_claim_id() const;
  PROTOBUF_ALWAYS_INLINE void _internal_set_claim_id(const ::std::string& value);
  ::std::string* PROTOBUF_NONNULL _internal_mutable_claim_id();

  public:
  // string node_id = 2;
  void clear_node_id() ;
  [[nodiscard]] const ::std::string& node_id() const;
  template <typename Arg_ = const ::std::string&, typename... Args_>
  void set_node_id(Arg_&& arg, Args_... args);
  ::std::string* PROTOBUF_NONNULL mutable_node_id();
  [[nodiscard]] ::std::string* PROTOBUF_NULLABLE release_node_id();
  void set_allocated_node_id(::std::string* PROTOBUF_NULLABLE value);

  private:
  const ::std::string& _internal_node_id() const;
  PROTOBUF_ALWAYS_INLINE void _internal_set_node_id(const ::std::string& value);
  ::std::string* PROTOBUF_NONNULL _internal_mutable_node_id();

  public:
  // .google.protobuf.Timestamp updated_at = 3;
  [[nodiscard]] bool has_updated_at()
      const;
  void clear_updated_at() ;
  [[nodiscard]] const ::google::protobuf::Timestamp& updated_at() const;
  [[nodiscard]] ::google::protobuf::Timestamp* PROTOBUF_NULLABLE release_updated_at();
  ::google::protobuf::Timestamp* PROTOBUF_NONNULL mutable_updated_at();
  void set_allocated_updated_at(::google::protobuf::Timestamp* PROTOBUF_NULLABLE value);
  void unsafe_arena_set_allocated_updated_at(::google::protobuf::Timestamp* PROTOBUF_NULLABLE value);
  ::google::protobuf::Timestamp* PROTOBUF_NULLABLE unsafe_arena_release_updated_at();

  private:
  const ::google::protobuf::Timestamp& _internal_updated_at() const;
  ::google::protobuf::Timestamp* PROTOBUF_NONNULL _internal_mutable_updated_at();

  public:
  // .nodeinstance.manifest.v1.Functions functions = 10;
  [[nodiscard]] bool has_functions()
      const;
  private:
  bool _internal_has_functions() const;

  public:
  void clear_functions() ;
  [[nodiscard]] const ::nodeinstance::manifest::v1::Functions& functions() const;
  [[nodiscard]] ::nodeinstance::manifest::v1::Functions* PROTOBUF_NULLABLE release_functions();
  ::nodeinstance::manifest::v1::Functions* PROTOBUF_NONNULL mutable_functions();
  void set_allocated_functions(::nodeinstance::manifest::v1::Functions* PROTOBUF_NULLABLE value);
  void unsafe_arena_set_allocated_functions(::nodeinstance::manifest::v1::Functions* PROTOBUF_NULLABLE value);
  ::nodeinstance::manifest::v1::Functions* PROTOBUF_NULLABLE unsafe_arena_release_functions();

  private:
  const ::nodeinstance::manifest::v1::Functions& _internal_functions() const;
  ::nodeinstance::manifest::v1::Functions* PROTOBUF_NONNULL _internal_mutable_functions();

  public:
  void clear_payload();
  PayloadCase payload_case() const;
  // @@protoc_insertion_point(class_scope:nodeinstance.manifest.v1.UpdateNodeInstanceManifest)
 private:
  class _Internal;
  void set_has_functions();
  [[nodiscard]] inline bool has_payload() const;
  inline void clear_has_payload();
  using ParseTableT_ =
      ::google::protobuf::internal::TcParseTable<2, 4,
                          2, 75,
                          2>;
  static constexpr ParseTableT_ InternalGenerateParseTable_(
      const ::google::protobuf::internal::ClassData* PROTOBUF_NONNULL class_data);
  friend class ::google::protobuf::internal::TcParser;
  #ifndef PROTOBUF_MESSAGE_GLOBALS
  static const ParseTableT_ _table_;
  #endif

  friend class ::google::protobuf::MessageLite;
  friend class ::google::protobuf::Arena;
  friend ::google::protobuf::internal::PrivateAccess;
  template <typename T>
  friend class ::google::protobuf::Arena::InternalHelper;
  using InternalArenaConstructable_ = void;
  using DestructorSkippable_ = void;
  struct Impl_ {
    inline explicit constexpr Impl_(::google::protobuf::internal::InternalVisibility visibility,
                                    ::google::protobuf::internal::ConstantInitialized) noexcept;
    inline explicit Impl_(
        ::google::protobuf::internal::InternalVisibility visibility,
        ::google::protobuf::Arena* PROTOBUF_NULLABLE arena);
    inline explicit Impl_(
        ::google::protobuf::internal::InternalVisibility visibility,
        ::google::protobuf::Arena* PROTOBUF_NULLABLE arena, const Impl_& from,
        const UpdateNodeInstanceManifest& from_msg);
    ::google::protobuf::internal::HasBits<1> _has_bits_;
    ::google::protobuf::internal::CachedSize _cached_size_;
    ::google::protobuf::internal::ArenaStringPtr claim_id_;
    ::google::protobuf::internal::ArenaStringPtr node_id_;
    ::google::protobuf::Timestamp* PROTOBUF_NULLABLE updated_at_;
    union PayloadUnion {
      constexpr PayloadUnion() : _constinit_{} {}
      ::google::protobuf::internal::ConstantInitialized _constinit_;
      ::nodeinstance::manifest::v1::Functions* PROTOBUF_NULLABLE functions_;
    } payload_;
    ::uint32_t _oneof_case_[1];
    PROTOBUF_TSAN_DECLARE_MEMBER
  };
  union { Impl_ _impl_; };
  friend struct ::TableStruct_nodeinstance_2fmanifest_2fv1_2fmanifest_2eproto;
};

// ===================================================================




// ===================================================================


#ifdef __GNUC__
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wstrict-aliasing"
#endif  // __GNUC__
// -------------------------------------------------------------------

// UpdateNodeInstanceManifest

// string claim_id = 1;
inline void UpdateNodeInstanceManifest::clear_claim_id() {
  ::google::protobuf::internal::TSanWrite(&_impl_);
  _impl_.claim_id_.ClearToEmpty();
  ClearHasBit(_impl_._has_bits_[0], 0x00000001U);
}
inline const ::std::string& UpdateNodeInstanceManifest::claim_id() const
    ABSL_ATTRIBUTE_LIFETIME_BOUND {
  // @@protoc_insertion_point(field_get:nodeinstance.manifest.v1.UpdateNodeInstanceManifest.claim_id)
  return _internal_claim_id();
}
template <typename Arg_, typename... Args_>
PROTOBUF_ALWAYS_INLINE void UpdateNodeInstanceManifest::set_claim_id(Arg_&& arg, Args_... args) {
  ::google::protobuf::internal::TSanWrite(&_impl_);
  SetHasBit(_impl_._has_bits_[0], 0x00000001U);
  _impl_.claim_id_.Set(static_cast<Arg_&&>(arg), args..., GetArena());
  // @@protoc_insertion_point(field_set:nodeinstance.manifest.v1.UpdateNodeInstanceManifest.claim_id)
}
inline ::std::string* PROTOBUF_NONNULL UpdateNodeInstanceManifest::mutable_claim_id()
    ABSL_ATTRIBUTE_LIFETIME_BOUND {
  SetHasBit(_impl_._has_bits_[0], 0x00000001U);
  ::std::string* _s = _internal_mutable_claim_id();
  // @@protoc_insertion_point(field_mutable:nodeinstance.manifest.v1.UpdateNodeInstanceManifest.claim_id)
  return _s;
}
inline const ::std::string& UpdateNodeInstanceManifest::_internal_claim_id() const {
  ::google::protobuf::internal::TSanRead(&_impl_);
  return _impl_.claim_id_.Get();
}
inline void UpdateNodeInstanceManifest::_internal_set_claim_id(const ::std::string& value) {
  ::google::protobuf::internal::TSanWrite(&_impl_);
  _impl_.claim_id_.Set(value, GetArena());
}
inline ::std::string* PROTOBUF_NONNULL UpdateNodeInstanceManifest::_internal_mutable_claim_id() {
  ::google::protobuf::internal::TSanWrite(&_impl_);
  return _impl_.claim_id_.Mutable( GetArena());
}
inline ::std::string* PROTOBUF_NULLABLE UpdateNodeInstanceManifest::release_claim_id() {
  ::google::protobuf::internal::TSanWrite(&_impl_);
  // @@protoc_insertion_point(field_release:nodeinstance.manifest.v1.UpdateNodeInstanceManifest.claim_id)
  if (!CheckHasBit(_impl_._has_bits_[0], 0x00000001U)) {
    return nullptr;
  }
  ClearHasBit(_impl_._has_bits_[0], 0x00000001U);
  auto* released = _impl_.claim_id_.Release();
  if (::google::protobuf::internal::DebugHardenForceCopyDefaultString()) {
    _impl_.claim_id_.Set("", GetArena());
  }
  return released;
}
inline void UpdateNodeInstanceManifest::set_allocated_claim_id(::std::string* PROTOBUF_NULLABLE value) {
  ::google::protobuf::internal::TSanWrite(&_impl_);
  if (value != nullptr) {
    SetHasBit(_impl_._has_bits_[0], 0x00000001U);
  } else {
    ClearHasBit(_impl_._has_bits_[0], 0x00000001U);
  }
  _impl_.claim_id_.SetAllocated(value, GetArena());
  if (::google::protobuf::internal::DebugHardenForceCopyDefaultString() && _impl_.claim_id_.IsDefault()) {
    _impl_.claim_id_.Set("", GetArena());
  }
  // @@protoc_insertion_point(field_set_allocated:nodeinstance.manifest.v1.UpdateNodeInstanceManifest.claim_id)
}

// string node_id = 2;
inline void UpdateNodeInstanceManifest::clear_node_id() {
  ::google::protobuf::internal::TSanWrite(&_impl_);
  _impl_.node_id_.ClearToEmpty();
  ClearHasBit(_impl_._has_bits_[0], 0x00000002U);
}
inline const ::std::string& UpdateNodeInstanceManifest::node_id() const
    ABSL_ATTRIBUTE_LIFETIME_BOUND {
  // @@protoc_insertion_point(field_get:nodeinstance.manifest.v1.UpdateNodeInstanceManifest.node_id)
  return _internal_node_id();
}
template <typename Arg_, typename... Args_>
PROTOBUF_ALWAYS_INLINE void UpdateNodeInstanceManifest::set_node_id(Arg_&& arg, Args_... args) {
  ::google::protobuf::internal::TSanWrite(&_impl_);
  SetHasBit(_impl_._has_bits_[0], 0x00000002U);
  _impl_.node_id_.Set(static_cast<Arg_&&>(arg), args..., GetArena());
  // @@protoc_insertion_point(field_set:nodeinstance.manifest.v1.UpdateNodeInstanceManifest.node_id)
}
inline ::std::string* PROTOBUF_NONNULL UpdateNodeInstanceManifest::mutable_node_id()
    ABSL_ATTRIBUTE_LIFETIME_BOUND {
  SetHasBit(_impl_._has_bits_[0], 0x00000002U);
  ::std::string* _s = _internal_mutable_node_id();
  // @@protoc_insertion_point(field_mutable:nodeinstance.manifest.v1.UpdateNodeInstanceManifest.node_id)
  return _s;
}
inline const ::std::string& UpdateNodeInstanceManifest::_internal_node_id() const {
  ::google::protobuf::internal::TSanRead(&_impl_);
  return _impl_.node_id_.Get();
}
inline void UpdateNodeInstanceManifest::_internal_set_node_id(const ::std::string& value) {
  ::google::protobuf::internal::TSanWrite(&_impl_);
  _impl_.node_id_.Set(value, GetArena());
}
inline ::std::string* PROTOBUF_NONNULL UpdateNodeInstanceManifest::_internal_mutable_node_id() {
  ::google::protobuf::internal::TSanWrite(&_impl_);
  return _impl_.node_id_.Mutable( GetArena());
}
inline ::std::string* PROTOBUF_NULLABLE UpdateNodeInstanceManifest::release_node_id() {
  ::google::protobuf::internal::TSanWrite(&_impl_);
  // @@protoc_insertion_point(field_release:nodeinstance.manifest.v1.UpdateNodeInstanceManifest.node_id)
  if (!CheckHasBit(_impl_._has_bits_[0], 0x00000002U)) {
    return nullptr;
  }
  ClearHasBit(_impl_._has_bits_[0], 0x00000002U);
  auto* released = _impl_.node_id_.Release();
  if (::google::protobuf::internal::DebugHardenForceCopyDefaultString()) {
    _impl_.node_id_.Set("", GetArena());
  }
  return released;
}
inline void UpdateNodeInstanceManifest::set_allocated_node_id(::std::string* PROTOBUF_NULLABLE value) {
  ::google::protobuf::internal::TSanWrite(&_impl_);
  if (value != nullptr) {
    SetHasBit(_impl_._has_bits_[0], 0x00000002U);
  } else {
    ClearHasBit(_impl_._has_bits_[0], 0x00000002U);
  }
  _impl_.node_id_.SetAllocated(value, GetArena());
  if (::google::protobuf::internal::DebugHardenForceCopyDefaultString() && _impl_.node_id_.IsDefault()) {
    _impl_.node_id_.Set("", GetArena());
  }
  // @@protoc_insertion_point(field_set_allocated:nodeinstance.manifest.v1.UpdateNodeInstanceManifest.node_id)
}

// .google.protobuf.Timestamp updated_at = 3;
inline bool UpdateNodeInstanceManifest::has_updated_at() const {
  bool value = CheckHasBit(_impl_._has_bits_[0], 0x00000004U);
  PROTOBUF_ASSUME(!value || _impl_.updated_at_ != nullptr);
  return value;
}
inline const ::google::protobuf::Timestamp& UpdateNodeInstanceManifest::_internal_updated_at() const {
  ::google::protobuf::internal::TSanRead(&_impl_);
  const ::google::protobuf::Timestamp* p = _impl_.updated_at_;
  return p != nullptr ? *p : *::google::protobuf::internal::MessageGlobalsBase::ToDefaultInstance<::google::protobuf::Timestamp>(&::google::protobuf::Timestamp_globals_);
}
inline const ::google::protobuf::Timestamp& UpdateNodeInstanceManifest::updated_at() const ABSL_ATTRIBUTE_LIFETIME_BOUND {
  // @@protoc_insertion_point(field_get:nodeinstance.manifest.v1.UpdateNodeInstanceManifest.updated_at)
  return _internal_updated_at();
}
inline void UpdateNodeInstanceManifest::unsafe_arena_set_allocated_updated_at(
    ::google::protobuf::Timestamp* PROTOBUF_NULLABLE value) {
  ::google::protobuf::internal::TSanWrite(&_impl_);
  if (GetArena() == nullptr) {
    delete reinterpret_cast<::google::protobuf::MessageLite*>(_impl_.updated_at_);
  }
  _impl_.updated_at_ = reinterpret_cast<::google::protobuf::Timestamp*>(value);
  if (value != nullptr) {
    SetHasBit(_impl_._has_bits_[0], 0x00000004U);
  } else {
    ClearHasBit(_impl_._has_bits_[0], 0x00000004U);
  }
  // @@protoc_insertion_point(field_unsafe_arena_set_allocated:nodeinstance.manifest.v1.UpdateNodeInstanceManifest.updated_at)
}
inline ::google::protobuf::Timestamp* PROTOBUF_NULLABLE UpdateNodeInstanceManifest::release_updated_at() {
  ::google::protobuf::internal::TSanWrite(&_impl_);

  ClearHasBit(_impl_._has_bits_[0], 0x00000004U);
  ::google::protobuf::Timestamp* released = _impl_.updated_at_;
  _impl_.updated_at_ = nullptr;
  if (::google::protobuf::internal::DebugHardenForceCopyInRelease()) {
    auto* old = reinterpret_cast<::google::protobuf::MessageLite*>(released);
    released = ::google::protobuf::internal::DuplicateIfNonNull(released);
    if (GetArena() == nullptr) {
      delete old;
    }
  } else {
    if (GetArena() != nullptr) {
      released = ::google::protobuf::internal::DuplicateIfNonNull(released);
    }
  }
  return released;
}
inline ::google::protobuf::Timestamp* PROTOBUF_NULLABLE UpdateNodeInstanceManifest::unsafe_arena_release_updated_at() {
  ::google::protobuf::internal::TSanWrite(&_impl_);
  // @@protoc_insertion_point(field_release:nodeinstance.manifest.v1.UpdateNodeInstanceManifest.updated_at)

  ClearHasBit(_impl_._has_bits_[0], 0x00000004U);
  ::google::protobuf::Timestamp* temp = _impl_.updated_at_;
  _impl_.updated_at_ = nullptr;
  return temp;
}
inline ::google::protobuf::Timestamp* PROTOBUF_NONNULL UpdateNodeInstanceManifest::_internal_mutable_updated_at() {
  ::google::protobuf::internal::TSanWrite(&_impl_);
  if (_impl_.updated_at_ == nullptr) {
    auto* p = ::google::protobuf::Message::DefaultConstruct<::google::protobuf::Timestamp>(GetArena());
    _impl_.updated_at_ = reinterpret_cast<::google::protobuf::Timestamp*>(p);
  }
  return _impl_.updated_at_;
}
inline ::google::protobuf::Timestamp* PROTOBUF_NONNULL UpdateNodeInstanceManifest::mutable_updated_at()
    ABSL_ATTRIBUTE_LIFETIME_BOUND {
  SetHasBit(_impl_._has_bits_[0], 0x00000004U);
  ::google::protobuf::Timestamp* _msg = _internal_mutable_updated_at();
  // @@protoc_insertion_point(field_mutable:nodeinstance.manifest.v1.UpdateNodeInstanceManifest.updated_at)
  return _msg;
}
inline void UpdateNodeInstanceManifest::set_allocated_updated_at(::google::protobuf::Timestamp* PROTOBUF_NULLABLE value) {
  ::google::protobuf::Arena* message_arena = GetArena();
  ::google::protobuf::internal::TSanWrite(&_impl_);
  if (message_arena == nullptr) {
    delete reinterpret_cast<::google::protobuf::MessageLite*>(_impl_.updated_at_);
  }

  if (value != nullptr) {
    ::google::protobuf::Arena* submessage_arena = reinterpret_cast<::google::protobuf::Message*>(value)->GetArena();
    if (message_arena != submessage_arena) {
      value = ::google::protobuf::internal::GetOwnedMessage(message_arena, value, submessage_arena);
    }
    SetHasBit(_impl_._has_bits_[0], 0x00000004U);
  } else {
    ClearHasBit(_impl_._has_bits_[0], 0x00000004U);
  }

  _impl_.updated_at_ = reinterpret_cast<::google::protobuf::Timestamp*>(value);
  // @@protoc_insertion_point(field_set_allocated:nodeinstance.manifest.v1.UpdateNodeInstanceManifest.updated_at)
}

// .nodeinstance.manifest.v1.Functions functions = 10;
inline bool UpdateNodeInstanceManifest::has_functions() const {
  return payload_case() == kFunctions;
}
inline bool UpdateNodeInstanceManifest::_internal_has_functions() const {
  return payload_case() == kFunctions;
}
inline void UpdateNodeInstanceManifest::set_has_functions() {
  _impl_._oneof_case_[0] = kFunctions;
}
inline void UpdateNodeInstanceManifest::clear_functions() {
  ::google::protobuf::internal::TSanWrite(&_impl_);
  if (payload_case() == kFunctions) {
    if (GetArena() == nullptr) {
      delete _impl_.payload_.functions_;
    } else if (::google::protobuf::internal::DebugHardenClearOneofMessageOnArena()) {
      ::google::protobuf::internal::MaybePoisonAfterClear(_impl_.payload_.functions_);
    }
    clear_has_payload();
  }
}
inline ::nodeinstance::manifest::v1::Functions* PROTOBUF_NULLABLE UpdateNodeInstanceManifest::release_functions() {
  // @@protoc_insertion_point(field_release:nodeinstance.manifest.v1.UpdateNodeInstanceManifest.functions)
  if (payload_case() == kFunctions) {
    clear_has_payload();
    auto* temp = _impl_.payload_.functions_;
    if (GetArena() != nullptr) {
      temp = ::google::protobuf::internal::DuplicateIfNonNull(temp);
    }
    _impl_.payload_.functions_ = nullptr;
    return temp;
  } else {
    return nullptr;
  }
}
inline const ::nodeinstance::manifest::v1::Functions& UpdateNodeInstanceManifest::_internal_functions() const {
  return payload_case() == kFunctions ? static_cast<const ::nodeinstance::manifest::v1::Functions&>(*_impl_.payload_.functions_)
                     : *::google::protobuf::internal::MessageGlobalsBase::ToDefaultInstance<::nodeinstance::manifest::v1::Functions>(&::nodeinstance::manifest::v1::Functions_globals_);
}
inline const ::nodeinstance::manifest::v1::Functions& UpdateNodeInstanceManifest::functions() const ABSL_ATTRIBUTE_LIFETIME_BOUND {
  // @@protoc_insertion_point(field_get:nodeinstance.manifest.v1.UpdateNodeInstanceManifest.functions)
  return _internal_functions();
}
inline ::nodeinstance::manifest::v1::Functions* PROTOBUF_NULLABLE UpdateNodeInstanceManifest::unsafe_arena_release_functions() {
  // @@protoc_insertion_point(field_unsafe_arena_release:nodeinstance.manifest.v1.UpdateNodeInstanceManifest.functions)
  if (payload_case() == kFunctions) {
    clear_has_payload();
    auto* temp = _impl_.payload_.functions_;
    _impl_.payload_.functions_ = nullptr;
    return temp;
  } else {
    return nullptr;
  }
}
inline void UpdateNodeInstanceManifest::unsafe_arena_set_allocated_functions(
    ::nodeinstance::manifest::v1::Functions* PROTOBUF_NULLABLE value) {
  // We rely on the oneof clear method to free the earlier contents
  // of this oneof. We can directly use the pointer we're given to
  // set the new value.
  clear_payload();
  if (value) {
    set_has_functions();
    _impl_.payload_.functions_ = value;
  }
  // @@protoc_insertion_point(field_unsafe_arena_set_allocated:nodeinstance.manifest.v1.UpdateNodeInstanceManifest.functions)
}
inline ::nodeinstance::manifest::v1::Functions* PROTOBUF_NONNULL UpdateNodeInstanceManifest::_internal_mutable_functions() {
  if (payload_case() != kFunctions) {
    clear_payload();
    set_has_functions();
    _impl_.payload_.functions_ = 
        ::google::protobuf::Message::DefaultConstruct<::nodeinstance::manifest::v1::Functions>(GetArena());
  }
  return _impl_.payload_.functions_;
}
inline ::nodeinstance::manifest::v1::Functions* PROTOBUF_NONNULL UpdateNodeInstanceManifest::mutable_functions()
    ABSL_ATTRIBUTE_LIFETIME_BOUND {
  ::nodeinstance::manifest::v1::Functions* _msg = _internal_mutable_functions();
  // @@protoc_insertion_point(field_mutable:nodeinstance.manifest.v1.UpdateNodeInstanceManifest.functions)
  return _msg;
}

inline bool UpdateNodeInstanceManifest::has_payload() const {
  return payload_case() != PAYLOAD_NOT_SET;
}
inline void UpdateNodeInstanceManifest::clear_has_payload() {
  _impl_._oneof_case_[0] = PAYLOAD_NOT_SET;
}
inline UpdateNodeInstanceManifest::PayloadCase UpdateNodeInstanceManifest::payload_case() const {
  return UpdateNodeInstanceManifest::PayloadCase(_impl_._oneof_case_[0]);
}
// -------------------------------------------------------------------

// Functions

// repeated .nodeinstance.manifest.v1.Function items = 1;
inline int Functions::_internal_items_size() const {
  return _internal_items().size();
}
inline int Functions::items_size() const {
  return _internal_items_size();
}
inline void Functions::clear_items() {
  ::google::protobuf::internal::TSanWrite(&_impl_);
  _impl_.items_.Clear();
  ClearHasBit(_impl_._has_bits_[0], 0x00000001U);
}
inline const ::nodeinstance::manifest::v1::Function& Functions::items(int index) const
    ABSL_ATTRIBUTE_LIFETIME_BOUND {
  // @@protoc_insertion_point(field_get:nodeinstance.manifest.v1.Functions.items)
  return _internal_items().Get(index);
}
inline ::nodeinstance::manifest::v1::Function* PROTOBUF_NONNULL Functions::mutable_items(int index)
    ABSL_ATTRIBUTE_LIFETIME_BOUND {
  // @@protoc_insertion_point(field_mutable:nodeinstance.manifest.v1.Functions.items)
  return _internal_mutable_items()->Mutable(index);
}
inline ::nodeinstance::manifest::v1::Function* PROTOBUF_NONNULL Functions::add_items()
    ABSL_ATTRIBUTE_LIFETIME_BOUND {
  ::google::protobuf::internal::TSanWrite(&_impl_);
  ::nodeinstance::manifest::v1::Function* _add =
      _internal_mutable_items()->InternalAddWithArena(
          ::google::protobuf::MessageLite::internal_visibility(), GetArena());
  SetHasBit(_impl_._has_bits_[0], 0x00000001U);
  // @@protoc_insertion_point(field_add:nodeinstance.manifest.v1.Functions.items)
  return _add;
}
inline const ::google::protobuf::RepeatedPtrField<::nodeinstance::manifest::v1::Function>& Functions::items() const
    ABSL_ATTRIBUTE_LIFETIME_BOUND {
  // @@protoc_insertion_point(field_list:nodeinstance.manifest.v1.Functions.items)
  return _internal_items();
}
inline ::google::protobuf::RepeatedPtrField<::nodeinstance::manifest::v1::Function>* PROTOBUF_NONNULL
Functions::mutable_items() ABSL_ATTRIBUTE_LIFETIME_BOUND {
  SetHasBit(_impl_._has_bits_[0], 0x00000001U);
  // @@protoc_insertion_point(field_mutable_list:nodeinstance.manifest.v1.Functions.items)
  ::google::protobuf::internal::TSanWrite(&_impl_);
  return _internal_mutable_items();
}
inline const ::google::protobuf::RepeatedPtrField<::nodeinstance::manifest::v1::Function>&
Functions::_internal_items() const {
  ::google::protobuf::internal::TSanRead(&_impl_);
  return _impl_.items_;
}
inline ::google::protobuf::RepeatedPtrField<::nodeinstance::manifest::v1::Function>* PROTOBUF_NONNULL
Functions::_internal_mutable_items() {
  ::google::protobuf::internal::TSanRead(&_impl_);
  return &_impl_.items_;
}

// -------------------------------------------------------------------

// Function

// string name = 1;
inline void Function::clear_name() {
  ::google::protobuf::internal::TSanWrite(&_impl_);
  _impl_.name_.ClearToEmpty();
  ClearHasBit(_impl_._has_bits_[0], 0x00000004U);
}
inline const ::std::string& Function::name() const
    ABSL_ATTRIBUTE_LIFETIME_BOUND {
  // @@protoc_insertion_point(field_get:nodeinstance.manifest.v1.Function.name)
  return _internal_name();
}
template <typename Arg_, typename... Args_>
PROTOBUF_ALWAYS_INLINE void Function::set_name(Arg_&& arg, Args_... args) {
  ::google::protobuf::internal::TSanWrite(&_impl_);
  SetHasBit(_impl_._has_bits_[0], 0x00000004U);
  _impl_.name_.Set(static_cast<Arg_&&>(arg), args..., GetArena());
  // @@protoc_insertion_point(field_set:nodeinstance.manifest.v1.Function.name)
}
inline ::std::string* PROTOBUF_NONNULL Function::mutable_name()
    ABSL_ATTRIBUTE_LIFETIME_BOUND {
  SetHasBit(_impl_._has_bits_[0], 0x00000004U);
  ::std::string* _s = _internal_mutable_name();
  // @@protoc_insertion_point(field_mutable:nodeinstance.manifest.v1.Function.name)
  return _s;
}
inline const ::std::string& Function::_internal_name() const {
  ::google::protobuf::internal::TSanRead(&_impl_);
  return _impl_.name_.Get();
}
inline void Function::_internal_set_name(const ::std::string& value) {
  ::google::protobuf::internal::TSanWrite(&_impl_);
  _impl_.name_.Set(value, GetArena());
}
inline ::std::string* PROTOBUF_NONNULL Function::_internal_mutable_name() {
  ::google::protobuf::internal::TSanWrite(&_impl_);
  return _impl_.name_.Mutable( GetArena());
}
inline ::std::string* PROTOBUF_NULLABLE Function::release_name() {
  ::google::protobuf::internal::TSanWrite(&_impl_);
  // @@protoc_insertion_point(field_release:nodeinstance.manifest.v1.Function.name)
  if (!CheckHasBit(_impl_._has_bits_[0], 0x00000004U)) {
    return nullptr;
  }
  ClearHasBit(_impl_._has_bits_[0], 0x00000004U);
  auto* released = _impl_.name_.Release();
  if (::google::protobuf::internal::DebugHardenForceCopyDefaultString()) {
    _impl_.name_.Set("", GetArena());
  }
  return released;
}
inline void Function::set_allocated_name(::std::string* PROTOBUF_NULLABLE value) {
  ::google::protobuf::internal::TSanWrite(&_impl_);
  if (value != nullptr) {
    SetHasBit(_impl_._has_bits_[0], 0x00000004U);
  } else {
    ClearHasBit(_impl_._has_bits_[0], 0x00000004U);
  }
  _impl_.name_.SetAllocated(value, GetArena());
  if (::google::protobuf::internal::DebugHardenForceCopyDefaultString() && _impl_.name_.IsDefault()) {
    _impl_.name_.Set("", GetArena());
  }
  // @@protoc_insertion_point(field_set_allocated:nodeinstance.manifest.v1.Function.name)
}

// string help = 2;
inline void Function::clear_help() {
  ::google::protobuf::internal::TSanWrite(&_impl_);
  _impl_.help_.ClearToEmpty();
  ClearHasBit(_impl_._has_bits_[0], 0x00000008U);
}
inline const ::std::string& Function::help() const
    ABSL_ATTRIBUTE_LIFETIME_BOUND {
  // @@protoc_insertion_point(field_get:nodeinstance.manifest.v1.Function.help)
  return _internal_help();
}
template <typename Arg_, typename... Args_>
PROTOBUF_ALWAYS_INLINE void Function::set_help(Arg_&& arg, Args_... args) {
  ::google::protobuf::internal::TSanWrite(&_impl_);
  SetHasBit(_impl_._has_bits_[0], 0x00000008U);
  _impl_.help_.Set(static_cast<Arg_&&>(arg), args..., GetArena());
  // @@protoc_insertion_point(field_set:nodeinstance.manifest.v1.Function.help)
}
inline ::std::string* PROTOBUF_NONNULL Function::mutable_help()
    ABSL_ATTRIBUTE_LIFETIME_BOUND {
  SetHasBit(_impl_._has_bits_[0], 0x00000008U);
  ::std::string* _s = _internal_mutable_help();
  // @@protoc_insertion_point(field_mutable:nodeinstance.manifest.v1.Function.help)
  return _s;
}
inline const ::std::string& Function::_internal_help() const {
  ::google::protobuf::internal::TSanRead(&_impl_);
  return _impl_.help_.Get();
}
inline void Function::_internal_set_help(const ::std::string& value) {
  ::google::protobuf::internal::TSanWrite(&_impl_);
  _impl_.help_.Set(value, GetArena());
}
inline ::std::string* PROTOBUF_NONNULL Function::_internal_mutable_help() {
  ::google::protobuf::internal::TSanWrite(&_impl_);
  return _impl_.help_.Mutable( GetArena());
}
inline ::std::string* PROTOBUF_NULLABLE Function::release_help() {
  ::google::protobuf::internal::TSanWrite(&_impl_);
  // @@protoc_insertion_point(field_release:nodeinstance.manifest.v1.Function.help)
  if (!CheckHasBit(_impl_._has_bits_[0], 0x00000008U)) {
    return nullptr;
  }
  ClearHasBit(_impl_._has_bits_[0], 0x00000008U);
  auto* released = _impl_.help_.Release();
  if (::google::protobuf::internal::DebugHardenForceCopyDefaultString()) {
    _impl_.help_.Set("", GetArena());
  }
  return released;
}
inline void Function::set_allocated_help(::std::string* PROTOBUF_NULLABLE value) {
  ::google::protobuf::internal::TSanWrite(&_impl_);
  if (value != nullptr) {
    SetHasBit(_impl_._has_bits_[0], 0x00000008U);
  } else {
    ClearHasBit(_impl_._has_bits_[0], 0x00000008U);
  }
  _impl_.help_.SetAllocated(value, GetArena());
  if (::google::protobuf::internal::DebugHardenForceCopyDefaultString() && _impl_.help_.IsDefault()) {
    _impl_.help_.Set("", GetArena());
  }
  // @@protoc_insertion_point(field_set_allocated:nodeinstance.manifest.v1.Function.help)
}

// repeated string tags = 3;
inline int Function::_internal_tags_size() const {
  return _internal_tags().size();
}
inline int Function::tags_size() const {
  return _internal_tags_size();
}
inline void Function::clear_tags() {
  ::google::protobuf::internal::TSanWrite(&_impl_);
  _impl_.tags_.Clear();
  ClearHasBit(_impl_._has_bits_[0], 0x00000001U);
}
inline ::std::string* PROTOBUF_NONNULL Function::add_tags()
    ABSL_ATTRIBUTE_LIFETIME_BOUND {
  ::google::protobuf::internal::TSanWrite(&_impl_);
  ::std::string* _s =
      _internal_mutable_tags()->InternalAddWithArena(
          ::google::protobuf::MessageLite::internal_visibility(), GetArena());
  SetHasBit(_impl_._has_bits_[0], 0x00000001U);
  // @@protoc_insertion_point(field_add_mutable:nodeinstance.manifest.v1.Function.tags)
  return _s;
}
inline const ::std::string& Function::tags(int index) const
    ABSL_ATTRIBUTE_LIFETIME_BOUND {
  // @@protoc_insertion_point(field_get:nodeinstance.manifest.v1.Function.tags)
  return _internal_tags().Get(index);
}
inline ::std::string* PROTOBUF_NONNULL Function::mutable_tags(int index)
    ABSL_ATTRIBUTE_LIFETIME_BOUND {
  // @@protoc_insertion_point(field_mutable:nodeinstance.manifest.v1.Function.tags)
  return _internal_mutable_tags()->Mutable(index);
}
template <typename Arg_, typename... Args_>
inline void Function::set_tags(int index, Arg_&& value, Args_... args) {
  ::google::protobuf::internal::AssignToString(
      *_internal_mutable_tags()->Mutable(index),
      ::std::forward<Arg_>(value), args... );
  // @@protoc_insertion_point(field_set:nodeinstance.manifest.v1.Function.tags)
}
template <typename Arg_, typename... Args_>
inline void Function::add_tags(Arg_&& value, Args_... args) {
  ::google::protobuf::internal::TSanWrite(&_impl_);
  ::google::protobuf::internal::AddToRepeatedPtrField(
      ::google::protobuf::MessageLite::internal_visibility(), GetArena(),
      *_internal_mutable_tags(), ::std::forward<Arg_>(value),
      args... );
  SetHasBit(_impl_._has_bits_[0], 0x00000001U);
  // @@protoc_insertion_point(field_add:nodeinstance.manifest.v1.Function.tags)
}
inline const ::google::protobuf::RepeatedPtrField<::std::string>& Function::tags()
    const ABSL_ATTRIBUTE_LIFETIME_BOUND {
  // @@protoc_insertion_point(field_list:nodeinstance.manifest.v1.Function.tags)
  return _internal_tags();
}
inline ::google::protobuf::RepeatedPtrField<::std::string>* PROTOBUF_NONNULL
Function::mutable_tags() ABSL_ATTRIBUTE_LIFETIME_BOUND {
  SetHasBit(_impl_._has_bits_[0], 0x00000001U);
  // @@protoc_insertion_point(field_mutable_list:nodeinstance.manifest.v1.Function.tags)
  ::google::protobuf::internal::TSanWrite(&_impl_);
  return _internal_mutable_tags();
}
inline const ::google::protobuf::RepeatedPtrField<::std::string>&
Function::_internal_tags() const {
  ::google::protobuf::internal::TSanRead(&_impl_);
  return _impl_.tags_;
}
inline ::google::protobuf::RepeatedPtrField<::std::string>* PROTOBUF_NONNULL
Function::_internal_mutable_tags() {
  ::google::protobuf::internal::TSanRead(&_impl_);
  return &_impl_.tags_;
}

// uint32 priority = 4;
inline void Function::clear_priority() {
  ::google::protobuf::internal::TSanWrite(&_impl_);
  _impl_.priority_ = 0u;
  ClearHasBit(_impl_._has_bits_[0], 0x00000010U);
}
inline ::uint32_t Function::priority() const {
  // @@protoc_insertion_point(field_get:nodeinstance.manifest.v1.Function.priority)
  return _internal_priority();
}
inline void Function::set_priority(::uint32_t value) {
  _internal_set_priority(value);
  SetHasBit(_impl_._has_bits_[0], 0x00000010U);
  // @@protoc_insertion_point(field_set:nodeinstance.manifest.v1.Function.priority)
}
inline ::uint32_t Function::_internal_priority() const {
  ::google::protobuf::internal::TSanRead(&_impl_);
  return _impl_.priority_;
}
inline void Function::_internal_set_priority(::uint32_t value) {
  ::google::protobuf::internal::TSanWrite(&_impl_);
  _impl_.priority_ = value;
}

// uint32 version = 5;
inline void Function::clear_version() {
  ::google::protobuf::internal::TSanWrite(&_impl_);
  _impl_.version_ = 0u;
  ClearHasBit(_impl_._has_bits_[0], 0x00000020U);
}
inline ::uint32_t Function::version() const {
  // @@protoc_insertion_point(field_get:nodeinstance.manifest.v1.Function.version)
  return _internal_version();
}
inline void Function::set_version(::uint32_t value) {
  _internal_set_version(value);
  SetHasBit(_impl_._has_bits_[0], 0x00000020U);
  // @@protoc_insertion_point(field_set:nodeinstance.manifest.v1.Function.version)
}
inline ::uint32_t Function::_internal_version() const {
  ::google::protobuf::internal::TSanRead(&_impl_);
  return _impl_.version_;
}
inline void Function::_internal_set_version(::uint32_t value) {
  ::google::protobuf::internal::TSanWrite(&_impl_);
  _impl_.version_ = value;
}

// repeated .nodeinstance.manifest.v1.HttpAccess access = 6;
inline int Function::_internal_access_size() const {
  return _internal_access().size();
}
inline int Function::access_size() const {
  return _internal_access_size();
}
inline void Function::clear_access() {
  ::google::protobuf::internal::TSanWrite(&_impl_);
  _impl_.access_.Clear();
  ClearHasBit(_impl_._has_bits_[0], 0x00000002U);
}
inline ::nodeinstance::manifest::v1::HttpAccess Function::access(int index) const {
  // @@protoc_insertion_point(field_get:nodeinstance.manifest.v1.Function.access)
  return static_cast<::nodeinstance::manifest::v1::HttpAccess>(_internal_access().Get(index));
}
inline void Function::set_access(int index, ::nodeinstance::manifest::v1::HttpAccess value) {
  _internal_mutable_access()->Set(index, value);
  // @@protoc_insertion_point(field_set:nodeinstance.manifest.v1.Function.access)
}
inline void Function::add_access(::nodeinstance::manifest::v1::HttpAccess value) {
  ::google::protobuf::internal::TSanWrite(&_impl_);
  _internal_mutable_access()
      ->InternalAddWithArena<const ::google::protobuf::MessageLite*>(
          internal_visibility(), this, value);
  SetHasBit(_impl_._has_bits_[0], 0x00000002U);
  // @@protoc_insertion_point(field_add:nodeinstance.manifest.v1.Function.access)
}
inline const ::google::protobuf::RepeatedField<int>& Function::access() const
    ABSL_ATTRIBUTE_LIFETIME_BOUND {
  // @@protoc_insertion_point(field_list:nodeinstance.manifest.v1.Function.access)
  return _internal_access();
}
inline ::google::protobuf::RepeatedField<int>* PROTOBUF_NONNULL Function::mutable_access()
    ABSL_ATTRIBUTE_LIFETIME_BOUND {
  SetHasBit(_impl_._has_bits_[0], 0x00000002U);
  // @@protoc_insertion_point(field_mutable_list:nodeinstance.manifest.v1.Function.access)
  ::google::protobuf::internal::TSanWrite(&_impl_);
  return _internal_mutable_access();
}
inline const ::google::protobuf::RepeatedField<int>& Function::_internal_access()
    const {
  ::google::protobuf::internal::TSanRead(&_impl_);
  return _impl_.access_;
}
inline ::google::protobuf::RepeatedField<int>* PROTOBUF_NONNULL
Function::_internal_mutable_access() {
  ::google::protobuf::internal::TSanRead(&_impl_);
  return &_impl_.access_;
}

#ifdef __GNUC__
#pragma GCC diagnostic pop
#endif  // __GNUC__

// @@protoc_insertion_point(namespace_scope)
}  // namespace v1
}  // namespace manifest
}  // namespace nodeinstance


namespace google {
namespace protobuf {

template <>
struct is_proto_enum<::nodeinstance::manifest::v1::HttpAccess> : std::true_type {};
template <>
inline const EnumDescriptor* PROTOBUF_NONNULL GetEnumDescriptor<::nodeinstance::manifest::v1::HttpAccess>() {
  return ::nodeinstance::manifest::v1::HttpAccess_descriptor();
}

}  // namespace protobuf
}  // namespace google

// @@protoc_insertion_point(global_scope)

#include "google/protobuf/port_undef.inc"
// clang-format on

#endif  // nodeinstance_2fmanifest_2fv1_2fmanifest_2eproto_2epb_2eh
