// Generated by the protocol buffer compiler.  DO NOT EDIT!
// NO CHECKED-IN PROTOBUF GENCODE
// source: nodeinstance/connection/v1/connection.proto
// Protobuf C++ Version: 7.35.1

#ifndef nodeinstance_2fconnection_2fv1_2fconnection_2eproto_2epb_2eh
#define nodeinstance_2fconnection_2fv1_2fconnection_2eproto_2epb_2eh

#include <limits>
#include <string>
#include <type_traits>
#include <utility>

// clang-format off
#include "google/protobuf/runtime_version.h"
#if PROTOBUF_VERSION != 7035001
#error "Protobuf C++ gencode is built with an incompatible version of"
#error "Protobuf C++ headers/runtime. See"
#error "https://protobuf.dev/support/cross-version-runtime-guarantee/#cpp"
#endif
#include "google/protobuf/io/coded_stream.h"
#include "google/protobuf/arena.h"
#include "google/protobuf/arenastring.h"
#include "google/protobuf/generated_message_tctable_decl.h"
#include "google/protobuf/generated_message_util.h"
#include "google/protobuf/metadata_lite.h"
#include "google/protobuf/generated_message_reflection.h"
#include "google/protobuf/message.h"
#include "google/protobuf/message_lite.h"
#include "google/protobuf/repeated_field.h"  // IWYU pragma: export
#include "google/protobuf/extension_set.h"  // IWYU pragma: export
#include "google/protobuf/unknown_field_set.h"
#include "google/protobuf/timestamp.pb.h"
#include "aclk/v1/lib.pb.h"
// @@protoc_insertion_point(includes)

// Must be included last.
#include "google/protobuf/port_def.inc"

#define PROTOBUF_INTERNAL_EXPORT_nodeinstance_2fconnection_2fv1_2fconnection_2eproto

namespace google {
namespace protobuf {
namespace internal {
template <typename T>
::absl::string_view GetAnyMessageName();
}  // namespace internal
}  // namespace protobuf
}  // namespace google

// Internal implementation detail -- do not use these members.
struct TableStruct_nodeinstance_2fconnection_2fv1_2fconnection_2eproto {
  static const ::uint32_t offsets[];
};
extern "C" {
extern const ::google::protobuf::internal::DescriptorTable descriptor_table_nodeinstance_2fconnection_2fv1_2fconnection_2eproto;
}  // extern "C"
namespace nodeinstance {
namespace v1 {
class UpdateNodeInstanceConnection;
struct UpdateNodeInstanceConnectionGlobalsTypeInternal;
#ifndef PROTOBUF_MESSAGE_GLOBALS
extern UpdateNodeInstanceConnectionGlobalsTypeInternal UpdateNodeInstanceConnection_globals_;
extern const ::google::protobuf::internal::ClassDataFull UpdateNodeInstanceConnection_class_data_;
#else
extern const UpdateNodeInstanceConnectionGlobalsTypeInternal UpdateNodeInstanceConnection_globals_;
#endif  // PROTOBUF_MESSAGE_GLOBALS
}  // namespace v1
}  // namespace nodeinstance
namespace google {
namespace protobuf {
}  // namespace protobuf
}  // namespace google

namespace nodeinstance {
namespace v1 {

// ===================================================================


// -------------------------------------------------------------------

class  PROTOBUF_FUTURE_ADD_EARLY_WARN_UNUSED UpdateNodeInstanceConnection final : public ::google::protobuf::Message
/* @@protoc_insertion_point(class_definition:nodeinstance.v1.UpdateNodeInstanceConnection) */ {
 public:
  inline UpdateNodeInstanceConnection() : UpdateNodeInstanceConnection(nullptr) {}
  ~UpdateNodeInstanceConnection() PROTOBUF_FINAL;

#if defined(PROTOBUF_CUSTOM_VTABLE)
  void operator delete(UpdateNodeInstanceConnection* PROTOBUF_NONNULL msg, ::std::destroying_delete_t) {
    SharedDtor(*msg);
    ::google::protobuf::internal::SizedDelete(msg, sizeof(UpdateNodeInstanceConnection));
  }
#endif

  template <typename = void>
  explicit constexpr UpdateNodeInstanceConnection(::google::protobuf::internal::ConstantInitialized,
                           const ::google::protobuf::internal::ClassData* PROTOBUF_NONNULL
                               class_data);

  inline UpdateNodeInstanceConnection(const UpdateNodeInstanceConnection& from) : UpdateNodeInstanceConnection(nullptr, from) {}
  inline UpdateNodeInstanceConnection(UpdateNodeInstanceConnection&& from) noexcept : UpdateNodeInstanceConnection(nullptr, ::std::move(from)) {}
  inline UpdateNodeInstanceConnection& operator=(const UpdateNodeInstanceConnection& from) {
    CopyFrom(from);
    return *this;
  }
  inline UpdateNodeInstanceConnection& operator=(UpdateNodeInstanceConnection&& from) noexcept {
    if (this == &from) return *this;
    if (::google::protobuf::internal::CanMoveWithInternalSwap(GetArena(), from.GetArena())) {
      InternalSwap(&from);
    } else {
      CopyFrom(from);
    }
    return *this;
  }

  [[nodiscard]] inline const ::google::protobuf::UnknownFieldSet& unknown_fields() const
      ABSL_ATTRIBUTE_LIFETIME_BOUND {
    return _internal_metadata_.unknown_fields<::google::protobuf::UnknownFieldSet>(::google::protobuf::UnknownFieldSet::default_instance);
  }
  [[nodiscard]] inline ::google::protobuf::UnknownFieldSet* PROTOBUF_NONNULL
  mutable_unknown_fields() ABSL_ATTRIBUTE_LIFETIME_BOUND {
    return _internal_metadata_.mutable_unknown_fields<::google::protobuf::UnknownFieldSet>();
  }

  [[nodiscard]] static const ::google::protobuf::Descriptor* PROTOBUF_NONNULL descriptor() {
    return GetDescriptor();
  }
  [[nodiscard]] static const ::google::protobuf::Descriptor* PROTOBUF_NONNULL
  GetDescriptor() {
    return default_instance().GetMetadata().descriptor;
  }
  [[nodiscard]] static const ::google::protobuf::Reflection* PROTOBUF_NONNULL GetReflection() {
    return default_instance().GetMetadata().reflection;
  }
  [[nodiscard]] static const UpdateNodeInstanceConnection& default_instance() {
    return *::google::protobuf::internal::MessageGlobalsBase::ToDefaultInstance<UpdateNodeInstanceConnection>(&UpdateNodeInstanceConnection_globals_);
  }
  static constexpr int kIndexInFileMessages = 0;
  friend void swap(UpdateNodeInstanceConnection& a, UpdateNodeInstanceConnection& b) { a.Swap(&b); }
  inline void Swap(UpdateNodeInstanceConnection* PROTOBUF_NONNULL other) {
    if (other == this) return;
    if (::google::protobuf::internal::CanUseInternalSwap(GetArena(), other->GetArena())) {
      InternalSwap(other);
    } else {
      ::google::protobuf::internal::GenericSwap(this, other);
    }
  }
  void UnsafeArenaSwap(UpdateNodeInstanceConnection* PROTOBUF_NONNULL other) {
    if (other == this) return;
    ABSL_DCHECK(GetArena() == other->GetArena());
    InternalSwap(other);
  }

  // implements Message ----------------------------------------------

  [[nodiscard]] UpdateNodeInstanceConnection* PROTOBUF_NONNULL
  New(::google::protobuf::Arena* PROTOBUF_NULLABLE arena = nullptr) const {
    return ::google::protobuf::Message::DefaultConstruct<UpdateNodeInstanceConnection>(arena);
  }
  using ::google::protobuf::Message::CopyFrom;
  void CopyFrom(const UpdateNodeInstanceConnection& from);
  using ::google::protobuf::Message::MergeFrom;
  void MergeFrom(const UpdateNodeInstanceConnection& from) { UpdateNodeInstanceConnection::MergeImpl(*this, from); }

  private:
  static void MergeImpl(::google::protobuf::MessageLite& to_msg,
                        const ::google::protobuf::MessageLite& from_msg);

  public:
  [[nodiscard]] bool IsInitialized() const {
    return true;
  }
  ABSL_ATTRIBUTE_REINITIALIZES void Clear() PROTOBUF_FINAL;
  #if defined(PROTOBUF_CUSTOM_VTABLE)
  private:
  [[nodiscard]] static ::size_t ByteSizeLong(const ::google::protobuf::MessageLite& msg);
  [[nodiscard]] static ::uint8_t* PROTOBUF_NONNULL _InternalSerialize(
      const ::google::protobuf::MessageLite& msg, ::uint8_t* PROTOBUF_NONNULL target,
      ::google::protobuf::io::EpsCopyOutputStream* PROTOBUF_NONNULL stream);

  public:
  [[nodiscard]] ::size_t ByteSizeLong() const { return ByteSizeLong(*this); }
  [[nodiscard]] ::uint8_t* PROTOBUF_NONNULL _InternalSerialize(
      ::uint8_t* PROTOBUF_NONNULL target,
      ::google::protobuf::io::EpsCopyOutputStream* PROTOBUF_NONNULL stream) const {
    return _InternalSerialize(*this, target, stream);
  }
  #else   // PROTOBUF_CUSTOM_VTABLE
  [[nodiscard]] ::size_t ByteSizeLong() const final;
  [[nodiscard]] ::uint8_t* PROTOBUF_NONNULL _InternalSerialize(
      ::uint8_t* PROTOBUF_NONNULL target,
      ::google::protobuf::io::EpsCopyOutputStream* PROTOBUF_NONNULL stream) const final;
  #endif  // PROTOBUF_CUSTOM_VTABLE
  [[nodiscard]] int GetCachedSize() const {
    return _impl_._cached_size_.Get();
  }

  private:
  void SharedCtor(::google::protobuf::Arena* PROTOBUF_NULLABLE arena);
  static void SharedDtor(MessageLite& self);
  void InternalSwap(UpdateNodeInstanceConnection* PROTOBUF_NONNULL other);
 private:
  template <typename T>
  friend ::absl::string_view(::google::protobuf::internal::GetAnyMessageName)();
  static ::absl::string_view FullMessageName() { return "nodeinstance.v1.UpdateNodeInstanceConnection"; }

  explicit UpdateNodeInstanceConnection(::google::protobuf::Arena* PROTOBUF_NULLABLE arena);
  UpdateNodeInstanceConnection(::google::protobuf::Arena* PROTOBUF_NULLABLE arena, const UpdateNodeInstanceConnection& from);
  UpdateNodeInstanceConnection(
      ::google::protobuf::Arena* PROTOBUF_NULLABLE arena, UpdateNodeInstanceConnection&& from) noexcept
      : UpdateNodeInstanceConnection(arena) {
    *this = ::std::move(from);
  }
  const ::google::protobuf::internal::ClassData* PROTOBUF_NONNULL GetClassData() const PROTOBUF_FINAL;
  static void* PROTOBUF_NONNULL PlacementNew_(
      const void* PROTOBUF_NONNULL, void* PROTOBUF_NONNULL mem,
      ::google::protobuf::Arena* PROTOBUF_NULLABLE arena);
  static constexpr auto InternalNewImpl_();

 public:
  static constexpr auto InternalGenerateClassData_(
      const MessageLite& prototype,
      const ::google::protobuf::internal::TcParseTableBase* PROTOBUF_NULLABLE tc_table = nullptr);

  [[nodiscard]] ::google::protobuf::Metadata GetMetadata() const;
  // nested types ----------------------------------------------------

  // accessors -------------------------------------------------------
  enum : int {
    kCapabilitiesFieldNumber = 10,
    kClaimIdFieldNumber = 1,
    kNodeIdFieldNumber = 2,
    kMqttBrokerAddrFieldNumber = 7,
    kUpdatedAtFieldNumber = 6,
    kSessionIdFieldNumber = 5,
    kLivenessFieldNumber = 3,
    kQueryableFieldNumber = 4,
    kVmqInstanceIdFieldNumber = 8,
    kHopsFieldNumber = 9,
  };
  // repeated .aclk_lib.v1.Capability capabilities = 10;
  [[nodiscard]] int capabilities_size()
      const;
  private:
  int _internal_capabilities_size() const;

  public:
  void clear_capabilities() ;
  [[nodiscard]] const ::aclk_lib::v1::Capability& capabilities(int index) const;
  [[nodiscard]] ::aclk_lib::v1::Capability* PROTOBUF_NONNULL mutable_capabilities(int index);
  ::aclk_lib::v1::Capability* PROTOBUF_NONNULL add_capabilities();
  [[nodiscard]] const ::google::protobuf::RepeatedPtrField<::aclk_lib::v1::Capability>&
  capabilities() const;
  [[nodiscard]] ::google::protobuf::RepeatedPtrField<::aclk_lib::v1::Capability>* PROTOBUF_NONNULL
  mutable_capabilities();

  private:
  const ::google::protobuf::RepeatedPtrField<::aclk_lib::v1::Capability>& _internal_capabilities() const;
  ::google::protobuf::RepeatedPtrField<::aclk_lib::v1::Capability>* PROTOBUF_NONNULL _internal_mutable_capabilities();

  public:
  // string claim_id = 1;
  void clear_claim_id() ;
  [[nodiscard]] const ::std::string& claim_id() const;
  template <typename Arg_ = const ::std::string&, typename... Args_>
  void set_claim_id(Arg_&& arg, Args_... args);
  ::std::string* PROTOBUF_NONNULL mutable_claim_id();
  [[nodiscard]] ::std::string* PROTOBUF_NULLABLE release_claim_id();
  void set_allocated_claim_id(::std::string* PROTOBUF_NULLABLE value);

  private:
  const ::std::string& _internal_claim_id() const;
  PROTOBUF_ALWAYS_INLINE void _internal_set_claim_id(const ::std::string& value);
  ::std::string* PROTOBUF_NONNULL _internal_mutable_claim_id();

  public:
  // string node_id = 2;
  void clear_node_id() ;
  [[nodiscard]] const ::std::string& node_id() const;
  template <typename Arg_ = const ::std::string&, typename... Args_>
  void set_node_id(Arg_&& arg, Args_... args);
  ::std::string* PROTOBUF_NONNULL mutable_node_id();
  [[nodiscard]] ::std::string* PROTOBUF_NULLABLE release_node_id();
  void set_allocated_node_id(::std::string* PROTOBUF_NULLABLE value);

  private:
  const ::std::string& _internal_node_id() const;
  PROTOBUF_ALWAYS_INLINE void _internal_set_node_id(const ::std::string& value);
  ::std::string* PROTOBUF_NONNULL _internal_mutable_node_id();

  public:
  // string mqtt_broker_addr = 7;
  void clear_mqtt_broker_addr() ;
  [[nodiscard]] const ::std::string& mqtt_broker_addr() const;
  template <typename Arg_ = const ::std::string&, typename... Args_>
  void set_mqtt_broker_addr(Arg_&& arg, Args_... args);
  ::std::string* PROTOBUF_NONNULL mutable_mqtt_broker_addr();
  [[nodiscard]] ::std::string* PROTOBUF_NULLABLE release_mqtt_broker_addr();
  void set_allocated_mqtt_broker_addr(::std::string* PROTOBUF_NULLABLE value);

  private:
  const ::std::string& _internal_mqtt_broker_addr() const;
  PROTOBUF_ALWAYS_INLINE void _internal_set_mqtt_broker_addr(const ::std::string& value);
  ::std::string* PROTOBUF_NONNULL _internal_mutable_mqtt_broker_addr();

  public:
  // .google.protobuf.Timestamp updated_at = 6;
  [[nodiscard]] bool has_updated_at()
      const;
  void clear_updated_at() ;
  [[nodiscard]] const ::google::protobuf::Timestamp& updated_at() const;
  [[nodiscard]] ::google::protobuf::Timestamp* PROTOBUF_NULLABLE release_updated_at();
  ::google::protobuf::Timestamp* PROTOBUF_NONNULL mutable_updated_at();
  void set_allocated_updated_at(::google::protobuf::Timestamp* PROTOBUF_NULLABLE value);
  void unsafe_arena_set_allocated_updated_at(::google::protobuf::Timestamp* PROTOBUF_NULLABLE value);
  ::google::protobuf::Timestamp* PROTOBUF_NULLABLE unsafe_arena_release_updated_at();

  private:
  const ::google::protobuf::Timestamp& _internal_updated_at() const;
  ::google::protobuf::Timestamp* PROTOBUF_NONNULL _internal_mutable_updated_at();

  public:
  // int64 session_id = 5;
  void clear_session_id() ;
  [[nodiscard]] ::int64_t session_id() const;
  void set_session_id(::int64_t value);

  private:
  ::int64_t _internal_session_id() const;
  void _internal_set_session_id(::int64_t value);

  public:
  // bool liveness = 3;
  void clear_liveness() ;
  [[nodiscard]] bool liveness() const;
  void set_liveness(bool value);

  private:
  bool _internal_liveness() const;
  void _internal_set_liveness(bool value);

  public:
  // bool queryable = 4;
  void clear_queryable() ;
  [[nodiscard]] bool queryable() const;
  void set_queryable(bool value);

  private:
  bool _internal_queryable() const;
  void _internal_set_queryable(bool value);

  public:
  // int32 vmq_instance_id = 8;
  void clear_vmq_instance_id() ;
  [[nodiscard]] ::int32_t vmq_instance_id() const;
  void set_vmq_instance_id(::int32_t value);

  private:
  ::int32_t _internal_vmq_instance_id() const;
  void _internal_set_vmq_instance_id(::int32_t value);

  public:
  // int32 hops = 9;
  void clear_hops() ;
  [[nodiscard]] ::int32_t hops() const;
  void set_hops(::int32_t value);

  private:
  ::int32_t _internal_hops() const;
  void _internal_set_hops(::int32_t value);

  public:
  // @@protoc_insertion_point(class_scope:nodeinstance.v1.UpdateNodeInstanceConnection)
 private:
  class _Internal;
  using ParseTableT_ =
      ::google::protobuf::internal::TcParseTable<4, 10,
                          2, 92,
                          2>;
  static constexpr ParseTableT_ InternalGenerateParseTable_(
      const ::google::protobuf::internal::ClassData* PROTOBUF_NONNULL class_data);
  friend class ::google::protobuf::internal::TcParser;
  #ifndef PROTOBUF_MESSAGE_GLOBALS
  static const ParseTableT_ _table_;
  #endif

  friend class ::google::protobuf::MessageLite;
  friend class ::google::protobuf::Arena;
  friend ::google::protobuf::internal::PrivateAccess;
  template <typename T>
  friend class ::google::protobuf::Arena::InternalHelper;
  using InternalArenaConstructable_ = void;
  using DestructorSkippable_ = void;
  struct Impl_ {
    inline explicit constexpr Impl_(::google::protobuf::internal::InternalVisibility visibility,
                                    ::google::protobuf::internal::ConstantInitialized) noexcept;
    inline explicit Impl_(
        ::google::protobuf::internal::InternalVisibility visibility,
        ::google::protobuf::Arena* PROTOBUF_NULLABLE arena);
    inline explicit Impl_(
        ::google::protobuf::internal::InternalVisibility visibility,
        ::google::protobuf::Arena* PROTOBUF_NULLABLE arena, const Impl_& from,
        const UpdateNodeInstanceConnection& from_msg);
    ::google::protobuf::internal::HasBits<1> _has_bits_;
    ::google::protobuf::internal::CachedSize _cached_size_;
    ::google::protobuf::RepeatedPtrField< ::aclk_lib::v1::Capability > capabilities_;
    ::google::protobuf::internal::ArenaStringPtr claim_id_;
    ::google::protobuf::internal::ArenaStringPtr node_id_;
    ::google::protobuf::internal::ArenaStringPtr mqtt_broker_addr_;
    ::google::protobuf::Timestamp* PROTOBUF_NULLABLE updated_at_;
    ::int64_t session_id_;
    bool liveness_;
    bool queryable_;
    ::int32_t vmq_instance_id_;
    ::int32_t hops_;
    PROTOBUF_TSAN_DECLARE_MEMBER
  };
  union { Impl_ _impl_; };
  friend struct ::TableStruct_nodeinstance_2fconnection_2fv1_2fconnection_2eproto;
};

// ===================================================================




// ===================================================================


#ifdef __GNUC__
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wstrict-aliasing"
#endif  // __GNUC__
// -------------------------------------------------------------------

// UpdateNodeInstanceConnection

// string claim_id = 1;
inline void UpdateNodeInstanceConnection::clear_claim_id() {
  ::google::protobuf::internal::TSanWrite(&_impl_);
  _impl_.claim_id_.ClearToEmpty();
  ClearHasBit(_impl_._has_bits_[0], 0x00000002U);
}
inline const ::std::string& UpdateNodeInstanceConnection::claim_id() const
    ABSL_ATTRIBUTE_LIFETIME_BOUND {
  // @@protoc_insertion_point(field_get:nodeinstance.v1.UpdateNodeInstanceConnection.claim_id)
  return _internal_claim_id();
}
template <typename Arg_, typename... Args_>
PROTOBUF_ALWAYS_INLINE void UpdateNodeInstanceConnection::set_claim_id(Arg_&& arg, Args_... args) {
  ::google::protobuf::internal::TSanWrite(&_impl_);
  SetHasBit(_impl_._has_bits_[0], 0x00000002U);
  _impl_.claim_id_.Set(static_cast<Arg_&&>(arg), args..., GetArena());
  // @@protoc_insertion_point(field_set:nodeinstance.v1.UpdateNodeInstanceConnection.claim_id)
}
inline ::std::string* PROTOBUF_NONNULL UpdateNodeInstanceConnection::mutable_claim_id()
    ABSL_ATTRIBUTE_LIFETIME_BOUND {
  SetHasBit(_impl_._has_bits_[0], 0x00000002U);
  ::std::string* _s = _internal_mutable_claim_id();
  // @@protoc_insertion_point(field_mutable:nodeinstance.v1.UpdateNodeInstanceConnection.claim_id)
  return _s;
}
inline const ::std::string& UpdateNodeInstanceConnection::_internal_claim_id() const {
  ::google::protobuf::internal::TSanRead(&_impl_);
  return _impl_.claim_id_.Get();
}
inline void UpdateNodeInstanceConnection::_internal_set_claim_id(const ::std::string& value) {
  ::google::protobuf::internal::TSanWrite(&_impl_);
  _impl_.claim_id_.Set(value, GetArena());
}
inline ::std::string* PROTOBUF_NONNULL UpdateNodeInstanceConnection::_internal_mutable_claim_id() {
  ::google::protobuf::internal::TSanWrite(&_impl_);
  return _impl_.claim_id_.Mutable( GetArena());
}
inline ::std::string* PROTOBUF_NULLABLE UpdateNodeInstanceConnection::release_claim_id() {
  ::google::protobuf::internal::TSanWrite(&_impl_);
  // @@protoc_insertion_point(field_release:nodeinstance.v1.UpdateNodeInstanceConnection.claim_id)
  if (!CheckHasBit(_impl_._has_bits_[0], 0x00000002U)) {
    return nullptr;
  }
  ClearHasBit(_impl_._has_bits_[0], 0x00000002U);
  auto* released = _impl_.claim_id_.Release();
  if (::google::protobuf::internal::DebugHardenForceCopyDefaultString()) {
    _impl_.claim_id_.Set("", GetArena());
  }
  return released;
}
inline void UpdateNodeInstanceConnection::set_allocated_claim_id(::std::string* PROTOBUF_NULLABLE value) {
  ::google::protobuf::internal::TSanWrite(&_impl_);
  if (value != nullptr) {
    SetHasBit(_impl_._has_bits_[0], 0x00000002U);
  } else {
    ClearHasBit(_impl_._has_bits_[0], 0x00000002U);
  }
  _impl_.claim_id_.SetAllocated(value, GetArena());
  if (::google::protobuf::internal::DebugHardenForceCopyDefaultString() && _impl_.claim_id_.IsDefault()) {
    _impl_.claim_id_.Set("", GetArena());
  }
  // @@protoc_insertion_point(field_set_allocated:nodeinstance.v1.UpdateNodeInstanceConnection.claim_id)
}

// string node_id = 2;
inline void UpdateNodeInstanceConnection::clear_node_id() {
  ::google::protobuf::internal::TSanWrite(&_impl_);
  _impl_.node_id_.ClearToEmpty();
  ClearHasBit(_impl_._has_bits_[0], 0x00000004U);
}
inline const ::std::string& UpdateNodeInstanceConnection::node_id() const
    ABSL_ATTRIBUTE_LIFETIME_BOUND {
  // @@protoc_insertion_point(field_get:nodeinstance.v1.UpdateNodeInstanceConnection.node_id)
  return _internal_node_id();
}
template <typename Arg_, typename... Args_>
PROTOBUF_ALWAYS_INLINE void UpdateNodeInstanceConnection::set_node_id(Arg_&& arg, Args_... args) {
  ::google::protobuf::internal::TSanWrite(&_impl_);
  SetHasBit(_impl_._has_bits_[0], 0x00000004U);
  _impl_.node_id_.Set(static_cast<Arg_&&>(arg), args..., GetArena());
  // @@protoc_insertion_point(field_set:nodeinstance.v1.UpdateNodeInstanceConnection.node_id)
}
inline ::std::string* PROTOBUF_NONNULL UpdateNodeInstanceConnection::mutable_node_id()
    ABSL_ATTRIBUTE_LIFETIME_BOUND {
  SetHasBit(_impl_._has_bits_[0], 0x00000004U);
  ::std::string* _s = _internal_mutable_node_id();
  // @@protoc_insertion_point(field_mutable:nodeinstance.v1.UpdateNodeInstanceConnection.node_id)
  return _s;
}
inline const ::std::string& UpdateNodeInstanceConnection::_internal_node_id() const {
  ::google::protobuf::internal::TSanRead(&_impl_);
  return _impl_.node_id_.Get();
}
inline void UpdateNodeInstanceConnection::_internal_set_node_id(const ::std::string& value) {
  ::google::protobuf::internal::TSanWrite(&_impl_);
  _impl_.node_id_.Set(value, GetArena());
}
inline ::std::string* PROTOBUF_NONNULL UpdateNodeInstanceConnection::_internal_mutable_node_id() {
  ::google::protobuf::internal::TSanWrite(&_impl_);
  return _impl_.node_id_.Mutable( GetArena());
}
inline ::std::string* PROTOBUF_NULLABLE UpdateNodeInstanceConnection::release_node_id() {
  ::google::protobuf::internal::TSanWrite(&_impl_);
  // @@protoc_insertion_point(field_release:nodeinstance.v1.UpdateNodeInstanceConnection.node_id)
  if (!CheckHasBit(_impl_._has_bits_[0], 0x00000004U)) {
    return nullptr;
  }
  ClearHasBit(_impl_._has_bits_[0], 0x00000004U);
  auto* released = _impl_.node_id_.Release();
  if (::google::protobuf::internal::DebugHardenForceCopyDefaultString()) {
    _impl_.node_id_.Set("", GetArena());
  }
  return released;
}
inline void UpdateNodeInstanceConnection::set_allocated_node_id(::std::string* PROTOBUF_NULLABLE value) {
  ::google::protobuf::internal::TSanWrite(&_impl_);
  if (value != nullptr) {
    SetHasBit(_impl_._has_bits_[0], 0x00000004U);
  } else {
    ClearHasBit(_impl_._has_bits_[0], 0x00000004U);
  }
  _impl_.node_id_.SetAllocated(value, GetArena());
  if (::google::protobuf::internal::DebugHardenForceCopyDefaultString() && _impl_.node_id_.IsDefault()) {
    _impl_.node_id_.Set("", GetArena());
  }
  // @@protoc_insertion_point(field_set_allocated:nodeinstance.v1.UpdateNodeInstanceConnection.node_id)
}

// bool liveness = 3;
inline void UpdateNodeInstanceConnection::clear_liveness() {
  ::google::protobuf::internal::TSanWrite(&_impl_);
  _impl_.liveness_ = false;
  ClearHasBit(_impl_._has_bits_[0], 0x00000040U);
}
inline bool UpdateNodeInstanceConnection::liveness() const {
  // @@protoc_insertion_point(field_get:nodeinstance.v1.UpdateNodeInstanceConnection.liveness)
  return _internal_liveness();
}
inline void UpdateNodeInstanceConnection::set_liveness(bool value) {
  _internal_set_liveness(value);
  SetHasBit(_impl_._has_bits_[0], 0x00000040U);
  // @@protoc_insertion_point(field_set:nodeinstance.v1.UpdateNodeInstanceConnection.liveness)
}
inline bool UpdateNodeInstanceConnection::_internal_liveness() const {
  ::google::protobuf::internal::TSanRead(&_impl_);
  return _impl_.liveness_;
}
inline void UpdateNodeInstanceConnection::_internal_set_liveness(bool value) {
  ::google::protobuf::internal::TSanWrite(&_impl_);
  _impl_.liveness_ = value;
}

// bool queryable = 4;
inline void UpdateNodeInstanceConnection::clear_queryable() {
  ::google::protobuf::internal::TSanWrite(&_impl_);
  _impl_.queryable_ = false;
  ClearHasBit(_impl_._has_bits_[0], 0x00000080U);
}
inline bool UpdateNodeInstanceConnection::queryable() const {
  // @@protoc_insertion_point(field_get:nodeinstance.v1.UpdateNodeInstanceConnection.queryable)
  return _internal_queryable();
}
inline void UpdateNodeInstanceConnection::set_queryable(bool value) {
  _internal_set_queryable(value);
  SetHasBit(_impl_._has_bits_[0], 0x00000080U);
  // @@protoc_insertion_point(field_set:nodeinstance.v1.UpdateNodeInstanceConnection.queryable)
}
inline bool UpdateNodeInstanceConnection::_internal_queryable() const {
  ::google::protobuf::internal::TSanRead(&_impl_);
  return _impl_.queryable_;
}
inline void UpdateNodeInstanceConnection::_internal_set_queryable(bool value) {
  ::google::protobuf::internal::TSanWrite(&_impl_);
  _impl_.queryable_ = value;
}

// int64 session_id = 5;
inline void UpdateNodeInstanceConnection::clear_session_id() {
  ::google::protobuf::internal::TSanWrite(&_impl_);
  _impl_.session_id_ = ::int64_t{0};
  ClearHasBit(_impl_._has_bits_[0], 0x00000020U);
}
inline ::int64_t UpdateNodeInstanceConnection::session_id() const {
  // @@protoc_insertion_point(field_get:nodeinstance.v1.UpdateNodeInstanceConnection.session_id)
  return _internal_session_id();
}
inline void UpdateNodeInstanceConnection::set_session_id(::int64_t value) {
  _internal_set_session_id(value);
  SetHasBit(_impl_._has_bits_[0], 0x00000020U);
  // @@protoc_insertion_point(field_set:nodeinstance.v1.UpdateNodeInstanceConnection.session_id)
}
inline ::int64_t UpdateNodeInstanceConnection::_internal_session_id() const {
  ::google::protobuf::internal::TSanRead(&_impl_);
  return _impl_.session_id_;
}
inline void UpdateNodeInstanceConnection::_internal_set_session_id(::int64_t value) {
  ::google::protobuf::internal::TSanWrite(&_impl_);
  _impl_.session_id_ = value;
}

// .google.protobuf.Timestamp updated_at = 6;
inline bool UpdateNodeInstanceConnection::has_updated_at() const {
  bool value = CheckHasBit(_impl_._has_bits_[0], 0x00000010U);
  PROTOBUF_ASSUME(!value || _impl_.updated_at_ != nullptr);
  return value;
}
inline const ::google::protobuf::Timestamp& UpdateNodeInstanceConnection::_internal_updated_at() const {
  ::google::protobuf::internal::TSanRead(&_impl_);
  const ::google::protobuf::Timestamp* p = _impl_.updated_at_;
  return p != nullptr ? *p : *::google::protobuf::internal::MessageGlobalsBase::ToDefaultInstance<::google::protobuf::Timestamp>(&::google::protobuf::Timestamp_globals_);
}
inline const ::google::protobuf::Timestamp& UpdateNodeInstanceConnection::updated_at() const ABSL_ATTRIBUTE_LIFETIME_BOUND {
  // @@protoc_insertion_point(field_get:nodeinstance.v1.UpdateNodeInstanceConnection.updated_at)
  return _internal_updated_at();
}
inline void UpdateNodeInstanceConnection::unsafe_arena_set_allocated_updated_at(
    ::google::protobuf::Timestamp* PROTOBUF_NULLABLE value) {
  ::google::protobuf::internal::TSanWrite(&_impl_);
  if (GetArena() == nullptr) {
    delete reinterpret_cast<::google::protobuf::MessageLite*>(_impl_.updated_at_);
  }
  _impl_.updated_at_ = reinterpret_cast<::google::protobuf::Timestamp*>(value);
  if (value != nullptr) {
    SetHasBit(_impl_._has_bits_[0], 0x00000010U);
  } else {
    ClearHasBit(_impl_._has_bits_[0], 0x00000010U);
  }
  // @@protoc_insertion_point(field_unsafe_arena_set_allocated:nodeinstance.v1.UpdateNodeInstanceConnection.updated_at)
}
inline ::google::protobuf::Timestamp* PROTOBUF_NULLABLE UpdateNodeInstanceConnection::release_updated_at() {
  ::google::protobuf::internal::TSanWrite(&_impl_);

  ClearHasBit(_impl_._has_bits_[0], 0x00000010U);
  ::google::protobuf::Timestamp* released = _impl_.updated_at_;
  _impl_.updated_at_ = nullptr;
  if (::google::protobuf::internal::DebugHardenForceCopyInRelease()) {
    auto* old = reinterpret_cast<::google::protobuf::MessageLite*>(released);
    released = ::google::protobuf::internal::DuplicateIfNonNull(released);
    if (GetArena() == nullptr) {
      delete old;
    }
  } else {
    if (GetArena() != nullptr) {
      released = ::google::protobuf::internal::DuplicateIfNonNull(released);
    }
  }
  return released;
}
inline ::google::protobuf::Timestamp* PROTOBUF_NULLABLE UpdateNodeInstanceConnection::unsafe_arena_release_updated_at() {
  ::google::protobuf::internal::TSanWrite(&_impl_);
  // @@protoc_insertion_point(field_release:nodeinstance.v1.UpdateNodeInstanceConnection.updated_at)

  ClearHasBit(_impl_._has_bits_[0], 0x00000010U);
  ::google::protobuf::Timestamp* temp = _impl_.updated_at_;
  _impl_.updated_at_ = nullptr;
  return temp;
}
inline ::google::protobuf::Timestamp* PROTOBUF_NONNULL UpdateNodeInstanceConnection::_internal_mutable_updated_at() {
  ::google::protobuf::internal::TSanWrite(&_impl_);
  if (_impl_.updated_at_ == nullptr) {
    auto* p = ::google::protobuf::Message::DefaultConstruct<::google::protobuf::Timestamp>(GetArena());
    _impl_.updated_at_ = reinterpret_cast<::google::protobuf::Timestamp*>(p);
  }
  return _impl_.updated_at_;
}
inline ::google::protobuf::Timestamp* PROTOBUF_NONNULL UpdateNodeInstanceConnection::mutable_updated_at()
    ABSL_ATTRIBUTE_LIFETIME_BOUND {
  SetHasBit(_impl_._has_bits_[0], 0x00000010U);
  ::google::protobuf::Timestamp* _msg = _internal_mutable_updated_at();
  // @@protoc_insertion_point(field_mutable:nodeinstance.v1.UpdateNodeInstanceConnection.updated_at)
  return _msg;
}
inline void UpdateNodeInstanceConnection::set_allocated_updated_at(::google::protobuf::Timestamp* PROTOBUF_NULLABLE value) {
  ::google::protobuf::Arena* message_arena = GetArena();
  ::google::protobuf::internal::TSanWrite(&_impl_);
  if (message_arena == nullptr) {
    delete reinterpret_cast<::google::protobuf::MessageLite*>(_impl_.updated_at_);
  }

  if (value != nullptr) {
    ::google::protobuf::Arena* submessage_arena = reinterpret_cast<::google::protobuf::Message*>(value)->GetArena();
    if (message_arena != submessage_arena) {
      value = ::google::protobuf::internal::GetOwnedMessage(message_arena, value, submessage_arena);
    }
    SetHasBit(_impl_._has_bits_[0], 0x00000010U);
  } else {
    ClearHasBit(_impl_._has_bits_[0], 0x00000010U);
  }

  _impl_.updated_at_ = reinterpret_cast<::google::protobuf::Timestamp*>(value);
  // @@protoc_insertion_point(field_set_allocated:nodeinstance.v1.UpdateNodeInstanceConnection.updated_at)
}

// string mqtt_broker_addr = 7;
inline void UpdateNodeInstanceConnection::clear_mqtt_broker_addr() {
  ::google::protobuf::internal::TSanWrite(&_impl_);
  _impl_.mqtt_broker_addr_.ClearToEmpty();
  ClearHasBit(_impl_._has_bits_[0], 0x00000008U);
}
inline const ::std::string& UpdateNodeInstanceConnection::mqtt_broker_addr() const
    ABSL_ATTRIBUTE_LIFETIME_BOUND {
  // @@protoc_insertion_point(field_get:nodeinstance.v1.UpdateNodeInstanceConnection.mqtt_broker_addr)
  return _internal_mqtt_broker_addr();
}
template <typename Arg_, typename... Args_>
PROTOBUF_ALWAYS_INLINE void UpdateNodeInstanceConnection::set_mqtt_broker_addr(Arg_&& arg, Args_... args) {
  ::google::protobuf::internal::TSanWrite(&_impl_);
  SetHasBit(_impl_._has_bits_[0], 0x00000008U);
  _impl_.mqtt_broker_addr_.Set(static_cast<Arg_&&>(arg), args..., GetArena());
  // @@protoc_insertion_point(field_set:nodeinstance.v1.UpdateNodeInstanceConnection.mqtt_broker_addr)
}
inline ::std::string* PROTOBUF_NONNULL UpdateNodeInstanceConnection::mutable_mqtt_broker_addr()
    ABSL_ATTRIBUTE_LIFETIME_BOUND {
  SetHasBit(_impl_._has_bits_[0], 0x00000008U);
  ::std::string* _s = _internal_mutable_mqtt_broker_addr();
  // @@protoc_insertion_point(field_mutable:nodeinstance.v1.UpdateNodeInstanceConnection.mqtt_broker_addr)
  return _s;
}
inline const ::std::string& UpdateNodeInstanceConnection::_internal_mqtt_broker_addr() const {
  ::google::protobuf::internal::TSanRead(&_impl_);
  return _impl_.mqtt_broker_addr_.Get();
}
inline void UpdateNodeInstanceConnection::_internal_set_mqtt_broker_addr(const ::std::string& value) {
  ::google::protobuf::internal::TSanWrite(&_impl_);
  _impl_.mqtt_broker_addr_.Set(value, GetArena());
}
inline ::std::string* PROTOBUF_NONNULL UpdateNodeInstanceConnection::_internal_mutable_mqtt_broker_addr() {
  ::google::protobuf::internal::TSanWrite(&_impl_);
  return _impl_.mqtt_broker_addr_.Mutable( GetArena());
}
inline ::std::string* PROTOBUF_NULLABLE UpdateNodeInstanceConnection::release_mqtt_broker_addr() {
  ::google::protobuf::internal::TSanWrite(&_impl_);
  // @@protoc_insertion_point(field_release:nodeinstance.v1.UpdateNodeInstanceConnection.mqtt_broker_addr)
  if (!CheckHasBit(_impl_._has_bits_[0], 0x00000008U)) {
    return nullptr;
  }
  ClearHasBit(_impl_._has_bits_[0], 0x00000008U);
  auto* released = _impl_.mqtt_broker_addr_.Release();
  if (::google::protobuf::internal::DebugHardenForceCopyDefaultString()) {
    _impl_.mqtt_broker_addr_.Set("", GetArena());
  }
  return released;
}
inline void UpdateNodeInstanceConnection::set_allocated_mqtt_broker_addr(::std::string* PROTOBUF_NULLABLE value) {
  ::google::protobuf::internal::TSanWrite(&_impl_);
  if (value != nullptr) {
    SetHasBit(_impl_._has_bits_[0], 0x00000008U);
  } else {
    ClearHasBit(_impl_._has_bits_[0], 0x00000008U);
  }
  _impl_.mqtt_broker_addr_.SetAllocated(value, GetArena());
  if (::google::protobuf::internal::DebugHardenForceCopyDefaultString() && _impl_.mqtt_broker_addr_.IsDefault()) {
    _impl_.mqtt_broker_addr_.Set("", GetArena());
  }
  // @@protoc_insertion_point(field_set_allocated:nodeinstance.v1.UpdateNodeInstanceConnection.mqtt_broker_addr)
}

// int32 vmq_instance_id = 8;
inline void UpdateNodeInstanceConnection::clear_vmq_instance_id() {
  ::google::protobuf::internal::TSanWrite(&_impl_);
  _impl_.vmq_instance_id_ = 0;
  ClearHasBit(_impl_._has_bits_[0], 0x00000100U);
}
inline ::int32_t UpdateNodeInstanceConnection::vmq_instance_id() const {
  // @@protoc_insertion_point(field_get:nodeinstance.v1.UpdateNodeInstanceConnection.vmq_instance_id)
  return _internal_vmq_instance_id();
}
inline void UpdateNodeInstanceConnection::set_vmq_instance_id(::int32_t value) {
  _internal_set_vmq_instance_id(value);
  SetHasBit(_impl_._has_bits_[0], 0x00000100U);
  // @@protoc_insertion_point(field_set:nodeinstance.v1.UpdateNodeInstanceConnection.vmq_instance_id)
}
inline ::int32_t UpdateNodeInstanceConnection::_internal_vmq_instance_id() const {
  ::google::protobuf::internal::TSanRead(&_impl_);
  return _impl_.vmq_instance_id_;
}
inline void UpdateNodeInstanceConnection::_internal_set_vmq_instance_id(::int32_t value) {
  ::google::protobuf::internal::TSanWrite(&_impl_);
  _impl_.vmq_instance_id_ = value;
}

// int32 hops = 9;
inline void UpdateNodeInstanceConnection::clear_hops() {
  ::google::protobuf::internal::TSanWrite(&_impl_);
  _impl_.hops_ = 0;
  ClearHasBit(_impl_._has_bits_[0], 0x00000200U);
}
inline ::int32_t UpdateNodeInstanceConnection::hops() const {
  // @@protoc_insertion_point(field_get:nodeinstance.v1.UpdateNodeInstanceConnection.hops)
  return _internal_hops();
}
inline void UpdateNodeInstanceConnection::set_hops(::int32_t value) {
  _internal_set_hops(value);
  SetHasBit(_impl_._has_bits_[0], 0x00000200U);
  // @@protoc_insertion_point(field_set:nodeinstance.v1.UpdateNodeInstanceConnection.hops)
}
inline ::int32_t UpdateNodeInstanceConnection::_internal_hops() const {
  ::google::protobuf::internal::TSanRead(&_impl_);
  return _impl_.hops_;
}
inline void UpdateNodeInstanceConnection::_internal_set_hops(::int32_t value) {
  ::google::protobuf::internal::TSanWrite(&_impl_);
  _impl_.hops_ = value;
}

// repeated .aclk_lib.v1.Capability capabilities = 10;
inline int UpdateNodeInstanceConnection::_internal_capabilities_size() const {
  return _internal_capabilities().size();
}
inline int UpdateNodeInstanceConnection::capabilities_size() const {
  return _internal_capabilities_size();
}
inline const ::aclk_lib::v1::Capability& UpdateNodeInstanceConnection::capabilities(int index) const
    ABSL_ATTRIBUTE_LIFETIME_BOUND {
  // @@protoc_insertion_point(field_get:nodeinstance.v1.UpdateNodeInstanceConnection.capabilities)
  return _internal_capabilities().Get(index);
}
inline ::aclk_lib::v1::Capability* PROTOBUF_NONNULL UpdateNodeInstanceConnection::mutable_capabilities(int index)
    ABSL_ATTRIBUTE_LIFETIME_BOUND {
  // @@protoc_insertion_point(field_mutable:nodeinstance.v1.UpdateNodeInstanceConnection.capabilities)
  return _internal_mutable_capabilities()->Mutable(index);
}
inline ::aclk_lib::v1::Capability* PROTOBUF_NONNULL UpdateNodeInstanceConnection::add_capabilities()
    ABSL_ATTRIBUTE_LIFETIME_BOUND {
  ::google::protobuf::internal::TSanWrite(&_impl_);
  ::aclk_lib::v1::Capability* _add =
      _internal_mutable_capabilities()->InternalAddWithArena(
          ::google::protobuf::MessageLite::internal_visibility(), GetArena());
  SetHasBit(_impl_._has_bits_[0], 0x00000001U);
  // @@protoc_insertion_point(field_add:nodeinstance.v1.UpdateNodeInstanceConnection.capabilities)
  return _add;
}
inline const ::google::protobuf::RepeatedPtrField<::aclk_lib::v1::Capability>& UpdateNodeInstanceConnection::capabilities() const
    ABSL_ATTRIBUTE_LIFETIME_BOUND {
  // @@protoc_insertion_point(field_list:nodeinstance.v1.UpdateNodeInstanceConnection.capabilities)
  return _internal_capabilities();
}
inline ::google::protobuf::RepeatedPtrField<::aclk_lib::v1::Capability>* PROTOBUF_NONNULL
UpdateNodeInstanceConnection::mutable_capabilities() ABSL_ATTRIBUTE_LIFETIME_BOUND {
  SetHasBit(_impl_._has_bits_[0], 0x00000001U);
  // @@protoc_insertion_point(field_mutable_list:nodeinstance.v1.UpdateNodeInstanceConnection.capabilities)
  ::google::protobuf::internal::TSanWrite(&_impl_);
  return _internal_mutable_capabilities();
}
inline const ::google::protobuf::RepeatedPtrField<::aclk_lib::v1::Capability>&
UpdateNodeInstanceConnection::_internal_capabilities() const {
  ::google::protobuf::internal::TSanRead(&_impl_);
  return _impl_.capabilities_;
}
inline ::google::protobuf::RepeatedPtrField<::aclk_lib::v1::Capability>* PROTOBUF_NONNULL
UpdateNodeInstanceConnection::_internal_mutable_capabilities() {
  ::google::protobuf::internal::TSanRead(&_impl_);
  return &_impl_.capabilities_;
}

#ifdef __GNUC__
#pragma GCC diagnostic pop
#endif  // __GNUC__

// @@protoc_insertion_point(namespace_scope)
}  // namespace v1
}  // namespace nodeinstance


// @@protoc_insertion_point(global_scope)

#include "google/protobuf/port_undef.inc"
// clang-format on

#endif  // nodeinstance_2fconnection_2fv1_2fconnection_2eproto_2epb_2eh
