// Generated by the protocol buffer compiler.  DO NOT EDIT!
// NO CHECKED-IN PROTOBUF GENCODE
// source: context/v1/context.proto
// Protobuf C++ Version: 7.35.1

#include "context/v1/context.pb.h"

#include <algorithm>
#include <type_traits>
#include "google/protobuf/io/coded_stream.h"
#include "google/protobuf/generated_message_tctable_impl.h"
#include "google/protobuf/internal_visibility.h"
#include "google/protobuf/extension_set.h"
#include "google/protobuf/generated_message_util.h"
#include "google/protobuf/wire_format_lite.h"
#include "google/protobuf/descriptor.h"
#include "google/protobuf/generated_message_reflection.h"
#include "google/protobuf/reflection_ops.h"
#include "google/protobuf/wire_format.h"
// @@protoc_insertion_point(includes)

// Must be included last.
#include "google/protobuf/port_def.inc"
PROTOBUF_PRAGMA_INIT_SEG
namespace _pb = ::google::protobuf;
namespace _pbi = ::google::protobuf::internal;
namespace _fl = ::google::protobuf::internal::field_layout;
#ifdef PROTOBUF_MESSAGE_GLOBALS
namespace {
PROTOBUF_CONSTINIT ::google::protobuf::internal::ReflectionData
    file_reflection_data[] = {
        // ::context::v1::ContextsUpdated
        {&::_pbi::kDescriptorMethods, &::descriptor_table_context_2fv1_2fcontext_2eproto, /* tracker*/ nullptr,},
        // ::context::v1::ContextUpdated
        {&::_pbi::kDescriptorMethods, &::descriptor_table_context_2fv1_2fcontext_2eproto, /* tracker*/ nullptr,},
        // ::context::v1::ContextsSnapshot
        {&::_pbi::kDescriptorMethods, &::descriptor_table_context_2fv1_2fcontext_2eproto, /* tracker*/ nullptr,},
};
}  // namespace
#endif
namespace context {
namespace v1 {
class ContextUpdated::_Internal {
 public:
  using HasBits = decltype(::std::declval<ContextUpdated>()._impl_._has_bits_);
  static constexpr ::int32_t kHasBitsOffset =
      8 * PROTOBUF_FIELD_OFFSET(ContextUpdated, _impl_._has_bits_);
};

constexpr ContextUpdated::ParseTableT_ ContextUpdated::InternalGenerateParseTable_(const ::_pbi::ClassData* class_data) {
  return ParseTableT_{
    {
      PROTOBUF_FIELD_OFFSET(ContextUpdated, _impl_._has_bits_),
      0, // no _extensions_
      10, 120,  // max_field_number, fast_idx_mask
      offsetof(ParseTableT_, field_lookup_table),
      4294966272,  // skipmap
      offsetof(ParseTableT_, field_entries),
      10,  // num_field_entries
      0,  // num_aux_entries
      offsetof(ParseTableT_, field_names),  // no aux_entries
      class_data,
      nullptr,  // post_loop_handler
      ::_pbi::TcParser::GenericFallback,  // fallback
      #ifdef PROTOBUF_PREFETCH_PARSE_TABLE
      ::_pbi::TcParser::GetTable<::context::v1::ContextUpdated>(),  // to_prefetch
      #endif  // PROTOBUF_PREFETCH_PARSE_TABLE
    }, {{
      {::_pbi::TcParser::MiniParse, {}},
      // string id = 1;
      {::_pbi::TcParser::FastUS1,
       {10, 0, 0,
        PROTOBUF_FIELD_OFFSET(ContextUpdated, _impl_.id_)}},
      // uint64 version = 2;
      {::_pbi::TcParser::SingularVarintNoZag1<::uint64_t, offsetof(ContextUpdated, _impl_.version_), 5>(),
       {16, 5, 0,
        PROTOBUF_FIELD_OFFSET(ContextUpdated, _impl_.version_)}},
      // uint64 first_entry = 3;
      {::_pbi::TcParser::SingularVarintNoZag1<::uint64_t, offsetof(ContextUpdated, _impl_.first_entry_), 6>(),
       {24, 6, 0,
        PROTOBUF_FIELD_OFFSET(ContextUpdated, _impl_.first_entry_)}},
      // uint64 last_entry = 4;
      {::_pbi::TcParser::SingularVarintNoZag1<::uint64_t, offsetof(ContextUpdated, _impl_.last_entry_), 7>(),
       {32, 7, 0,
        PROTOBUF_FIELD_OFFSET(ContextUpdated, _impl_.last_entry_)}},
      // bool deleted = 5;
      {::_pbi::TcParser::SingularVarintNoZag1<bool, offsetof(ContextUpdated, _impl_.deleted_), 9>(),
       {40, 9, 0,
        PROTOBUF_FIELD_OFFSET(ContextUpdated, _impl_.deleted_)}},
      // string title = 6;
      {::_pbi::TcParser::FastUS1,
       {50, 1, 0,
        PROTOBUF_FIELD_OFFSET(ContextUpdated, _impl_.title_)}},
      // uint64 priority = 7;
      {::_pbi::TcParser::SingularVarintNoZag1<::uint64_t, offsetof(ContextUpdated, _impl_.priority_), 8>(),
       {56, 8, 0,
        PROTOBUF_FIELD_OFFSET(ContextUpdated, _impl_.priority_)}},
      // string chart_type = 8;
      {::_pbi::TcParser::FastUS1,
       {66, 2, 0,
        PROTOBUF_FIELD_OFFSET(ContextUpdated, _impl_.chart_type_)}},
      // string units = 9;
      {::_pbi::TcParser::FastUS1,
       {74, 3, 0,
        PROTOBUF_FIELD_OFFSET(ContextUpdated, _impl_.units_)}},
      // string family = 10;
      {::_pbi::TcParser::FastUS1,
       {82, 4, 0,
        PROTOBUF_FIELD_OFFSET(ContextUpdated, _impl_.family_)}},
      {::_pbi::TcParser::MiniParse, {}},
      {::_pbi::TcParser::MiniParse, {}},
      {::_pbi::TcParser::MiniParse, {}},
      {::_pbi::TcParser::MiniParse, {}},
      {::_pbi::TcParser::MiniParse, {}},
    }}, {{
      65535, 65535
    }}, {{
      // string id = 1;
      {PROTOBUF_FIELD_OFFSET(ContextUpdated, _impl_.id_), _Internal::kHasBitsOffset + 0, 0, (0 | ::_fl::kFcOptional | ::_fl::kUtf8String | ::_fl::kRepAString)},
      // uint64 version = 2;
      {PROTOBUF_FIELD_OFFSET(ContextUpdated, _impl_.version_), _Internal::kHasBitsOffset + 5, 0, (0 | ::_fl::kFcOptional | ::_fl::kUInt64)},
      // uint64 first_entry = 3;
      {PROTOBUF_FIELD_OFFSET(ContextUpdated, _impl_.first_entry_), _Internal::kHasBitsOffset + 6, 0, (0 | ::_fl::kFcOptional | ::_fl::kUInt64)},
      // uint64 last_entry = 4;
      {PROTOBUF_FIELD_OFFSET(ContextUpdated, _impl_.last_entry_), _Internal::kHasBitsOffset + 7, 0, (0 | ::_fl::kFcOptional | ::_fl::kUInt64)},
      // bool deleted = 5;
      {PROTOBUF_FIELD_OFFSET(ContextUpdated, _impl_.deleted_), _Internal::kHasBitsOffset + 9, 0, (0 | ::_fl::kFcOptional | ::_fl::kBool)},
      // string title = 6;
      {PROTOBUF_FIELD_OFFSET(ContextUpdated, _impl_.title_), _Internal::kHasBitsOffset + 1, 0, (0 | ::_fl::kFcOptional | ::_fl::kUtf8String | ::_fl::kRepAString)},
      // uint64 priority = 7;
      {PROTOBUF_FIELD_OFFSET(ContextUpdated, _impl_.priority_), _Internal::kHasBitsOffset + 8, 0, (0 | ::_fl::kFcOptional | ::_fl::kUInt64)},
      // string chart_type = 8;
      {PROTOBUF_FIELD_OFFSET(ContextUpdated, _impl_.chart_type_), _Internal::kHasBitsOffset + 2, 0, (0 | ::_fl::kFcOptional | ::_fl::kUtf8String | ::_fl::kRepAString)},
      // string units = 9;
      {PROTOBUF_FIELD_OFFSET(ContextUpdated, _impl_.units_), _Internal::kHasBitsOffset + 3, 0, (0 | ::_fl::kFcOptional | ::_fl::kUtf8String | ::_fl::kRepAString)},
      // string family = 10;
      {PROTOBUF_FIELD_OFFSET(ContextUpdated, _impl_.family_), _Internal::kHasBitsOffset + 4, 0, (0 | ::_fl::kFcOptional | ::_fl::kUtf8String | ::_fl::kRepAString)},
    }},
    // no aux_entries
    {{
      "\31\2\0\0\0\0\5\0\12\5\6\0\0\0\0\0"
      "context.v1.ContextUpdated"
      "id"
      "title"
      "chart_type"
      "units"
      "family"
    }},
  };
}


inline constexpr ContextUpdated::Impl_::Impl_(
    [[maybe_unused]] ::google::protobuf::internal::InternalVisibility visibility,
    ::_pbi::ConstantInitialized) noexcept
      : _cached_size_{0},
        id_(
            &::google::protobuf::internal::fixed_address_empty_string,
            ::_pbi::ConstantInitialized()),
        title_(
            &::google::protobuf::internal::fixed_address_empty_string,
            ::_pbi::ConstantInitialized()),
        chart_type_(
            &::google::protobuf::internal::fixed_address_empty_string,
            ::_pbi::ConstantInitialized()),
        units_(
            &::google::protobuf::internal::fixed_address_empty_string,
            ::_pbi::ConstantInitialized()),
        family_(
            &::google::protobuf::internal::fixed_address_empty_string,
            ::_pbi::ConstantInitialized()),
        version_{::uint64_t{0u}},
        first_entry_{::uint64_t{0u}},
        last_entry_{::uint64_t{0u}},
        priority_{::uint64_t{0u}},
        deleted_{false} {}

template <typename>
constexpr ContextUpdated::ContextUpdated(::_pbi::ConstantInitialized,
                       const ::_pbi::ClassData* PROTOBUF_NONNULL class_data)
    : ::google::protobuf::Message(
#if defined(PROTOBUF_CUSTOM_VTABLE)
          class_data
#endif  // PROTOBUF_CUSTOM_VTABLE
          ),
      _impl_(internal_visibility(), ::_pbi::ConstantInitialized()) {
}
inline void* PROTOBUF_NONNULL ContextUpdated::PlacementNew_(
    const void* PROTOBUF_NONNULL, void* PROTOBUF_NONNULL mem,
    ::google::protobuf::Arena* PROTOBUF_NULLABLE arena) {
  return ::new (mem) ContextUpdated(arena);
}
constexpr auto ContextUpdated::InternalNewImpl_() {
  return ::google::protobuf::internal::MessageCreator::CopyInit(sizeof(ContextUpdated), alignof(ContextUpdated));
}
constexpr auto ContextUpdated::InternalGenerateClassData_(
    const MessageLite& prototype,
    const ::google::protobuf::internal::TcParseTableBase* tc_table) {
  return ::google::protobuf::internal::ClassDataFull{
      ::google::protobuf::internal::ClassData{
          &prototype,
#ifndef PROTOBUF_MESSAGE_GLOBALS
          &_table_.header,
#else
          tc_table,
#endif
          nullptr,  // IsInitialized
          &ContextUpdated::MergeImpl,
          ::google::protobuf::Message::GetNewImpl<ContextUpdated>(),
#if defined(PROTOBUF_CUSTOM_VTABLE)
          &ContextUpdated::SharedDtor,
          ::google::protobuf::Message::GetClearImpl<ContextUpdated>(), &ContextUpdated::ByteSizeLong,
              &ContextUpdated::_InternalSerialize,
#endif  // PROTOBUF_CUSTOM_VTABLE
          PROTOBUF_FIELD_OFFSET(ContextUpdated, _impl_._cached_size_),
          false,
      },
#ifdef PROTOBUF_MESSAGE_GLOBALS
      &file_reflection_data[1],
#else   // !PROTOBUF_MESSAGE_GLOBALS
      &::_pbi::kDescriptorMethods,
      &descriptor_table_context_2fv1_2fcontext_2eproto,
      nullptr,  // tracker
#endif  // PROTOBUF_MESSAGE_GLOBALS
  };
}
struct ContextUpdatedGlobalsTypeInternal : ::_pbi::MessageGlobalsBase {
  constexpr ContextUpdatedGlobalsTypeInternal()
      :
#ifndef PROTOBUF_MESSAGE_GLOBALS
        _default(::_pbi::ConstantInitialized{},
                 ContextUpdated_class_data_.base())
#else   // !PROTOBUF_MESSAGE_GLOBALS
        MessageGlobalsBase(ContextUpdated::InternalGenerateClassData_(
            _default, &ContextUpdated_globals_._table.header)),
        _default(::_pbi::ConstantInitialized{}, GetClassData()),
        _table(::_pbi::PrivateAccess::GenerateParseTable<ContextUpdated>(
            GetClassData()))
#endif  // PROTOBUF_MESSAGE_GLOBALS
  {
  }
  ~ContextUpdatedGlobalsTypeInternal() {}
  union {
    alignas(::_pbi::kMaxMessageAlignment) ContextUpdated _default;
  };
#ifdef PROTOBUF_MESSAGE_GLOBALS
  decltype(::_pbi::PrivateAccess::GenerateParseTable<ContextUpdated>(
      ::std::declval<const ::_pbi::ClassData*>())) _table;
#endif
};
#ifdef PROTOBUF_MESSAGE_GLOBALS
static_assert(PROTOBUF_FIELD_OFFSET(ContextUpdatedGlobalsTypeInternal, _default) ==
              ::_pbi::MessageGlobalsBase::OffsetToDefault());
#endif  // PROTOBUF_MESSAGE_GLOBALS

PROTOBUF_ATTRIBUTE_NO_DESTROY PROTOBUF_CONSTINIT
    PROTOBUF_ATTRIBUTE_INIT_PRIORITY1 PROTOBUF_MESSAGE_GLOBALS_CONST ContextUpdatedGlobalsTypeInternal ContextUpdated_globals_
        PROTOBUF_MESSAGE_GLOBALS_SECTION(.data.rel.ro);
#if defined(PROTOBUF_CUSTOM_VTABLE)
namespace {
const ::_pbi::ClassData* ContextUpdated_get_class_data() {
#ifdef PROTOBUF_MESSAGE_GLOBALS
  return ContextUpdated_globals_.GetClassData();
#else
  return ContextUpdated_class_data_.base();
#endif  // PROTOBUF_MESSAGE_GLOBALS
}
}  // namespace
#endif  // PROTOBUF_CUSTOM_VTABLE
class ContextsUpdated::_Internal {
 public:
  using HasBits = decltype(::std::declval<ContextsUpdated>()._impl_._has_bits_);
  static constexpr ::int32_t kHasBitsOffset =
      8 * PROTOBUF_FIELD_OFFSET(ContextsUpdated, _impl_._has_bits_);
};

constexpr ContextsUpdated::ParseTableT_ ContextsUpdated::InternalGenerateParseTable_(const ::_pbi::ClassData* class_data) {
  return ParseTableT_{
    {
      PROTOBUF_FIELD_OFFSET(ContextsUpdated, _impl_._has_bits_),
      0, // no _extensions_
      5, 56,  // max_field_number, fast_idx_mask
      offsetof(ParseTableT_, field_lookup_table),
      4294967264,  // skipmap
      offsetof(ParseTableT_, field_entries),
      5,  // num_field_entries
      1,  // num_aux_entries
      offsetof(ParseTableT_, aux_entries),
      class_data,
      nullptr,  // post_loop_handler
      ::_pbi::TcParser::GenericFallback,  // fallback
      #ifdef PROTOBUF_PREFETCH_PARSE_TABLE
      ::_pbi::TcParser::GetTable<::context::v1::ContextsUpdated>(),  // to_prefetch
      #endif  // PROTOBUF_PREFETCH_PARSE_TABLE
    }, {{
      {::_pbi::TcParser::MiniParse, {}},
      // repeated .context.v1.ContextUpdated contextUpdates = 1;
      {::_pbi::TcParser::FastMtR1,
       {10, 0, 0,
        PROTOBUF_FIELD_OFFSET(ContextsUpdated, _impl_.contextupdates_)}},
      // string claim_id = 2;
      {::_pbi::TcParser::FastUS1,
       {18, 1, 0,
        PROTOBUF_FIELD_OFFSET(ContextsUpdated, _impl_.claim_id_)}},
      // string node_id = 3;
      {::_pbi::TcParser::FastUS1,
       {26, 2, 0,
        PROTOBUF_FIELD_OFFSET(ContextsUpdated, _impl_.node_id_)}},
      // uint64 version_hash = 4;
      {::_pbi::TcParser::SingularVarintNoZag1<::uint64_t, offsetof(ContextsUpdated, _impl_.version_hash_), 3>(),
       {32, 3, 0,
        PROTOBUF_FIELD_OFFSET(ContextsUpdated, _impl_.version_hash_)}},
      // uint64 created_at = 5;
      {::_pbi::TcParser::SingularVarintNoZag1<::uint64_t, offsetof(ContextsUpdated, _impl_.created_at_), 4>(),
       {40, 4, 0,
        PROTOBUF_FIELD_OFFSET(ContextsUpdated, _impl_.created_at_)}},
      {::_pbi::TcParser::MiniParse, {}},
      {::_pbi::TcParser::MiniParse, {}},
    }}, {{
      65535, 65535
    }}, {{
      // repeated .context.v1.ContextUpdated contextUpdates = 1;
      {PROTOBUF_FIELD_OFFSET(ContextsUpdated, _impl_.contextupdates_), _Internal::kHasBitsOffset + 0, 0, (0 | ::_fl::kFcRepeated | ::_fl::kMessage | ::_fl::kTvTable)},
      // string claim_id = 2;
      {PROTOBUF_FIELD_OFFSET(ContextsUpdated, _impl_.claim_id_), _Internal::kHasBitsOffset + 1, 0, (0 | ::_fl::kFcOptional | ::_fl::kUtf8String | ::_fl::kRepAString)},
      // string node_id = 3;
      {PROTOBUF_FIELD_OFFSET(ContextsUpdated, _impl_.node_id_), _Internal::kHasBitsOffset + 2, 0, (0 | ::_fl::kFcOptional | ::_fl::kUtf8String | ::_fl::kRepAString)},
      // uint64 version_hash = 4;
      {PROTOBUF_FIELD_OFFSET(ContextsUpdated, _impl_.version_hash_), _Internal::kHasBitsOffset + 3, 0, (0 | ::_fl::kFcOptional | ::_fl::kUInt64)},
      // uint64 created_at = 5;
      {PROTOBUF_FIELD_OFFSET(ContextsUpdated, _impl_.created_at_), _Internal::kHasBitsOffset + 4, 0, (0 | ::_fl::kFcOptional | ::_fl::kUInt64)},
    }},
    {{
        #ifndef PROTOBUF_MESSAGE_GLOBALS
        {::_pbi::TcParser::GetTable<::context::v1::ContextUpdated>()},
        #else
        {::_pbi::FieldAuxMessageGlobals(), &::context::v1::ContextUpdated_globals_},
        #endif
    }},
    {{
      "\32\0\10\7\0\0\0\0"
      "context.v1.ContextsUpdated"
      "claim_id"
      "node_id"
    }},
  };
}


inline constexpr ContextsUpdated::Impl_::Impl_(
    [[maybe_unused]] ::google::protobuf::internal::InternalVisibility visibility,
    ::_pbi::ConstantInitialized) noexcept
      : _cached_size_{0},
        contextupdates_ { visibility, ::_pbi::InternalMetadataOffset::Build<
            ::context::v1::ContextsUpdated,
            PROTOBUF_FIELD_OFFSET(::context::v1::ContextsUpdated, _impl_.contextupdates_)>()
         }
        ,
        claim_id_(
            &::google::protobuf::internal::fixed_address_empty_string,
            ::_pbi::ConstantInitialized()),
        node_id_(
            &::google::protobuf::internal::fixed_address_empty_string,
            ::_pbi::ConstantInitialized()),
        version_hash_{::uint64_t{0u}},
        created_at_{::uint64_t{0u}} {}

template <typename>
constexpr ContextsUpdated::ContextsUpdated(::_pbi::ConstantInitialized,
                       const ::_pbi::ClassData* PROTOBUF_NONNULL class_data)
    : ::google::protobuf::Message(
#if defined(PROTOBUF_CUSTOM_VTABLE)
          class_data
#endif  // PROTOBUF_CUSTOM_VTABLE
          ),
      _impl_(internal_visibility(), ::_pbi::ConstantInitialized()) {
}
inline void* PROTOBUF_NONNULL ContextsUpdated::PlacementNew_(
    const void* PROTOBUF_NONNULL, void* PROTOBUF_NONNULL mem,
    ::google::protobuf::Arena* PROTOBUF_NULLABLE arena) {
  return ::new (mem) ContextsUpdated(arena);
}
constexpr auto ContextsUpdated::InternalNewImpl_() {
  return ::google::protobuf::internal::MessageCreator::CopyInit(sizeof(ContextsUpdated), alignof(ContextsUpdated));
}
constexpr auto ContextsUpdated::InternalGenerateClassData_(
    const MessageLite& prototype,
    const ::google::protobuf::internal::TcParseTableBase* tc_table) {
  return ::google::protobuf::internal::ClassDataFull{
      ::google::protobuf::internal::ClassData{
          &prototype,
#ifndef PROTOBUF_MESSAGE_GLOBALS
          &_table_.header,
#else
          tc_table,
#endif
          nullptr,  // IsInitialized
          &ContextsUpdated::MergeImpl,
          ::google::protobuf::Message::GetNewImpl<ContextsUpdated>(),
#if defined(PROTOBUF_CUSTOM_VTABLE)
          &ContextsUpdated::SharedDtor,
          ::google::protobuf::Message::GetClearImpl<ContextsUpdated>(), &ContextsUpdated::ByteSizeLong,
              &ContextsUpdated::_InternalSerialize,
#endif  // PROTOBUF_CUSTOM_VTABLE
          PROTOBUF_FIELD_OFFSET(ContextsUpdated, _impl_._cached_size_),
          false,
      },
#ifdef PROTOBUF_MESSAGE_GLOBALS
      &file_reflection_data[0],
#else   // !PROTOBUF_MESSAGE_GLOBALS
      &::_pbi::kDescriptorMethods,
      &descriptor_table_context_2fv1_2fcontext_2eproto,
      nullptr,  // tracker
#endif  // PROTOBUF_MESSAGE_GLOBALS
  };
}
struct ContextsUpdatedGlobalsTypeInternal : ::_pbi::MessageGlobalsBase {
  constexpr ContextsUpdatedGlobalsTypeInternal()
      :
#ifndef PROTOBUF_MESSAGE_GLOBALS
        _default(::_pbi::ConstantInitialized{},
                 ContextsUpdated_class_data_.base())
#else   // !PROTOBUF_MESSAGE_GLOBALS
        MessageGlobalsBase(ContextsUpdated::InternalGenerateClassData_(
            _default, &ContextsUpdated_globals_._table.header)),
        _default(::_pbi::ConstantInitialized{}, GetClassData()),
        _table(::_pbi::PrivateAccess::GenerateParseTable<ContextsUpdated>(
            GetClassData()))
#endif  // PROTOBUF_MESSAGE_GLOBALS
  {
  }
  ~ContextsUpdatedGlobalsTypeInternal() {}
  union {
    alignas(::_pbi::kMaxMessageAlignment) ContextsUpdated _default;
  };
#ifdef PROTOBUF_MESSAGE_GLOBALS
  decltype(::_pbi::PrivateAccess::GenerateParseTable<ContextsUpdated>(
      ::std::declval<const ::_pbi::ClassData*>())) _table;
#endif
};
#ifdef PROTOBUF_MESSAGE_GLOBALS
static_assert(PROTOBUF_FIELD_OFFSET(ContextsUpdatedGlobalsTypeInternal, _default) ==
              ::_pbi::MessageGlobalsBase::OffsetToDefault());
#endif  // PROTOBUF_MESSAGE_GLOBALS

PROTOBUF_ATTRIBUTE_NO_DESTROY PROTOBUF_CONSTINIT
    PROTOBUF_ATTRIBUTE_INIT_PRIORITY1 PROTOBUF_MESSAGE_GLOBALS_CONST ContextsUpdatedGlobalsTypeInternal ContextsUpdated_globals_
        PROTOBUF_MESSAGE_GLOBALS_SECTION(.data.rel.ro);
#if defined(PROTOBUF_CUSTOM_VTABLE)
namespace {
const ::_pbi::ClassData* ContextsUpdated_get_class_data() {
#ifdef PROTOBUF_MESSAGE_GLOBALS
  return ContextsUpdated_globals_.GetClassData();
#else
  return ContextsUpdated_class_data_.base();
#endif  // PROTOBUF_MESSAGE_GLOBALS
}
}  // namespace
#endif  // PROTOBUF_CUSTOM_VTABLE
class ContextsSnapshot::_Internal {
 public:
  using HasBits = decltype(::std::declval<ContextsSnapshot>()._impl_._has_bits_);
  static constexpr ::int32_t kHasBitsOffset =
      8 * PROTOBUF_FIELD_OFFSET(ContextsSnapshot, _impl_._has_bits_);
};

constexpr ContextsSnapshot::ParseTableT_ ContextsSnapshot::InternalGenerateParseTable_(const ::_pbi::ClassData* class_data) {
  return ParseTableT_{
    {
      PROTOBUF_FIELD_OFFSET(ContextsSnapshot, _impl_._has_bits_),
      0, // no _extensions_
      4, 24,  // max_field_number, fast_idx_mask
      offsetof(ParseTableT_, field_lookup_table),
      4294967280,  // skipmap
      offsetof(ParseTableT_, field_entries),
      4,  // num_field_entries
      1,  // num_aux_entries
      offsetof(ParseTableT_, aux_entries),
      class_data,
      nullptr,  // post_loop_handler
      ::_pbi::TcParser::GenericFallback,  // fallback
      #ifdef PROTOBUF_PREFETCH_PARSE_TABLE
      ::_pbi::TcParser::GetTable<::context::v1::ContextsSnapshot>(),  // to_prefetch
      #endif  // PROTOBUF_PREFETCH_PARSE_TABLE
    }, {{
      // uint64 version = 4;
      {::_pbi::TcParser::SingularVarintNoZag1<::uint64_t, offsetof(ContextsSnapshot, _impl_.version_), 3>(),
       {32, 3, 0,
        PROTOBUF_FIELD_OFFSET(ContextsSnapshot, _impl_.version_)}},
      // repeated .context.v1.ContextUpdated contexts = 1;
      {::_pbi::TcParser::FastMtR1,
       {10, 0, 0,
        PROTOBUF_FIELD_OFFSET(ContextsSnapshot, _impl_.contexts_)}},
      // string claim_id = 2;
      {::_pbi::TcParser::FastUS1,
       {18, 1, 0,
        PROTOBUF_FIELD_OFFSET(ContextsSnapshot, _impl_.claim_id_)}},
      // string node_id = 3;
      {::_pbi::TcParser::FastUS1,
       {26, 2, 0,
        PROTOBUF_FIELD_OFFSET(ContextsSnapshot, _impl_.node_id_)}},
    }}, {{
      65535, 65535
    }}, {{
      // repeated .context.v1.ContextUpdated contexts = 1;
      {PROTOBUF_FIELD_OFFSET(ContextsSnapshot, _impl_.contexts_), _Internal::kHasBitsOffset + 0, 0, (0 | ::_fl::kFcRepeated | ::_fl::kMessage | ::_fl::kTvTable)},
      // string claim_id = 2;
      {PROTOBUF_FIELD_OFFSET(ContextsSnapshot, _impl_.claim_id_), _Internal::kHasBitsOffset + 1, 0, (0 | ::_fl::kFcOptional | ::_fl::kUtf8String | ::_fl::kRepAString)},
      // string node_id = 3;
      {PROTOBUF_FIELD_OFFSET(ContextsSnapshot, _impl_.node_id_), _Internal::kHasBitsOffset + 2, 0, (0 | ::_fl::kFcOptional | ::_fl::kUtf8String | ::_fl::kRepAString)},
      // uint64 version = 4;
      {PROTOBUF_FIELD_OFFSET(ContextsSnapshot, _impl_.version_), _Internal::kHasBitsOffset + 3, 0, (0 | ::_fl::kFcOptional | ::_fl::kUInt64)},
    }},
    {{
        #ifndef PROTOBUF_MESSAGE_GLOBALS
        {::_pbi::TcParser::GetTable<::context::v1::ContextUpdated>()},
        #else
        {::_pbi::FieldAuxMessageGlobals(), &::context::v1::ContextUpdated_globals_},
        #endif
    }},
    {{
      "\33\0\10\7\0\0\0\0"
      "context.v1.ContextsSnapshot"
      "claim_id"
      "node_id"
    }},
  };
}


inline constexpr ContextsSnapshot::Impl_::Impl_(
    [[maybe_unused]] ::google::protobuf::internal::InternalVisibility visibility,
    ::_pbi::ConstantInitialized) noexcept
      : _cached_size_{0},
        contexts_ { visibility, ::_pbi::InternalMetadataOffset::Build<
            ::context::v1::ContextsSnapshot,
            PROTOBUF_FIELD_OFFSET(::context::v1::ContextsSnapshot, _impl_.contexts_)>()
         }
        ,
        claim_id_(
            &::google::protobuf::internal::fixed_address_empty_string,
            ::_pbi::ConstantInitialized()),
        node_id_(
            &::google::protobuf::internal::fixed_address_empty_string,
            ::_pbi::ConstantInitialized()),
        version_{::uint64_t{0u}} {}

template <typename>
constexpr ContextsSnapshot::ContextsSnapshot(::_pbi::ConstantInitialized,
                       const ::_pbi::ClassData* PROTOBUF_NONNULL class_data)
    : ::google::protobuf::Message(
#if defined(PROTOBUF_CUSTOM_VTABLE)
          class_data
#endif  // PROTOBUF_CUSTOM_VTABLE
          ),
      _impl_(internal_visibility(), ::_pbi::ConstantInitialized()) {
}
inline void* PROTOBUF_NONNULL ContextsSnapshot::PlacementNew_(
    const void* PROTOBUF_NONNULL, void* PROTOBUF_NONNULL mem,
    ::google::protobuf::Arena* PROTOBUF_NULLABLE arena) {
  return ::new (mem) ContextsSnapshot(arena);
}
constexpr auto ContextsSnapshot::InternalNewImpl_() {
  return ::google::protobuf::internal::MessageCreator::CopyInit(sizeof(ContextsSnapshot), alignof(ContextsSnapshot));
}
constexpr auto ContextsSnapshot::InternalGenerateClassData_(
    const MessageLite& prototype,
    const ::google::protobuf::internal::TcParseTableBase* tc_table) {
  return ::google::protobuf::internal::ClassDataFull{
      ::google::protobuf::internal::ClassData{
          &prototype,
#ifndef PROTOBUF_MESSAGE_GLOBALS
          &_table_.header,
#else
          tc_table,
#endif
          nullptr,  // IsInitialized
          &ContextsSnapshot::MergeImpl,
          ::google::protobuf::Message::GetNewImpl<ContextsSnapshot>(),
#if defined(PROTOBUF_CUSTOM_VTABLE)
          &ContextsSnapshot::SharedDtor,
          ::google::protobuf::Message::GetClearImpl<ContextsSnapshot>(), &ContextsSnapshot::ByteSizeLong,
              &ContextsSnapshot::_InternalSerialize,
#endif  // PROTOBUF_CUSTOM_VTABLE
          PROTOBUF_FIELD_OFFSET(ContextsSnapshot, _impl_._cached_size_),
          false,
      },
#ifdef PROTOBUF_MESSAGE_GLOBALS
      &file_reflection_data[2],
#else   // !PROTOBUF_MESSAGE_GLOBALS
      &::_pbi::kDescriptorMethods,
      &descriptor_table_context_2fv1_2fcontext_2eproto,
      nullptr,  // tracker
#endif  // PROTOBUF_MESSAGE_GLOBALS
  };
}
struct ContextsSnapshotGlobalsTypeInternal : ::_pbi::MessageGlobalsBase {
  constexpr ContextsSnapshotGlobalsTypeInternal()
      :
#ifndef PROTOBUF_MESSAGE_GLOBALS
        _default(::_pbi::ConstantInitialized{},
                 ContextsSnapshot_class_data_.base())
#else   // !PROTOBUF_MESSAGE_GLOBALS
        MessageGlobalsBase(ContextsSnapshot::InternalGenerateClassData_(
            _default, &ContextsSnapshot_globals_._table.header)),
        _default(::_pbi::ConstantInitialized{}, GetClassData()),
        _table(::_pbi::PrivateAccess::GenerateParseTable<ContextsSnapshot>(
            GetClassData()))
#endif  // PROTOBUF_MESSAGE_GLOBALS
  {
  }
  ~ContextsSnapshotGlobalsTypeInternal() {}
  union {
    alignas(::_pbi::kMaxMessageAlignment) ContextsSnapshot _default;
  };
#ifdef PROTOBUF_MESSAGE_GLOBALS
  decltype(::_pbi::PrivateAccess::GenerateParseTable<ContextsSnapshot>(
      ::std::declval<const ::_pbi::ClassData*>())) _table;
#endif
};
#ifdef PROTOBUF_MESSAGE_GLOBALS
static_assert(PROTOBUF_FIELD_OFFSET(ContextsSnapshotGlobalsTypeInternal, _default) ==
              ::_pbi::MessageGlobalsBase::OffsetToDefault());
#endif  // PROTOBUF_MESSAGE_GLOBALS

PROTOBUF_ATTRIBUTE_NO_DESTROY PROTOBUF_CONSTINIT
    PROTOBUF_ATTRIBUTE_INIT_PRIORITY1 PROTOBUF_MESSAGE_GLOBALS_CONST ContextsSnapshotGlobalsTypeInternal ContextsSnapshot_globals_
        PROTOBUF_MESSAGE_GLOBALS_SECTION(.data.rel.ro);
#if defined(PROTOBUF_CUSTOM_VTABLE)
namespace {
const ::_pbi::ClassData* ContextsSnapshot_get_class_data() {
#ifdef PROTOBUF_MESSAGE_GLOBALS
  return ContextsSnapshot_globals_.GetClassData();
#else
  return ContextsSnapshot_class_data_.base();
#endif  // PROTOBUF_MESSAGE_GLOBALS
}
}  // namespace
#endif  // PROTOBUF_CUSTOM_VTABLE
}  // namespace v1
}  // namespace context
static constexpr const ::_pb::EnumDescriptor* PROTOBUF_NONNULL* PROTOBUF_NULLABLE
    file_level_enum_descriptors_context_2fv1_2fcontext_2eproto = nullptr;
static constexpr const ::_pb::ServiceDescriptor* PROTOBUF_NONNULL* PROTOBUF_NULLABLE
    file_level_service_descriptors_context_2fv1_2fcontext_2eproto = nullptr;
const ::uint32_t
    TableStruct_context_2fv1_2fcontext_2eproto::offsets[] ABSL_ATTRIBUTE_SECTION_VARIABLE(
        protodesc_cold) = {
        0x081, // bitmap
        PROTOBUF_FIELD_OFFSET(::context::v1::ContextsUpdated, _impl_._has_bits_),
        8, // hasbit index offset
        PROTOBUF_FIELD_OFFSET(::context::v1::ContextsUpdated, _impl_.contextupdates_),
        PROTOBUF_FIELD_OFFSET(::context::v1::ContextsUpdated, _impl_.claim_id_),
        PROTOBUF_FIELD_OFFSET(::context::v1::ContextsUpdated, _impl_.node_id_),
        PROTOBUF_FIELD_OFFSET(::context::v1::ContextsUpdated, _impl_.version_hash_),
        PROTOBUF_FIELD_OFFSET(::context::v1::ContextsUpdated, _impl_.created_at_),
        0,
        1,
        2,
        3,
        4,
        0x081, // bitmap
        PROTOBUF_FIELD_OFFSET(::context::v1::ContextUpdated, _impl_._has_bits_),
        13, // hasbit index offset
        PROTOBUF_FIELD_OFFSET(::context::v1::ContextUpdated, _impl_.id_),
        PROTOBUF_FIELD_OFFSET(::context::v1::ContextUpdated, _impl_.version_),
        PROTOBUF_FIELD_OFFSET(::context::v1::ContextUpdated, _impl_.first_entry_),
        PROTOBUF_FIELD_OFFSET(::context::v1::ContextUpdated, _impl_.last_entry_),
        PROTOBUF_FIELD_OFFSET(::context::v1::ContextUpdated, _impl_.deleted_),
        PROTOBUF_FIELD_OFFSET(::context::v1::ContextUpdated, _impl_.title_),
        PROTOBUF_FIELD_OFFSET(::context::v1::ContextUpdated, _impl_.priority_),
        PROTOBUF_FIELD_OFFSET(::context::v1::ContextUpdated, _impl_.chart_type_),
        PROTOBUF_FIELD_OFFSET(::context::v1::ContextUpdated, _impl_.units_),
        PROTOBUF_FIELD_OFFSET(::context::v1::ContextUpdated, _impl_.family_),
        0,
        5,
        6,
        7,
        9,
        1,
        8,
        2,
        3,
        4,
        0x081, // bitmap
        PROTOBUF_FIELD_OFFSET(::context::v1::ContextsSnapshot, _impl_._has_bits_),
        7, // hasbit index offset
        PROTOBUF_FIELD_OFFSET(::context::v1::ContextsSnapshot, _impl_.contexts_),
        PROTOBUF_FIELD_OFFSET(::context::v1::ContextsSnapshot, _impl_.claim_id_),
        PROTOBUF_FIELD_OFFSET(::context::v1::ContextsSnapshot, _impl_.node_id_),
        PROTOBUF_FIELD_OFFSET(::context::v1::ContextsSnapshot, _impl_.version_),
        0,
        1,
        2,
        3,
};

static const ::_pbi::MigrationSchema
    schemas[] ABSL_ATTRIBUTE_SECTION_VARIABLE(protodesc_cold) = {
        {0, sizeof(::context::v1::ContextsUpdated)},
        {13, sizeof(::context::v1::ContextUpdated)},
        {36, sizeof(::context::v1::ContextsSnapshot)},
};
static const ::_pbi::MessageGlobalsBase* PROTOBUF_NONNULL const
    file_message_globals[] = {
        &::context::v1::ContextsUpdated_globals_,
        &::context::v1::ContextUpdated_globals_,
        &::context::v1::ContextsSnapshot_globals_,
};
const char descriptor_table_protodef_context_2fv1_2fcontext_2eproto[] ABSL_ATTRIBUTE_SECTION_VARIABLE(
    protodesc_cold) = {
    "\n\030context/v1/context.proto\022\ncontext.v1\"\222"
    "\001\n\017ContextsUpdated\0222\n\016contextUpdates\030\001 \003"
    "(\0132\032.context.v1.ContextUpdated\022\020\n\010claim_"
    "id\030\002 \001(\t\022\017\n\007node_id\030\003 \001(\t\022\024\n\014version_has"
    "h\030\004 \001(\004\022\022\n\ncreated_at\030\005 \001(\004\"\273\001\n\016ContextU"
    "pdated\022\n\n\002id\030\001 \001(\t\022\017\n\007version\030\002 \001(\004\022\023\n\013f"
    "irst_entry\030\003 \001(\004\022\022\n\nlast_entry\030\004 \001(\004\022\017\n\007"
    "deleted\030\005 \001(\010\022\r\n\005title\030\006 \001(\t\022\020\n\010priority"
    "\030\007 \001(\004\022\022\n\nchart_type\030\010 \001(\t\022\r\n\005units\030\t \001("
    "\t\022\016\n\006family\030\n \001(\t\"t\n\020ContextsSnapshot\022,\n"
    "\010contexts\030\001 \003(\0132\032.context.v1.ContextUpda"
    "ted\022\020\n\010claim_id\030\002 \001(\t\022\017\n\007node_id\030\003 \001(\t\022\017"
    "\n\007version\030\004 \001(\004B\024Z\022context/v1;contextb\006p"
    "roto3"
};
static ::absl::once_flag descriptor_table_context_2fv1_2fcontext_2eproto_once;
PROTOBUF_CONSTINIT const ::_pbi::DescriptorTable descriptor_table_context_2fv1_2fcontext_2eproto = {
    false,
    false,
    525,
    descriptor_table_protodef_context_2fv1_2fcontext_2eproto,
    "context/v1/context.proto",
    &descriptor_table_context_2fv1_2fcontext_2eproto_once,
    nullptr,
    0,
    3,
    schemas,
    file_message_globals,
    TableStruct_context_2fv1_2fcontext_2eproto::offsets,
    file_level_enum_descriptors_context_2fv1_2fcontext_2eproto,
    file_level_service_descriptors_context_2fv1_2fcontext_2eproto,
};
namespace context {
namespace v1 {
// ===================================================================

ContextsUpdated::ContextsUpdated(::google::protobuf::Arena* PROTOBUF_NULLABLE arena)
#if defined(PROTOBUF_CUSTOM_VTABLE)
    : ::google::protobuf::Message(arena, ContextsUpdated_get_class_data()) {
#else   // PROTOBUF_CUSTOM_VTABLE
    : ::google::protobuf::Message(arena) {
#endif  // PROTOBUF_CUSTOM_VTABLE
  SharedCtor(arena);
  // @@protoc_insertion_point(arena_constructor:context.v1.ContextsUpdated)
}
PROTOBUF_NDEBUG_INLINE ContextsUpdated::Impl_::Impl_(
    [[maybe_unused]] ::google::protobuf::internal::InternalVisibility visibility,
    [[maybe_unused]] ::google::protobuf::Arena* PROTOBUF_NULLABLE arena, const Impl_& from,
    [[maybe_unused]] const ::context::v1::ContextsUpdated& from_msg)
      : _has_bits_{from._has_bits_},
        _cached_size_{0},
        contextupdates_ {
          visibility, ::_pbi::InternalMetadataOffset::Build<
              ::context::v1::ContextsUpdated,
              PROTOBUF_FIELD_OFFSET(::context::v1::ContextsUpdated, _impl_.contextupdates_)>()
          , from.contextupdates_
        }
        ,
        claim_id_(arena, from.claim_id_),
        node_id_(arena, from.node_id_) {}

ContextsUpdated::ContextsUpdated(
    ::google::protobuf::Arena* PROTOBUF_NULLABLE arena,
    const ContextsUpdated& from)
#if defined(PROTOBUF_CUSTOM_VTABLE)
    : ::google::protobuf::Message(arena, ContextsUpdated_get_class_data()) {

#else   // PROTOBUF_CUSTOM_VTABLE
    : ::google::protobuf::Message(arena) {
#endif  // PROTOBUF_CUSTOM_VTABLE
  ContextsUpdated* const _this = this;
  (void)_this;
  _internal_metadata_.MergeFrom<::google::protobuf::UnknownFieldSet>(
      from._internal_metadata_);
  new (&_impl_) Impl_(internal_visibility(), arena, from._impl_, from);
  ::memcpy(reinterpret_cast<char*>(&_impl_) +
               offsetof(Impl_, version_hash_),
           reinterpret_cast<const char*>(&from._impl_) +
               offsetof(Impl_, version_hash_),
           offsetof(Impl_, created_at_) -
               offsetof(Impl_, version_hash_) +
               sizeof(Impl_::created_at_));

  // @@protoc_insertion_point(copy_constructor:context.v1.ContextsUpdated)
}
PROTOBUF_NDEBUG_INLINE ContextsUpdated::Impl_::Impl_(
    [[maybe_unused]] ::google::protobuf::internal::InternalVisibility visibility,
    [[maybe_unused]] ::google::protobuf::Arena* PROTOBUF_NULLABLE arena)
      : _cached_size_{0},
        contextupdates_ { visibility, ::_pbi::InternalMetadataOffset::Build<
            ::context::v1::ContextsUpdated,
            PROTOBUF_FIELD_OFFSET(::context::v1::ContextsUpdated, _impl_.contextupdates_)>()
         }
        ,
        claim_id_(arena),
        node_id_(arena) {}

inline void ContextsUpdated::SharedCtor(::_pb::Arena* PROTOBUF_NULLABLE arena) {
  new (&_impl_) Impl_(internal_visibility(), arena);
  ::memset(reinterpret_cast<char*>(&_impl_) +
               offsetof(Impl_, version_hash_),
           0,
           offsetof(Impl_, created_at_) -
               offsetof(Impl_, version_hash_) +
               sizeof(Impl_::created_at_));
}
ContextsUpdated::~ContextsUpdated() {
  // @@protoc_insertion_point(destructor:context.v1.ContextsUpdated)
  SharedDtor(*this);
}
inline void ContextsUpdated::SharedDtor(MessageLite& self) {
  ContextsUpdated& this_ = static_cast<ContextsUpdated&>(self);
  if constexpr (::_pbi::DebugHardenCheckHasBitConsistency()) {
    this_.CheckHasBitConsistency();
  }
  this_._internal_metadata_.Delete<::google::protobuf::UnknownFieldSet>();
  ABSL_DCHECK(this_.GetArena() == nullptr);
  this_._impl_.claim_id_.Destroy();
  this_._impl_.node_id_.Destroy();
  this_._impl_.~Impl_();
}

#ifndef PROTOBUF_MESSAGE_GLOBALS
PROTOBUF_CONSTINIT PROTOBUF_ATTRIBUTE_INIT_PRIORITY1 const
    ::google::protobuf::internal::ClassDataFull ContextsUpdated_class_data_ =
        ContextsUpdated::InternalGenerateClassData_(ContextsUpdated_globals_._default);

PROTOBUF_ATTRIBUTE_WEAK const ::google::protobuf::internal::ClassData* PROTOBUF_NONNULL
ContextsUpdated::GetClassData() const {
  ::google::protobuf::internal::PrefetchToLocalCache(&ContextsUpdated_class_data_);
  ::google::protobuf::internal::PrefetchToLocalCache(ContextsUpdated_class_data_.tc_table);
  return ContextsUpdated_class_data_.base();
}
#else
PROTOBUF_ATTRIBUTE_WEAK const ::google::protobuf::internal::ClassData* PROTOBUF_NONNULL
ContextsUpdated::GetClassData() const {
  ::google::protobuf::internal::PrefetchToLocalCache(&ContextsUpdated_globals_);
  ::google::protobuf::internal::PrefetchToLocalCache(
      ::google::protobuf::internal::MessageGlobalsBase::ToParseTableBase(&ContextsUpdated_globals_));
  return ContextsUpdated_globals_.GetClassData();
}
#endif  // !PROTOBUF_MESSAGE_GLOBALS
#ifndef PROTOBUF_MESSAGE_GLOBALS
PROTOBUF_CONSTINIT
PROTOBUF_ATTRIBUTE_INIT_PRIORITY1 const ContextsUpdated::ParseTableT_
    ContextsUpdated::_table_ =
        ContextsUpdated::InternalGenerateParseTable_(ContextsUpdated_class_data_.base());
#endif  // !PROTOBUF_MESSAGE_GLOBALS
PROTOBUF_NOINLINE void ContextsUpdated::Clear() {
// @@protoc_insertion_point(message_clear_start:context.v1.ContextsUpdated)
  ::google::protobuf::internal::TSanWrite(&_impl_);
  ::uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void) cached_has_bits;

  cached_has_bits = _impl_._has_bits_[0];
  if (BatchCheckHasBit(cached_has_bits, 0x00000007U)) {
    if (CheckHasBit(cached_has_bits, 0x00000001U)) {
      _impl_.contextupdates_.Clear();
    }
    if (CheckHasBit(cached_has_bits, 0x00000002U)) {
      _impl_.claim_id_.ClearNonDefaultToEmpty();
    }
    if (CheckHasBit(cached_has_bits, 0x00000004U)) {
      _impl_.node_id_.ClearNonDefaultToEmpty();
    }
  }
  if (BatchCheckHasBit(cached_has_bits, 0x00000018U)) {
    ::memset(&_impl_.version_hash_, 0, static_cast<::size_t>(
        reinterpret_cast<char*>(&_impl_.created_at_) -
        reinterpret_cast<char*>(&_impl_.version_hash_)) + sizeof(_impl_.created_at_));
  }
  _impl_._has_bits_.Clear();
  _internal_metadata_.Clear<::google::protobuf::UnknownFieldSet>();
}

#if defined(PROTOBUF_CUSTOM_VTABLE)
::uint8_t* PROTOBUF_NONNULL ContextsUpdated::_InternalSerialize(
    const ::google::protobuf::MessageLite& base, ::uint8_t* PROTOBUF_NONNULL target,
    ::google::protobuf::io::EpsCopyOutputStream* PROTOBUF_NONNULL stream) {
  const ContextsUpdated& this_ = static_cast<const ContextsUpdated&>(base);
#else   // PROTOBUF_CUSTOM_VTABLE
::uint8_t* PROTOBUF_NONNULL ContextsUpdated::_InternalSerialize(
    ::uint8_t* PROTOBUF_NONNULL target,
    ::google::protobuf::io::EpsCopyOutputStream* PROTOBUF_NONNULL stream) const {
  const ContextsUpdated& this_ = *this;
#endif  // PROTOBUF_CUSTOM_VTABLE
  if constexpr (::_pbi::DebugHardenCheckHasBitConsistency()) {
    this_.CheckHasBitConsistency();
  }
  // @@protoc_insertion_point(serialize_to_array_start:context.v1.ContextsUpdated)
  ::uint32_t cached_has_bits = 0;
  (void)cached_has_bits;

  cached_has_bits = this_._impl_._has_bits_[0];
  // repeated .context.v1.ContextUpdated contextUpdates = 1;
  if (CheckHasBit(cached_has_bits, 0x00000001U)) {
    for (unsigned i = 0, n = static_cast<unsigned>(
                             this_._internal_contextupdates_size());
         i < n; i++) {
      const auto& repfield = this_._internal_contextupdates().Get(i);
      target =
          ::google::protobuf::internal::WireFormatLite::InternalWriteMessage(
              1, repfield, repfield.GetCachedSize(),
              target, stream);
    }
  }

  // string claim_id = 2;
  if (CheckHasBit(cached_has_bits, 0x00000002U)) {
    if (!this_._internal_claim_id().empty()) {
      const ::std::string& _s = this_._internal_claim_id();
      ::google::protobuf::internal::WireFormatLite::VerifyUtf8String(
          _s.data(), static_cast<int>(_s.length()), ::google::protobuf::internal::WireFormatLite::SERIALIZE, "context.v1.ContextsUpdated.claim_id");
      target = stream->WriteStringMaybeAliased(2, _s, target);
    }
  }

  // string node_id = 3;
  if (CheckHasBit(cached_has_bits, 0x00000004U)) {
    if (!this_._internal_node_id().empty()) {
      const ::std::string& _s = this_._internal_node_id();
      ::google::protobuf::internal::WireFormatLite::VerifyUtf8String(
          _s.data(), static_cast<int>(_s.length()), ::google::protobuf::internal::WireFormatLite::SERIALIZE, "context.v1.ContextsUpdated.node_id");
      target = stream->WriteStringMaybeAliased(3, _s, target);
    }
  }

  // uint64 version_hash = 4;
  if (CheckHasBit(cached_has_bits, 0x00000008U)) {
    if (this_._internal_version_hash() != 0) {
      target = stream->EnsureSpace(target);
      target = ::_pbi::WireFormatLite::WriteUInt64ToArray(
          4, this_._internal_version_hash(), target);
    }
  }

  // uint64 created_at = 5;
  if (CheckHasBit(cached_has_bits, 0x00000010U)) {
    if (this_._internal_created_at() != 0) {
      target = stream->EnsureSpace(target);
      target = ::_pbi::WireFormatLite::WriteUInt64ToArray(
          5, this_._internal_created_at(), target);
    }
  }

  if (ABSL_PREDICT_FALSE(this_._internal_metadata_.have_unknown_fields())) {
    target =
        ::_pbi::WireFormat::InternalSerializeUnknownFieldsToArray(
            this_._internal_metadata_.unknown_fields<::google::protobuf::UnknownFieldSet>(::google::protobuf::UnknownFieldSet::default_instance), target, stream);
  }
  // @@protoc_insertion_point(serialize_to_array_end:context.v1.ContextsUpdated)
  return target;
}

#if defined(PROTOBUF_CUSTOM_VTABLE)
::size_t ContextsUpdated::ByteSizeLong(const MessageLite& base) {
  const ContextsUpdated& this_ = static_cast<const ContextsUpdated&>(base);
#else   // PROTOBUF_CUSTOM_VTABLE
::size_t ContextsUpdated::ByteSizeLong() const {
  const ContextsUpdated& this_ = *this;
#endif  // PROTOBUF_CUSTOM_VTABLE
  // @@protoc_insertion_point(message_byte_size_start:context.v1.ContextsUpdated)
  ::size_t total_size = 0;

  ::uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void)cached_has_bits;

  ::_pbi::Prefetch5LinesFrom7Lines(&this_);
  cached_has_bits = this_._impl_._has_bits_[0];
  if (BatchCheckHasBit(cached_has_bits, 0x0000001fU)) {
    // repeated .context.v1.ContextUpdated contextUpdates = 1;
    if (CheckHasBit(cached_has_bits, 0x00000001U)) {
      total_size += 1UL * this_._internal_contextupdates_size();
      for (const auto& msg : this_._internal_contextupdates()) {
        total_size += ::google::protobuf::internal::WireFormatLite::MessageSize(msg);
      }
    }
    // string claim_id = 2;
    if (CheckHasBit(cached_has_bits, 0x00000002U)) {
      if (!this_._internal_claim_id().empty()) {
        total_size += 1 + ::google::protobuf::internal::WireFormatLite::StringSize(
                                        this_._internal_claim_id());
      }
    }
    // string node_id = 3;
    if (CheckHasBit(cached_has_bits, 0x00000004U)) {
      if (!this_._internal_node_id().empty()) {
        total_size += 1 + ::google::protobuf::internal::WireFormatLite::StringSize(
                                        this_._internal_node_id());
      }
    }
    // uint64 version_hash = 4;
    if (CheckHasBit(cached_has_bits, 0x00000008U)) {
      if (this_._internal_version_hash() != 0) {
        total_size += ::_pbi::WireFormatLite::UInt64SizePlusOne(
            this_._internal_version_hash());
      }
    }
    // uint64 created_at = 5;
    if (CheckHasBit(cached_has_bits, 0x00000010U)) {
      if (this_._internal_created_at() != 0) {
        total_size += ::_pbi::WireFormatLite::UInt64SizePlusOne(
            this_._internal_created_at());
      }
    }
  }
  return this_.MaybeComputeUnknownFieldsSize(total_size,
                                             &this_._impl_._cached_size_);
}

void ContextsUpdated::MergeImpl(::google::protobuf::MessageLite& to_msg,
                      const ::google::protobuf::MessageLite& from_msg) {
   auto* const _this = static_cast<ContextsUpdated*>(&to_msg);
  auto& from = static_cast<const ContextsUpdated&>(from_msg);
  if constexpr (::_pbi::DebugHardenCheckHasBitConsistency()) {
    from.CheckHasBitConsistency();
  }
  ::google::protobuf::Arena* arena = _this->GetArena();
  // @@protoc_insertion_point(class_specific_merge_from_start:context.v1.ContextsUpdated)
  ABSL_DCHECK_NE(&from, _this);
  ::uint32_t cached_has_bits = 0;
  (void)cached_has_bits;

  cached_has_bits = from._impl_._has_bits_[0];
  if (BatchCheckHasBit(cached_has_bits, 0x0000001fU)) {
    if (CheckHasBit(cached_has_bits, 0x00000001U)) {
      _this->_internal_mutable_contextupdates()->InternalMergeFromWithArena(
          ::google::protobuf::MessageLite::internal_visibility(), arena,
          from._internal_contextupdates());
    }
    if (CheckHasBit(cached_has_bits, 0x00000002U)) {
      if (!from._internal_claim_id().empty()) {
        _this->_internal_set_claim_id(from._internal_claim_id());
      } else {
        if (_this->_impl_.claim_id_.IsDefault()) {
          _this->_internal_set_claim_id("");
        }
      }
    }
    if (CheckHasBit(cached_has_bits, 0x00000004U)) {
      if (!from._internal_node_id().empty()) {
        _this->_internal_set_node_id(from._internal_node_id());
      } else {
        if (_this->_impl_.node_id_.IsDefault()) {
          _this->_internal_set_node_id("");
        }
      }
    }
    if (CheckHasBit(cached_has_bits, 0x00000008U)) {
      if (from._internal_version_hash() != 0) {
        _this->_impl_.version_hash_ = from._impl_.version_hash_;
      }
    }
    if (CheckHasBit(cached_has_bits, 0x00000010U)) {
      if (from._internal_created_at() != 0) {
        _this->_impl_.created_at_ = from._impl_.created_at_;
      }
    }
  }
  _this->_impl_._has_bits_[0] |= cached_has_bits;
  _this->_internal_metadata_.MergeFrom<::google::protobuf::UnknownFieldSet>(
      from._internal_metadata_);
}

void ContextsUpdated::CopyFrom(const ContextsUpdated& from) {
  // @@protoc_insertion_point(class_specific_copy_from_start:context.v1.ContextsUpdated)
  if (&from == this) return;
  Clear();
  MergeFrom(from);
}


void ContextsUpdated::InternalSwap(ContextsUpdated* PROTOBUF_RESTRICT PROTOBUF_NONNULL other) {
  using ::std::swap;
  auto* arena = GetArena();
  ABSL_DCHECK_EQ(arena, other->GetArena());
  _internal_metadata_.InternalSwap(&other->_internal_metadata_);
  swap(_impl_._has_bits_[0], other->_impl_._has_bits_[0]);
  _impl_.contextupdates_.InternalSwap(&other->_impl_.contextupdates_);
  ::_pbi::ArenaStringPtr::InternalSwap(&_impl_.claim_id_, &other->_impl_.claim_id_, arena);
  ::_pbi::ArenaStringPtr::InternalSwap(&_impl_.node_id_, &other->_impl_.node_id_, arena);
  ::google::protobuf::internal::memswap<
      PROTOBUF_FIELD_OFFSET(ContextsUpdated, _impl_.created_at_)
      + sizeof(ContextsUpdated::_impl_.created_at_)
      - PROTOBUF_FIELD_OFFSET(ContextsUpdated, _impl_.version_hash_)>(
          reinterpret_cast<char*>(&_impl_.version_hash_),
          reinterpret_cast<char*>(&other->_impl_.version_hash_));
}

::google::protobuf::Metadata ContextsUpdated::GetMetadata() const {
  return ::google::protobuf::Message::GetMetadataImpl(GetClassData()->full());
}
// ===================================================================

ContextUpdated::ContextUpdated(::google::protobuf::Arena* PROTOBUF_NULLABLE arena)
#if defined(PROTOBUF_CUSTOM_VTABLE)
    : ::google::protobuf::Message(arena, ContextUpdated_get_class_data()) {
#else   // PROTOBUF_CUSTOM_VTABLE
    : ::google::protobuf::Message(arena) {
#endif  // PROTOBUF_CUSTOM_VTABLE
  SharedCtor(arena);
  // @@protoc_insertion_point(arena_constructor:context.v1.ContextUpdated)
}
PROTOBUF_NDEBUG_INLINE ContextUpdated::Impl_::Impl_(
    [[maybe_unused]] ::google::protobuf::internal::InternalVisibility visibility,
    [[maybe_unused]] ::google::protobuf::Arena* PROTOBUF_NULLABLE arena, const Impl_& from,
    [[maybe_unused]] const ::context::v1::ContextUpdated& from_msg)
      : _has_bits_{from._has_bits_},
        _cached_size_{0},
        id_(arena, from.id_),
        title_(arena, from.title_),
        chart_type_(arena, from.chart_type_),
        units_(arena, from.units_),
        family_(arena, from.family_) {}

ContextUpdated::ContextUpdated(
    ::google::protobuf::Arena* PROTOBUF_NULLABLE arena,
    const ContextUpdated& from)
#if defined(PROTOBUF_CUSTOM_VTABLE)
    : ::google::protobuf::Message(arena, ContextUpdated_get_class_data()) {

#else   // PROTOBUF_CUSTOM_VTABLE
    : ::google::protobuf::Message(arena) {
#endif  // PROTOBUF_CUSTOM_VTABLE
  ContextUpdated* const _this = this;
  (void)_this;
  _internal_metadata_.MergeFrom<::google::protobuf::UnknownFieldSet>(
      from._internal_metadata_);
  new (&_impl_) Impl_(internal_visibility(), arena, from._impl_, from);
  ::memcpy(reinterpret_cast<char*>(&_impl_) +
               offsetof(Impl_, version_),
           reinterpret_cast<const char*>(&from._impl_) +
               offsetof(Impl_, version_),
           offsetof(Impl_, deleted_) -
               offsetof(Impl_, version_) +
               sizeof(Impl_::deleted_));

  // @@protoc_insertion_point(copy_constructor:context.v1.ContextUpdated)
}
PROTOBUF_NDEBUG_INLINE ContextUpdated::Impl_::Impl_(
    [[maybe_unused]] ::google::protobuf::internal::InternalVisibility visibility,
    [[maybe_unused]] ::google::protobuf::Arena* PROTOBUF_NULLABLE arena)
      : _cached_size_{0},
        id_(arena),
        title_(arena),
        chart_type_(arena),
        units_(arena),
        family_(arena) {}

inline void ContextUpdated::SharedCtor(::_pb::Arena* PROTOBUF_NULLABLE arena) {
  new (&_impl_) Impl_(internal_visibility(), arena);
  ::memset(reinterpret_cast<char*>(&_impl_) +
               offsetof(Impl_, version_),
           0,
           offsetof(Impl_, deleted_) -
               offsetof(Impl_, version_) +
               sizeof(Impl_::deleted_));
}
ContextUpdated::~ContextUpdated() {
  // @@protoc_insertion_point(destructor:context.v1.ContextUpdated)
  SharedDtor(*this);
}
inline void ContextUpdated::SharedDtor(MessageLite& self) {
  ContextUpdated& this_ = static_cast<ContextUpdated&>(self);
  if constexpr (::_pbi::DebugHardenCheckHasBitConsistency()) {
    this_.CheckHasBitConsistency();
  }
  this_._internal_metadata_.Delete<::google::protobuf::UnknownFieldSet>();
  ABSL_DCHECK(this_.GetArena() == nullptr);
  this_._impl_.id_.Destroy();
  this_._impl_.title_.Destroy();
  this_._impl_.chart_type_.Destroy();
  this_._impl_.units_.Destroy();
  this_._impl_.family_.Destroy();
  this_._impl_.~Impl_();
}

#ifndef PROTOBUF_MESSAGE_GLOBALS
PROTOBUF_CONSTINIT PROTOBUF_ATTRIBUTE_INIT_PRIORITY1 const
    ::google::protobuf::internal::ClassDataFull ContextUpdated_class_data_ =
        ContextUpdated::InternalGenerateClassData_(ContextUpdated_globals_._default);

PROTOBUF_ATTRIBUTE_WEAK const ::google::protobuf::internal::ClassData* PROTOBUF_NONNULL
ContextUpdated::GetClassData() const {
  ::google::protobuf::internal::PrefetchToLocalCache(&ContextUpdated_class_data_);
  ::google::protobuf::internal::PrefetchToLocalCache(ContextUpdated_class_data_.tc_table);
  return ContextUpdated_class_data_.base();
}
#else
PROTOBUF_ATTRIBUTE_WEAK const ::google::protobuf::internal::ClassData* PROTOBUF_NONNULL
ContextUpdated::GetClassData() const {
  ::google::protobuf::internal::PrefetchToLocalCache(&ContextUpdated_globals_);
  ::google::protobuf::internal::PrefetchToLocalCache(
      ::google::protobuf::internal::MessageGlobalsBase::ToParseTableBase(&ContextUpdated_globals_));
  return ContextUpdated_globals_.GetClassData();
}
#endif  // !PROTOBUF_MESSAGE_GLOBALS
#ifndef PROTOBUF_MESSAGE_GLOBALS
PROTOBUF_CONSTINIT
PROTOBUF_ATTRIBUTE_INIT_PRIORITY1 const ContextUpdated::ParseTableT_
    ContextUpdated::_table_ =
        ContextUpdated::InternalGenerateParseTable_(ContextUpdated_class_data_.base());
#endif  // !PROTOBUF_MESSAGE_GLOBALS
PROTOBUF_NOINLINE void ContextUpdated::Clear() {
// @@protoc_insertion_point(message_clear_start:context.v1.ContextUpdated)
  ::google::protobuf::internal::TSanWrite(&_impl_);
  ::uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void) cached_has_bits;

  cached_has_bits = _impl_._has_bits_[0];
  if (BatchCheckHasBit(cached_has_bits, 0x0000001fU)) {
    if (CheckHasBit(cached_has_bits, 0x00000001U)) {
      _impl_.id_.ClearNonDefaultToEmpty();
    }
    if (CheckHasBit(cached_has_bits, 0x00000002U)) {
      _impl_.title_.ClearNonDefaultToEmpty();
    }
    if (CheckHasBit(cached_has_bits, 0x00000004U)) {
      _impl_.chart_type_.ClearNonDefaultToEmpty();
    }
    if (CheckHasBit(cached_has_bits, 0x00000008U)) {
      _impl_.units_.ClearNonDefaultToEmpty();
    }
    if (CheckHasBit(cached_has_bits, 0x00000010U)) {
      _impl_.family_.ClearNonDefaultToEmpty();
    }
  }
  if (BatchCheckHasBit(cached_has_bits, 0x000000e0U)) {
    ::memset(&_impl_.version_, 0, static_cast<::size_t>(
        reinterpret_cast<char*>(&_impl_.last_entry_) -
        reinterpret_cast<char*>(&_impl_.version_)) + sizeof(_impl_.last_entry_));
  }
  if (BatchCheckHasBit(cached_has_bits, 0x00000300U)) {
    ::memset(&_impl_.priority_, 0, static_cast<::size_t>(
        reinterpret_cast<char*>(&_impl_.deleted_) -
        reinterpret_cast<char*>(&_impl_.priority_)) + sizeof(_impl_.deleted_));
  }
  _impl_._has_bits_.Clear();
  _internal_metadata_.Clear<::google::protobuf::UnknownFieldSet>();
}

#if defined(PROTOBUF_CUSTOM_VTABLE)
::uint8_t* PROTOBUF_NONNULL ContextUpdated::_InternalSerialize(
    const ::google::protobuf::MessageLite& base, ::uint8_t* PROTOBUF_NONNULL target,
    ::google::protobuf::io::EpsCopyOutputStream* PROTOBUF_NONNULL stream) {
  const ContextUpdated& this_ = static_cast<const ContextUpdated&>(base);
#else   // PROTOBUF_CUSTOM_VTABLE
::uint8_t* PROTOBUF_NONNULL ContextUpdated::_InternalSerialize(
    ::uint8_t* PROTOBUF_NONNULL target,
    ::google::protobuf::io::EpsCopyOutputStream* PROTOBUF_NONNULL stream) const {
  const ContextUpdated& this_ = *this;
#endif  // PROTOBUF_CUSTOM_VTABLE
  if constexpr (::_pbi::DebugHardenCheckHasBitConsistency()) {
    this_.CheckHasBitConsistency();
  }
  // @@protoc_insertion_point(serialize_to_array_start:context.v1.ContextUpdated)
  ::uint32_t cached_has_bits = 0;
  (void)cached_has_bits;

  cached_has_bits = this_._impl_._has_bits_[0];
  // string id = 1;
  if (CheckHasBit(cached_has_bits, 0x00000001U)) {
    if (!this_._internal_id().empty()) {
      const ::std::string& _s = this_._internal_id();
      ::google::protobuf::internal::WireFormatLite::VerifyUtf8String(
          _s.data(), static_cast<int>(_s.length()), ::google::protobuf::internal::WireFormatLite::SERIALIZE, "context.v1.ContextUpdated.id");
      target = stream->WriteStringMaybeAliased(1, _s, target);
    }
  }

  // uint64 version = 2;
  if (CheckHasBit(cached_has_bits, 0x00000020U)) {
    if (this_._internal_version() != 0) {
      target = stream->EnsureSpace(target);
      target = ::_pbi::WireFormatLite::WriteUInt64ToArray(
          2, this_._internal_version(), target);
    }
  }

  // uint64 first_entry = 3;
  if (CheckHasBit(cached_has_bits, 0x00000040U)) {
    if (this_._internal_first_entry() != 0) {
      target = stream->EnsureSpace(target);
      target = ::_pbi::WireFormatLite::WriteUInt64ToArray(
          3, this_._internal_first_entry(), target);
    }
  }

  // uint64 last_entry = 4;
  if (CheckHasBit(cached_has_bits, 0x00000080U)) {
    if (this_._internal_last_entry() != 0) {
      target = stream->EnsureSpace(target);
      target = ::_pbi::WireFormatLite::WriteUInt64ToArray(
          4, this_._internal_last_entry(), target);
    }
  }

  // bool deleted = 5;
  if (CheckHasBit(cached_has_bits, 0x00000200U)) {
    if (this_._internal_deleted() != 0) {
      target = stream->EnsureSpace(target);
      target = ::_pbi::WireFormatLite::WriteBoolToArray(
          5, this_._internal_deleted(), target);
    }
  }

  // string title = 6;
  if (CheckHasBit(cached_has_bits, 0x00000002U)) {
    if (!this_._internal_title().empty()) {
      const ::std::string& _s = this_._internal_title();
      ::google::protobuf::internal::WireFormatLite::VerifyUtf8String(
          _s.data(), static_cast<int>(_s.length()), ::google::protobuf::internal::WireFormatLite::SERIALIZE, "context.v1.ContextUpdated.title");
      target = stream->WriteStringMaybeAliased(6, _s, target);
    }
  }

  // uint64 priority = 7;
  if (CheckHasBit(cached_has_bits, 0x00000100U)) {
    if (this_._internal_priority() != 0) {
      target = stream->EnsureSpace(target);
      target = ::_pbi::WireFormatLite::WriteUInt64ToArray(
          7, this_._internal_priority(), target);
    }
  }

  // string chart_type = 8;
  if (CheckHasBit(cached_has_bits, 0x00000004U)) {
    if (!this_._internal_chart_type().empty()) {
      const ::std::string& _s = this_._internal_chart_type();
      ::google::protobuf::internal::WireFormatLite::VerifyUtf8String(
          _s.data(), static_cast<int>(_s.length()), ::google::protobuf::internal::WireFormatLite::SERIALIZE, "context.v1.ContextUpdated.chart_type");
      target = stream->WriteStringMaybeAliased(8, _s, target);
    }
  }

  // string units = 9;
  if (CheckHasBit(cached_has_bits, 0x00000008U)) {
    if (!this_._internal_units().empty()) {
      const ::std::string& _s = this_._internal_units();
      ::google::protobuf::internal::WireFormatLite::VerifyUtf8String(
          _s.data(), static_cast<int>(_s.length()), ::google::protobuf::internal::WireFormatLite::SERIALIZE, "context.v1.ContextUpdated.units");
      target = stream->WriteStringMaybeAliased(9, _s, target);
    }
  }

  // string family = 10;
  if (CheckHasBit(cached_has_bits, 0x00000010U)) {
    if (!this_._internal_family().empty()) {
      const ::std::string& _s = this_._internal_family();
      ::google::protobuf::internal::WireFormatLite::VerifyUtf8String(
          _s.data(), static_cast<int>(_s.length()), ::google::protobuf::internal::WireFormatLite::SERIALIZE, "context.v1.ContextUpdated.family");
      target = stream->WriteStringMaybeAliased(10, _s, target);
    }
  }

  if (ABSL_PREDICT_FALSE(this_._internal_metadata_.have_unknown_fields())) {
    target =
        ::_pbi::WireFormat::InternalSerializeUnknownFieldsToArray(
            this_._internal_metadata_.unknown_fields<::google::protobuf::UnknownFieldSet>(::google::protobuf::UnknownFieldSet::default_instance), target, stream);
  }
  // @@protoc_insertion_point(serialize_to_array_end:context.v1.ContextUpdated)
  return target;
}

#if defined(PROTOBUF_CUSTOM_VTABLE)
::size_t ContextUpdated::ByteSizeLong(const MessageLite& base) {
  const ContextUpdated& this_ = static_cast<const ContextUpdated&>(base);
#else   // PROTOBUF_CUSTOM_VTABLE
::size_t ContextUpdated::ByteSizeLong() const {
  const ContextUpdated& this_ = *this;
#endif  // PROTOBUF_CUSTOM_VTABLE
  // @@protoc_insertion_point(message_byte_size_start:context.v1.ContextUpdated)
  ::size_t total_size = 0;

  ::uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void)cached_has_bits;

  ::_pbi::Prefetch5LinesFrom7Lines(&this_);
  cached_has_bits = this_._impl_._has_bits_[0];
  if (BatchCheckHasBit(cached_has_bits, 0x000000ffU)) {
    // string id = 1;
    if (CheckHasBit(cached_has_bits, 0x00000001U)) {
      if (!this_._internal_id().empty()) {
        total_size += 1 + ::google::protobuf::internal::WireFormatLite::StringSize(
                                        this_._internal_id());
      }
    }
    // string title = 6;
    if (CheckHasBit(cached_has_bits, 0x00000002U)) {
      if (!this_._internal_title().empty()) {
        total_size += 1 + ::google::protobuf::internal::WireFormatLite::StringSize(
                                        this_._internal_title());
      }
    }
    // string chart_type = 8;
    if (CheckHasBit(cached_has_bits, 0x00000004U)) {
      if (!this_._internal_chart_type().empty()) {
        total_size += 1 + ::google::protobuf::internal::WireFormatLite::StringSize(
                                        this_._internal_chart_type());
      }
    }
    // string units = 9;
    if (CheckHasBit(cached_has_bits, 0x00000008U)) {
      if (!this_._internal_units().empty()) {
        total_size += 1 + ::google::protobuf::internal::WireFormatLite::StringSize(
                                        this_._internal_units());
      }
    }
    // string family = 10;
    if (CheckHasBit(cached_has_bits, 0x00000010U)) {
      if (!this_._internal_family().empty()) {
        total_size += 1 + ::google::protobuf::internal::WireFormatLite::StringSize(
                                        this_._internal_family());
      }
    }
    // uint64 version = 2;
    if (CheckHasBit(cached_has_bits, 0x00000020U)) {
      if (this_._internal_version() != 0) {
        total_size += ::_pbi::WireFormatLite::UInt64SizePlusOne(
            this_._internal_version());
      }
    }
    // uint64 first_entry = 3;
    if (CheckHasBit(cached_has_bits, 0x00000040U)) {
      if (this_._internal_first_entry() != 0) {
        total_size += ::_pbi::WireFormatLite::UInt64SizePlusOne(
            this_._internal_first_entry());
      }
    }
    // uint64 last_entry = 4;
    if (CheckHasBit(cached_has_bits, 0x00000080U)) {
      if (this_._internal_last_entry() != 0) {
        total_size += ::_pbi::WireFormatLite::UInt64SizePlusOne(
            this_._internal_last_entry());
      }
    }
  }
  if (BatchCheckHasBit(cached_has_bits, 0x00000300U)) {
    // uint64 priority = 7;
    if (CheckHasBit(cached_has_bits, 0x00000100U)) {
      if (this_._internal_priority() != 0) {
        total_size += ::_pbi::WireFormatLite::UInt64SizePlusOne(
            this_._internal_priority());
      }
    }
    // bool deleted = 5;
    if (CheckHasBit(cached_has_bits, 0x00000200U)) {
      if (this_._internal_deleted() != 0) {
        total_size += 2;
      }
    }
  }
  return this_.MaybeComputeUnknownFieldsSize(total_size,
                                             &this_._impl_._cached_size_);
}

void ContextUpdated::MergeImpl(::google::protobuf::MessageLite& to_msg,
                      const ::google::protobuf::MessageLite& from_msg) {
   auto* const _this = static_cast<ContextUpdated*>(&to_msg);
  auto& from = static_cast<const ContextUpdated&>(from_msg);
  if constexpr (::_pbi::DebugHardenCheckHasBitConsistency()) {
    from.CheckHasBitConsistency();
  }
  // @@protoc_insertion_point(class_specific_merge_from_start:context.v1.ContextUpdated)
  ABSL_DCHECK_NE(&from, _this);
  ::uint32_t cached_has_bits = 0;
  (void)cached_has_bits;

  cached_has_bits = from._impl_._has_bits_[0];
  if (BatchCheckHasBit(cached_has_bits, 0x000000ffU)) {
    if (CheckHasBit(cached_has_bits, 0x00000001U)) {
      if (!from._internal_id().empty()) {
        _this->_internal_set_id(from._internal_id());
      } else {
        if (_this->_impl_.id_.IsDefault()) {
          _this->_internal_set_id("");
        }
      }
    }
    if (CheckHasBit(cached_has_bits, 0x00000002U)) {
      if (!from._internal_title().empty()) {
        _this->_internal_set_title(from._internal_title());
      } else {
        if (_this->_impl_.title_.IsDefault()) {
          _this->_internal_set_title("");
        }
      }
    }
    if (CheckHasBit(cached_has_bits, 0x00000004U)) {
      if (!from._internal_chart_type().empty()) {
        _this->_internal_set_chart_type(from._internal_chart_type());
      } else {
        if (_this->_impl_.chart_type_.IsDefault()) {
          _this->_internal_set_chart_type("");
        }
      }
    }
    if (CheckHasBit(cached_has_bits, 0x00000008U)) {
      if (!from._internal_units().empty()) {
        _this->_internal_set_units(from._internal_units());
      } else {
        if (_this->_impl_.units_.IsDefault()) {
          _this->_internal_set_units("");
        }
      }
    }
    if (CheckHasBit(cached_has_bits, 0x00000010U)) {
      if (!from._internal_family().empty()) {
        _this->_internal_set_family(from._internal_family());
      } else {
        if (_this->_impl_.family_.IsDefault()) {
          _this->_internal_set_family("");
        }
      }
    }
    if (CheckHasBit(cached_has_bits, 0x00000020U)) {
      if (from._internal_version() != 0) {
        _this->_impl_.version_ = from._impl_.version_;
      }
    }
    if (CheckHasBit(cached_has_bits, 0x00000040U)) {
      if (from._internal_first_entry() != 0) {
        _this->_impl_.first_entry_ = from._impl_.first_entry_;
      }
    }
    if (CheckHasBit(cached_has_bits, 0x00000080U)) {
      if (from._internal_last_entry() != 0) {
        _this->_impl_.last_entry_ = from._impl_.last_entry_;
      }
    }
  }
  if (BatchCheckHasBit(cached_has_bits, 0x00000300U)) {
    if (CheckHasBit(cached_has_bits, 0x00000100U)) {
      if (from._internal_priority() != 0) {
        _this->_impl_.priority_ = from._impl_.priority_;
      }
    }
    if (CheckHasBit(cached_has_bits, 0x00000200U)) {
      if (from._internal_deleted() != 0) {
        _this->_impl_.deleted_ = from._impl_.deleted_;
      }
    }
  }
  _this->_impl_._has_bits_[0] |= cached_has_bits;
  _this->_internal_metadata_.MergeFrom<::google::protobuf::UnknownFieldSet>(
      from._internal_metadata_);
}

void ContextUpdated::CopyFrom(const ContextUpdated& from) {
  // @@protoc_insertion_point(class_specific_copy_from_start:context.v1.ContextUpdated)
  if (&from == this) return;
  Clear();
  MergeFrom(from);
}


void ContextUpdated::InternalSwap(ContextUpdated* PROTOBUF_RESTRICT PROTOBUF_NONNULL other) {
  using ::std::swap;
  auto* arena = GetArena();
  ABSL_DCHECK_EQ(arena, other->GetArena());
  _internal_metadata_.InternalSwap(&other->_internal_metadata_);
  swap(_impl_._has_bits_[0], other->_impl_._has_bits_[0]);
  ::_pbi::ArenaStringPtr::InternalSwap(&_impl_.id_, &other->_impl_.id_, arena);
  ::_pbi::ArenaStringPtr::InternalSwap(&_impl_.title_, &other->_impl_.title_, arena);
  ::_pbi::ArenaStringPtr::InternalSwap(&_impl_.chart_type_, &other->_impl_.chart_type_, arena);
  ::_pbi::ArenaStringPtr::InternalSwap(&_impl_.units_, &other->_impl_.units_, arena);
  ::_pbi::ArenaStringPtr::InternalSwap(&_impl_.family_, &other->_impl_.family_, arena);
  ::google::protobuf::internal::memswap<
      PROTOBUF_FIELD_OFFSET(ContextUpdated, _impl_.deleted_)
      + sizeof(ContextUpdated::_impl_.deleted_)
      - PROTOBUF_FIELD_OFFSET(ContextUpdated, _impl_.version_)>(
          reinterpret_cast<char*>(&_impl_.version_),
          reinterpret_cast<char*>(&other->_impl_.version_));
}

::google::protobuf::Metadata ContextUpdated::GetMetadata() const {
  return ::google::protobuf::Message::GetMetadataImpl(GetClassData()->full());
}
// ===================================================================

ContextsSnapshot::ContextsSnapshot(::google::protobuf::Arena* PROTOBUF_NULLABLE arena)
#if defined(PROTOBUF_CUSTOM_VTABLE)
    : ::google::protobuf::Message(arena, ContextsSnapshot_get_class_data()) {
#else   // PROTOBUF_CUSTOM_VTABLE
    : ::google::protobuf::Message(arena) {
#endif  // PROTOBUF_CUSTOM_VTABLE
  SharedCtor(arena);
  // @@protoc_insertion_point(arena_constructor:context.v1.ContextsSnapshot)
}
PROTOBUF_NDEBUG_INLINE ContextsSnapshot::Impl_::Impl_(
    [[maybe_unused]] ::google::protobuf::internal::InternalVisibility visibility,
    [[maybe_unused]] ::google::protobuf::Arena* PROTOBUF_NULLABLE arena, const Impl_& from,
    [[maybe_unused]] const ::context::v1::ContextsSnapshot& from_msg)
      : _has_bits_{from._has_bits_},
        _cached_size_{0},
        contexts_ {
          visibility, ::_pbi::InternalMetadataOffset::Build<
              ::context::v1::ContextsSnapshot,
              PROTOBUF_FIELD_OFFSET(::context::v1::ContextsSnapshot, _impl_.contexts_)>()
          , from.contexts_
        }
        ,
        claim_id_(arena, from.claim_id_),
        node_id_(arena, from.node_id_) {}

ContextsSnapshot::ContextsSnapshot(
    ::google::protobuf::Arena* PROTOBUF_NULLABLE arena,
    const ContextsSnapshot& from)
#if defined(PROTOBUF_CUSTOM_VTABLE)
    : ::google::protobuf::Message(arena, ContextsSnapshot_get_class_data()) {

#else   // PROTOBUF_CUSTOM_VTABLE
    : ::google::protobuf::Message(arena) {
#endif  // PROTOBUF_CUSTOM_VTABLE
  ContextsSnapshot* const _this = this;
  (void)_this;
  _internal_metadata_.MergeFrom<::google::protobuf::UnknownFieldSet>(
      from._internal_metadata_);
  new (&_impl_) Impl_(internal_visibility(), arena, from._impl_, from);
  _impl_.version_ = from._impl_.version_;

  // @@protoc_insertion_point(copy_constructor:context.v1.ContextsSnapshot)
}
PROTOBUF_NDEBUG_INLINE ContextsSnapshot::Impl_::Impl_(
    [[maybe_unused]] ::google::protobuf::internal::InternalVisibility visibility,
    [[maybe_unused]] ::google::protobuf::Arena* PROTOBUF_NULLABLE arena)
      : _cached_size_{0},
        contexts_ { visibility, ::_pbi::InternalMetadataOffset::Build<
            ::context::v1::ContextsSnapshot,
            PROTOBUF_FIELD_OFFSET(::context::v1::ContextsSnapshot, _impl_.contexts_)>()
         }
        ,
        claim_id_(arena),
        node_id_(arena) {}

inline void ContextsSnapshot::SharedCtor(::_pb::Arena* PROTOBUF_NULLABLE arena) {
  new (&_impl_) Impl_(internal_visibility(), arena);
  _impl_.version_ = {};
}
ContextsSnapshot::~ContextsSnapshot() {
  // @@protoc_insertion_point(destructor:context.v1.ContextsSnapshot)
  SharedDtor(*this);
}
inline void ContextsSnapshot::SharedDtor(MessageLite& self) {
  ContextsSnapshot& this_ = static_cast<ContextsSnapshot&>(self);
  if constexpr (::_pbi::DebugHardenCheckHasBitConsistency()) {
    this_.CheckHasBitConsistency();
  }
  this_._internal_metadata_.Delete<::google::protobuf::UnknownFieldSet>();
  ABSL_DCHECK(this_.GetArena() == nullptr);
  this_._impl_.claim_id_.Destroy();
  this_._impl_.node_id_.Destroy();
  this_._impl_.~Impl_();
}

#ifndef PROTOBUF_MESSAGE_GLOBALS
PROTOBUF_CONSTINIT PROTOBUF_ATTRIBUTE_INIT_PRIORITY1 const
    ::google::protobuf::internal::ClassDataFull ContextsSnapshot_class_data_ =
        ContextsSnapshot::InternalGenerateClassData_(ContextsSnapshot_globals_._default);

PROTOBUF_ATTRIBUTE_WEAK const ::google::protobuf::internal::ClassData* PROTOBUF_NONNULL
ContextsSnapshot::GetClassData() const {
  ::google::protobuf::internal::PrefetchToLocalCache(&ContextsSnapshot_class_data_);
  ::google::protobuf::internal::PrefetchToLocalCache(ContextsSnapshot_class_data_.tc_table);
  return ContextsSnapshot_class_data_.base();
}
#else
PROTOBUF_ATTRIBUTE_WEAK const ::google::protobuf::internal::ClassData* PROTOBUF_NONNULL
ContextsSnapshot::GetClassData() const {
  ::google::protobuf::internal::PrefetchToLocalCache(&ContextsSnapshot_globals_);
  ::google::protobuf::internal::PrefetchToLocalCache(
      ::google::protobuf::internal::MessageGlobalsBase::ToParseTableBase(&ContextsSnapshot_globals_));
  return ContextsSnapshot_globals_.GetClassData();
}
#endif  // !PROTOBUF_MESSAGE_GLOBALS
#ifndef PROTOBUF_MESSAGE_GLOBALS
PROTOBUF_CONSTINIT
PROTOBUF_ATTRIBUTE_INIT_PRIORITY1 const ContextsSnapshot::ParseTableT_
    ContextsSnapshot::_table_ =
        ContextsSnapshot::InternalGenerateParseTable_(ContextsSnapshot_class_data_.base());
#endif  // !PROTOBUF_MESSAGE_GLOBALS
PROTOBUF_NOINLINE void ContextsSnapshot::Clear() {
// @@protoc_insertion_point(message_clear_start:context.v1.ContextsSnapshot)
  ::google::protobuf::internal::TSanWrite(&_impl_);
  ::uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void) cached_has_bits;

  cached_has_bits = _impl_._has_bits_[0];
  if (BatchCheckHasBit(cached_has_bits, 0x00000007U)) {
    if (CheckHasBit(cached_has_bits, 0x00000001U)) {
      _impl_.contexts_.Clear();
    }
    if (CheckHasBit(cached_has_bits, 0x00000002U)) {
      _impl_.claim_id_.ClearNonDefaultToEmpty();
    }
    if (CheckHasBit(cached_has_bits, 0x00000004U)) {
      _impl_.node_id_.ClearNonDefaultToEmpty();
    }
  }
  _impl_.version_ = ::uint64_t{0u};
  _impl_._has_bits_.Clear();
  _internal_metadata_.Clear<::google::protobuf::UnknownFieldSet>();
}

#if defined(PROTOBUF_CUSTOM_VTABLE)
::uint8_t* PROTOBUF_NONNULL ContextsSnapshot::_InternalSerialize(
    const ::google::protobuf::MessageLite& base, ::uint8_t* PROTOBUF_NONNULL target,
    ::google::protobuf::io::EpsCopyOutputStream* PROTOBUF_NONNULL stream) {
  const ContextsSnapshot& this_ = static_cast<const ContextsSnapshot&>(base);
#else   // PROTOBUF_CUSTOM_VTABLE
::uint8_t* PROTOBUF_NONNULL ContextsSnapshot::_InternalSerialize(
    ::uint8_t* PROTOBUF_NONNULL target,
    ::google::protobuf::io::EpsCopyOutputStream* PROTOBUF_NONNULL stream) const {
  const ContextsSnapshot& this_ = *this;
#endif  // PROTOBUF_CUSTOM_VTABLE
  if constexpr (::_pbi::DebugHardenCheckHasBitConsistency()) {
    this_.CheckHasBitConsistency();
  }
  // @@protoc_insertion_point(serialize_to_array_start:context.v1.ContextsSnapshot)
  ::uint32_t cached_has_bits = 0;
  (void)cached_has_bits;

  cached_has_bits = this_._impl_._has_bits_[0];
  // repeated .context.v1.ContextUpdated contexts = 1;
  if (CheckHasBit(cached_has_bits, 0x00000001U)) {
    for (unsigned i = 0, n = static_cast<unsigned>(
                             this_._internal_contexts_size());
         i < n; i++) {
      const auto& repfield = this_._internal_contexts().Get(i);
      target =
          ::google::protobuf::internal::WireFormatLite::InternalWriteMessage(
              1, repfield, repfield.GetCachedSize(),
              target, stream);
    }
  }

  // string claim_id = 2;
  if (CheckHasBit(cached_has_bits, 0x00000002U)) {
    if (!this_._internal_claim_id().empty()) {
      const ::std::string& _s = this_._internal_claim_id();
      ::google::protobuf::internal::WireFormatLite::VerifyUtf8String(
          _s.data(), static_cast<int>(_s.length()), ::google::protobuf::internal::WireFormatLite::SERIALIZE, "context.v1.ContextsSnapshot.claim_id");
      target = stream->WriteStringMaybeAliased(2, _s, target);
    }
  }

  // string node_id = 3;
  if (CheckHasBit(cached_has_bits, 0x00000004U)) {
    if (!this_._internal_node_id().empty()) {
      const ::std::string& _s = this_._internal_node_id();
      ::google::protobuf::internal::WireFormatLite::VerifyUtf8String(
          _s.data(), static_cast<int>(_s.length()), ::google::protobuf::internal::WireFormatLite::SERIALIZE, "context.v1.ContextsSnapshot.node_id");
      target = stream->WriteStringMaybeAliased(3, _s, target);
    }
  }

  // uint64 version = 4;
  if (CheckHasBit(cached_has_bits, 0x00000008U)) {
    if (this_._internal_version() != 0) {
      target = stream->EnsureSpace(target);
      target = ::_pbi::WireFormatLite::WriteUInt64ToArray(
          4, this_._internal_version(), target);
    }
  }

  if (ABSL_PREDICT_FALSE(this_._internal_metadata_.have_unknown_fields())) {
    target =
        ::_pbi::WireFormat::InternalSerializeUnknownFieldsToArray(
            this_._internal_metadata_.unknown_fields<::google::protobuf::UnknownFieldSet>(::google::protobuf::UnknownFieldSet::default_instance), target, stream);
  }
  // @@protoc_insertion_point(serialize_to_array_end:context.v1.ContextsSnapshot)
  return target;
}

#if defined(PROTOBUF_CUSTOM_VTABLE)
::size_t ContextsSnapshot::ByteSizeLong(const MessageLite& base) {
  const ContextsSnapshot& this_ = static_cast<const ContextsSnapshot&>(base);
#else   // PROTOBUF_CUSTOM_VTABLE
::size_t ContextsSnapshot::ByteSizeLong() const {
  const ContextsSnapshot& this_ = *this;
#endif  // PROTOBUF_CUSTOM_VTABLE
  // @@protoc_insertion_point(message_byte_size_start:context.v1.ContextsSnapshot)
  ::size_t total_size = 0;

  ::uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void)cached_has_bits;

  ::_pbi::Prefetch5LinesFrom7Lines(&this_);
  cached_has_bits = this_._impl_._has_bits_[0];
  if (BatchCheckHasBit(cached_has_bits, 0x0000000fU)) {
    // repeated .context.v1.ContextUpdated contexts = 1;
    if (CheckHasBit(cached_has_bits, 0x00000001U)) {
      total_size += 1UL * this_._internal_contexts_size();
      for (const auto& msg : this_._internal_contexts()) {
        total_size += ::google::protobuf::internal::WireFormatLite::MessageSize(msg);
      }
    }
    // string claim_id = 2;
    if (CheckHasBit(cached_has_bits, 0x00000002U)) {
      if (!this_._internal_claim_id().empty()) {
        total_size += 1 + ::google::protobuf::internal::WireFormatLite::StringSize(
                                        this_._internal_claim_id());
      }
    }
    // string node_id = 3;
    if (CheckHasBit(cached_has_bits, 0x00000004U)) {
      if (!this_._internal_node_id().empty()) {
        total_size += 1 + ::google::protobuf::internal::WireFormatLite::StringSize(
                                        this_._internal_node_id());
      }
    }
    // uint64 version = 4;
    if (CheckHasBit(cached_has_bits, 0x00000008U)) {
      if (this_._internal_version() != 0) {
        total_size += ::_pbi::WireFormatLite::UInt64SizePlusOne(
            this_._internal_version());
      }
    }
  }
  return this_.MaybeComputeUnknownFieldsSize(total_size,
                                             &this_._impl_._cached_size_);
}

void ContextsSnapshot::MergeImpl(::google::protobuf::MessageLite& to_msg,
                      const ::google::protobuf::MessageLite& from_msg) {
   auto* const _this = static_cast<ContextsSnapshot*>(&to_msg);
  auto& from = static_cast<const ContextsSnapshot&>(from_msg);
  if constexpr (::_pbi::DebugHardenCheckHasBitConsistency()) {
    from.CheckHasBitConsistency();
  }
  ::google::protobuf::Arena* arena = _this->GetArena();
  // @@protoc_insertion_point(class_specific_merge_from_start:context.v1.ContextsSnapshot)
  ABSL_DCHECK_NE(&from, _this);
  ::uint32_t cached_has_bits = 0;
  (void)cached_has_bits;

  cached_has_bits = from._impl_._has_bits_[0];
  if (BatchCheckHasBit(cached_has_bits, 0x0000000fU)) {
    if (CheckHasBit(cached_has_bits, 0x00000001U)) {
      _this->_internal_mutable_contexts()->InternalMergeFromWithArena(
          ::google::protobuf::MessageLite::internal_visibility(), arena,
          from._internal_contexts());
    }
    if (CheckHasBit(cached_has_bits, 0x00000002U)) {
      if (!from._internal_claim_id().empty()) {
        _this->_internal_set_claim_id(from._internal_claim_id());
      } else {
        if (_this->_impl_.claim_id_.IsDefault()) {
          _this->_internal_set_claim_id("");
        }
      }
    }
    if (CheckHasBit(cached_has_bits, 0x00000004U)) {
      if (!from._internal_node_id().empty()) {
        _this->_internal_set_node_id(from._internal_node_id());
      } else {
        if (_this->_impl_.node_id_.IsDefault()) {
          _this->_internal_set_node_id("");
        }
      }
    }
    if (CheckHasBit(cached_has_bits, 0x00000008U)) {
      if (from._internal_version() != 0) {
        _this->_impl_.version_ = from._impl_.version_;
      }
    }
  }
  _this->_impl_._has_bits_[0] |= cached_has_bits;
  _this->_internal_metadata_.MergeFrom<::google::protobuf::UnknownFieldSet>(
      from._internal_metadata_);
}

void ContextsSnapshot::CopyFrom(const ContextsSnapshot& from) {
  // @@protoc_insertion_point(class_specific_copy_from_start:context.v1.ContextsSnapshot)
  if (&from == this) return;
  Clear();
  MergeFrom(from);
}


void ContextsSnapshot::InternalSwap(ContextsSnapshot* PROTOBUF_RESTRICT PROTOBUF_NONNULL other) {
  using ::std::swap;
  auto* arena = GetArena();
  ABSL_DCHECK_EQ(arena, other->GetArena());
  _internal_metadata_.InternalSwap(&other->_internal_metadata_);
  swap(_impl_._has_bits_[0], other->_impl_._has_bits_[0]);
  _impl_.contexts_.InternalSwap(&other->_impl_.contexts_);
  ::_pbi::ArenaStringPtr::InternalSwap(&_impl_.claim_id_, &other->_impl_.claim_id_, arena);
  ::_pbi::ArenaStringPtr::InternalSwap(&_impl_.node_id_, &other->_impl_.node_id_, arena);
  swap(_impl_.version_, other->_impl_.version_);
}

::google::protobuf::Metadata ContextsSnapshot::GetMetadata() const {
  return ::google::protobuf::Message::GetMetadataImpl(GetClassData()->full());
}
// @@protoc_insertion_point(namespace_scope)
}  // namespace v1
}  // namespace context
namespace google {
namespace protobuf {
}  // namespace protobuf
}  // namespace google
// @@protoc_insertion_point(global_scope)
PROTOBUF_ATTRIBUTE_INIT_PRIORITY2 static ::std::false_type
    _static_init2_ [[maybe_unused]] =
        (::_pbi::AddDescriptors(&descriptor_table_context_2fv1_2fcontext_2eproto),
         ::std::false_type{});
#include "google/protobuf/port_undef.inc"
