// Generated by the protocol buffer compiler.  DO NOT EDIT!
// NO CHECKED-IN PROTOBUF GENCODE
// source: nodeinstance/manifest/v1/manifest.proto
// Protobuf C++ Version: 7.35.1

#include "nodeinstance/manifest/v1/manifest.pb.h"

#include <algorithm>
#include <type_traits>
#include "google/protobuf/io/coded_stream.h"
#include "google/protobuf/generated_message_tctable_impl.h"
#include "google/protobuf/internal_visibility.h"
#include "google/protobuf/extension_set.h"
#include "google/protobuf/generated_message_util.h"
#include "google/protobuf/wire_format_lite.h"
#include "google/protobuf/descriptor.h"
#include "google/protobuf/generated_message_reflection.h"
#include "google/protobuf/reflection_ops.h"
#include "google/protobuf/wire_format.h"
// @@protoc_insertion_point(includes)

// Must be included last.
#include "google/protobuf/port_def.inc"
PROTOBUF_PRAGMA_INIT_SEG
namespace _pb = ::google::protobuf;
namespace _pbi = ::google::protobuf::internal;
namespace _fl = ::google::protobuf::internal::field_layout;
#ifdef PROTOBUF_MESSAGE_GLOBALS
namespace {
PROTOBUF_CONSTINIT ::google::protobuf::internal::ReflectionData
    file_reflection_data[] = {
        // ::nodeinstance::manifest::v1::UpdateNodeInstanceManifest
        {&::_pbi::kDescriptorMethods, &::descriptor_table_nodeinstance_2fmanifest_2fv1_2fmanifest_2eproto, /* tracker*/ nullptr,},
        // ::nodeinstance::manifest::v1::Functions
        {&::_pbi::kDescriptorMethods, &::descriptor_table_nodeinstance_2fmanifest_2fv1_2fmanifest_2eproto, /* tracker*/ nullptr,},
        // ::nodeinstance::manifest::v1::Function
        {&::_pbi::kDescriptorMethods, &::descriptor_table_nodeinstance_2fmanifest_2fv1_2fmanifest_2eproto, /* tracker*/ nullptr,},
};
}  // namespace
#endif
namespace nodeinstance {
namespace manifest {
namespace v1 {
class Function::_Internal {
 public:
  using HasBits = decltype(::std::declval<Function>()._impl_._has_bits_);
  static constexpr ::int32_t kHasBitsOffset =
      8 * PROTOBUF_FIELD_OFFSET(Function, _impl_._has_bits_);
};

constexpr Function::ParseTableT_ Function::InternalGenerateParseTable_(const ::_pbi::ClassData* class_data) {
  return ParseTableT_{
    {
      PROTOBUF_FIELD_OFFSET(Function, _impl_._has_bits_),
      0, // no _extensions_
      6, 56,  // max_field_number, fast_idx_mask
      offsetof(ParseTableT_, field_lookup_table),
      4294967232,  // skipmap
      offsetof(ParseTableT_, field_entries),
      6,  // num_field_entries
      0,  // num_aux_entries
      offsetof(ParseTableT_, field_names),  // no aux_entries
      class_data,
      nullptr,  // post_loop_handler
      ::_pbi::TcParser::GenericFallback,  // fallback
      #ifdef PROTOBUF_PREFETCH_PARSE_TABLE
      ::_pbi::TcParser::GetTable<::nodeinstance::manifest::v1::Function>(),  // to_prefetch
      #endif  // PROTOBUF_PREFETCH_PARSE_TABLE
    }, {{
      {::_pbi::TcParser::MiniParse, {}},
      // string name = 1;
      {::_pbi::TcParser::FastUS1,
       {10, 2, 0,
        PROTOBUF_FIELD_OFFSET(Function, _impl_.name_)}},
      // string help = 2;
      {::_pbi::TcParser::FastUS1,
       {18, 3, 0,
        PROTOBUF_FIELD_OFFSET(Function, _impl_.help_)}},
      // repeated string tags = 3;
      {::_pbi::TcParser::FastUR1,
       {26, 0, 0,
        PROTOBUF_FIELD_OFFSET(Function, _impl_.tags_)}},
      // uint32 priority = 4;
      {::_pbi::TcParser::SingularVarintNoZag1<::uint32_t, offsetof(Function, _impl_.priority_), 4>(),
       {32, 4, 0,
        PROTOBUF_FIELD_OFFSET(Function, _impl_.priority_)}},
      // uint32 version = 5;
      {::_pbi::TcParser::SingularVarintNoZag1<::uint32_t, offsetof(Function, _impl_.version_), 5>(),
       {40, 5, 0,
        PROTOBUF_FIELD_OFFSET(Function, _impl_.version_)}},
      // repeated .nodeinstance.manifest.v1.HttpAccess access = 6;
      {::_pbi::TcParser::FastV32P1,
       {50, 1, 0,
        PROTOBUF_FIELD_OFFSET(Function, _impl_.access_)}},
      {::_pbi::TcParser::MiniParse, {}},
    }}, {{
      65535, 65535
    }}, {{
      // string name = 1;
      {PROTOBUF_FIELD_OFFSET(Function, _impl_.name_), _Internal::kHasBitsOffset + 2, 0, (0 | ::_fl::kFcOptional | ::_fl::kUtf8String | ::_fl::kRepAString)},
      // string help = 2;
      {PROTOBUF_FIELD_OFFSET(Function, _impl_.help_), _Internal::kHasBitsOffset + 3, 0, (0 | ::_fl::kFcOptional | ::_fl::kUtf8String | ::_fl::kRepAString)},
      // repeated string tags = 3;
      {PROTOBUF_FIELD_OFFSET(Function, _impl_.tags_), _Internal::kHasBitsOffset + 0, 0, (0 | ::_fl::kFcRepeated | ::_fl::kUtf8String | ::_fl::kRepSString)},
      // uint32 priority = 4;
      {PROTOBUF_FIELD_OFFSET(Function, _impl_.priority_), _Internal::kHasBitsOffset + 4, 0, (0 | ::_fl::kFcOptional | ::_fl::kUInt32)},
      // uint32 version = 5;
      {PROTOBUF_FIELD_OFFSET(Function, _impl_.version_), _Internal::kHasBitsOffset + 5, 0, (0 | ::_fl::kFcOptional | ::_fl::kUInt32)},
      // repeated .nodeinstance.manifest.v1.HttpAccess access = 6;
      {PROTOBUF_FIELD_OFFSET(Function, _impl_.access_), _Internal::kHasBitsOffset + 1, 0, (0 | ::_fl::kFcRepeated | ::_fl::kPackedOpenEnum)},
    }},
    // no aux_entries
    {{
      "\41\4\4\4\0\0\0\0"
      "nodeinstance.manifest.v1.Function"
      "name"
      "help"
      "tags"
    }},
  };
}


inline constexpr Function::Impl_::Impl_(
    [[maybe_unused]] ::google::protobuf::internal::InternalVisibility visibility,
    ::_pbi::ConstantInitialized) noexcept
      : _cached_size_{0},
        tags_ { visibility, ::_pbi::InternalMetadataOffset::Build<
            ::nodeinstance::manifest::v1::Function,
            PROTOBUF_FIELD_OFFSET(::nodeinstance::manifest::v1::Function, _impl_.tags_)>()
         }
        ,
        access_ { visibility, ::_pbi::InternalMetadataOffset::Build<
            ::nodeinstance::manifest::v1::Function,
            PROTOBUF_FIELD_OFFSET(::nodeinstance::manifest::v1::Function, _impl_.access_)>()
         }
        ,
        _access_cached_byte_size_{0},
        name_(
            &::google::protobuf::internal::fixed_address_empty_string,
            ::_pbi::ConstantInitialized()),
        help_(
            &::google::protobuf::internal::fixed_address_empty_string,
            ::_pbi::ConstantInitialized()),
        priority_{0u},
        version_{0u} {}

template <typename>
constexpr Function::Function(::_pbi::ConstantInitialized,
                       const ::_pbi::ClassData* PROTOBUF_NONNULL class_data)
    : ::google::protobuf::Message(
#if defined(PROTOBUF_CUSTOM_VTABLE)
          class_data
#endif  // PROTOBUF_CUSTOM_VTABLE
          ),
      _impl_(internal_visibility(), ::_pbi::ConstantInitialized()) {
}
inline void* PROTOBUF_NONNULL Function::PlacementNew_(
    const void* PROTOBUF_NONNULL, void* PROTOBUF_NONNULL mem,
    ::google::protobuf::Arena* PROTOBUF_NULLABLE arena) {
  return ::new (mem) Function(arena);
}
constexpr auto Function::InternalNewImpl_() {
  return ::google::protobuf::internal::MessageCreator::CopyInit(sizeof(Function), alignof(Function));
}
constexpr auto Function::InternalGenerateClassData_(
    const MessageLite& prototype,
    const ::google::protobuf::internal::TcParseTableBase* tc_table) {
  return ::google::protobuf::internal::ClassDataFull{
      ::google::protobuf::internal::ClassData{
          &prototype,
#ifndef PROTOBUF_MESSAGE_GLOBALS
          &_table_.header,
#else
          tc_table,
#endif
          nullptr,  // IsInitialized
          &Function::MergeImpl,
          ::google::protobuf::Message::GetNewImpl<Function>(),
#if defined(PROTOBUF_CUSTOM_VTABLE)
          &Function::SharedDtor,
          ::google::protobuf::Message::GetClearImpl<Function>(), &Function::ByteSizeLong,
              &Function::_InternalSerialize,
#endif  // PROTOBUF_CUSTOM_VTABLE
          PROTOBUF_FIELD_OFFSET(Function, _impl_._cached_size_),
          false,
      },
#ifdef PROTOBUF_MESSAGE_GLOBALS
      &file_reflection_data[2],
#else   // !PROTOBUF_MESSAGE_GLOBALS
      &::_pbi::kDescriptorMethods,
      &descriptor_table_nodeinstance_2fmanifest_2fv1_2fmanifest_2eproto,
      nullptr,  // tracker
#endif  // PROTOBUF_MESSAGE_GLOBALS
  };
}
struct FunctionGlobalsTypeInternal : ::_pbi::MessageGlobalsBase {
  constexpr FunctionGlobalsTypeInternal()
      :
#ifndef PROTOBUF_MESSAGE_GLOBALS
        _default(::_pbi::ConstantInitialized{},
                 Function_class_data_.base())
#else   // !PROTOBUF_MESSAGE_GLOBALS
        MessageGlobalsBase(Function::InternalGenerateClassData_(
            _default, &Function_globals_._table.header)),
        _default(::_pbi::ConstantInitialized{}, GetClassData()),
        _table(::_pbi::PrivateAccess::GenerateParseTable<Function>(
            GetClassData()))
#endif  // PROTOBUF_MESSAGE_GLOBALS
  {
  }
  ~FunctionGlobalsTypeInternal() {}
  union {
    alignas(::_pbi::kMaxMessageAlignment) Function _default;
  };
#ifdef PROTOBUF_MESSAGE_GLOBALS
  decltype(::_pbi::PrivateAccess::GenerateParseTable<Function>(
      ::std::declval<const ::_pbi::ClassData*>())) _table;
#endif
};
#ifdef PROTOBUF_MESSAGE_GLOBALS
static_assert(PROTOBUF_FIELD_OFFSET(FunctionGlobalsTypeInternal, _default) ==
              ::_pbi::MessageGlobalsBase::OffsetToDefault());
#endif  // PROTOBUF_MESSAGE_GLOBALS

PROTOBUF_ATTRIBUTE_NO_DESTROY PROTOBUF_CONSTINIT
    PROTOBUF_ATTRIBUTE_INIT_PRIORITY1 PROTOBUF_MESSAGE_GLOBALS_CONST FunctionGlobalsTypeInternal Function_globals_
        PROTOBUF_MESSAGE_GLOBALS_SECTION(.data.rel.ro);
#if defined(PROTOBUF_CUSTOM_VTABLE)
namespace {
const ::_pbi::ClassData* Function_get_class_data() {
#ifdef PROTOBUF_MESSAGE_GLOBALS
  return Function_globals_.GetClassData();
#else
  return Function_class_data_.base();
#endif  // PROTOBUF_MESSAGE_GLOBALS
}
}  // namespace
#endif  // PROTOBUF_CUSTOM_VTABLE
class Functions::_Internal {
 public:
  using HasBits = decltype(::std::declval<Functions>()._impl_._has_bits_);
  static constexpr ::int32_t kHasBitsOffset =
      8 * PROTOBUF_FIELD_OFFSET(Functions, _impl_._has_bits_);
};

constexpr Functions::ParseTableT_ Functions::InternalGenerateParseTable_(const ::_pbi::ClassData* class_data) {
  return ParseTableT_{
    {
      PROTOBUF_FIELD_OFFSET(Functions, _impl_._has_bits_),
      0, // no _extensions_
      1, 0,  // max_field_number, fast_idx_mask
      offsetof(ParseTableT_, field_lookup_table),
      4294967294,  // skipmap
      offsetof(ParseTableT_, field_entries),
      1,  // num_field_entries
      1,  // num_aux_entries
      offsetof(ParseTableT_, aux_entries),
      class_data,
      nullptr,  // post_loop_handler
      ::_pbi::TcParser::GenericFallback,  // fallback
      #ifdef PROTOBUF_PREFETCH_PARSE_TABLE
      ::_pbi::TcParser::GetTable<::nodeinstance::manifest::v1::Functions>(),  // to_prefetch
      #endif  // PROTOBUF_PREFETCH_PARSE_TABLE
    }, {{
      // repeated .nodeinstance.manifest.v1.Function items = 1;
      {::_pbi::TcParser::FastMtR1,
       {10, 0, 0,
        PROTOBUF_FIELD_OFFSET(Functions, _impl_.items_)}},
    }}, {{
      65535, 65535
    }}, {{
      // repeated .nodeinstance.manifest.v1.Function items = 1;
      {PROTOBUF_FIELD_OFFSET(Functions, _impl_.items_), _Internal::kHasBitsOffset + 0, 0, (0 | ::_fl::kFcRepeated | ::_fl::kMessage | ::_fl::kTvTable)},
    }},
    {{
        #ifndef PROTOBUF_MESSAGE_GLOBALS
        {::_pbi::TcParser::GetTable<::nodeinstance::manifest::v1::Function>()},
        #else
        {::_pbi::FieldAuxMessageGlobals(), &::nodeinstance::manifest::v1::Function_globals_},
        #endif
    }},
    {{
    }},
  };
}


inline constexpr Functions::Impl_::Impl_(
    [[maybe_unused]] ::google::protobuf::internal::InternalVisibility visibility,
    ::_pbi::ConstantInitialized) noexcept
      : _cached_size_{0},
        items_ { visibility, ::_pbi::InternalMetadataOffset::Build<
            ::nodeinstance::manifest::v1::Functions,
            PROTOBUF_FIELD_OFFSET(::nodeinstance::manifest::v1::Functions, _impl_.items_)>()
         }
     {}

template <typename>
constexpr Functions::Functions(::_pbi::ConstantInitialized,
                       const ::_pbi::ClassData* PROTOBUF_NONNULL class_data)
    : ::google::protobuf::Message(
#if defined(PROTOBUF_CUSTOM_VTABLE)
          class_data
#endif  // PROTOBUF_CUSTOM_VTABLE
          ),
      _impl_(internal_visibility(), ::_pbi::ConstantInitialized()) {
}
inline void* PROTOBUF_NONNULL Functions::PlacementNew_(
    const void* PROTOBUF_NONNULL, void* PROTOBUF_NONNULL mem,
    ::google::protobuf::Arena* PROTOBUF_NULLABLE arena) {
  return ::new (mem) Functions(arena);
}
constexpr auto Functions::InternalNewImpl_() {
  return ::google::protobuf::internal::MessageCreator::CopyInit(sizeof(Functions), alignof(Functions));
}
constexpr auto Functions::InternalGenerateClassData_(
    const MessageLite& prototype,
    const ::google::protobuf::internal::TcParseTableBase* tc_table) {
  return ::google::protobuf::internal::ClassDataFull{
      ::google::protobuf::internal::ClassData{
          &prototype,
#ifndef PROTOBUF_MESSAGE_GLOBALS
          &_table_.header,
#else
          tc_table,
#endif
          nullptr,  // IsInitialized
          &Functions::MergeImpl,
          ::google::protobuf::Message::GetNewImpl<Functions>(),
#if defined(PROTOBUF_CUSTOM_VTABLE)
          &Functions::SharedDtor,
          ::google::protobuf::Message::GetClearImpl<Functions>(), &Functions::ByteSizeLong,
              &Functions::_InternalSerialize,
#endif  // PROTOBUF_CUSTOM_VTABLE
          PROTOBUF_FIELD_OFFSET(Functions, _impl_._cached_size_),
          false,
      },
#ifdef PROTOBUF_MESSAGE_GLOBALS
      &file_reflection_data[1],
#else   // !PROTOBUF_MESSAGE_GLOBALS
      &::_pbi::kDescriptorMethods,
      &descriptor_table_nodeinstance_2fmanifest_2fv1_2fmanifest_2eproto,
      nullptr,  // tracker
#endif  // PROTOBUF_MESSAGE_GLOBALS
  };
}
struct FunctionsGlobalsTypeInternal : ::_pbi::MessageGlobalsBase {
  constexpr FunctionsGlobalsTypeInternal()
      :
#ifndef PROTOBUF_MESSAGE_GLOBALS
        _default(::_pbi::ConstantInitialized{},
                 Functions_class_data_.base())
#else   // !PROTOBUF_MESSAGE_GLOBALS
        MessageGlobalsBase(Functions::InternalGenerateClassData_(
            _default, &Functions_globals_._table.header)),
        _default(::_pbi::ConstantInitialized{}, GetClassData()),
        _table(::_pbi::PrivateAccess::GenerateParseTable<Functions>(
            GetClassData()))
#endif  // PROTOBUF_MESSAGE_GLOBALS
  {
  }
  ~FunctionsGlobalsTypeInternal() {}
  union {
    alignas(::_pbi::kMaxMessageAlignment) Functions _default;
  };
#ifdef PROTOBUF_MESSAGE_GLOBALS
  decltype(::_pbi::PrivateAccess::GenerateParseTable<Functions>(
      ::std::declval<const ::_pbi::ClassData*>())) _table;
#endif
};
#ifdef PROTOBUF_MESSAGE_GLOBALS
static_assert(PROTOBUF_FIELD_OFFSET(FunctionsGlobalsTypeInternal, _default) ==
              ::_pbi::MessageGlobalsBase::OffsetToDefault());
#endif  // PROTOBUF_MESSAGE_GLOBALS

PROTOBUF_ATTRIBUTE_NO_DESTROY PROTOBUF_CONSTINIT
    PROTOBUF_ATTRIBUTE_INIT_PRIORITY1 PROTOBUF_MESSAGE_GLOBALS_CONST FunctionsGlobalsTypeInternal Functions_globals_
        PROTOBUF_MESSAGE_GLOBALS_SECTION(.data.rel.ro);
#if defined(PROTOBUF_CUSTOM_VTABLE)
namespace {
const ::_pbi::ClassData* Functions_get_class_data() {
#ifdef PROTOBUF_MESSAGE_GLOBALS
  return Functions_globals_.GetClassData();
#else
  return Functions_class_data_.base();
#endif  // PROTOBUF_MESSAGE_GLOBALS
}
}  // namespace
#endif  // PROTOBUF_CUSTOM_VTABLE
class UpdateNodeInstanceManifest::_Internal {
 public:
  using HasBits = decltype(::std::declval<UpdateNodeInstanceManifest>()._impl_._has_bits_);
  static constexpr ::int32_t kHasBitsOffset =
      8 * PROTOBUF_FIELD_OFFSET(UpdateNodeInstanceManifest, _impl_._has_bits_);
  static constexpr ::int32_t kOneofCaseOffset =
      PROTOBUF_FIELD_OFFSET(::nodeinstance::manifest::v1::UpdateNodeInstanceManifest, _impl_._oneof_case_);
};

constexpr UpdateNodeInstanceManifest::ParseTableT_ UpdateNodeInstanceManifest::InternalGenerateParseTable_(const ::_pbi::ClassData* class_data) {
  return ParseTableT_{
    {
      PROTOBUF_FIELD_OFFSET(UpdateNodeInstanceManifest, _impl_._has_bits_),
      0, // no _extensions_
      10, 24,  // max_field_number, fast_idx_mask
      offsetof(ParseTableT_, field_lookup_table),
      4294966776,  // skipmap
      offsetof(ParseTableT_, field_entries),
      4,  // num_field_entries
      2,  // num_aux_entries
      offsetof(ParseTableT_, aux_entries),
      class_data,
      nullptr,  // post_loop_handler
      ::_pbi::TcParser::GenericFallback,  // fallback
      #ifdef PROTOBUF_PREFETCH_PARSE_TABLE
      ::_pbi::TcParser::GetTable<::nodeinstance::manifest::v1::UpdateNodeInstanceManifest>(),  // to_prefetch
      #endif  // PROTOBUF_PREFETCH_PARSE_TABLE
    }, {{
      {::_pbi::TcParser::MiniParse, {}},
      // string claim_id = 1;
      {::_pbi::TcParser::FastUS1,
       {10, 0, 0,
        PROTOBUF_FIELD_OFFSET(UpdateNodeInstanceManifest, _impl_.claim_id_)}},
      // string node_id = 2;
      {::_pbi::TcParser::FastUS1,
       {18, 1, 0,
        PROTOBUF_FIELD_OFFSET(UpdateNodeInstanceManifest, _impl_.node_id_)}},
      // .google.protobuf.Timestamp updated_at = 3;
      {::_pbi::TcParser::FastMtS1,
       {26, 2, 0,
        PROTOBUF_FIELD_OFFSET(UpdateNodeInstanceManifest, _impl_.updated_at_)}},
    }}, {{
      65535, 65535
    }}, {{
      // string claim_id = 1;
      {PROTOBUF_FIELD_OFFSET(UpdateNodeInstanceManifest, _impl_.claim_id_), _Internal::kHasBitsOffset + 0, 0, (0 | ::_fl::kFcOptional | ::_fl::kUtf8String | ::_fl::kRepAString)},
      // string node_id = 2;
      {PROTOBUF_FIELD_OFFSET(UpdateNodeInstanceManifest, _impl_.node_id_), _Internal::kHasBitsOffset + 1, 0, (0 | ::_fl::kFcOptional | ::_fl::kUtf8String | ::_fl::kRepAString)},
      // .google.protobuf.Timestamp updated_at = 3;
      {PROTOBUF_FIELD_OFFSET(UpdateNodeInstanceManifest, _impl_.updated_at_), _Internal::kHasBitsOffset + 2, 0, (0 | ::_fl::kFcOptional | ::_fl::kMessage | ::_fl::kTvTable)},
      // .nodeinstance.manifest.v1.Functions functions = 10;
      {PROTOBUF_FIELD_OFFSET(UpdateNodeInstanceManifest, _impl_.payload_.functions_), _Internal::kOneofCaseOffset + 0, 1, (0 | ::_fl::kFcOneof | ::_fl::kMessage | ::_fl::kTvTable)},
    }},
    {{
        #ifndef PROTOBUF_MESSAGE_GLOBALS
        {::_pbi::TcParser::GetTable<::google::protobuf::Timestamp>()},
        #else
        {::_pbi::FieldAuxMessageGlobals(), &::google::protobuf::Timestamp_globals_},
        #endif
        #ifndef PROTOBUF_MESSAGE_GLOBALS
        {::_pbi::TcParser::GetTable<::nodeinstance::manifest::v1::Functions>()},
        #else
        {::_pbi::FieldAuxMessageGlobals(), &::nodeinstance::manifest::v1::Functions_globals_},
        #endif
    }},
    {{
      "\63\10\7\0\0\0\0\0"
      "nodeinstance.manifest.v1.UpdateNodeInstanceManifest"
      "claim_id"
      "node_id"
    }},
  };
}


inline constexpr UpdateNodeInstanceManifest::Impl_::Impl_(
    [[maybe_unused]] ::google::protobuf::internal::InternalVisibility visibility,
    ::_pbi::ConstantInitialized) noexcept
      : _cached_size_{0},
        claim_id_(
            &::google::protobuf::internal::fixed_address_empty_string,
            ::_pbi::ConstantInitialized()),
        node_id_(
            &::google::protobuf::internal::fixed_address_empty_string,
            ::_pbi::ConstantInitialized()),
        updated_at_{nullptr},
        payload_{},
        _oneof_case_{} {}

template <typename>
constexpr UpdateNodeInstanceManifest::UpdateNodeInstanceManifest(::_pbi::ConstantInitialized,
                       const ::_pbi::ClassData* PROTOBUF_NONNULL class_data)
    : ::google::protobuf::Message(
#if defined(PROTOBUF_CUSTOM_VTABLE)
          class_data
#endif  // PROTOBUF_CUSTOM_VTABLE
          ),
      _impl_(internal_visibility(), ::_pbi::ConstantInitialized()) {
}
inline void* PROTOBUF_NONNULL UpdateNodeInstanceManifest::PlacementNew_(
    const void* PROTOBUF_NONNULL, void* PROTOBUF_NONNULL mem,
    ::google::protobuf::Arena* PROTOBUF_NULLABLE arena) {
  return ::new (mem) UpdateNodeInstanceManifest(arena);
}
constexpr auto UpdateNodeInstanceManifest::InternalNewImpl_() {
  return ::google::protobuf::internal::MessageCreator::CopyInit(sizeof(UpdateNodeInstanceManifest), alignof(UpdateNodeInstanceManifest));
}
constexpr auto UpdateNodeInstanceManifest::InternalGenerateClassData_(
    const MessageLite& prototype,
    const ::google::protobuf::internal::TcParseTableBase* tc_table) {
  return ::google::protobuf::internal::ClassDataFull{
      ::google::protobuf::internal::ClassData{
          &prototype,
#ifndef PROTOBUF_MESSAGE_GLOBALS
          &_table_.header,
#else
          tc_table,
#endif
          nullptr,  // IsInitialized
          &UpdateNodeInstanceManifest::MergeImpl,
          ::google::protobuf::Message::GetNewImpl<UpdateNodeInstanceManifest>(),
#if defined(PROTOBUF_CUSTOM_VTABLE)
          &UpdateNodeInstanceManifest::SharedDtor,
          ::google::protobuf::Message::GetClearImpl<UpdateNodeInstanceManifest>(), &UpdateNodeInstanceManifest::ByteSizeLong,
              &UpdateNodeInstanceManifest::_InternalSerialize,
#endif  // PROTOBUF_CUSTOM_VTABLE
          PROTOBUF_FIELD_OFFSET(UpdateNodeInstanceManifest, _impl_._cached_size_),
          false,
      },
#ifdef PROTOBUF_MESSAGE_GLOBALS
      &file_reflection_data[0],
#else   // !PROTOBUF_MESSAGE_GLOBALS
      &::_pbi::kDescriptorMethods,
      &descriptor_table_nodeinstance_2fmanifest_2fv1_2fmanifest_2eproto,
      nullptr,  // tracker
#endif  // PROTOBUF_MESSAGE_GLOBALS
  };
}
struct UpdateNodeInstanceManifestGlobalsTypeInternal : ::_pbi::MessageGlobalsBase {
  constexpr UpdateNodeInstanceManifestGlobalsTypeInternal()
      :
#ifndef PROTOBUF_MESSAGE_GLOBALS
        _default(::_pbi::ConstantInitialized{},
                 UpdateNodeInstanceManifest_class_data_.base())
#else   // !PROTOBUF_MESSAGE_GLOBALS
        MessageGlobalsBase(UpdateNodeInstanceManifest::InternalGenerateClassData_(
            _default, &UpdateNodeInstanceManifest_globals_._table.header)),
        _default(::_pbi::ConstantInitialized{}, GetClassData()),
        _table(::_pbi::PrivateAccess::GenerateParseTable<UpdateNodeInstanceManifest>(
            GetClassData()))
#endif  // PROTOBUF_MESSAGE_GLOBALS
  {
  }
  ~UpdateNodeInstanceManifestGlobalsTypeInternal() {}
  union {
    alignas(::_pbi::kMaxMessageAlignment) UpdateNodeInstanceManifest _default;
  };
#ifdef PROTOBUF_MESSAGE_GLOBALS
  decltype(::_pbi::PrivateAccess::GenerateParseTable<UpdateNodeInstanceManifest>(
      ::std::declval<const ::_pbi::ClassData*>())) _table;
#endif
};
#ifdef PROTOBUF_MESSAGE_GLOBALS
static_assert(PROTOBUF_FIELD_OFFSET(UpdateNodeInstanceManifestGlobalsTypeInternal, _default) ==
              ::_pbi::MessageGlobalsBase::OffsetToDefault());
#endif  // PROTOBUF_MESSAGE_GLOBALS

PROTOBUF_ATTRIBUTE_NO_DESTROY PROTOBUF_CONSTINIT
    PROTOBUF_ATTRIBUTE_INIT_PRIORITY1 PROTOBUF_MESSAGE_GLOBALS_CONST UpdateNodeInstanceManifestGlobalsTypeInternal UpdateNodeInstanceManifest_globals_
        PROTOBUF_MESSAGE_GLOBALS_SECTION(.data.rel.ro);
#if defined(PROTOBUF_CUSTOM_VTABLE)
namespace {
const ::_pbi::ClassData* UpdateNodeInstanceManifest_get_class_data() {
#ifdef PROTOBUF_MESSAGE_GLOBALS
  return UpdateNodeInstanceManifest_globals_.GetClassData();
#else
  return UpdateNodeInstanceManifest_class_data_.base();
#endif  // PROTOBUF_MESSAGE_GLOBALS
}
}  // namespace
#endif  // PROTOBUF_CUSTOM_VTABLE
}  // namespace v1
}  // namespace manifest
}  // namespace nodeinstance
static const ::_pb::EnumDescriptor* PROTOBUF_NONNULL
    file_level_enum_descriptors_nodeinstance_2fmanifest_2fv1_2fmanifest_2eproto[1];
static constexpr const ::_pb::ServiceDescriptor* PROTOBUF_NONNULL* PROTOBUF_NULLABLE
    file_level_service_descriptors_nodeinstance_2fmanifest_2fv1_2fmanifest_2eproto = nullptr;
const ::uint32_t
    TableStruct_nodeinstance_2fmanifest_2fv1_2fmanifest_2eproto::offsets[] ABSL_ATTRIBUTE_SECTION_VARIABLE(
        protodesc_cold) = {
        0x085, // bitmap
        PROTOBUF_FIELD_OFFSET(::nodeinstance::manifest::v1::UpdateNodeInstanceManifest, _impl_._has_bits_),
        PROTOBUF_FIELD_OFFSET(::nodeinstance::manifest::v1::UpdateNodeInstanceManifest, _impl_._oneof_case_[0]),
        9, // hasbit index offset
        PROTOBUF_FIELD_OFFSET(::nodeinstance::manifest::v1::UpdateNodeInstanceManifest, _impl_.claim_id_),
        PROTOBUF_FIELD_OFFSET(::nodeinstance::manifest::v1::UpdateNodeInstanceManifest, _impl_.node_id_),
        PROTOBUF_FIELD_OFFSET(::nodeinstance::manifest::v1::UpdateNodeInstanceManifest, _impl_.updated_at_),
        PROTOBUF_FIELD_OFFSET(::nodeinstance::manifest::v1::UpdateNodeInstanceManifest, _impl_.payload_),
        PROTOBUF_FIELD_OFFSET(::nodeinstance::manifest::v1::UpdateNodeInstanceManifest, _impl_.payload_),
        0,
        1,
        2,
        ~0u,
        0x081, // bitmap
        PROTOBUF_FIELD_OFFSET(::nodeinstance::manifest::v1::Functions, _impl_._has_bits_),
        4, // hasbit index offset
        PROTOBUF_FIELD_OFFSET(::nodeinstance::manifest::v1::Functions, _impl_.items_),
        0,
        0x081, // bitmap
        PROTOBUF_FIELD_OFFSET(::nodeinstance::manifest::v1::Function, _impl_._has_bits_),
        9, // hasbit index offset
        PROTOBUF_FIELD_OFFSET(::nodeinstance::manifest::v1::Function, _impl_.name_),
        PROTOBUF_FIELD_OFFSET(::nodeinstance::manifest::v1::Function, _impl_.help_),
        PROTOBUF_FIELD_OFFSET(::nodeinstance::manifest::v1::Function, _impl_.tags_),
        PROTOBUF_FIELD_OFFSET(::nodeinstance::manifest::v1::Function, _impl_.priority_),
        PROTOBUF_FIELD_OFFSET(::nodeinstance::manifest::v1::Function, _impl_.version_),
        PROTOBUF_FIELD_OFFSET(::nodeinstance::manifest::v1::Function, _impl_.access_),
        2,
        3,
        0,
        4,
        5,
        1,
};

static const ::_pbi::MigrationSchema
    schemas[] ABSL_ATTRIBUTE_SECTION_VARIABLE(protodesc_cold) = {
        {0, sizeof(::nodeinstance::manifest::v1::UpdateNodeInstanceManifest)},
        {13, sizeof(::nodeinstance::manifest::v1::Functions)},
        {18, sizeof(::nodeinstance::manifest::v1::Function)},
};
static const ::_pbi::MessageGlobalsBase* PROTOBUF_NONNULL const
    file_message_globals[] = {
        &::nodeinstance::manifest::v1::UpdateNodeInstanceManifest_globals_,
        &::nodeinstance::manifest::v1::Functions_globals_,
        &::nodeinstance::manifest::v1::Function_globals_,
};
const char descriptor_table_protodef_nodeinstance_2fmanifest_2fv1_2fmanifest_2eproto[] ABSL_ATTRIBUTE_SECTION_VARIABLE(
    protodesc_cold) = {
    "\n\'nodeinstance/manifest/v1/manifest.prot"
    "o\022\030nodeinstance.manifest.v1\032\037google/prot"
    "obuf/timestamp.proto\"\264\001\n\032UpdateNodeInsta"
    "nceManifest\022\020\n\010claim_id\030\001 \001(\t\022\017\n\007node_id"
    "\030\002 \001(\t\022.\n\nupdated_at\030\003 \001(\0132\032.google.prot"
    "obuf.Timestamp\0228\n\tfunctions\030\n \001(\0132#.node"
    "instance.manifest.v1.FunctionsH\000B\t\n\007payl"
    "oad\">\n\tFunctions\0221\n\005items\030\001 \003(\0132\".nodein"
    "stance.manifest.v1.Function\"\215\001\n\010Function"
    "\022\014\n\004name\030\001 \001(\t\022\014\n\004help\030\002 \001(\t\022\014\n\004tags\030\003 \003"
    "(\t\022\020\n\010priority\030\004 \001(\r\022\017\n\007version\030\005 \001(\r\0224\n"
    "\006access\030\006 \003(\0162$.nodeinstance.manifest.v1"
    ".HttpAccess*\232\003\n\nHttpAccess\022\033\n\027HTTP_ACCES"
    "S_UNSPECIFIED\020\000\022\031\n\025HTTP_ACCESS_SIGNED_IN"
    "\020\001\022\032\n\026HTTP_ACCESS_SAME_SPACE\020\002\022\032\n\026HTTP_A"
    "CCESS_COMMERCIAL\020\003\022\036\n\032HTTP_ACCESS_ANONYM"
    "OUS_DATA\020\004\022\036\n\032HTTP_ACCESS_SENSITIVE_DATA"
    "\020\005\022\033\n\027HTTP_ACCESS_VIEW_CONFIG\020\006\022\033\n\027HTTP_"
    "ACCESS_EDIT_CONFIG\020\007\022)\n%HTTP_ACCESS_VIEW"
    "_NOTIFICATIONS_CONFIG\020\010\022)\n%HTTP_ACCESS_E"
    "DIT_NOTIFICATIONS_CONFIG\020\t\022%\n!HTTP_ACCES"
    "S_VIEW_ALERTS_SILENCING\020\n\022%\n!HTTP_ACCESS"
    "_EDIT_ALERTS_SILENCING\020\013B0Z.node_instanc"
    "e/manifest/v1;nodeinstancemanifestb\006prot"
    "o3"
};
static const ::_pbi::DescriptorTable* PROTOBUF_NONNULL const
    descriptor_table_nodeinstance_2fmanifest_2fv1_2fmanifest_2eproto_deps[1] = {
        &::descriptor_table_google_2fprotobuf_2ftimestamp_2eproto,
};
static ::absl::once_flag descriptor_table_nodeinstance_2fmanifest_2fv1_2fmanifest_2eproto_once;
PROTOBUF_CONSTINIT const ::_pbi::DescriptorTable descriptor_table_nodeinstance_2fmanifest_2fv1_2fmanifest_2eproto = {
    false,
    false,
    962,
    descriptor_table_protodef_nodeinstance_2fmanifest_2fv1_2fmanifest_2eproto,
    "nodeinstance/manifest/v1/manifest.proto",
    &descriptor_table_nodeinstance_2fmanifest_2fv1_2fmanifest_2eproto_once,
    descriptor_table_nodeinstance_2fmanifest_2fv1_2fmanifest_2eproto_deps,
    1,
    3,
    schemas,
    file_message_globals,
    TableStruct_nodeinstance_2fmanifest_2fv1_2fmanifest_2eproto::offsets,
    file_level_enum_descriptors_nodeinstance_2fmanifest_2fv1_2fmanifest_2eproto,
    file_level_service_descriptors_nodeinstance_2fmanifest_2fv1_2fmanifest_2eproto,
};
namespace nodeinstance {
namespace manifest {
namespace v1 {
[[nodiscard]] const ::google::protobuf::EnumDescriptor* PROTOBUF_NONNULL
HttpAccess_descriptor() {
  ::google::protobuf::internal::AssignDescriptors(&descriptor_table_nodeinstance_2fmanifest_2fv1_2fmanifest_2eproto);
  return file_level_enum_descriptors_nodeinstance_2fmanifest_2fv1_2fmanifest_2eproto[0];
}
PROTOBUF_CONSTINIT const uint32_t HttpAccess_internal_data_[] = {
    786432u, 0u, };
// ===================================================================

void UpdateNodeInstanceManifest::clear_updated_at() {
  ::google::protobuf::internal::TSanWrite(&_impl_);
  if (_impl_.updated_at_ != nullptr) _impl_.updated_at_->Clear();
  ClearHasBit(_impl_._has_bits_[0], 0x00000004U);
}
void UpdateNodeInstanceManifest::set_allocated_functions(::nodeinstance::manifest::v1::Functions* PROTOBUF_NULLABLE functions) {
  ::google::protobuf::Arena* message_arena = GetArena();
  clear_payload();
  if (functions) {
    ::google::protobuf::Arena* submessage_arena = functions->GetArena();
    if (message_arena != submessage_arena) {
      functions = ::google::protobuf::internal::GetOwnedMessage(message_arena, functions, submessage_arena);
    }
    set_has_functions();
    _impl_.payload_.functions_ = functions;
  }
  // @@protoc_insertion_point(field_set_allocated:nodeinstance.manifest.v1.UpdateNodeInstanceManifest.functions)
}
UpdateNodeInstanceManifest::UpdateNodeInstanceManifest(::google::protobuf::Arena* PROTOBUF_NULLABLE arena)
#if defined(PROTOBUF_CUSTOM_VTABLE)
    : ::google::protobuf::Message(arena, UpdateNodeInstanceManifest_get_class_data()) {
#else   // PROTOBUF_CUSTOM_VTABLE
    : ::google::protobuf::Message(arena) {
#endif  // PROTOBUF_CUSTOM_VTABLE
  SharedCtor(arena);
  // @@protoc_insertion_point(arena_constructor:nodeinstance.manifest.v1.UpdateNodeInstanceManifest)
}
PROTOBUF_NDEBUG_INLINE UpdateNodeInstanceManifest::Impl_::Impl_(
    [[maybe_unused]] ::google::protobuf::internal::InternalVisibility visibility,
    [[maybe_unused]] ::google::protobuf::Arena* PROTOBUF_NULLABLE arena, const Impl_& from,
    [[maybe_unused]] const ::nodeinstance::manifest::v1::UpdateNodeInstanceManifest& from_msg)
      : _has_bits_{from._has_bits_},
        _cached_size_{0},
        claim_id_(arena, from.claim_id_),
        node_id_(arena, from.node_id_),
        payload_{},
        _oneof_case_{from._oneof_case_[0]} {}

UpdateNodeInstanceManifest::UpdateNodeInstanceManifest(
    ::google::protobuf::Arena* PROTOBUF_NULLABLE arena,
    const UpdateNodeInstanceManifest& from)
#if defined(PROTOBUF_CUSTOM_VTABLE)
    : ::google::protobuf::Message(arena, UpdateNodeInstanceManifest_get_class_data()) {

#else   // PROTOBUF_CUSTOM_VTABLE
    : ::google::protobuf::Message(arena) {
#endif  // PROTOBUF_CUSTOM_VTABLE
  UpdateNodeInstanceManifest* const _this = this;
  (void)_this;
  _internal_metadata_.MergeFrom<::google::protobuf::UnknownFieldSet>(
      from._internal_metadata_);
  new (&_impl_) Impl_(internal_visibility(), arena, from._impl_, from);
  ::uint32_t cached_has_bits = _impl_._has_bits_[0];
  _impl_.updated_at_ = (CheckHasBit(cached_has_bits, 0x00000004U))
                ? ::google::protobuf::Message::CopyConstruct(arena, *from._impl_.updated_at_)
                : nullptr;
  switch (payload_case()) {
    case PAYLOAD_NOT_SET:
      break;
      case kFunctions:
        _impl_.payload_.functions_ = ::google::protobuf::Message::CopyConstruct(arena, *from._impl_.payload_.functions_);
        break;
  }

  // @@protoc_insertion_point(copy_constructor:nodeinstance.manifest.v1.UpdateNodeInstanceManifest)
}
PROTOBUF_NDEBUG_INLINE UpdateNodeInstanceManifest::Impl_::Impl_(
    [[maybe_unused]] ::google::protobuf::internal::InternalVisibility visibility,
    [[maybe_unused]] ::google::protobuf::Arena* PROTOBUF_NULLABLE arena)
      : _cached_size_{0},
        claim_id_(arena),
        node_id_(arena),
        payload_{},
        _oneof_case_{} {}

inline void UpdateNodeInstanceManifest::SharedCtor(::_pb::Arena* PROTOBUF_NULLABLE arena) {
  new (&_impl_) Impl_(internal_visibility(), arena);
  _impl_.updated_at_ = {};
}
UpdateNodeInstanceManifest::~UpdateNodeInstanceManifest() {
  // @@protoc_insertion_point(destructor:nodeinstance.manifest.v1.UpdateNodeInstanceManifest)
  SharedDtor(*this);
}
inline void UpdateNodeInstanceManifest::SharedDtor(MessageLite& self) {
  UpdateNodeInstanceManifest& this_ = static_cast<UpdateNodeInstanceManifest&>(self);
  if constexpr (::_pbi::DebugHardenCheckHasBitConsistency()) {
    this_.CheckHasBitConsistency();
  }
  this_._internal_metadata_.Delete<::google::protobuf::UnknownFieldSet>();
  ABSL_DCHECK(this_.GetArena() == nullptr);
  this_._impl_.claim_id_.Destroy();
  this_._impl_.node_id_.Destroy();
  delete this_._impl_.updated_at_;
  if (this_.has_payload()) {
    this_.clear_payload();
  }
  this_._impl_.~Impl_();
}

void UpdateNodeInstanceManifest::clear_payload() {
// @@protoc_insertion_point(one_of_clear_start:nodeinstance.manifest.v1.UpdateNodeInstanceManifest)
  ::google::protobuf::internal::TSanWrite(&_impl_);
  switch (payload_case()) {
    case kFunctions: {
      if (GetArena() == nullptr) {
        delete _impl_.payload_.functions_;
      } else if (::google::protobuf::internal::DebugHardenClearOneofMessageOnArena()) {
        ::google::protobuf::internal::MaybePoisonAfterClear(_impl_.payload_.functions_);
      }
      break;
    }
    case PAYLOAD_NOT_SET: {
      break;
    }
  }
  _impl_._oneof_case_[0] = PAYLOAD_NOT_SET;
}


#ifndef PROTOBUF_MESSAGE_GLOBALS
PROTOBUF_CONSTINIT PROTOBUF_ATTRIBUTE_INIT_PRIORITY1 const
    ::google::protobuf::internal::ClassDataFull UpdateNodeInstanceManifest_class_data_ =
        UpdateNodeInstanceManifest::InternalGenerateClassData_(UpdateNodeInstanceManifest_globals_._default);

PROTOBUF_ATTRIBUTE_WEAK const ::google::protobuf::internal::ClassData* PROTOBUF_NONNULL
UpdateNodeInstanceManifest::GetClassData() const {
  ::google::protobuf::internal::PrefetchToLocalCache(&UpdateNodeInstanceManifest_class_data_);
  ::google::protobuf::internal::PrefetchToLocalCache(UpdateNodeInstanceManifest_class_data_.tc_table);
  return UpdateNodeInstanceManifest_class_data_.base();
}
#else
PROTOBUF_ATTRIBUTE_WEAK const ::google::protobuf::internal::ClassData* PROTOBUF_NONNULL
UpdateNodeInstanceManifest::GetClassData() const {
  ::google::protobuf::internal::PrefetchToLocalCache(&UpdateNodeInstanceManifest_globals_);
  ::google::protobuf::internal::PrefetchToLocalCache(
      ::google::protobuf::internal::MessageGlobalsBase::ToParseTableBase(&UpdateNodeInstanceManifest_globals_));
  return UpdateNodeInstanceManifest_globals_.GetClassData();
}
#endif  // !PROTOBUF_MESSAGE_GLOBALS
#ifndef PROTOBUF_MESSAGE_GLOBALS
PROTOBUF_CONSTINIT
PROTOBUF_ATTRIBUTE_INIT_PRIORITY1 const UpdateNodeInstanceManifest::ParseTableT_
    UpdateNodeInstanceManifest::_table_ =
        UpdateNodeInstanceManifest::InternalGenerateParseTable_(UpdateNodeInstanceManifest_class_data_.base());
#endif  // !PROTOBUF_MESSAGE_GLOBALS
PROTOBUF_NOINLINE void UpdateNodeInstanceManifest::Clear() {
// @@protoc_insertion_point(message_clear_start:nodeinstance.manifest.v1.UpdateNodeInstanceManifest)
  ::google::protobuf::internal::TSanWrite(&_impl_);
  ::uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void) cached_has_bits;

  cached_has_bits = _impl_._has_bits_[0];
  if (BatchCheckHasBit(cached_has_bits, 0x00000007U)) {
    if (CheckHasBit(cached_has_bits, 0x00000001U)) {
      _impl_.claim_id_.ClearNonDefaultToEmpty();
    }
    if (CheckHasBit(cached_has_bits, 0x00000002U)) {
      _impl_.node_id_.ClearNonDefaultToEmpty();
    }
    if (CheckHasBit(cached_has_bits, 0x00000004U)) {
      ABSL_DCHECK(_impl_.updated_at_ != nullptr);
      _impl_.updated_at_->Clear();
    }
  }
  clear_payload();
  _impl_._has_bits_.Clear();
  _internal_metadata_.Clear<::google::protobuf::UnknownFieldSet>();
}

#if defined(PROTOBUF_CUSTOM_VTABLE)
::uint8_t* PROTOBUF_NONNULL UpdateNodeInstanceManifest::_InternalSerialize(
    const ::google::protobuf::MessageLite& base, ::uint8_t* PROTOBUF_NONNULL target,
    ::google::protobuf::io::EpsCopyOutputStream* PROTOBUF_NONNULL stream) {
  const UpdateNodeInstanceManifest& this_ = static_cast<const UpdateNodeInstanceManifest&>(base);
#else   // PROTOBUF_CUSTOM_VTABLE
::uint8_t* PROTOBUF_NONNULL UpdateNodeInstanceManifest::_InternalSerialize(
    ::uint8_t* PROTOBUF_NONNULL target,
    ::google::protobuf::io::EpsCopyOutputStream* PROTOBUF_NONNULL stream) const {
  const UpdateNodeInstanceManifest& this_ = *this;
#endif  // PROTOBUF_CUSTOM_VTABLE
  if constexpr (::_pbi::DebugHardenCheckHasBitConsistency()) {
    this_.CheckHasBitConsistency();
  }
  // @@protoc_insertion_point(serialize_to_array_start:nodeinstance.manifest.v1.UpdateNodeInstanceManifest)
  ::uint32_t cached_has_bits = 0;
  (void)cached_has_bits;

  cached_has_bits = this_._impl_._has_bits_[0];
  // string claim_id = 1;
  if (CheckHasBit(cached_has_bits, 0x00000001U)) {
    if (!this_._internal_claim_id().empty()) {
      const ::std::string& _s = this_._internal_claim_id();
      ::google::protobuf::internal::WireFormatLite::VerifyUtf8String(
          _s.data(), static_cast<int>(_s.length()), ::google::protobuf::internal::WireFormatLite::SERIALIZE, "nodeinstance.manifest.v1.UpdateNodeInstanceManifest.claim_id");
      target = stream->WriteStringMaybeAliased(1, _s, target);
    }
  }

  // string node_id = 2;
  if (CheckHasBit(cached_has_bits, 0x00000002U)) {
    if (!this_._internal_node_id().empty()) {
      const ::std::string& _s = this_._internal_node_id();
      ::google::protobuf::internal::WireFormatLite::VerifyUtf8String(
          _s.data(), static_cast<int>(_s.length()), ::google::protobuf::internal::WireFormatLite::SERIALIZE, "nodeinstance.manifest.v1.UpdateNodeInstanceManifest.node_id");
      target = stream->WriteStringMaybeAliased(2, _s, target);
    }
  }

  // .google.protobuf.Timestamp updated_at = 3;
  if (CheckHasBit(cached_has_bits, 0x00000004U)) {
    target = ::google::protobuf::internal::WireFormatLite::InternalWriteMessage(
        3, *this_._impl_.updated_at_, this_._impl_.updated_at_->GetCachedSize(), target,
        stream);
  }

  // .nodeinstance.manifest.v1.Functions functions = 10;
  if (this_.payload_case() == kFunctions) {
    target = ::google::protobuf::internal::WireFormatLite::InternalWriteMessage(
        10, *this_._impl_.payload_.functions_, this_._impl_.payload_.functions_->GetCachedSize(), target,
        stream);
  }

  if (ABSL_PREDICT_FALSE(this_._internal_metadata_.have_unknown_fields())) {
    target =
        ::_pbi::WireFormat::InternalSerializeUnknownFieldsToArray(
            this_._internal_metadata_.unknown_fields<::google::protobuf::UnknownFieldSet>(::google::protobuf::UnknownFieldSet::default_instance), target, stream);
  }
  // @@protoc_insertion_point(serialize_to_array_end:nodeinstance.manifest.v1.UpdateNodeInstanceManifest)
  return target;
}

#if defined(PROTOBUF_CUSTOM_VTABLE)
::size_t UpdateNodeInstanceManifest::ByteSizeLong(const MessageLite& base) {
  const UpdateNodeInstanceManifest& this_ = static_cast<const UpdateNodeInstanceManifest&>(base);
#else   // PROTOBUF_CUSTOM_VTABLE
::size_t UpdateNodeInstanceManifest::ByteSizeLong() const {
  const UpdateNodeInstanceManifest& this_ = *this;
#endif  // PROTOBUF_CUSTOM_VTABLE
  // @@protoc_insertion_point(message_byte_size_start:nodeinstance.manifest.v1.UpdateNodeInstanceManifest)
  ::size_t total_size = 0;

  ::uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void)cached_has_bits;

  ::_pbi::Prefetch5LinesFrom7Lines(&this_);
  cached_has_bits = this_._impl_._has_bits_[0];
  if (BatchCheckHasBit(cached_has_bits, 0x00000007U)) {
    // string claim_id = 1;
    if (CheckHasBit(cached_has_bits, 0x00000001U)) {
      if (!this_._internal_claim_id().empty()) {
        total_size += 1 + ::google::protobuf::internal::WireFormatLite::StringSize(
                                        this_._internal_claim_id());
      }
    }
    // string node_id = 2;
    if (CheckHasBit(cached_has_bits, 0x00000002U)) {
      if (!this_._internal_node_id().empty()) {
        total_size += 1 + ::google::protobuf::internal::WireFormatLite::StringSize(
                                        this_._internal_node_id());
      }
    }
    // .google.protobuf.Timestamp updated_at = 3;
    if (CheckHasBit(cached_has_bits, 0x00000004U)) {
      total_size += 1 +
                    ::google::protobuf::internal::WireFormatLite::MessageSize(*this_._impl_.updated_at_);
    }
  }
  switch (this_.payload_case()) {
    // .nodeinstance.manifest.v1.Functions functions = 10;
    case kFunctions: {
      total_size += 1 +
                    ::google::protobuf::internal::WireFormatLite::MessageSize(*this_._impl_.payload_.functions_);
      break;
    }
    case PAYLOAD_NOT_SET: {
      break;
    }
  }
  return this_.MaybeComputeUnknownFieldsSize(total_size,
                                             &this_._impl_._cached_size_);
}

void UpdateNodeInstanceManifest::MergeImpl(::google::protobuf::MessageLite& to_msg,
                      const ::google::protobuf::MessageLite& from_msg) {
   auto* const _this = static_cast<UpdateNodeInstanceManifest*>(&to_msg);
  auto& from = static_cast<const UpdateNodeInstanceManifest&>(from_msg);
  if constexpr (::_pbi::DebugHardenCheckHasBitConsistency()) {
    from.CheckHasBitConsistency();
  }
  ::google::protobuf::Arena* arena = _this->GetArena();
  // @@protoc_insertion_point(class_specific_merge_from_start:nodeinstance.manifest.v1.UpdateNodeInstanceManifest)
  ABSL_DCHECK_NE(&from, _this);
  ::uint32_t cached_has_bits = 0;
  (void)cached_has_bits;

  cached_has_bits = from._impl_._has_bits_[0];
  if (BatchCheckHasBit(cached_has_bits, 0x00000007U)) {
    if (CheckHasBit(cached_has_bits, 0x00000001U)) {
      if (!from._internal_claim_id().empty()) {
        _this->_internal_set_claim_id(from._internal_claim_id());
      } else {
        if (_this->_impl_.claim_id_.IsDefault()) {
          _this->_internal_set_claim_id("");
        }
      }
    }
    if (CheckHasBit(cached_has_bits, 0x00000002U)) {
      if (!from._internal_node_id().empty()) {
        _this->_internal_set_node_id(from._internal_node_id());
      } else {
        if (_this->_impl_.node_id_.IsDefault()) {
          _this->_internal_set_node_id("");
        }
      }
    }
    if (CheckHasBit(cached_has_bits, 0x00000004U)) {
      ABSL_DCHECK(from._impl_.updated_at_ != nullptr);
      if (_this->_impl_.updated_at_ == nullptr) {
        _this->_impl_.updated_at_ = ::google::protobuf::Message::CopyConstruct(arena, *from._impl_.updated_at_);
      } else {
        _this->_impl_.updated_at_->MergeFrom(*from._impl_.updated_at_);
      }
    }
  }
  _this->_impl_._has_bits_[0] |= cached_has_bits;
  if (const uint32_t oneof_from_case =
          from._impl_._oneof_case_[0]) {
    const uint32_t oneof_to_case = _this->_impl_._oneof_case_[0];
    const bool oneof_needs_init = oneof_to_case != oneof_from_case;
    if (oneof_needs_init) {
      if (oneof_to_case != 0) {
        _this->clear_payload();
      }
      _this->_impl_._oneof_case_[0] = oneof_from_case;
    }

    switch (oneof_from_case) {
      case kFunctions: {
        if (oneof_needs_init) {
          _this->_impl_.payload_.functions_ = ::google::protobuf::Message::CopyConstruct(arena, *from._impl_.payload_.functions_);
        } else {
          _this->_impl_.payload_.functions_->MergeFrom(*from._impl_.payload_.functions_);
        }
        break;
      }
      case PAYLOAD_NOT_SET:
        break;
    }
  }
  _this->_internal_metadata_.MergeFrom<::google::protobuf::UnknownFieldSet>(
      from._internal_metadata_);
}

void UpdateNodeInstanceManifest::CopyFrom(const UpdateNodeInstanceManifest& from) {
  // @@protoc_insertion_point(class_specific_copy_from_start:nodeinstance.manifest.v1.UpdateNodeInstanceManifest)
  if (&from == this) return;
  Clear();
  MergeFrom(from);
}


void UpdateNodeInstanceManifest::InternalSwap(UpdateNodeInstanceManifest* PROTOBUF_RESTRICT PROTOBUF_NONNULL other) {
  using ::std::swap;
  auto* arena = GetArena();
  ABSL_DCHECK_EQ(arena, other->GetArena());
  _internal_metadata_.InternalSwap(&other->_internal_metadata_);
  swap(_impl_._has_bits_[0], other->_impl_._has_bits_[0]);
  ::_pbi::ArenaStringPtr::InternalSwap(&_impl_.claim_id_, &other->_impl_.claim_id_, arena);
  ::_pbi::ArenaStringPtr::InternalSwap(&_impl_.node_id_, &other->_impl_.node_id_, arena);
  swap(_impl_.updated_at_, other->_impl_.updated_at_);
  swap(_impl_.payload_, other->_impl_.payload_);
  swap(_impl_._oneof_case_[0], other->_impl_._oneof_case_[0]);
}

::google::protobuf::Metadata UpdateNodeInstanceManifest::GetMetadata() const {
  return ::google::protobuf::Message::GetMetadataImpl(GetClassData()->full());
}
// ===================================================================

Functions::Functions(::google::protobuf::Arena* PROTOBUF_NULLABLE arena)
#if defined(PROTOBUF_CUSTOM_VTABLE)
    : ::google::protobuf::Message(arena, Functions_get_class_data()) {
#else   // PROTOBUF_CUSTOM_VTABLE
    : ::google::protobuf::Message(arena) {
#endif  // PROTOBUF_CUSTOM_VTABLE
  SharedCtor(arena);
  // @@protoc_insertion_point(arena_constructor:nodeinstance.manifest.v1.Functions)
}
PROTOBUF_NDEBUG_INLINE Functions::Impl_::Impl_(
    [[maybe_unused]] ::google::protobuf::internal::InternalVisibility visibility,
    [[maybe_unused]] ::google::protobuf::Arena* PROTOBUF_NULLABLE arena, const Impl_& from,
    [[maybe_unused]] const ::nodeinstance::manifest::v1::Functions& from_msg)
      : _has_bits_{from._has_bits_},
        _cached_size_{0},
        items_ {
          visibility, ::_pbi::InternalMetadataOffset::Build<
              ::nodeinstance::manifest::v1::Functions,
              PROTOBUF_FIELD_OFFSET(::nodeinstance::manifest::v1::Functions, _impl_.items_)>()
          , from.items_
        }
     {}

Functions::Functions(
    ::google::protobuf::Arena* PROTOBUF_NULLABLE arena,
    const Functions& from)
#if defined(PROTOBUF_CUSTOM_VTABLE)
    : ::google::protobuf::Message(arena, Functions_get_class_data()) {

#else   // PROTOBUF_CUSTOM_VTABLE
    : ::google::protobuf::Message(arena) {
#endif  // PROTOBUF_CUSTOM_VTABLE
  Functions* const _this = this;
  (void)_this;
  _internal_metadata_.MergeFrom<::google::protobuf::UnknownFieldSet>(
      from._internal_metadata_);
  new (&_impl_) Impl_(internal_visibility(), arena, from._impl_, from);

  // @@protoc_insertion_point(copy_constructor:nodeinstance.manifest.v1.Functions)
}
PROTOBUF_NDEBUG_INLINE Functions::Impl_::Impl_(
    [[maybe_unused]] ::google::protobuf::internal::InternalVisibility visibility,
    [[maybe_unused]] ::google::protobuf::Arena* PROTOBUF_NULLABLE arena)
      : _cached_size_{0},
        items_ { visibility, ::_pbi::InternalMetadataOffset::Build<
            ::nodeinstance::manifest::v1::Functions,
            PROTOBUF_FIELD_OFFSET(::nodeinstance::manifest::v1::Functions, _impl_.items_)>()
         }
     {}

inline void Functions::SharedCtor(::_pb::Arena* PROTOBUF_NULLABLE arena) {
  new (&_impl_) Impl_(internal_visibility(), arena);
}
Functions::~Functions() {
  // @@protoc_insertion_point(destructor:nodeinstance.manifest.v1.Functions)
  SharedDtor(*this);
}
inline void Functions::SharedDtor(MessageLite& self) {
  Functions& this_ = static_cast<Functions&>(self);
  if constexpr (::_pbi::DebugHardenCheckHasBitConsistency()) {
    this_.CheckHasBitConsistency();
  }
  this_._internal_metadata_.Delete<::google::protobuf::UnknownFieldSet>();
  ABSL_DCHECK(this_.GetArena() == nullptr);
  this_._impl_.~Impl_();
}

#ifndef PROTOBUF_MESSAGE_GLOBALS
PROTOBUF_CONSTINIT PROTOBUF_ATTRIBUTE_INIT_PRIORITY1 const
    ::google::protobuf::internal::ClassDataFull Functions_class_data_ =
        Functions::InternalGenerateClassData_(Functions_globals_._default);

PROTOBUF_ATTRIBUTE_WEAK const ::google::protobuf::internal::ClassData* PROTOBUF_NONNULL
Functions::GetClassData() const {
  ::google::protobuf::internal::PrefetchToLocalCache(&Functions_class_data_);
  ::google::protobuf::internal::PrefetchToLocalCache(Functions_class_data_.tc_table);
  return Functions_class_data_.base();
}
#else
PROTOBUF_ATTRIBUTE_WEAK const ::google::protobuf::internal::ClassData* PROTOBUF_NONNULL
Functions::GetClassData() const {
  ::google::protobuf::internal::PrefetchToLocalCache(&Functions_globals_);
  ::google::protobuf::internal::PrefetchToLocalCache(
      ::google::protobuf::internal::MessageGlobalsBase::ToParseTableBase(&Functions_globals_));
  return Functions_globals_.GetClassData();
}
#endif  // !PROTOBUF_MESSAGE_GLOBALS
#ifndef PROTOBUF_MESSAGE_GLOBALS
PROTOBUF_CONSTINIT
PROTOBUF_ATTRIBUTE_INIT_PRIORITY1 const Functions::ParseTableT_
    Functions::_table_ =
        Functions::InternalGenerateParseTable_(Functions_class_data_.base());
#endif  // !PROTOBUF_MESSAGE_GLOBALS
PROTOBUF_NOINLINE void Functions::Clear() {
// @@protoc_insertion_point(message_clear_start:nodeinstance.manifest.v1.Functions)
  ::google::protobuf::internal::TSanWrite(&_impl_);
  ::uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void) cached_has_bits;

  cached_has_bits = _impl_._has_bits_[0];
  if (CheckHasBit(cached_has_bits, 0x00000001U)) {
    _impl_.items_.Clear();
  }
  _impl_._has_bits_.Clear();
  _internal_metadata_.Clear<::google::protobuf::UnknownFieldSet>();
}

#if defined(PROTOBUF_CUSTOM_VTABLE)
::uint8_t* PROTOBUF_NONNULL Functions::_InternalSerialize(
    const ::google::protobuf::MessageLite& base, ::uint8_t* PROTOBUF_NONNULL target,
    ::google::protobuf::io::EpsCopyOutputStream* PROTOBUF_NONNULL stream) {
  const Functions& this_ = static_cast<const Functions&>(base);
#else   // PROTOBUF_CUSTOM_VTABLE
::uint8_t* PROTOBUF_NONNULL Functions::_InternalSerialize(
    ::uint8_t* PROTOBUF_NONNULL target,
    ::google::protobuf::io::EpsCopyOutputStream* PROTOBUF_NONNULL stream) const {
  const Functions& this_ = *this;
#endif  // PROTOBUF_CUSTOM_VTABLE
  if constexpr (::_pbi::DebugHardenCheckHasBitConsistency()) {
    this_.CheckHasBitConsistency();
  }
  // @@protoc_insertion_point(serialize_to_array_start:nodeinstance.manifest.v1.Functions)
  ::uint32_t cached_has_bits = 0;
  (void)cached_has_bits;

  cached_has_bits = this_._impl_._has_bits_[0];
  // repeated .nodeinstance.manifest.v1.Function items = 1;
  if (CheckHasBit(cached_has_bits, 0x00000001U)) {
    for (unsigned i = 0, n = static_cast<unsigned>(
                             this_._internal_items_size());
         i < n; i++) {
      const auto& repfield = this_._internal_items().Get(i);
      target =
          ::google::protobuf::internal::WireFormatLite::InternalWriteMessage(
              1, repfield, repfield.GetCachedSize(),
              target, stream);
    }
  }

  if (ABSL_PREDICT_FALSE(this_._internal_metadata_.have_unknown_fields())) {
    target =
        ::_pbi::WireFormat::InternalSerializeUnknownFieldsToArray(
            this_._internal_metadata_.unknown_fields<::google::protobuf::UnknownFieldSet>(::google::protobuf::UnknownFieldSet::default_instance), target, stream);
  }
  // @@protoc_insertion_point(serialize_to_array_end:nodeinstance.manifest.v1.Functions)
  return target;
}

#if defined(PROTOBUF_CUSTOM_VTABLE)
::size_t Functions::ByteSizeLong(const MessageLite& base) {
  const Functions& this_ = static_cast<const Functions&>(base);
#else   // PROTOBUF_CUSTOM_VTABLE
::size_t Functions::ByteSizeLong() const {
  const Functions& this_ = *this;
#endif  // PROTOBUF_CUSTOM_VTABLE
  // @@protoc_insertion_point(message_byte_size_start:nodeinstance.manifest.v1.Functions)
  ::size_t total_size = 0;

  ::uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void)cached_has_bits;

  ::_pbi::Prefetch5LinesFrom7Lines(&this_);
   {
    // repeated .nodeinstance.manifest.v1.Function items = 1;
    cached_has_bits = this_._impl_._has_bits_[0];
    if (CheckHasBit(cached_has_bits, 0x00000001U)) {
      total_size += 1UL * this_._internal_items_size();
      for (const auto& msg : this_._internal_items()) {
        total_size += ::google::protobuf::internal::WireFormatLite::MessageSize(msg);
      }
    }
  }
  return this_.MaybeComputeUnknownFieldsSize(total_size,
                                             &this_._impl_._cached_size_);
}

void Functions::MergeImpl(::google::protobuf::MessageLite& to_msg,
                      const ::google::protobuf::MessageLite& from_msg) {
   auto* const _this = static_cast<Functions*>(&to_msg);
  auto& from = static_cast<const Functions&>(from_msg);
  if constexpr (::_pbi::DebugHardenCheckHasBitConsistency()) {
    from.CheckHasBitConsistency();
  }
  ::google::protobuf::Arena* arena = _this->GetArena();
  // @@protoc_insertion_point(class_specific_merge_from_start:nodeinstance.manifest.v1.Functions)
  ABSL_DCHECK_NE(&from, _this);
  ::uint32_t cached_has_bits = 0;
  (void)cached_has_bits;

  cached_has_bits = from._impl_._has_bits_[0];
  if (CheckHasBit(cached_has_bits, 0x00000001U)) {
    _this->_internal_mutable_items()->InternalMergeFromWithArena(
        ::google::protobuf::MessageLite::internal_visibility(), arena,
        from._internal_items());
  }
  _this->_impl_._has_bits_[0] |= cached_has_bits;
  _this->_internal_metadata_.MergeFrom<::google::protobuf::UnknownFieldSet>(
      from._internal_metadata_);
}

void Functions::CopyFrom(const Functions& from) {
  // @@protoc_insertion_point(class_specific_copy_from_start:nodeinstance.manifest.v1.Functions)
  if (&from == this) return;
  Clear();
  MergeFrom(from);
}


void Functions::InternalSwap(Functions* PROTOBUF_RESTRICT PROTOBUF_NONNULL other) {
  using ::std::swap;
  _internal_metadata_.InternalSwap(&other->_internal_metadata_);
  swap(_impl_._has_bits_[0], other->_impl_._has_bits_[0]);
  _impl_.items_.InternalSwap(&other->_impl_.items_);
}

::google::protobuf::Metadata Functions::GetMetadata() const {
  return ::google::protobuf::Message::GetMetadataImpl(GetClassData()->full());
}
// ===================================================================

Function::Function(::google::protobuf::Arena* PROTOBUF_NULLABLE arena)
#if defined(PROTOBUF_CUSTOM_VTABLE)
    : ::google::protobuf::Message(arena, Function_get_class_data()) {
#else   // PROTOBUF_CUSTOM_VTABLE
    : ::google::protobuf::Message(arena) {
#endif  // PROTOBUF_CUSTOM_VTABLE
  SharedCtor(arena);
  // @@protoc_insertion_point(arena_constructor:nodeinstance.manifest.v1.Function)
}
PROTOBUF_NDEBUG_INLINE Function::Impl_::Impl_(
    [[maybe_unused]] ::google::protobuf::internal::InternalVisibility visibility,
    [[maybe_unused]] ::google::protobuf::Arena* PROTOBUF_NULLABLE arena, const Impl_& from,
    [[maybe_unused]] const ::nodeinstance::manifest::v1::Function& from_msg)
      : _has_bits_{from._has_bits_},
        _cached_size_{0},
        tags_ {
          visibility, ::_pbi::InternalMetadataOffset::Build<
              ::nodeinstance::manifest::v1::Function,
              PROTOBUF_FIELD_OFFSET(::nodeinstance::manifest::v1::Function, _impl_.tags_)>()
          , from.tags_
        }
        ,
        access_ {
          visibility, ::_pbi::InternalMetadataOffset::Build<
              ::nodeinstance::manifest::v1::Function,
              PROTOBUF_FIELD_OFFSET(::nodeinstance::manifest::v1::Function, _impl_.access_)>()
          , from.access_
        }
        ,
        _access_cached_byte_size_{0},
        name_(arena, from.name_),
        help_(arena, from.help_) {}

Function::Function(
    ::google::protobuf::Arena* PROTOBUF_NULLABLE arena,
    const Function& from)
#if defined(PROTOBUF_CUSTOM_VTABLE)
    : ::google::protobuf::Message(arena, Function_get_class_data()) {

#else   // PROTOBUF_CUSTOM_VTABLE
    : ::google::protobuf::Message(arena) {
#endif  // PROTOBUF_CUSTOM_VTABLE
  Function* const _this = this;
  (void)_this;
  _internal_metadata_.MergeFrom<::google::protobuf::UnknownFieldSet>(
      from._internal_metadata_);
  new (&_impl_) Impl_(internal_visibility(), arena, from._impl_, from);
  ::memcpy(reinterpret_cast<char*>(&_impl_) +
               offsetof(Impl_, priority_),
           reinterpret_cast<const char*>(&from._impl_) +
               offsetof(Impl_, priority_),
           offsetof(Impl_, version_) -
               offsetof(Impl_, priority_) +
               sizeof(Impl_::version_));

  // @@protoc_insertion_point(copy_constructor:nodeinstance.manifest.v1.Function)
}
PROTOBUF_NDEBUG_INLINE Function::Impl_::Impl_(
    [[maybe_unused]] ::google::protobuf::internal::InternalVisibility visibility,
    [[maybe_unused]] ::google::protobuf::Arena* PROTOBUF_NULLABLE arena)
      : _cached_size_{0},
        tags_ { visibility, ::_pbi::InternalMetadataOffset::Build<
            ::nodeinstance::manifest::v1::Function,
            PROTOBUF_FIELD_OFFSET(::nodeinstance::manifest::v1::Function, _impl_.tags_)>()
         }
        ,
        access_ { visibility, ::_pbi::InternalMetadataOffset::Build<
            ::nodeinstance::manifest::v1::Function,
            PROTOBUF_FIELD_OFFSET(::nodeinstance::manifest::v1::Function, _impl_.access_)>()
         }
        ,
        _access_cached_byte_size_{0},
        name_(arena),
        help_(arena) {}

inline void Function::SharedCtor(::_pb::Arena* PROTOBUF_NULLABLE arena) {
  new (&_impl_) Impl_(internal_visibility(), arena);
  ::memset(reinterpret_cast<char*>(&_impl_) +
               offsetof(Impl_, priority_),
           0,
           offsetof(Impl_, version_) -
               offsetof(Impl_, priority_) +
               sizeof(Impl_::version_));
}
Function::~Function() {
  // @@protoc_insertion_point(destructor:nodeinstance.manifest.v1.Function)
  SharedDtor(*this);
}
inline void Function::SharedDtor(MessageLite& self) {
  Function& this_ = static_cast<Function&>(self);
  if constexpr (::_pbi::DebugHardenCheckHasBitConsistency()) {
    this_.CheckHasBitConsistency();
  }
  this_._internal_metadata_.Delete<::google::protobuf::UnknownFieldSet>();
  ABSL_DCHECK(this_.GetArena() == nullptr);
  this_._impl_.name_.Destroy();
  this_._impl_.help_.Destroy();
  this_._impl_.~Impl_();
}

#ifndef PROTOBUF_MESSAGE_GLOBALS
PROTOBUF_CONSTINIT PROTOBUF_ATTRIBUTE_INIT_PRIORITY1 const
    ::google::protobuf::internal::ClassDataFull Function_class_data_ =
        Function::InternalGenerateClassData_(Function_globals_._default);

PROTOBUF_ATTRIBUTE_WEAK const ::google::protobuf::internal::ClassData* PROTOBUF_NONNULL
Function::GetClassData() const {
  ::google::protobuf::internal::PrefetchToLocalCache(&Function_class_data_);
  ::google::protobuf::internal::PrefetchToLocalCache(Function_class_data_.tc_table);
  return Function_class_data_.base();
}
#else
PROTOBUF_ATTRIBUTE_WEAK const ::google::protobuf::internal::ClassData* PROTOBUF_NONNULL
Function::GetClassData() const {
  ::google::protobuf::internal::PrefetchToLocalCache(&Function_globals_);
  ::google::protobuf::internal::PrefetchToLocalCache(
      ::google::protobuf::internal::MessageGlobalsBase::ToParseTableBase(&Function_globals_));
  return Function_globals_.GetClassData();
}
#endif  // !PROTOBUF_MESSAGE_GLOBALS
#ifndef PROTOBUF_MESSAGE_GLOBALS
PROTOBUF_CONSTINIT
PROTOBUF_ATTRIBUTE_INIT_PRIORITY1 const Function::ParseTableT_
    Function::_table_ =
        Function::InternalGenerateParseTable_(Function_class_data_.base());
#endif  // !PROTOBUF_MESSAGE_GLOBALS
PROTOBUF_NOINLINE void Function::Clear() {
// @@protoc_insertion_point(message_clear_start:nodeinstance.manifest.v1.Function)
  ::google::protobuf::internal::TSanWrite(&_impl_);
  ::uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void) cached_has_bits;

  cached_has_bits = _impl_._has_bits_[0];
  if (BatchCheckHasBit(cached_has_bits, 0x0000000fU)) {
    if (CheckHasBit(cached_has_bits, 0x00000001U)) {
      _impl_.tags_.Clear();
    }
    if (CheckHasBit(cached_has_bits, 0x00000002U)) {
      _impl_.access_.Clear();
    }
    if (CheckHasBit(cached_has_bits, 0x00000004U)) {
      _impl_.name_.ClearNonDefaultToEmpty();
    }
    if (CheckHasBit(cached_has_bits, 0x00000008U)) {
      _impl_.help_.ClearNonDefaultToEmpty();
    }
  }
  if (BatchCheckHasBit(cached_has_bits, 0x00000030U)) {
    ::memset(&_impl_.priority_, 0, static_cast<::size_t>(
        reinterpret_cast<char*>(&_impl_.version_) -
        reinterpret_cast<char*>(&_impl_.priority_)) + sizeof(_impl_.version_));
  }
  _impl_._has_bits_.Clear();
  _internal_metadata_.Clear<::google::protobuf::UnknownFieldSet>();
}

#if defined(PROTOBUF_CUSTOM_VTABLE)
::uint8_t* PROTOBUF_NONNULL Function::_InternalSerialize(
    const ::google::protobuf::MessageLite& base, ::uint8_t* PROTOBUF_NONNULL target,
    ::google::protobuf::io::EpsCopyOutputStream* PROTOBUF_NONNULL stream) {
  const Function& this_ = static_cast<const Function&>(base);
#else   // PROTOBUF_CUSTOM_VTABLE
::uint8_t* PROTOBUF_NONNULL Function::_InternalSerialize(
    ::uint8_t* PROTOBUF_NONNULL target,
    ::google::protobuf::io::EpsCopyOutputStream* PROTOBUF_NONNULL stream) const {
  const Function& this_ = *this;
#endif  // PROTOBUF_CUSTOM_VTABLE
  if constexpr (::_pbi::DebugHardenCheckHasBitConsistency()) {
    this_.CheckHasBitConsistency();
  }
  // @@protoc_insertion_point(serialize_to_array_start:nodeinstance.manifest.v1.Function)
  ::uint32_t cached_has_bits = 0;
  (void)cached_has_bits;

  cached_has_bits = this_._impl_._has_bits_[0];
  // string name = 1;
  if (CheckHasBit(cached_has_bits, 0x00000004U)) {
    if (!this_._internal_name().empty()) {
      const ::std::string& _s = this_._internal_name();
      ::google::protobuf::internal::WireFormatLite::VerifyUtf8String(
          _s.data(), static_cast<int>(_s.length()), ::google::protobuf::internal::WireFormatLite::SERIALIZE, "nodeinstance.manifest.v1.Function.name");
      target = stream->WriteStringMaybeAliased(1, _s, target);
    }
  }

  // string help = 2;
  if (CheckHasBit(cached_has_bits, 0x00000008U)) {
    if (!this_._internal_help().empty()) {
      const ::std::string& _s = this_._internal_help();
      ::google::protobuf::internal::WireFormatLite::VerifyUtf8String(
          _s.data(), static_cast<int>(_s.length()), ::google::protobuf::internal::WireFormatLite::SERIALIZE, "nodeinstance.manifest.v1.Function.help");
      target = stream->WriteStringMaybeAliased(2, _s, target);
    }
  }

  // repeated string tags = 3;
  if (CheckHasBit(cached_has_bits, 0x00000001U)) {
    for (int i = 0, n = this_._internal_tags_size(); i < n; ++i) {
      const auto& s = this_._internal_tags().Get(i);
      ::google::protobuf::internal::WireFormatLite::VerifyUtf8String(
          s.data(), static_cast<int>(s.length()), ::google::protobuf::internal::WireFormatLite::SERIALIZE, "nodeinstance.manifest.v1.Function.tags");
      target = stream->WriteString(3, s, target);
    }
  }

  // uint32 priority = 4;
  if (CheckHasBit(cached_has_bits, 0x00000010U)) {
    if (this_._internal_priority() != 0) {
      target = stream->EnsureSpace(target);
      target = ::_pbi::WireFormatLite::WriteUInt32ToArray(
          4, this_._internal_priority(), target);
    }
  }

  // uint32 version = 5;
  if (CheckHasBit(cached_has_bits, 0x00000020U)) {
    if (this_._internal_version() != 0) {
      target = stream->EnsureSpace(target);
      target = ::_pbi::WireFormatLite::WriteUInt32ToArray(
          5, this_._internal_version(), target);
    }
  }

  // repeated .nodeinstance.manifest.v1.HttpAccess access = 6;
  if (CheckHasBit(cached_has_bits, 0x00000002U)) {
    {
      ::size_t byte_size = this_._impl_._access_cached_byte_size_.Get();
      if (byte_size > 0) {
        target = stream->WriteEnumPacked(
            6, this_._internal_access(), byte_size, target);
      }
    }
  }

  if (ABSL_PREDICT_FALSE(this_._internal_metadata_.have_unknown_fields())) {
    target =
        ::_pbi::WireFormat::InternalSerializeUnknownFieldsToArray(
            this_._internal_metadata_.unknown_fields<::google::protobuf::UnknownFieldSet>(::google::protobuf::UnknownFieldSet::default_instance), target, stream);
  }
  // @@protoc_insertion_point(serialize_to_array_end:nodeinstance.manifest.v1.Function)
  return target;
}

#if defined(PROTOBUF_CUSTOM_VTABLE)
::size_t Function::ByteSizeLong(const MessageLite& base) {
  const Function& this_ = static_cast<const Function&>(base);
#else   // PROTOBUF_CUSTOM_VTABLE
::size_t Function::ByteSizeLong() const {
  const Function& this_ = *this;
#endif  // PROTOBUF_CUSTOM_VTABLE
  // @@protoc_insertion_point(message_byte_size_start:nodeinstance.manifest.v1.Function)
  ::size_t total_size = 0;

  ::uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void)cached_has_bits;

  ::_pbi::Prefetch5LinesFrom7Lines(&this_);
  cached_has_bits = this_._impl_._has_bits_[0];
  if (BatchCheckHasBit(cached_has_bits, 0x0000003fU)) {
    // repeated string tags = 3;
    if (CheckHasBit(cached_has_bits, 0x00000001U)) {
      total_size +=
          1 * ::google::protobuf::internal::FromIntSize(this_._internal_tags().size());
      for (int i = 0, n = this_._internal_tags().size(); i < n; ++i) {
        total_size += ::google::protobuf::internal::WireFormatLite::StringSize(
            this_._internal_tags().Get(i));
      }
    }
    // repeated .nodeinstance.manifest.v1.HttpAccess access = 6;
    if (CheckHasBit(cached_has_bits, 0x00000002U)) {
      total_size += ::_pbi::WireFormatLite::EnumSizeWithPackedTagSize(
          this_._internal_access(), 1, this_._impl_._access_cached_byte_size_);
    }
    // string name = 1;
    if (CheckHasBit(cached_has_bits, 0x00000004U)) {
      if (!this_._internal_name().empty()) {
        total_size += 1 + ::google::protobuf::internal::WireFormatLite::StringSize(
                                        this_._internal_name());
      }
    }
    // string help = 2;
    if (CheckHasBit(cached_has_bits, 0x00000008U)) {
      if (!this_._internal_help().empty()) {
        total_size += 1 + ::google::protobuf::internal::WireFormatLite::StringSize(
                                        this_._internal_help());
      }
    }
    // uint32 priority = 4;
    if (CheckHasBit(cached_has_bits, 0x00000010U)) {
      if (this_._internal_priority() != 0) {
        total_size += ::_pbi::WireFormatLite::UInt32SizePlusOne(
            this_._internal_priority());
      }
    }
    // uint32 version = 5;
    if (CheckHasBit(cached_has_bits, 0x00000020U)) {
      if (this_._internal_version() != 0) {
        total_size += ::_pbi::WireFormatLite::UInt32SizePlusOne(
            this_._internal_version());
      }
    }
  }
  return this_.MaybeComputeUnknownFieldsSize(total_size,
                                             &this_._impl_._cached_size_);
}

void Function::MergeImpl(::google::protobuf::MessageLite& to_msg,
                      const ::google::protobuf::MessageLite& from_msg) {
   auto* const _this = static_cast<Function*>(&to_msg);
  auto& from = static_cast<const Function&>(from_msg);
  if constexpr (::_pbi::DebugHardenCheckHasBitConsistency()) {
    from.CheckHasBitConsistency();
  }
  ::google::protobuf::Arena* arena = _this->GetArena();
  // @@protoc_insertion_point(class_specific_merge_from_start:nodeinstance.manifest.v1.Function)
  ABSL_DCHECK_NE(&from, _this);
  ::uint32_t cached_has_bits = 0;
  (void)cached_has_bits;

  cached_has_bits = from._impl_._has_bits_[0];
  if (BatchCheckHasBit(cached_has_bits, 0x0000003fU)) {
    if (CheckHasBit(cached_has_bits, 0x00000001U)) {
      _this->_internal_mutable_tags()->InternalMergeFromWithArena(
          ::google::protobuf::MessageLite::internal_visibility(), arena,
          from._internal_tags());
    }
    if (CheckHasBit(cached_has_bits, 0x00000002U)) {
      _this->_internal_mutable_access()->MergeFrom(from._internal_access());
    }
    if (CheckHasBit(cached_has_bits, 0x00000004U)) {
      if (!from._internal_name().empty()) {
        _this->_internal_set_name(from._internal_name());
      } else {
        if (_this->_impl_.name_.IsDefault()) {
          _this->_internal_set_name("");
        }
      }
    }
    if (CheckHasBit(cached_has_bits, 0x00000008U)) {
      if (!from._internal_help().empty()) {
        _this->_internal_set_help(from._internal_help());
      } else {
        if (_this->_impl_.help_.IsDefault()) {
          _this->_internal_set_help("");
        }
      }
    }
    if (CheckHasBit(cached_has_bits, 0x00000010U)) {
      if (from._internal_priority() != 0) {
        _this->_impl_.priority_ = from._impl_.priority_;
      }
    }
    if (CheckHasBit(cached_has_bits, 0x00000020U)) {
      if (from._internal_version() != 0) {
        _this->_impl_.version_ = from._impl_.version_;
      }
    }
  }
  _this->_impl_._has_bits_[0] |= cached_has_bits;
  _this->_internal_metadata_.MergeFrom<::google::protobuf::UnknownFieldSet>(
      from._internal_metadata_);
}

void Function::CopyFrom(const Function& from) {
  // @@protoc_insertion_point(class_specific_copy_from_start:nodeinstance.manifest.v1.Function)
  if (&from == this) return;
  Clear();
  MergeFrom(from);
}


void Function::InternalSwap(Function* PROTOBUF_RESTRICT PROTOBUF_NONNULL other) {
  using ::std::swap;
  auto* arena = GetArena();
  ABSL_DCHECK_EQ(arena, other->GetArena());
  _internal_metadata_.InternalSwap(&other->_internal_metadata_);
  swap(_impl_._has_bits_[0], other->_impl_._has_bits_[0]);
  _impl_.tags_.InternalSwap(&other->_impl_.tags_);
  _impl_.access_.InternalSwap(&other->_impl_.access_);
  ::_pbi::ArenaStringPtr::InternalSwap(&_impl_.name_, &other->_impl_.name_, arena);
  ::_pbi::ArenaStringPtr::InternalSwap(&_impl_.help_, &other->_impl_.help_, arena);
  ::google::protobuf::internal::memswap<
      PROTOBUF_FIELD_OFFSET(Function, _impl_.version_)
      + sizeof(Function::_impl_.version_)
      - PROTOBUF_FIELD_OFFSET(Function, _impl_.priority_)>(
          reinterpret_cast<char*>(&_impl_.priority_),
          reinterpret_cast<char*>(&other->_impl_.priority_));
}

::google::protobuf::Metadata Function::GetMetadata() const {
  return ::google::protobuf::Message::GetMetadataImpl(GetClassData()->full());
}
// @@protoc_insertion_point(namespace_scope)
}  // namespace v1
}  // namespace manifest
}  // namespace nodeinstance
namespace google {
namespace protobuf {
}  // namespace protobuf
}  // namespace google
// @@protoc_insertion_point(global_scope)
PROTOBUF_ATTRIBUTE_INIT_PRIORITY2 static ::std::false_type
    _static_init2_ [[maybe_unused]] =
        (::_pbi::AddDescriptors(&descriptor_table_nodeinstance_2fmanifest_2fv1_2fmanifest_2eproto),
         ::std::false_type{});
#include "google/protobuf/port_undef.inc"
