// Generated by the protocol buffer compiler.  DO NOT EDIT!
// NO CHECKED-IN PROTOBUF GENCODE
// source: nodeinstance/connection/v1/connection.proto
// Protobuf C++ Version: 7.35.1

#include "nodeinstance/connection/v1/connection.pb.h"

#include <algorithm>
#include <type_traits>
#include "google/protobuf/io/coded_stream.h"
#include "google/protobuf/generated_message_tctable_impl.h"
#include "google/protobuf/internal_visibility.h"
#include "google/protobuf/extension_set.h"
#include "google/protobuf/generated_message_util.h"
#include "google/protobuf/wire_format_lite.h"
#include "google/protobuf/descriptor.h"
#include "google/protobuf/generated_message_reflection.h"
#include "google/protobuf/reflection_ops.h"
#include "google/protobuf/wire_format.h"
// @@protoc_insertion_point(includes)

// Must be included last.
#include "google/protobuf/port_def.inc"
PROTOBUF_PRAGMA_INIT_SEG
namespace _pb = ::google::protobuf;
namespace _pbi = ::google::protobuf::internal;
namespace _fl = ::google::protobuf::internal::field_layout;
#ifdef PROTOBUF_MESSAGE_GLOBALS
namespace {
PROTOBUF_CONSTINIT ::google::protobuf::internal::ReflectionData
    file_reflection_data[] = {
        // ::nodeinstance::v1::UpdateNodeInstanceConnection
        {&::_pbi::kDescriptorMethods, &::descriptor_table_nodeinstance_2fconnection_2fv1_2fconnection_2eproto, /* tracker*/ nullptr,},
};
}  // namespace
#endif
namespace nodeinstance {
namespace v1 {
class UpdateNodeInstanceConnection::_Internal {
 public:
  using HasBits = decltype(::std::declval<UpdateNodeInstanceConnection>()._impl_._has_bits_);
  static constexpr ::int32_t kHasBitsOffset =
      8 * PROTOBUF_FIELD_OFFSET(UpdateNodeInstanceConnection, _impl_._has_bits_);
};

constexpr UpdateNodeInstanceConnection::ParseTableT_ UpdateNodeInstanceConnection::InternalGenerateParseTable_(const ::_pbi::ClassData* class_data) {
  return ParseTableT_{
    {
      PROTOBUF_FIELD_OFFSET(UpdateNodeInstanceConnection, _impl_._has_bits_),
      0, // no _extensions_
      10, 120,  // max_field_number, fast_idx_mask
      offsetof(ParseTableT_, field_lookup_table),
      4294966272,  // skipmap
      offsetof(ParseTableT_, field_entries),
      10,  // num_field_entries
      2,  // num_aux_entries
      offsetof(ParseTableT_, aux_entries),
      class_data,
      nullptr,  // post_loop_handler
      ::_pbi::TcParser::GenericFallback,  // fallback
      #ifdef PROTOBUF_PREFETCH_PARSE_TABLE
      ::_pbi::TcParser::GetTable<::nodeinstance::v1::UpdateNodeInstanceConnection>(),  // to_prefetch
      #endif  // PROTOBUF_PREFETCH_PARSE_TABLE
    }, {{
      {::_pbi::TcParser::MiniParse, {}},
      // string claim_id = 1;
      {::_pbi::TcParser::FastUS1,
       {10, 1, 0,
        PROTOBUF_FIELD_OFFSET(UpdateNodeInstanceConnection, _impl_.claim_id_)}},
      // string node_id = 2;
      {::_pbi::TcParser::FastUS1,
       {18, 2, 0,
        PROTOBUF_FIELD_OFFSET(UpdateNodeInstanceConnection, _impl_.node_id_)}},
      // bool liveness = 3;
      {::_pbi::TcParser::SingularVarintNoZag1<bool, offsetof(UpdateNodeInstanceConnection, _impl_.liveness_), 6>(),
       {24, 6, 0,
        PROTOBUF_FIELD_OFFSET(UpdateNodeInstanceConnection, _impl_.liveness_)}},
      // bool queryable = 4;
      {::_pbi::TcParser::SingularVarintNoZag1<bool, offsetof(UpdateNodeInstanceConnection, _impl_.queryable_), 7>(),
       {32, 7, 0,
        PROTOBUF_FIELD_OFFSET(UpdateNodeInstanceConnection, _impl_.queryable_)}},
      // int64 session_id = 5;
      {::_pbi::TcParser::SingularVarintNoZag1<::uint64_t, offsetof(UpdateNodeInstanceConnection, _impl_.session_id_), 5>(),
       {40, 5, 0,
        PROTOBUF_FIELD_OFFSET(UpdateNodeInstanceConnection, _impl_.session_id_)}},
      // .google.protobuf.Timestamp updated_at = 6;
      {::_pbi::TcParser::FastMtS1,
       {50, 4, 0,
        PROTOBUF_FIELD_OFFSET(UpdateNodeInstanceConnection, _impl_.updated_at_)}},
      // string mqtt_broker_addr = 7;
      {::_pbi::TcParser::FastUS1,
       {58, 3, 0,
        PROTOBUF_FIELD_OFFSET(UpdateNodeInstanceConnection, _impl_.mqtt_broker_addr_)}},
      // int32 vmq_instance_id = 8;
      {::_pbi::TcParser::SingularVarintNoZag1<::uint32_t, offsetof(UpdateNodeInstanceConnection, _impl_.vmq_instance_id_), 8>(),
       {64, 8, 0,
        PROTOBUF_FIELD_OFFSET(UpdateNodeInstanceConnection, _impl_.vmq_instance_id_)}},
      // int32 hops = 9;
      {::_pbi::TcParser::SingularVarintNoZag1<::uint32_t, offsetof(UpdateNodeInstanceConnection, _impl_.hops_), 9>(),
       {72, 9, 0,
        PROTOBUF_FIELD_OFFSET(UpdateNodeInstanceConnection, _impl_.hops_)}},
      // repeated .aclk_lib.v1.Capability capabilities = 10;
      {::_pbi::TcParser::FastMtR1,
       {82, 0, 1,
        PROTOBUF_FIELD_OFFSET(UpdateNodeInstanceConnection, _impl_.capabilities_)}},
      {::_pbi::TcParser::MiniParse, {}},
      {::_pbi::TcParser::MiniParse, {}},
      {::_pbi::TcParser::MiniParse, {}},
      {::_pbi::TcParser::MiniParse, {}},
      {::_pbi::TcParser::MiniParse, {}},
    }}, {{
      65535, 65535
    }}, {{
      // string claim_id = 1;
      {PROTOBUF_FIELD_OFFSET(UpdateNodeInstanceConnection, _impl_.claim_id_), _Internal::kHasBitsOffset + 1, 0, (0 | ::_fl::kFcOptional | ::_fl::kUtf8String | ::_fl::kRepAString)},
      // string node_id = 2;
      {PROTOBUF_FIELD_OFFSET(UpdateNodeInstanceConnection, _impl_.node_id_), _Internal::kHasBitsOffset + 2, 0, (0 | ::_fl::kFcOptional | ::_fl::kUtf8String | ::_fl::kRepAString)},
      // bool liveness = 3;
      {PROTOBUF_FIELD_OFFSET(UpdateNodeInstanceConnection, _impl_.liveness_), _Internal::kHasBitsOffset + 6, 0, (0 | ::_fl::kFcOptional | ::_fl::kBool)},
      // bool queryable = 4;
      {PROTOBUF_FIELD_OFFSET(UpdateNodeInstanceConnection, _impl_.queryable_), _Internal::kHasBitsOffset + 7, 0, (0 | ::_fl::kFcOptional | ::_fl::kBool)},
      // int64 session_id = 5;
      {PROTOBUF_FIELD_OFFSET(UpdateNodeInstanceConnection, _impl_.session_id_), _Internal::kHasBitsOffset + 5, 0, (0 | ::_fl::kFcOptional | ::_fl::kInt64)},
      // .google.protobuf.Timestamp updated_at = 6;
      {PROTOBUF_FIELD_OFFSET(UpdateNodeInstanceConnection, _impl_.updated_at_), _Internal::kHasBitsOffset + 4, 0, (0 | ::_fl::kFcOptional | ::_fl::kMessage | ::_fl::kTvTable)},
      // string mqtt_broker_addr = 7;
      {PROTOBUF_FIELD_OFFSET(UpdateNodeInstanceConnection, _impl_.mqtt_broker_addr_), _Internal::kHasBitsOffset + 3, 0, (0 | ::_fl::kFcOptional | ::_fl::kUtf8String | ::_fl::kRepAString)},
      // int32 vmq_instance_id = 8;
      {PROTOBUF_FIELD_OFFSET(UpdateNodeInstanceConnection, _impl_.vmq_instance_id_), _Internal::kHasBitsOffset + 8, 0, (0 | ::_fl::kFcOptional | ::_fl::kInt32)},
      // int32 hops = 9;
      {PROTOBUF_FIELD_OFFSET(UpdateNodeInstanceConnection, _impl_.hops_), _Internal::kHasBitsOffset + 9, 0, (0 | ::_fl::kFcOptional | ::_fl::kInt32)},
      // repeated .aclk_lib.v1.Capability capabilities = 10;
      {PROTOBUF_FIELD_OFFSET(UpdateNodeInstanceConnection, _impl_.capabilities_), _Internal::kHasBitsOffset + 0, 1, (0 | ::_fl::kFcRepeated | ::_fl::kMessage | ::_fl::kTvTable)},
    }},
    {{
        #ifndef PROTOBUF_MESSAGE_GLOBALS
        {::_pbi::TcParser::GetTable<::google::protobuf::Timestamp>()},
        #else
        {::_pbi::FieldAuxMessageGlobals(), &::google::protobuf::Timestamp_globals_},
        #endif
        #ifndef PROTOBUF_MESSAGE_GLOBALS
        {::_pbi::TcParser::GetTable<::aclk_lib::v1::Capability>()},
        #else
        {::_pbi::FieldAuxMessageGlobals(), &::aclk_lib::v1::Capability_globals_},
        #endif
    }},
    {{
      "\54\10\7\0\0\0\0\20\0\0\0\0\0\0\0\0"
      "nodeinstance.v1.UpdateNodeInstanceConnection"
      "claim_id"
      "node_id"
      "mqtt_broker_addr"
    }},
  };
}


inline constexpr UpdateNodeInstanceConnection::Impl_::Impl_(
    [[maybe_unused]] ::google::protobuf::internal::InternalVisibility visibility,
    ::_pbi::ConstantInitialized) noexcept
      : _cached_size_{0},
        capabilities_ { visibility, ::_pbi::InternalMetadataOffset::Build<
            ::nodeinstance::v1::UpdateNodeInstanceConnection,
            PROTOBUF_FIELD_OFFSET(::nodeinstance::v1::UpdateNodeInstanceConnection, _impl_.capabilities_)>()
         }
        ,
        claim_id_(
            &::google::protobuf::internal::fixed_address_empty_string,
            ::_pbi::ConstantInitialized()),
        node_id_(
            &::google::protobuf::internal::fixed_address_empty_string,
            ::_pbi::ConstantInitialized()),
        mqtt_broker_addr_(
            &::google::protobuf::internal::fixed_address_empty_string,
            ::_pbi::ConstantInitialized()),
        updated_at_{nullptr},
        session_id_{::int64_t{0}},
        liveness_{false},
        queryable_{false},
        vmq_instance_id_{0},
        hops_{0} {}

template <typename>
constexpr UpdateNodeInstanceConnection::UpdateNodeInstanceConnection(::_pbi::ConstantInitialized,
                       const ::_pbi::ClassData* PROTOBUF_NONNULL class_data)
    : ::google::protobuf::Message(
#if defined(PROTOBUF_CUSTOM_VTABLE)
          class_data
#endif  // PROTOBUF_CUSTOM_VTABLE
          ),
      _impl_(internal_visibility(), ::_pbi::ConstantInitialized()) {
}
inline void* PROTOBUF_NONNULL UpdateNodeInstanceConnection::PlacementNew_(
    const void* PROTOBUF_NONNULL, void* PROTOBUF_NONNULL mem,
    ::google::protobuf::Arena* PROTOBUF_NULLABLE arena) {
  return ::new (mem) UpdateNodeInstanceConnection(arena);
}
constexpr auto UpdateNodeInstanceConnection::InternalNewImpl_() {
  return ::google::protobuf::internal::MessageCreator::CopyInit(sizeof(UpdateNodeInstanceConnection), alignof(UpdateNodeInstanceConnection));
}
constexpr auto UpdateNodeInstanceConnection::InternalGenerateClassData_(
    const MessageLite& prototype,
    const ::google::protobuf::internal::TcParseTableBase* tc_table) {
  return ::google::protobuf::internal::ClassDataFull{
      ::google::protobuf::internal::ClassData{
          &prototype,
#ifndef PROTOBUF_MESSAGE_GLOBALS
          &_table_.header,
#else
          tc_table,
#endif
          nullptr,  // IsInitialized
          &UpdateNodeInstanceConnection::MergeImpl,
          ::google::protobuf::Message::GetNewImpl<UpdateNodeInstanceConnection>(),
#if defined(PROTOBUF_CUSTOM_VTABLE)
          &UpdateNodeInstanceConnection::SharedDtor,
          ::google::protobuf::Message::GetClearImpl<UpdateNodeInstanceConnection>(), &UpdateNodeInstanceConnection::ByteSizeLong,
              &UpdateNodeInstanceConnection::_InternalSerialize,
#endif  // PROTOBUF_CUSTOM_VTABLE
          PROTOBUF_FIELD_OFFSET(UpdateNodeInstanceConnection, _impl_._cached_size_),
          false,
      },
#ifdef PROTOBUF_MESSAGE_GLOBALS
      &file_reflection_data[0],
#else   // !PROTOBUF_MESSAGE_GLOBALS
      &::_pbi::kDescriptorMethods,
      &descriptor_table_nodeinstance_2fconnection_2fv1_2fconnection_2eproto,
      nullptr,  // tracker
#endif  // PROTOBUF_MESSAGE_GLOBALS
  };
}
struct UpdateNodeInstanceConnectionGlobalsTypeInternal : ::_pbi::MessageGlobalsBase {
  constexpr UpdateNodeInstanceConnectionGlobalsTypeInternal()
      :
#ifndef PROTOBUF_MESSAGE_GLOBALS
        _default(::_pbi::ConstantInitialized{},
                 UpdateNodeInstanceConnection_class_data_.base())
#else   // !PROTOBUF_MESSAGE_GLOBALS
        MessageGlobalsBase(UpdateNodeInstanceConnection::InternalGenerateClassData_(
            _default, &UpdateNodeInstanceConnection_globals_._table.header)),
        _default(::_pbi::ConstantInitialized{}, GetClassData()),
        _table(::_pbi::PrivateAccess::GenerateParseTable<UpdateNodeInstanceConnection>(
            GetClassData()))
#endif  // PROTOBUF_MESSAGE_GLOBALS
  {
  }
  ~UpdateNodeInstanceConnectionGlobalsTypeInternal() {}
  union {
    alignas(::_pbi::kMaxMessageAlignment) UpdateNodeInstanceConnection _default;
  };
#ifdef PROTOBUF_MESSAGE_GLOBALS
  decltype(::_pbi::PrivateAccess::GenerateParseTable<UpdateNodeInstanceConnection>(
      ::std::declval<const ::_pbi::ClassData*>())) _table;
#endif
};
#ifdef PROTOBUF_MESSAGE_GLOBALS
static_assert(PROTOBUF_FIELD_OFFSET(UpdateNodeInstanceConnectionGlobalsTypeInternal, _default) ==
              ::_pbi::MessageGlobalsBase::OffsetToDefault());
#endif  // PROTOBUF_MESSAGE_GLOBALS

PROTOBUF_ATTRIBUTE_NO_DESTROY PROTOBUF_CONSTINIT
    PROTOBUF_ATTRIBUTE_INIT_PRIORITY1 PROTOBUF_MESSAGE_GLOBALS_CONST UpdateNodeInstanceConnectionGlobalsTypeInternal UpdateNodeInstanceConnection_globals_
        PROTOBUF_MESSAGE_GLOBALS_SECTION(.data.rel.ro);
#if defined(PROTOBUF_CUSTOM_VTABLE)
namespace {
const ::_pbi::ClassData* UpdateNodeInstanceConnection_get_class_data() {
#ifdef PROTOBUF_MESSAGE_GLOBALS
  return UpdateNodeInstanceConnection_globals_.GetClassData();
#else
  return UpdateNodeInstanceConnection_class_data_.base();
#endif  // PROTOBUF_MESSAGE_GLOBALS
}
}  // namespace
#endif  // PROTOBUF_CUSTOM_VTABLE
}  // namespace v1
}  // namespace nodeinstance
static constexpr const ::_pb::EnumDescriptor* PROTOBUF_NONNULL* PROTOBUF_NULLABLE
    file_level_enum_descriptors_nodeinstance_2fconnection_2fv1_2fconnection_2eproto = nullptr;
static constexpr const ::_pb::ServiceDescriptor* PROTOBUF_NONNULL* PROTOBUF_NULLABLE
    file_level_service_descriptors_nodeinstance_2fconnection_2fv1_2fconnection_2eproto = nullptr;
const ::uint32_t
    TableStruct_nodeinstance_2fconnection_2fv1_2fconnection_2eproto::offsets[] ABSL_ATTRIBUTE_SECTION_VARIABLE(
        protodesc_cold) = {
        0x081, // bitmap
        PROTOBUF_FIELD_OFFSET(::nodeinstance::v1::UpdateNodeInstanceConnection, _impl_._has_bits_),
        13, // hasbit index offset
        PROTOBUF_FIELD_OFFSET(::nodeinstance::v1::UpdateNodeInstanceConnection, _impl_.claim_id_),
        PROTOBUF_FIELD_OFFSET(::nodeinstance::v1::UpdateNodeInstanceConnection, _impl_.node_id_),
        PROTOBUF_FIELD_OFFSET(::nodeinstance::v1::UpdateNodeInstanceConnection, _impl_.liveness_),
        PROTOBUF_FIELD_OFFSET(::nodeinstance::v1::UpdateNodeInstanceConnection, _impl_.queryable_),
        PROTOBUF_FIELD_OFFSET(::nodeinstance::v1::UpdateNodeInstanceConnection, _impl_.session_id_),
        PROTOBUF_FIELD_OFFSET(::nodeinstance::v1::UpdateNodeInstanceConnection, _impl_.updated_at_),
        PROTOBUF_FIELD_OFFSET(::nodeinstance::v1::UpdateNodeInstanceConnection, _impl_.mqtt_broker_addr_),
        PROTOBUF_FIELD_OFFSET(::nodeinstance::v1::UpdateNodeInstanceConnection, _impl_.vmq_instance_id_),
        PROTOBUF_FIELD_OFFSET(::nodeinstance::v1::UpdateNodeInstanceConnection, _impl_.hops_),
        PROTOBUF_FIELD_OFFSET(::nodeinstance::v1::UpdateNodeInstanceConnection, _impl_.capabilities_),
        1,
        2,
        6,
        7,
        5,
        4,
        3,
        8,
        9,
        0,
};

static const ::_pbi::MigrationSchema
    schemas[] ABSL_ATTRIBUTE_SECTION_VARIABLE(protodesc_cold) = {
        {0, sizeof(::nodeinstance::v1::UpdateNodeInstanceConnection)},
};
static const ::_pbi::MessageGlobalsBase* PROTOBUF_NONNULL const
    file_message_globals[] = {
        &::nodeinstance::v1::UpdateNodeInstanceConnection_globals_,
};
const char descriptor_table_protodef_nodeinstance_2fconnection_2fv1_2fconnection_2eproto[] ABSL_ATTRIBUTE_SECTION_VARIABLE(
    protodesc_cold) = {
    "\n+nodeinstance/connection/v1/connection."
    "proto\022\017nodeinstance.v1\032\037google/protobuf/"
    "timestamp.proto\032\021aclk/v1/lib.proto\"\232\002\n\034U"
    "pdateNodeInstanceConnection\022\020\n\010claim_id\030"
    "\001 \001(\t\022\017\n\007node_id\030\002 \001(\t\022\020\n\010liveness\030\003 \001(\010"
    "\022\021\n\tqueryable\030\004 \001(\010\022\022\n\nsession_id\030\005 \001(\003\022"
    ".\n\nupdated_at\030\006 \001(\0132\032.google.protobuf.Ti"
    "mestamp\022\030\n\020mqtt_broker_addr\030\007 \001(\t\022\027\n\017vmq"
    "_instance_id\030\010 \001(\005\022\014\n\004hops\030\t \001(\005\022-\n\014capa"
    "bilities\030\n \003(\0132\027.aclk_lib.v1.CapabilityB"
    "3Z1nodeinstance/connection/v1;nodeinstan"
    "ceconnectionb\006proto3"
};
static const ::_pbi::DescriptorTable* PROTOBUF_NONNULL const
    descriptor_table_nodeinstance_2fconnection_2fv1_2fconnection_2eproto_deps[2] = {
        &::descriptor_table_aclk_2fv1_2flib_2eproto,
        &::descriptor_table_google_2fprotobuf_2ftimestamp_2eproto,
};
static ::absl::once_flag descriptor_table_nodeinstance_2fconnection_2fv1_2fconnection_2eproto_once;
PROTOBUF_CONSTINIT const ::_pbi::DescriptorTable descriptor_table_nodeinstance_2fconnection_2fv1_2fconnection_2eproto = {
    false,
    false,
    460,
    descriptor_table_protodef_nodeinstance_2fconnection_2fv1_2fconnection_2eproto,
    "nodeinstance/connection/v1/connection.proto",
    &descriptor_table_nodeinstance_2fconnection_2fv1_2fconnection_2eproto_once,
    descriptor_table_nodeinstance_2fconnection_2fv1_2fconnection_2eproto_deps,
    2,
    1,
    schemas,
    file_message_globals,
    TableStruct_nodeinstance_2fconnection_2fv1_2fconnection_2eproto::offsets,
    file_level_enum_descriptors_nodeinstance_2fconnection_2fv1_2fconnection_2eproto,
    file_level_service_descriptors_nodeinstance_2fconnection_2fv1_2fconnection_2eproto,
};
namespace nodeinstance {
namespace v1 {
// ===================================================================

void UpdateNodeInstanceConnection::clear_updated_at() {
  ::google::protobuf::internal::TSanWrite(&_impl_);
  if (_impl_.updated_at_ != nullptr) _impl_.updated_at_->Clear();
  ClearHasBit(_impl_._has_bits_[0], 0x00000010U);
}
void UpdateNodeInstanceConnection::clear_capabilities() {
  ::google::protobuf::internal::TSanWrite(&_impl_);
  _impl_.capabilities_.Clear();
  ClearHasBit(_impl_._has_bits_[0], 0x00000001U);
}
UpdateNodeInstanceConnection::UpdateNodeInstanceConnection(::google::protobuf::Arena* PROTOBUF_NULLABLE arena)
#if defined(PROTOBUF_CUSTOM_VTABLE)
    : ::google::protobuf::Message(arena, UpdateNodeInstanceConnection_get_class_data()) {
#else   // PROTOBUF_CUSTOM_VTABLE
    : ::google::protobuf::Message(arena) {
#endif  // PROTOBUF_CUSTOM_VTABLE
  SharedCtor(arena);
  // @@protoc_insertion_point(arena_constructor:nodeinstance.v1.UpdateNodeInstanceConnection)
}
PROTOBUF_NDEBUG_INLINE UpdateNodeInstanceConnection::Impl_::Impl_(
    [[maybe_unused]] ::google::protobuf::internal::InternalVisibility visibility,
    [[maybe_unused]] ::google::protobuf::Arena* PROTOBUF_NULLABLE arena, const Impl_& from,
    [[maybe_unused]] const ::nodeinstance::v1::UpdateNodeInstanceConnection& from_msg)
      : _has_bits_{from._has_bits_},
        _cached_size_{0},
        capabilities_ {
          visibility, ::_pbi::InternalMetadataOffset::Build<
              ::nodeinstance::v1::UpdateNodeInstanceConnection,
              PROTOBUF_FIELD_OFFSET(::nodeinstance::v1::UpdateNodeInstanceConnection, _impl_.capabilities_)>()
          , from.capabilities_
        }
        ,
        claim_id_(arena, from.claim_id_),
        node_id_(arena, from.node_id_),
        mqtt_broker_addr_(arena, from.mqtt_broker_addr_) {}

UpdateNodeInstanceConnection::UpdateNodeInstanceConnection(
    ::google::protobuf::Arena* PROTOBUF_NULLABLE arena,
    const UpdateNodeInstanceConnection& from)
#if defined(PROTOBUF_CUSTOM_VTABLE)
    : ::google::protobuf::Message(arena, UpdateNodeInstanceConnection_get_class_data()) {

#else   // PROTOBUF_CUSTOM_VTABLE
    : ::google::protobuf::Message(arena) {
#endif  // PROTOBUF_CUSTOM_VTABLE
  UpdateNodeInstanceConnection* const _this = this;
  (void)_this;
  _internal_metadata_.MergeFrom<::google::protobuf::UnknownFieldSet>(
      from._internal_metadata_);
  new (&_impl_) Impl_(internal_visibility(), arena, from._impl_, from);
  ::uint32_t cached_has_bits = _impl_._has_bits_[0];
  _impl_.updated_at_ = (CheckHasBit(cached_has_bits, 0x00000010U))
                ? ::google::protobuf::Message::CopyConstruct(arena, *from._impl_.updated_at_)
                : nullptr;
  ::memcpy(reinterpret_cast<char*>(&_impl_) +
               offsetof(Impl_, session_id_),
           reinterpret_cast<const char*>(&from._impl_) +
               offsetof(Impl_, session_id_),
           offsetof(Impl_, hops_) -
               offsetof(Impl_, session_id_) +
               sizeof(Impl_::hops_));

  // @@protoc_insertion_point(copy_constructor:nodeinstance.v1.UpdateNodeInstanceConnection)
}
PROTOBUF_NDEBUG_INLINE UpdateNodeInstanceConnection::Impl_::Impl_(
    [[maybe_unused]] ::google::protobuf::internal::InternalVisibility visibility,
    [[maybe_unused]] ::google::protobuf::Arena* PROTOBUF_NULLABLE arena)
      : _cached_size_{0},
        capabilities_ { visibility, ::_pbi::InternalMetadataOffset::Build<
            ::nodeinstance::v1::UpdateNodeInstanceConnection,
            PROTOBUF_FIELD_OFFSET(::nodeinstance::v1::UpdateNodeInstanceConnection, _impl_.capabilities_)>()
         }
        ,
        claim_id_(arena),
        node_id_(arena),
        mqtt_broker_addr_(arena) {}

inline void UpdateNodeInstanceConnection::SharedCtor(::_pb::Arena* PROTOBUF_NULLABLE arena) {
  new (&_impl_) Impl_(internal_visibility(), arena);
  ::memset(reinterpret_cast<char*>(&_impl_) +
               offsetof(Impl_, updated_at_),
           0,
           offsetof(Impl_, hops_) -
               offsetof(Impl_, updated_at_) +
               sizeof(Impl_::hops_));
}
UpdateNodeInstanceConnection::~UpdateNodeInstanceConnection() {
  // @@protoc_insertion_point(destructor:nodeinstance.v1.UpdateNodeInstanceConnection)
  SharedDtor(*this);
}
inline void UpdateNodeInstanceConnection::SharedDtor(MessageLite& self) {
  UpdateNodeInstanceConnection& this_ = static_cast<UpdateNodeInstanceConnection&>(self);
  if constexpr (::_pbi::DebugHardenCheckHasBitConsistency()) {
    this_.CheckHasBitConsistency();
  }
  this_._internal_metadata_.Delete<::google::protobuf::UnknownFieldSet>();
  ABSL_DCHECK(this_.GetArena() == nullptr);
  this_._impl_.claim_id_.Destroy();
  this_._impl_.node_id_.Destroy();
  this_._impl_.mqtt_broker_addr_.Destroy();
  delete this_._impl_.updated_at_;
  this_._impl_.~Impl_();
}

#ifndef PROTOBUF_MESSAGE_GLOBALS
PROTOBUF_CONSTINIT PROTOBUF_ATTRIBUTE_INIT_PRIORITY1 const
    ::google::protobuf::internal::ClassDataFull UpdateNodeInstanceConnection_class_data_ =
        UpdateNodeInstanceConnection::InternalGenerateClassData_(UpdateNodeInstanceConnection_globals_._default);

PROTOBUF_ATTRIBUTE_WEAK const ::google::protobuf::internal::ClassData* PROTOBUF_NONNULL
UpdateNodeInstanceConnection::GetClassData() const {
  ::google::protobuf::internal::PrefetchToLocalCache(&UpdateNodeInstanceConnection_class_data_);
  ::google::protobuf::internal::PrefetchToLocalCache(UpdateNodeInstanceConnection_class_data_.tc_table);
  return UpdateNodeInstanceConnection_class_data_.base();
}
#else
PROTOBUF_ATTRIBUTE_WEAK const ::google::protobuf::internal::ClassData* PROTOBUF_NONNULL
UpdateNodeInstanceConnection::GetClassData() const {
  ::google::protobuf::internal::PrefetchToLocalCache(&UpdateNodeInstanceConnection_globals_);
  ::google::protobuf::internal::PrefetchToLocalCache(
      ::google::protobuf::internal::MessageGlobalsBase::ToParseTableBase(&UpdateNodeInstanceConnection_globals_));
  return UpdateNodeInstanceConnection_globals_.GetClassData();
}
#endif  // !PROTOBUF_MESSAGE_GLOBALS
#ifndef PROTOBUF_MESSAGE_GLOBALS
PROTOBUF_CONSTINIT
PROTOBUF_ATTRIBUTE_INIT_PRIORITY1 const UpdateNodeInstanceConnection::ParseTableT_
    UpdateNodeInstanceConnection::_table_ =
        UpdateNodeInstanceConnection::InternalGenerateParseTable_(UpdateNodeInstanceConnection_class_data_.base());
#endif  // !PROTOBUF_MESSAGE_GLOBALS
PROTOBUF_NOINLINE void UpdateNodeInstanceConnection::Clear() {
// @@protoc_insertion_point(message_clear_start:nodeinstance.v1.UpdateNodeInstanceConnection)
  ::google::protobuf::internal::TSanWrite(&_impl_);
  ::uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void) cached_has_bits;

  cached_has_bits = _impl_._has_bits_[0];
  if (BatchCheckHasBit(cached_has_bits, 0x0000001fU)) {
    if (CheckHasBit(cached_has_bits, 0x00000001U)) {
      _impl_.capabilities_.Clear();
    }
    if (CheckHasBit(cached_has_bits, 0x00000002U)) {
      _impl_.claim_id_.ClearNonDefaultToEmpty();
    }
    if (CheckHasBit(cached_has_bits, 0x00000004U)) {
      _impl_.node_id_.ClearNonDefaultToEmpty();
    }
    if (CheckHasBit(cached_has_bits, 0x00000008U)) {
      _impl_.mqtt_broker_addr_.ClearNonDefaultToEmpty();
    }
    if (CheckHasBit(cached_has_bits, 0x00000010U)) {
      ABSL_DCHECK(_impl_.updated_at_ != nullptr);
      _impl_.updated_at_->Clear();
    }
  }
  if (BatchCheckHasBit(cached_has_bits, 0x000000e0U)) {
    ::memset(&_impl_.session_id_, 0, static_cast<::size_t>(
        reinterpret_cast<char*>(&_impl_.queryable_) -
        reinterpret_cast<char*>(&_impl_.session_id_)) + sizeof(_impl_.queryable_));
  }
  if (BatchCheckHasBit(cached_has_bits, 0x00000300U)) {
    ::memset(&_impl_.vmq_instance_id_, 0, static_cast<::size_t>(
        reinterpret_cast<char*>(&_impl_.hops_) -
        reinterpret_cast<char*>(&_impl_.vmq_instance_id_)) + sizeof(_impl_.hops_));
  }
  _impl_._has_bits_.Clear();
  _internal_metadata_.Clear<::google::protobuf::UnknownFieldSet>();
}

#if defined(PROTOBUF_CUSTOM_VTABLE)
::uint8_t* PROTOBUF_NONNULL UpdateNodeInstanceConnection::_InternalSerialize(
    const ::google::protobuf::MessageLite& base, ::uint8_t* PROTOBUF_NONNULL target,
    ::google::protobuf::io::EpsCopyOutputStream* PROTOBUF_NONNULL stream) {
  const UpdateNodeInstanceConnection& this_ = static_cast<const UpdateNodeInstanceConnection&>(base);
#else   // PROTOBUF_CUSTOM_VTABLE
::uint8_t* PROTOBUF_NONNULL UpdateNodeInstanceConnection::_InternalSerialize(
    ::uint8_t* PROTOBUF_NONNULL target,
    ::google::protobuf::io::EpsCopyOutputStream* PROTOBUF_NONNULL stream) const {
  const UpdateNodeInstanceConnection& this_ = *this;
#endif  // PROTOBUF_CUSTOM_VTABLE
  if constexpr (::_pbi::DebugHardenCheckHasBitConsistency()) {
    this_.CheckHasBitConsistency();
  }
  // @@protoc_insertion_point(serialize_to_array_start:nodeinstance.v1.UpdateNodeInstanceConnection)
  ::uint32_t cached_has_bits = 0;
  (void)cached_has_bits;

  cached_has_bits = this_._impl_._has_bits_[0];
  // string claim_id = 1;
  if (CheckHasBit(cached_has_bits, 0x00000002U)) {
    if (!this_._internal_claim_id().empty()) {
      const ::std::string& _s = this_._internal_claim_id();
      ::google::protobuf::internal::WireFormatLite::VerifyUtf8String(
          _s.data(), static_cast<int>(_s.length()), ::google::protobuf::internal::WireFormatLite::SERIALIZE, "nodeinstance.v1.UpdateNodeInstanceConnection.claim_id");
      target = stream->WriteStringMaybeAliased(1, _s, target);
    }
  }

  // string node_id = 2;
  if (CheckHasBit(cached_has_bits, 0x00000004U)) {
    if (!this_._internal_node_id().empty()) {
      const ::std::string& _s = this_._internal_node_id();
      ::google::protobuf::internal::WireFormatLite::VerifyUtf8String(
          _s.data(), static_cast<int>(_s.length()), ::google::protobuf::internal::WireFormatLite::SERIALIZE, "nodeinstance.v1.UpdateNodeInstanceConnection.node_id");
      target = stream->WriteStringMaybeAliased(2, _s, target);
    }
  }

  // bool liveness = 3;
  if (CheckHasBit(cached_has_bits, 0x00000040U)) {
    if (this_._internal_liveness() != 0) {
      target = stream->EnsureSpace(target);
      target = ::_pbi::WireFormatLite::WriteBoolToArray(
          3, this_._internal_liveness(), target);
    }
  }

  // bool queryable = 4;
  if (CheckHasBit(cached_has_bits, 0x00000080U)) {
    if (this_._internal_queryable() != 0) {
      target = stream->EnsureSpace(target);
      target = ::_pbi::WireFormatLite::WriteBoolToArray(
          4, this_._internal_queryable(), target);
    }
  }

  // int64 session_id = 5;
  if (CheckHasBit(cached_has_bits, 0x00000020U)) {
    if (this_._internal_session_id() != 0) {
      target =
          ::google::protobuf::internal::WireFormatLite::WriteInt64ToArrayWithField<5>(
              stream, this_._internal_session_id(), target);
    }
  }

  // .google.protobuf.Timestamp updated_at = 6;
  if (CheckHasBit(cached_has_bits, 0x00000010U)) {
    target = ::google::protobuf::internal::WireFormatLite::InternalWriteMessage(
        6, *this_._impl_.updated_at_, this_._impl_.updated_at_->GetCachedSize(), target,
        stream);
  }

  // string mqtt_broker_addr = 7;
  if (CheckHasBit(cached_has_bits, 0x00000008U)) {
    if (!this_._internal_mqtt_broker_addr().empty()) {
      const ::std::string& _s = this_._internal_mqtt_broker_addr();
      ::google::protobuf::internal::WireFormatLite::VerifyUtf8String(
          _s.data(), static_cast<int>(_s.length()), ::google::protobuf::internal::WireFormatLite::SERIALIZE, "nodeinstance.v1.UpdateNodeInstanceConnection.mqtt_broker_addr");
      target = stream->WriteStringMaybeAliased(7, _s, target);
    }
  }

  // int32 vmq_instance_id = 8;
  if (CheckHasBit(cached_has_bits, 0x00000100U)) {
    if (this_._internal_vmq_instance_id() != 0) {
      target =
          ::google::protobuf::internal::WireFormatLite::WriteInt32ToArrayWithField<8>(
              stream, this_._internal_vmq_instance_id(), target);
    }
  }

  // int32 hops = 9;
  if (CheckHasBit(cached_has_bits, 0x00000200U)) {
    if (this_._internal_hops() != 0) {
      target =
          ::google::protobuf::internal::WireFormatLite::WriteInt32ToArrayWithField<9>(
              stream, this_._internal_hops(), target);
    }
  }

  // repeated .aclk_lib.v1.Capability capabilities = 10;
  if (CheckHasBit(cached_has_bits, 0x00000001U)) {
    for (unsigned i = 0, n = static_cast<unsigned>(
                             this_._internal_capabilities_size());
         i < n; i++) {
      const auto& repfield = this_._internal_capabilities().Get(i);
      target =
          ::google::protobuf::internal::WireFormatLite::InternalWriteMessage(
              10, repfield, repfield.GetCachedSize(),
              target, stream);
    }
  }

  if (ABSL_PREDICT_FALSE(this_._internal_metadata_.have_unknown_fields())) {
    target =
        ::_pbi::WireFormat::InternalSerializeUnknownFieldsToArray(
            this_._internal_metadata_.unknown_fields<::google::protobuf::UnknownFieldSet>(::google::protobuf::UnknownFieldSet::default_instance), target, stream);
  }
  // @@protoc_insertion_point(serialize_to_array_end:nodeinstance.v1.UpdateNodeInstanceConnection)
  return target;
}

#if defined(PROTOBUF_CUSTOM_VTABLE)
::size_t UpdateNodeInstanceConnection::ByteSizeLong(const MessageLite& base) {
  const UpdateNodeInstanceConnection& this_ = static_cast<const UpdateNodeInstanceConnection&>(base);
#else   // PROTOBUF_CUSTOM_VTABLE
::size_t UpdateNodeInstanceConnection::ByteSizeLong() const {
  const UpdateNodeInstanceConnection& this_ = *this;
#endif  // PROTOBUF_CUSTOM_VTABLE
  // @@protoc_insertion_point(message_byte_size_start:nodeinstance.v1.UpdateNodeInstanceConnection)
  ::size_t total_size = 0;

  ::uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void)cached_has_bits;

  ::_pbi::Prefetch5LinesFrom7Lines(&this_);
  cached_has_bits = this_._impl_._has_bits_[0];
  if (BatchCheckHasBit(cached_has_bits, 0x000000ffU)) {
    // repeated .aclk_lib.v1.Capability capabilities = 10;
    if (CheckHasBit(cached_has_bits, 0x00000001U)) {
      total_size += 1UL * this_._internal_capabilities_size();
      for (const auto& msg : this_._internal_capabilities()) {
        total_size += ::google::protobuf::internal::WireFormatLite::MessageSize(msg);
      }
    }
    // string claim_id = 1;
    if (CheckHasBit(cached_has_bits, 0x00000002U)) {
      if (!this_._internal_claim_id().empty()) {
        total_size += 1 + ::google::protobuf::internal::WireFormatLite::StringSize(
                                        this_._internal_claim_id());
      }
    }
    // string node_id = 2;
    if (CheckHasBit(cached_has_bits, 0x00000004U)) {
      if (!this_._internal_node_id().empty()) {
        total_size += 1 + ::google::protobuf::internal::WireFormatLite::StringSize(
                                        this_._internal_node_id());
      }
    }
    // string mqtt_broker_addr = 7;
    if (CheckHasBit(cached_has_bits, 0x00000008U)) {
      if (!this_._internal_mqtt_broker_addr().empty()) {
        total_size += 1 + ::google::protobuf::internal::WireFormatLite::StringSize(
                                        this_._internal_mqtt_broker_addr());
      }
    }
    // .google.protobuf.Timestamp updated_at = 6;
    if (CheckHasBit(cached_has_bits, 0x00000010U)) {
      total_size += 1 +
                    ::google::protobuf::internal::WireFormatLite::MessageSize(*this_._impl_.updated_at_);
    }
    // int64 session_id = 5;
    if (CheckHasBit(cached_has_bits, 0x00000020U)) {
      if (this_._internal_session_id() != 0) {
        total_size += ::_pbi::WireFormatLite::Int64SizePlusOne(
            this_._internal_session_id());
      }
    }
    // bool liveness = 3;
    if (CheckHasBit(cached_has_bits, 0x00000040U)) {
      if (this_._internal_liveness() != 0) {
        total_size += 2;
      }
    }
    // bool queryable = 4;
    if (CheckHasBit(cached_has_bits, 0x00000080U)) {
      if (this_._internal_queryable() != 0) {
        total_size += 2;
      }
    }
  }
  if (BatchCheckHasBit(cached_has_bits, 0x00000300U)) {
    // int32 vmq_instance_id = 8;
    if (CheckHasBit(cached_has_bits, 0x00000100U)) {
      if (this_._internal_vmq_instance_id() != 0) {
        total_size += ::_pbi::WireFormatLite::Int32SizePlusOne(
            this_._internal_vmq_instance_id());
      }
    }
    // int32 hops = 9;
    if (CheckHasBit(cached_has_bits, 0x00000200U)) {
      if (this_._internal_hops() != 0) {
        total_size += ::_pbi::WireFormatLite::Int32SizePlusOne(
            this_._internal_hops());
      }
    }
  }
  return this_.MaybeComputeUnknownFieldsSize(total_size,
                                             &this_._impl_._cached_size_);
}

void UpdateNodeInstanceConnection::MergeImpl(::google::protobuf::MessageLite& to_msg,
                      const ::google::protobuf::MessageLite& from_msg) {
   auto* const _this = static_cast<UpdateNodeInstanceConnection*>(&to_msg);
  auto& from = static_cast<const UpdateNodeInstanceConnection&>(from_msg);
  if constexpr (::_pbi::DebugHardenCheckHasBitConsistency()) {
    from.CheckHasBitConsistency();
  }
  ::google::protobuf::Arena* arena = _this->GetArena();
  // @@protoc_insertion_point(class_specific_merge_from_start:nodeinstance.v1.UpdateNodeInstanceConnection)
  ABSL_DCHECK_NE(&from, _this);
  ::uint32_t cached_has_bits = 0;
  (void)cached_has_bits;

  cached_has_bits = from._impl_._has_bits_[0];
  if (BatchCheckHasBit(cached_has_bits, 0x000000ffU)) {
    if (CheckHasBit(cached_has_bits, 0x00000001U)) {
      _this->_internal_mutable_capabilities()->InternalMergeFromWithArena(
          ::google::protobuf::MessageLite::internal_visibility(), arena,
          from._internal_capabilities());
    }
    if (CheckHasBit(cached_has_bits, 0x00000002U)) {
      if (!from._internal_claim_id().empty()) {
        _this->_internal_set_claim_id(from._internal_claim_id());
      } else {
        if (_this->_impl_.claim_id_.IsDefault()) {
          _this->_internal_set_claim_id("");
        }
      }
    }
    if (CheckHasBit(cached_has_bits, 0x00000004U)) {
      if (!from._internal_node_id().empty()) {
        _this->_internal_set_node_id(from._internal_node_id());
      } else {
        if (_this->_impl_.node_id_.IsDefault()) {
          _this->_internal_set_node_id("");
        }
      }
    }
    if (CheckHasBit(cached_has_bits, 0x00000008U)) {
      if (!from._internal_mqtt_broker_addr().empty()) {
        _this->_internal_set_mqtt_broker_addr(from._internal_mqtt_broker_addr());
      } else {
        if (_this->_impl_.mqtt_broker_addr_.IsDefault()) {
          _this->_internal_set_mqtt_broker_addr("");
        }
      }
    }
    if (CheckHasBit(cached_has_bits, 0x00000010U)) {
      ABSL_DCHECK(from._impl_.updated_at_ != nullptr);
      if (_this->_impl_.updated_at_ == nullptr) {
        _this->_impl_.updated_at_ = ::google::protobuf::Message::CopyConstruct(arena, *from._impl_.updated_at_);
      } else {
        _this->_impl_.updated_at_->MergeFrom(*from._impl_.updated_at_);
      }
    }
    if (CheckHasBit(cached_has_bits, 0x00000020U)) {
      if (from._internal_session_id() != 0) {
        _this->_impl_.session_id_ = from._impl_.session_id_;
      }
    }
    if (CheckHasBit(cached_has_bits, 0x00000040U)) {
      if (from._internal_liveness() != 0) {
        _this->_impl_.liveness_ = from._impl_.liveness_;
      }
    }
    if (CheckHasBit(cached_has_bits, 0x00000080U)) {
      if (from._internal_queryable() != 0) {
        _this->_impl_.queryable_ = from._impl_.queryable_;
      }
    }
  }
  if (BatchCheckHasBit(cached_has_bits, 0x00000300U)) {
    if (CheckHasBit(cached_has_bits, 0x00000100U)) {
      if (from._internal_vmq_instance_id() != 0) {
        _this->_impl_.vmq_instance_id_ = from._impl_.vmq_instance_id_;
      }
    }
    if (CheckHasBit(cached_has_bits, 0x00000200U)) {
      if (from._internal_hops() != 0) {
        _this->_impl_.hops_ = from._impl_.hops_;
      }
    }
  }
  _this->_impl_._has_bits_[0] |= cached_has_bits;
  _this->_internal_metadata_.MergeFrom<::google::protobuf::UnknownFieldSet>(
      from._internal_metadata_);
}

void UpdateNodeInstanceConnection::CopyFrom(const UpdateNodeInstanceConnection& from) {
  // @@protoc_insertion_point(class_specific_copy_from_start:nodeinstance.v1.UpdateNodeInstanceConnection)
  if (&from == this) return;
  Clear();
  MergeFrom(from);
}


void UpdateNodeInstanceConnection::InternalSwap(UpdateNodeInstanceConnection* PROTOBUF_RESTRICT PROTOBUF_NONNULL other) {
  using ::std::swap;
  auto* arena = GetArena();
  ABSL_DCHECK_EQ(arena, other->GetArena());
  _internal_metadata_.InternalSwap(&other->_internal_metadata_);
  swap(_impl_._has_bits_[0], other->_impl_._has_bits_[0]);
  _impl_.capabilities_.InternalSwap(&other->_impl_.capabilities_);
  ::_pbi::ArenaStringPtr::InternalSwap(&_impl_.claim_id_, &other->_impl_.claim_id_, arena);
  ::_pbi::ArenaStringPtr::InternalSwap(&_impl_.node_id_, &other->_impl_.node_id_, arena);
  ::_pbi::ArenaStringPtr::InternalSwap(&_impl_.mqtt_broker_addr_, &other->_impl_.mqtt_broker_addr_, arena);
  ::google::protobuf::internal::memswap<
      PROTOBUF_FIELD_OFFSET(UpdateNodeInstanceConnection, _impl_.hops_)
      + sizeof(UpdateNodeInstanceConnection::_impl_.hops_)
      - PROTOBUF_FIELD_OFFSET(UpdateNodeInstanceConnection, _impl_.updated_at_)>(
          reinterpret_cast<char*>(&_impl_.updated_at_),
          reinterpret_cast<char*>(&other->_impl_.updated_at_));
}

::google::protobuf::Metadata UpdateNodeInstanceConnection::GetMetadata() const {
  return ::google::protobuf::Message::GetMetadataImpl(GetClassData()->full());
}
// @@protoc_insertion_point(namespace_scope)
}  // namespace v1
}  // namespace nodeinstance
namespace google {
namespace protobuf {
}  // namespace protobuf
}  // namespace google
// @@protoc_insertion_point(global_scope)
PROTOBUF_ATTRIBUTE_INIT_PRIORITY2 static ::std::false_type
    _static_init2_ [[maybe_unused]] =
        (::_pbi::AddDescriptors(&descriptor_table_nodeinstance_2fconnection_2fv1_2fconnection_2eproto),
         ::std::false_type{});
#include "google/protobuf/port_undef.inc"
