// Generated by the protocol buffer compiler.  DO NOT EDIT!
// NO CHECKED-IN PROTOBUF GENCODE
// source: agent/v1/disconnect.proto
// Protobuf C++ Version: 7.35.1

#ifndef agent_2fv1_2fdisconnect_2eproto_2epb_2eh
#define agent_2fv1_2fdisconnect_2eproto_2epb_2eh

#include <limits>
#include <string>
#include <type_traits>
#include <utility>

// clang-format off
#include "google/protobuf/runtime_version.h"
#if PROTOBUF_VERSION != 7035001
#error "Protobuf C++ gencode is built with an incompatible version of"
#error "Protobuf C++ headers/runtime. See"
#error "https://protobuf.dev/support/cross-version-runtime-guarantee/#cpp"
#endif
#include "google/protobuf/io/coded_stream.h"
#include "google/protobuf/arena.h"
#include "google/protobuf/arenastring.h"
#include "google/protobuf/generated_message_tctable_decl.h"
#include "google/protobuf/generated_message_util.h"
#include "google/protobuf/metadata_lite.h"
#include "google/protobuf/generated_message_reflection.h"
#include "google/protobuf/message.h"
#include "google/protobuf/message_lite.h"
#include "google/protobuf/repeated_field.h"  // IWYU pragma: export
#include "google/protobuf/extension_set.h"  // IWYU pragma: export
#include "google/protobuf/unknown_field_set.h"
#include "google/protobuf/timestamp.pb.h"
// @@protoc_insertion_point(includes)

// Must be included last.
#include "google/protobuf/port_def.inc"

#define PROTOBUF_INTERNAL_EXPORT_agent_2fv1_2fdisconnect_2eproto

namespace google {
namespace protobuf {
namespace internal {
template <typename T>
::absl::string_view GetAnyMessageName();
}  // namespace internal
}  // namespace protobuf
}  // namespace google

// Internal implementation detail -- do not use these members.
struct TableStruct_agent_2fv1_2fdisconnect_2eproto {
  static const ::uint32_t offsets[];
};
extern "C" {
extern const ::google::protobuf::internal::DescriptorTable descriptor_table_agent_2fv1_2fdisconnect_2eproto;
}  // extern "C"
namespace agent {
namespace v1 {
class DisconnectReq;
struct DisconnectReqGlobalsTypeInternal;
#ifndef PROTOBUF_MESSAGE_GLOBALS
extern DisconnectReqGlobalsTypeInternal DisconnectReq_globals_;
extern const ::google::protobuf::internal::ClassDataFull DisconnectReq_class_data_;
#else
extern const DisconnectReqGlobalsTypeInternal DisconnectReq_globals_;
#endif  // PROTOBUF_MESSAGE_GLOBALS
}  // namespace v1
}  // namespace agent
namespace google {
namespace protobuf {
}  // namespace protobuf
}  // namespace google

namespace agent {
namespace v1 {

// ===================================================================


// -------------------------------------------------------------------

class  PROTOBUF_FUTURE_ADD_EARLY_WARN_UNUSED DisconnectReq final : public ::google::protobuf::Message
/* @@protoc_insertion_point(class_definition:agent.v1.DisconnectReq) */ {
 public:
  inline DisconnectReq() : DisconnectReq(nullptr) {}
  ~DisconnectReq() PROTOBUF_FINAL;

#if defined(PROTOBUF_CUSTOM_VTABLE)
  void operator delete(DisconnectReq* PROTOBUF_NONNULL msg, ::std::destroying_delete_t) {
    SharedDtor(*msg);
    ::google::protobuf::internal::SizedDelete(msg, sizeof(DisconnectReq));
  }
#endif

  template <typename = void>
  explicit constexpr DisconnectReq(::google::protobuf::internal::ConstantInitialized,
                           const ::google::protobuf::internal::ClassData* PROTOBUF_NONNULL
                               class_data);

  inline DisconnectReq(const DisconnectReq& from) : DisconnectReq(nullptr, from) {}
  inline DisconnectReq(DisconnectReq&& from) noexcept : DisconnectReq(nullptr, ::std::move(from)) {}
  inline DisconnectReq& operator=(const DisconnectReq& from) {
    CopyFrom(from);
    return *this;
  }
  inline DisconnectReq& operator=(DisconnectReq&& from) noexcept {
    if (this == &from) return *this;
    if (::google::protobuf::internal::CanMoveWithInternalSwap(GetArena(), from.GetArena())) {
      InternalSwap(&from);
    } else {
      CopyFrom(from);
    }
    return *this;
  }

  [[nodiscard]] inline const ::google::protobuf::UnknownFieldSet& unknown_fields() const
      ABSL_ATTRIBUTE_LIFETIME_BOUND {
    return _internal_metadata_.unknown_fields<::google::protobuf::UnknownFieldSet>(::google::protobuf::UnknownFieldSet::default_instance);
  }
  [[nodiscard]] inline ::google::protobuf::UnknownFieldSet* PROTOBUF_NONNULL
  mutable_unknown_fields() ABSL_ATTRIBUTE_LIFETIME_BOUND {
    return _internal_metadata_.mutable_unknown_fields<::google::protobuf::UnknownFieldSet>();
  }

  [[nodiscard]] static const ::google::protobuf::Descriptor* PROTOBUF_NONNULL descriptor() {
    return GetDescriptor();
  }
  [[nodiscard]] static const ::google::protobuf::Descriptor* PROTOBUF_NONNULL
  GetDescriptor() {
    return default_instance().GetMetadata().descriptor;
  }
  [[nodiscard]] static const ::google::protobuf::Reflection* PROTOBUF_NONNULL GetReflection() {
    return default_instance().GetMetadata().reflection;
  }
  [[nodiscard]] static const DisconnectReq& default_instance() {
    return *::google::protobuf::internal::MessageGlobalsBase::ToDefaultInstance<DisconnectReq>(&DisconnectReq_globals_);
  }
  static constexpr int kIndexInFileMessages = 0;
  friend void swap(DisconnectReq& a, DisconnectReq& b) { a.Swap(&b); }
  inline void Swap(DisconnectReq* PROTOBUF_NONNULL other) {
    if (other == this) return;
    if (::google::protobuf::internal::CanUseInternalSwap(GetArena(), other->GetArena())) {
      InternalSwap(other);
    } else {
      ::google::protobuf::internal::GenericSwap(this, other);
    }
  }
  void UnsafeArenaSwap(DisconnectReq* PROTOBUF_NONNULL other) {
    if (other == this) return;
    ABSL_DCHECK(GetArena() == other->GetArena());
    InternalSwap(other);
  }

  // implements Message ----------------------------------------------

  [[nodiscard]] DisconnectReq* PROTOBUF_NONNULL
  New(::google::protobuf::Arena* PROTOBUF_NULLABLE arena = nullptr) const {
    return ::google::protobuf::Message::DefaultConstruct<DisconnectReq>(arena);
  }
  using ::google::protobuf::Message::CopyFrom;
  void CopyFrom(const DisconnectReq& from);
  using ::google::protobuf::Message::MergeFrom;
  void MergeFrom(const DisconnectReq& from) { DisconnectReq::MergeImpl(*this, from); }

  private:
  static void MergeImpl(::google::protobuf::MessageLite& to_msg,
                        const ::google::protobuf::MessageLite& from_msg);

  public:
  [[nodiscard]] bool IsInitialized() const {
    return true;
  }
  ABSL_ATTRIBUTE_REINITIALIZES void Clear() PROTOBUF_FINAL;
  #if defined(PROTOBUF_CUSTOM_VTABLE)
  private:
  [[nodiscard]] static ::size_t ByteSizeLong(const ::google::protobuf::MessageLite& msg);
  [[nodiscard]] static ::uint8_t* PROTOBUF_NONNULL _InternalSerialize(
      const ::google::protobuf::MessageLite& msg, ::uint8_t* PROTOBUF_NONNULL target,
      ::google::protobuf::io::EpsCopyOutputStream* PROTOBUF_NONNULL stream);

  public:
  [[nodiscard]] ::size_t ByteSizeLong() const { return ByteSizeLong(*this); }
  [[nodiscard]] ::uint8_t* PROTOBUF_NONNULL _InternalSerialize(
      ::uint8_t* PROTOBUF_NONNULL target,
      ::google::protobuf::io::EpsCopyOutputStream* PROTOBUF_NONNULL stream) const {
    return _InternalSerialize(*this, target, stream);
  }
  #else   // PROTOBUF_CUSTOM_VTABLE
  [[nodiscard]] ::size_t ByteSizeLong() const final;
  [[nodiscard]] ::uint8_t* PROTOBUF_NONNULL _InternalSerialize(
      ::uint8_t* PROTOBUF_NONNULL target,
      ::google::protobuf::io::EpsCopyOutputStream* PROTOBUF_NONNULL stream) const final;
  #endif  // PROTOBUF_CUSTOM_VTABLE
  [[nodiscard]] int GetCachedSize() const {
    return _impl_._cached_size_.Get();
  }

  private:
  void SharedCtor(::google::protobuf::Arena* PROTOBUF_NULLABLE arena);
  static void SharedDtor(MessageLite& self);
  void InternalSwap(DisconnectReq* PROTOBUF_NONNULL other);
 private:
  template <typename T>
  friend ::absl::string_view(::google::protobuf::internal::GetAnyMessageName)();
  static ::absl::string_view FullMessageName() { return "agent.v1.DisconnectReq"; }

  explicit DisconnectReq(::google::protobuf::Arena* PROTOBUF_NULLABLE arena);
  DisconnectReq(::google::protobuf::Arena* PROTOBUF_NULLABLE arena, const DisconnectReq& from);
  DisconnectReq(
      ::google::protobuf::Arena* PROTOBUF_NULLABLE arena, DisconnectReq&& from) noexcept
      : DisconnectReq(arena) {
    *this = ::std::move(from);
  }
  const ::google::protobuf::internal::ClassData* PROTOBUF_NONNULL GetClassData() const PROTOBUF_FINAL;
  static void* PROTOBUF_NONNULL PlacementNew_(
      const void* PROTOBUF_NONNULL, void* PROTOBUF_NONNULL mem,
      ::google::protobuf::Arena* PROTOBUF_NULLABLE arena);
  static constexpr auto InternalNewImpl_();

 public:
  static constexpr auto InternalGenerateClassData_(
      const MessageLite& prototype,
      const ::google::protobuf::internal::TcParseTableBase* PROTOBUF_NULLABLE tc_table = nullptr);

  [[nodiscard]] ::google::protobuf::Metadata GetMetadata() const;
  // nested types ----------------------------------------------------

  // accessors -------------------------------------------------------
  enum : int {
    kErrorDescriptionFieldNumber = 5,
    kCreatedAtFieldNumber = 3,
    kReconnectAfterSecondsFieldNumber = 1,
    kPermabanFieldNumber = 2,
    kErrorCodeFieldNumber = 4,
  };
  // string error_description = 5;
  void clear_error_description() ;
  [[nodiscard]] const ::std::string& error_description() const;
  template <typename Arg_ = const ::std::string&, typename... Args_>
  void set_error_description(Arg_&& arg, Args_... args);
  ::std::string* PROTOBUF_NONNULL mutable_error_description();
  [[nodiscard]] ::std::string* PROTOBUF_NULLABLE release_error_description();
  void set_allocated_error_description(::std::string* PROTOBUF_NULLABLE value);

  private:
  const ::std::string& _internal_error_description() const;
  PROTOBUF_ALWAYS_INLINE void _internal_set_error_description(const ::std::string& value);
  ::std::string* PROTOBUF_NONNULL _internal_mutable_error_description();

  public:
  // .google.protobuf.Timestamp created_at = 3;
  [[nodiscard]] bool has_created_at()
      const;
  void clear_created_at() ;
  [[nodiscard]] const ::google::protobuf::Timestamp& created_at() const;
  [[nodiscard]] ::google::protobuf::Timestamp* PROTOBUF_NULLABLE release_created_at();
  ::google::protobuf::Timestamp* PROTOBUF_NONNULL mutable_created_at();
  void set_allocated_created_at(::google::protobuf::Timestamp* PROTOBUF_NULLABLE value);
  void unsafe_arena_set_allocated_created_at(::google::protobuf::Timestamp* PROTOBUF_NULLABLE value);
  ::google::protobuf::Timestamp* PROTOBUF_NULLABLE unsafe_arena_release_created_at();

  private:
  const ::google::protobuf::Timestamp& _internal_created_at() const;
  ::google::protobuf::Timestamp* PROTOBUF_NONNULL _internal_mutable_created_at();

  public:
  // uint64 reconnect_after_seconds = 1;
  void clear_reconnect_after_seconds() ;
  [[nodiscard]] ::uint64_t reconnect_after_seconds() const;
  void set_reconnect_after_seconds(::uint64_t value);

  private:
  ::uint64_t _internal_reconnect_after_seconds() const;
  void _internal_set_reconnect_after_seconds(::uint64_t value);

  public:
  // bool permaban = 2;
  void clear_permaban() ;
  [[nodiscard]] bool permaban() const;
  void set_permaban(bool value);

  private:
  bool _internal_permaban() const;
  void _internal_set_permaban(bool value);

  public:
  // uint32 error_code = 4;
  void clear_error_code() ;
  [[nodiscard]] ::uint32_t error_code() const;
  void set_error_code(::uint32_t value);

  private:
  ::uint32_t _internal_error_code() const;
  void _internal_set_error_code(::uint32_t value);

  public:
  // @@protoc_insertion_point(class_scope:agent.v1.DisconnectReq)
 private:
  class _Internal;
  using ParseTableT_ =
      ::google::protobuf::internal::TcParseTable<3, 5,
                          1, 48,
                          2>;
  static constexpr ParseTableT_ InternalGenerateParseTable_(
      const ::google::protobuf::internal::ClassData* PROTOBUF_NONNULL class_data);
  friend class ::google::protobuf::internal::TcParser;
  #ifndef PROTOBUF_MESSAGE_GLOBALS
  static const ParseTableT_ _table_;
  #endif

  friend class ::google::protobuf::MessageLite;
  friend class ::google::protobuf::Arena;
  friend ::google::protobuf::internal::PrivateAccess;
  template <typename T>
  friend class ::google::protobuf::Arena::InternalHelper;
  using InternalArenaConstructable_ = void;
  using DestructorSkippable_ = void;
  struct Impl_ {
    inline explicit constexpr Impl_(::google::protobuf::internal::InternalVisibility visibility,
                                    ::google::protobuf::internal::ConstantInitialized) noexcept;
    inline explicit Impl_(
        ::google::protobuf::internal::InternalVisibility visibility,
        ::google::protobuf::Arena* PROTOBUF_NULLABLE arena);
    inline explicit Impl_(
        ::google::protobuf::internal::InternalVisibility visibility,
        ::google::protobuf::Arena* PROTOBUF_NULLABLE arena, const Impl_& from,
        const DisconnectReq& from_msg);
    ::google::protobuf::internal::HasBits<1> _has_bits_;
    ::google::protobuf::internal::CachedSize _cached_size_;
    ::google::protobuf::internal::ArenaStringPtr error_description_;
    ::google::protobuf::Timestamp* PROTOBUF_NULLABLE created_at_;
    ::uint64_t reconnect_after_seconds_;
    bool permaban_;
    ::uint32_t error_code_;
    PROTOBUF_TSAN_DECLARE_MEMBER
  };
  union { Impl_ _impl_; };
  friend struct ::TableStruct_agent_2fv1_2fdisconnect_2eproto;
};

// ===================================================================




// ===================================================================


#ifdef __GNUC__
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wstrict-aliasing"
#endif  // __GNUC__
// -------------------------------------------------------------------

// DisconnectReq

// uint64 reconnect_after_seconds = 1;
inline void DisconnectReq::clear_reconnect_after_seconds() {
  ::google::protobuf::internal::TSanWrite(&_impl_);
  _impl_.reconnect_after_seconds_ = ::uint64_t{0u};
  ClearHasBit(_impl_._has_bits_[0], 0x00000004U);
}
inline ::uint64_t DisconnectReq::reconnect_after_seconds() const {
  // @@protoc_insertion_point(field_get:agent.v1.DisconnectReq.reconnect_after_seconds)
  return _internal_reconnect_after_seconds();
}
inline void DisconnectReq::set_reconnect_after_seconds(::uint64_t value) {
  _internal_set_reconnect_after_seconds(value);
  SetHasBit(_impl_._has_bits_[0], 0x00000004U);
  // @@protoc_insertion_point(field_set:agent.v1.DisconnectReq.reconnect_after_seconds)
}
inline ::uint64_t DisconnectReq::_internal_reconnect_after_seconds() const {
  ::google::protobuf::internal::TSanRead(&_impl_);
  return _impl_.reconnect_after_seconds_;
}
inline void DisconnectReq::_internal_set_reconnect_after_seconds(::uint64_t value) {
  ::google::protobuf::internal::TSanWrite(&_impl_);
  _impl_.reconnect_after_seconds_ = value;
}

// bool permaban = 2;
inline void DisconnectReq::clear_permaban() {
  ::google::protobuf::internal::TSanWrite(&_impl_);
  _impl_.permaban_ = false;
  ClearHasBit(_impl_._has_bits_[0], 0x00000008U);
}
inline bool DisconnectReq::permaban() const {
  // @@protoc_insertion_point(field_get:agent.v1.DisconnectReq.permaban)
  return _internal_permaban();
}
inline void DisconnectReq::set_permaban(bool value) {
  _internal_set_permaban(value);
  SetHasBit(_impl_._has_bits_[0], 0x00000008U);
  // @@protoc_insertion_point(field_set:agent.v1.DisconnectReq.permaban)
}
inline bool DisconnectReq::_internal_permaban() const {
  ::google::protobuf::internal::TSanRead(&_impl_);
  return _impl_.permaban_;
}
inline void DisconnectReq::_internal_set_permaban(bool value) {
  ::google::protobuf::internal::TSanWrite(&_impl_);
  _impl_.permaban_ = value;
}

// .google.protobuf.Timestamp created_at = 3;
inline bool DisconnectReq::has_created_at() const {
  bool value = CheckHasBit(_impl_._has_bits_[0], 0x00000002U);
  PROTOBUF_ASSUME(!value || _impl_.created_at_ != nullptr);
  return value;
}
inline const ::google::protobuf::Timestamp& DisconnectReq::_internal_created_at() const {
  ::google::protobuf::internal::TSanRead(&_impl_);
  const ::google::protobuf::Timestamp* p = _impl_.created_at_;
  return p != nullptr ? *p : *::google::protobuf::internal::MessageGlobalsBase::ToDefaultInstance<::google::protobuf::Timestamp>(&::google::protobuf::Timestamp_globals_);
}
inline const ::google::protobuf::Timestamp& DisconnectReq::created_at() const ABSL_ATTRIBUTE_LIFETIME_BOUND {
  // @@protoc_insertion_point(field_get:agent.v1.DisconnectReq.created_at)
  return _internal_created_at();
}
inline void DisconnectReq::unsafe_arena_set_allocated_created_at(
    ::google::protobuf::Timestamp* PROTOBUF_NULLABLE value) {
  ::google::protobuf::internal::TSanWrite(&_impl_);
  if (GetArena() == nullptr) {
    delete reinterpret_cast<::google::protobuf::MessageLite*>(_impl_.created_at_);
  }
  _impl_.created_at_ = reinterpret_cast<::google::protobuf::Timestamp*>(value);
  if (value != nullptr) {
    SetHasBit(_impl_._has_bits_[0], 0x00000002U);
  } else {
    ClearHasBit(_impl_._has_bits_[0], 0x00000002U);
  }
  // @@protoc_insertion_point(field_unsafe_arena_set_allocated:agent.v1.DisconnectReq.created_at)
}
inline ::google::protobuf::Timestamp* PROTOBUF_NULLABLE DisconnectReq::release_created_at() {
  ::google::protobuf::internal::TSanWrite(&_impl_);

  ClearHasBit(_impl_._has_bits_[0], 0x00000002U);
  ::google::protobuf::Timestamp* released = _impl_.created_at_;
  _impl_.created_at_ = nullptr;
  if (::google::protobuf::internal::DebugHardenForceCopyInRelease()) {
    auto* old = reinterpret_cast<::google::protobuf::MessageLite*>(released);
    released = ::google::protobuf::internal::DuplicateIfNonNull(released);
    if (GetArena() == nullptr) {
      delete old;
    }
  } else {
    if (GetArena() != nullptr) {
      released = ::google::protobuf::internal::DuplicateIfNonNull(released);
    }
  }
  return released;
}
inline ::google::protobuf::Timestamp* PROTOBUF_NULLABLE DisconnectReq::unsafe_arena_release_created_at() {
  ::google::protobuf::internal::TSanWrite(&_impl_);
  // @@protoc_insertion_point(field_release:agent.v1.DisconnectReq.created_at)

  ClearHasBit(_impl_._has_bits_[0], 0x00000002U);
  ::google::protobuf::Timestamp* temp = _impl_.created_at_;
  _impl_.created_at_ = nullptr;
  return temp;
}
inline ::google::protobuf::Timestamp* PROTOBUF_NONNULL DisconnectReq::_internal_mutable_created_at() {
  ::google::protobuf::internal::TSanWrite(&_impl_);
  if (_impl_.created_at_ == nullptr) {
    auto* p = ::google::protobuf::Message::DefaultConstruct<::google::protobuf::Timestamp>(GetArena());
    _impl_.created_at_ = reinterpret_cast<::google::protobuf::Timestamp*>(p);
  }
  return _impl_.created_at_;
}
inline ::google::protobuf::Timestamp* PROTOBUF_NONNULL DisconnectReq::mutable_created_at()
    ABSL_ATTRIBUTE_LIFETIME_BOUND {
  SetHasBit(_impl_._has_bits_[0], 0x00000002U);
  ::google::protobuf::Timestamp* _msg = _internal_mutable_created_at();
  // @@protoc_insertion_point(field_mutable:agent.v1.DisconnectReq.created_at)
  return _msg;
}
inline void DisconnectReq::set_allocated_created_at(::google::protobuf::Timestamp* PROTOBUF_NULLABLE value) {
  ::google::protobuf::Arena* message_arena = GetArena();
  ::google::protobuf::internal::TSanWrite(&_impl_);
  if (message_arena == nullptr) {
    delete reinterpret_cast<::google::protobuf::MessageLite*>(_impl_.created_at_);
  }

  if (value != nullptr) {
    ::google::protobuf::Arena* submessage_arena = reinterpret_cast<::google::protobuf::Message*>(value)->GetArena();
    if (message_arena != submessage_arena) {
      value = ::google::protobuf::internal::GetOwnedMessage(message_arena, value, submessage_arena);
    }
    SetHasBit(_impl_._has_bits_[0], 0x00000002U);
  } else {
    ClearHasBit(_impl_._has_bits_[0], 0x00000002U);
  }

  _impl_.created_at_ = reinterpret_cast<::google::protobuf::Timestamp*>(value);
  // @@protoc_insertion_point(field_set_allocated:agent.v1.DisconnectReq.created_at)
}

// uint32 error_code = 4;
inline void DisconnectReq::clear_error_code() {
  ::google::protobuf::internal::TSanWrite(&_impl_);
  _impl_.error_code_ = 0u;
  ClearHasBit(_impl_._has_bits_[0], 0x00000010U);
}
inline ::uint32_t DisconnectReq::error_code() const {
  // @@protoc_insertion_point(field_get:agent.v1.DisconnectReq.error_code)
  return _internal_error_code();
}
inline void DisconnectReq::set_error_code(::uint32_t value) {
  _internal_set_error_code(value);
  SetHasBit(_impl_._has_bits_[0], 0x00000010U);
  // @@protoc_insertion_point(field_set:agent.v1.DisconnectReq.error_code)
}
inline ::uint32_t DisconnectReq::_internal_error_code() const {
  ::google::protobuf::internal::TSanRead(&_impl_);
  return _impl_.error_code_;
}
inline void DisconnectReq::_internal_set_error_code(::uint32_t value) {
  ::google::protobuf::internal::TSanWrite(&_impl_);
  _impl_.error_code_ = value;
}

// string error_description = 5;
inline void DisconnectReq::clear_error_description() {
  ::google::protobuf::internal::TSanWrite(&_impl_);
  _impl_.error_description_.ClearToEmpty();
  ClearHasBit(_impl_._has_bits_[0], 0x00000001U);
}
inline const ::std::string& DisconnectReq::error_description() const
    ABSL_ATTRIBUTE_LIFETIME_BOUND {
  // @@protoc_insertion_point(field_get:agent.v1.DisconnectReq.error_description)
  return _internal_error_description();
}
template <typename Arg_, typename... Args_>
PROTOBUF_ALWAYS_INLINE void DisconnectReq::set_error_description(Arg_&& arg, Args_... args) {
  ::google::protobuf::internal::TSanWrite(&_impl_);
  SetHasBit(_impl_._has_bits_[0], 0x00000001U);
  _impl_.error_description_.Set(static_cast<Arg_&&>(arg), args..., GetArena());
  // @@protoc_insertion_point(field_set:agent.v1.DisconnectReq.error_description)
}
inline ::std::string* PROTOBUF_NONNULL DisconnectReq::mutable_error_description()
    ABSL_ATTRIBUTE_LIFETIME_BOUND {
  SetHasBit(_impl_._has_bits_[0], 0x00000001U);
  ::std::string* _s = _internal_mutable_error_description();
  // @@protoc_insertion_point(field_mutable:agent.v1.DisconnectReq.error_description)
  return _s;
}
inline const ::std::string& DisconnectReq::_internal_error_description() const {
  ::google::protobuf::internal::TSanRead(&_impl_);
  return _impl_.error_description_.Get();
}
inline void DisconnectReq::_internal_set_error_description(const ::std::string& value) {
  ::google::protobuf::internal::TSanWrite(&_impl_);
  _impl_.error_description_.Set(value, GetArena());
}
inline ::std::string* PROTOBUF_NONNULL DisconnectReq::_internal_mutable_error_description() {
  ::google::protobuf::internal::TSanWrite(&_impl_);
  return _impl_.error_description_.Mutable( GetArena());
}
inline ::std::string* PROTOBUF_NULLABLE DisconnectReq::release_error_description() {
  ::google::protobuf::internal::TSanWrite(&_impl_);
  // @@protoc_insertion_point(field_release:agent.v1.DisconnectReq.error_description)
  if (!CheckHasBit(_impl_._has_bits_[0], 0x00000001U)) {
    return nullptr;
  }
  ClearHasBit(_impl_._has_bits_[0], 0x00000001U);
  auto* released = _impl_.error_description_.Release();
  if (::google::protobuf::internal::DebugHardenForceCopyDefaultString()) {
    _impl_.error_description_.Set("", GetArena());
  }
  return released;
}
inline void DisconnectReq::set_allocated_error_description(::std::string* PROTOBUF_NULLABLE value) {
  ::google::protobuf::internal::TSanWrite(&_impl_);
  if (value != nullptr) {
    SetHasBit(_impl_._has_bits_[0], 0x00000001U);
  } else {
    ClearHasBit(_impl_._has_bits_[0], 0x00000001U);
  }
  _impl_.error_description_.SetAllocated(value, GetArena());
  if (::google::protobuf::internal::DebugHardenForceCopyDefaultString() && _impl_.error_description_.IsDefault()) {
    _impl_.error_description_.Set("", GetArena());
  }
  // @@protoc_insertion_point(field_set_allocated:agent.v1.DisconnectReq.error_description)
}

#ifdef __GNUC__
#pragma GCC diagnostic pop
#endif  // __GNUC__

// @@protoc_insertion_point(namespace_scope)
}  // namespace v1
}  // namespace agent


// @@protoc_insertion_point(global_scope)

#include "google/protobuf/port_undef.inc"
// clang-format on

#endif  // agent_2fv1_2fdisconnect_2eproto_2epb_2eh
