// SPDX-License-Identifier: GPL-3.0-or-later

#ifndef NETDATA_BENCHMARK_RW_H
#define NETDATA_BENCHMARK_RW_H

#include "../libnetdata.h"

int rwlocks_stress_test(void);

#endif //NETDATA_BENCHMARK_RW_H
