// SPDX-License-Identifier: GPL-3.0-or-later

#ifndef NETDATA_NETDATA_CONF_CLOUD_H
#define NETDATA_NETDATA_CONF_CLOUD_H

#include "libnetdata/libnetdata.h"

size_t netdata_conf_cloud_query_threads(void);

#endif //NETDATA_NETDATA_CONF_CLOUD_H
