// Generated by the protocol buffer compiler.  DO NOT EDIT!
// NO CHECKED-IN PROTOBUF GENCODE
// source: agent/v1/disconnect.proto
// Protobuf C++ Version: 7.35.1

#include "agent/v1/disconnect.pb.h"

#include <algorithm>
#include <type_traits>
#include "google/protobuf/io/coded_stream.h"
#include "google/protobuf/generated_message_tctable_impl.h"
#include "google/protobuf/internal_visibility.h"
#include "google/protobuf/extension_set.h"
#include "google/protobuf/generated_message_util.h"
#include "google/protobuf/wire_format_lite.h"
#include "google/protobuf/descriptor.h"
#include "google/protobuf/generated_message_reflection.h"
#include "google/protobuf/reflection_ops.h"
#include "google/protobuf/wire_format.h"
// @@protoc_insertion_point(includes)

// Must be included last.
#include "google/protobuf/port_def.inc"
PROTOBUF_PRAGMA_INIT_SEG
namespace _pb = ::google::protobuf;
namespace _pbi = ::google::protobuf::internal;
namespace _fl = ::google::protobuf::internal::field_layout;
#ifdef PROTOBUF_MESSAGE_GLOBALS
namespace {
PROTOBUF_CONSTINIT ::google::protobuf::internal::ReflectionData
    file_reflection_data[] = {
        // ::agent::v1::DisconnectReq
        {&::_pbi::kDescriptorMethods, &::descriptor_table_agent_2fv1_2fdisconnect_2eproto, /* tracker*/ nullptr,},
};
}  // namespace
#endif
namespace agent {
namespace v1 {
class DisconnectReq::_Internal {
 public:
  using HasBits = decltype(::std::declval<DisconnectReq>()._impl_._has_bits_);
  static constexpr ::int32_t kHasBitsOffset =
      8 * PROTOBUF_FIELD_OFFSET(DisconnectReq, _impl_._has_bits_);
};

constexpr DisconnectReq::ParseTableT_ DisconnectReq::InternalGenerateParseTable_(const ::_pbi::ClassData* class_data) {
  return ParseTableT_{
    {
      PROTOBUF_FIELD_OFFSET(DisconnectReq, _impl_._has_bits_),
      0, // no _extensions_
      5, 56,  // max_field_number, fast_idx_mask
      offsetof(ParseTableT_, field_lookup_table),
      4294967264,  // skipmap
      offsetof(ParseTableT_, field_entries),
      5,  // num_field_entries
      1,  // num_aux_entries
      offsetof(ParseTableT_, aux_entries),
      class_data,
      nullptr,  // post_loop_handler
      ::_pbi::TcParser::GenericFallback,  // fallback
      #ifdef PROTOBUF_PREFETCH_PARSE_TABLE
      ::_pbi::TcParser::GetTable<::agent::v1::DisconnectReq>(),  // to_prefetch
      #endif  // PROTOBUF_PREFETCH_PARSE_TABLE
    }, {{
      {::_pbi::TcParser::MiniParse, {}},
      // uint64 reconnect_after_seconds = 1;
      {::_pbi::TcParser::SingularVarintNoZag1<::uint64_t, offsetof(DisconnectReq, _impl_.reconnect_after_seconds_), 2>(),
       {8, 2, 0,
        PROTOBUF_FIELD_OFFSET(DisconnectReq, _impl_.reconnect_after_seconds_)}},
      // bool permaban = 2;
      {::_pbi::TcParser::SingularVarintNoZag1<bool, offsetof(DisconnectReq, _impl_.permaban_), 3>(),
       {16, 3, 0,
        PROTOBUF_FIELD_OFFSET(DisconnectReq, _impl_.permaban_)}},
      // .google.protobuf.Timestamp created_at = 3;
      {::_pbi::TcParser::FastMtS1,
       {26, 1, 0,
        PROTOBUF_FIELD_OFFSET(DisconnectReq, _impl_.created_at_)}},
      // uint32 error_code = 4;
      {::_pbi::TcParser::SingularVarintNoZag1<::uint32_t, offsetof(DisconnectReq, _impl_.error_code_), 4>(),
       {32, 4, 0,
        PROTOBUF_FIELD_OFFSET(DisconnectReq, _impl_.error_code_)}},
      // string error_description = 5;
      {::_pbi::TcParser::FastUS1,
       {42, 0, 0,
        PROTOBUF_FIELD_OFFSET(DisconnectReq, _impl_.error_description_)}},
      {::_pbi::TcParser::MiniParse, {}},
      {::_pbi::TcParser::MiniParse, {}},
    }}, {{
      65535, 65535
    }}, {{
      // uint64 reconnect_after_seconds = 1;
      {PROTOBUF_FIELD_OFFSET(DisconnectReq, _impl_.reconnect_after_seconds_), _Internal::kHasBitsOffset + 2, 0, (0 | ::_fl::kFcOptional | ::_fl::kUInt64)},
      // bool permaban = 2;
      {PROTOBUF_FIELD_OFFSET(DisconnectReq, _impl_.permaban_), _Internal::kHasBitsOffset + 3, 0, (0 | ::_fl::kFcOptional | ::_fl::kBool)},
      // .google.protobuf.Timestamp created_at = 3;
      {PROTOBUF_FIELD_OFFSET(DisconnectReq, _impl_.created_at_), _Internal::kHasBitsOffset + 1, 0, (0 | ::_fl::kFcOptional | ::_fl::kMessage | ::_fl::kTvTable)},
      // uint32 error_code = 4;
      {PROTOBUF_FIELD_OFFSET(DisconnectReq, _impl_.error_code_), _Internal::kHasBitsOffset + 4, 0, (0 | ::_fl::kFcOptional | ::_fl::kUInt32)},
      // string error_description = 5;
      {PROTOBUF_FIELD_OFFSET(DisconnectReq, _impl_.error_description_), _Internal::kHasBitsOffset + 0, 0, (0 | ::_fl::kFcOptional | ::_fl::kUtf8String | ::_fl::kRepAString)},
    }},
    {{
        #ifndef PROTOBUF_MESSAGE_GLOBALS
        {::_pbi::TcParser::GetTable<::google::protobuf::Timestamp>()},
        #else
        {::_pbi::FieldAuxMessageGlobals(), &::google::protobuf::Timestamp_globals_},
        #endif
    }},
    {{
      "\26\0\0\0\0\21\0\0"
      "agent.v1.DisconnectReq"
      "error_description"
    }},
  };
}


inline constexpr DisconnectReq::Impl_::Impl_(
    [[maybe_unused]] ::google::protobuf::internal::InternalVisibility visibility,
    ::_pbi::ConstantInitialized) noexcept
      : _cached_size_{0},
        error_description_(
            &::google::protobuf::internal::fixed_address_empty_string,
            ::_pbi::ConstantInitialized()),
        created_at_{nullptr},
        reconnect_after_seconds_{::uint64_t{0u}},
        permaban_{false},
        error_code_{0u} {}

template <typename>
constexpr DisconnectReq::DisconnectReq(::_pbi::ConstantInitialized,
                       const ::_pbi::ClassData* PROTOBUF_NONNULL class_data)
    : ::google::protobuf::Message(
#if defined(PROTOBUF_CUSTOM_VTABLE)
          class_data
#endif  // PROTOBUF_CUSTOM_VTABLE
          ),
      _impl_(internal_visibility(), ::_pbi::ConstantInitialized()) {
}
inline void* PROTOBUF_NONNULL DisconnectReq::PlacementNew_(
    const void* PROTOBUF_NONNULL, void* PROTOBUF_NONNULL mem,
    ::google::protobuf::Arena* PROTOBUF_NULLABLE arena) {
  return ::new (mem) DisconnectReq(arena);
}
constexpr auto DisconnectReq::InternalNewImpl_() {
  return ::google::protobuf::internal::MessageCreator::CopyInit(sizeof(DisconnectReq), alignof(DisconnectReq));
}
constexpr auto DisconnectReq::InternalGenerateClassData_(
    const MessageLite& prototype,
    const ::google::protobuf::internal::TcParseTableBase* tc_table) {
  return ::google::protobuf::internal::ClassDataFull{
      ::google::protobuf::internal::ClassData{
          &prototype,
#ifndef PROTOBUF_MESSAGE_GLOBALS
          &_table_.header,
#else
          tc_table,
#endif
          nullptr,  // IsInitialized
          &DisconnectReq::MergeImpl,
          ::google::protobuf::Message::GetNewImpl<DisconnectReq>(),
#if defined(PROTOBUF_CUSTOM_VTABLE)
          &DisconnectReq::SharedDtor,
          ::google::protobuf::Message::GetClearImpl<DisconnectReq>(), &DisconnectReq::ByteSizeLong,
              &DisconnectReq::_InternalSerialize,
#endif  // PROTOBUF_CUSTOM_VTABLE
          PROTOBUF_FIELD_OFFSET(DisconnectReq, _impl_._cached_size_),
          false,
      },
#ifdef PROTOBUF_MESSAGE_GLOBALS
      &file_reflection_data[0],
#else   // !PROTOBUF_MESSAGE_GLOBALS
      &::_pbi::kDescriptorMethods,
      &descriptor_table_agent_2fv1_2fdisconnect_2eproto,
      nullptr,  // tracker
#endif  // PROTOBUF_MESSAGE_GLOBALS
  };
}
struct DisconnectReqGlobalsTypeInternal : ::_pbi::MessageGlobalsBase {
  constexpr DisconnectReqGlobalsTypeInternal()
      :
#ifndef PROTOBUF_MESSAGE_GLOBALS
        _default(::_pbi::ConstantInitialized{},
                 DisconnectReq_class_data_.base())
#else   // !PROTOBUF_MESSAGE_GLOBALS
        MessageGlobalsBase(DisconnectReq::InternalGenerateClassData_(
            _default, &DisconnectReq_globals_._table.header)),
        _default(::_pbi::ConstantInitialized{}, GetClassData()),
        _table(::_pbi::PrivateAccess::GenerateParseTable<DisconnectReq>(
            GetClassData()))
#endif  // PROTOBUF_MESSAGE_GLOBALS
  {
  }
  ~DisconnectReqGlobalsTypeInternal() {}
  union {
    alignas(::_pbi::kMaxMessageAlignment) DisconnectReq _default;
  };
#ifdef PROTOBUF_MESSAGE_GLOBALS
  decltype(::_pbi::PrivateAccess::GenerateParseTable<DisconnectReq>(
      ::std::declval<const ::_pbi::ClassData*>())) _table;
#endif
};
#ifdef PROTOBUF_MESSAGE_GLOBALS
static_assert(PROTOBUF_FIELD_OFFSET(DisconnectReqGlobalsTypeInternal, _default) ==
              ::_pbi::MessageGlobalsBase::OffsetToDefault());
#endif  // PROTOBUF_MESSAGE_GLOBALS

PROTOBUF_ATTRIBUTE_NO_DESTROY PROTOBUF_CONSTINIT
    PROTOBUF_ATTRIBUTE_INIT_PRIORITY1 PROTOBUF_MESSAGE_GLOBALS_CONST DisconnectReqGlobalsTypeInternal DisconnectReq_globals_
        PROTOBUF_MESSAGE_GLOBALS_SECTION(.data.rel.ro);
#if defined(PROTOBUF_CUSTOM_VTABLE)
namespace {
const ::_pbi::ClassData* DisconnectReq_get_class_data() {
#ifdef PROTOBUF_MESSAGE_GLOBALS
  return DisconnectReq_globals_.GetClassData();
#else
  return DisconnectReq_class_data_.base();
#endif  // PROTOBUF_MESSAGE_GLOBALS
}
}  // namespace
#endif  // PROTOBUF_CUSTOM_VTABLE
}  // namespace v1
}  // namespace agent
static constexpr const ::_pb::EnumDescriptor* PROTOBUF_NONNULL* PROTOBUF_NULLABLE
    file_level_enum_descriptors_agent_2fv1_2fdisconnect_2eproto = nullptr;
static constexpr const ::_pb::ServiceDescriptor* PROTOBUF_NONNULL* PROTOBUF_NULLABLE
    file_level_service_descriptors_agent_2fv1_2fdisconnect_2eproto = nullptr;
const ::uint32_t
    TableStruct_agent_2fv1_2fdisconnect_2eproto::offsets[] ABSL_ATTRIBUTE_SECTION_VARIABLE(
        protodesc_cold) = {
        0x081, // bitmap
        PROTOBUF_FIELD_OFFSET(::agent::v1::DisconnectReq, _impl_._has_bits_),
        8, // hasbit index offset
        PROTOBUF_FIELD_OFFSET(::agent::v1::DisconnectReq, _impl_.reconnect_after_seconds_),
        PROTOBUF_FIELD_OFFSET(::agent::v1::DisconnectReq, _impl_.permaban_),
        PROTOBUF_FIELD_OFFSET(::agent::v1::DisconnectReq, _impl_.created_at_),
        PROTOBUF_FIELD_OFFSET(::agent::v1::DisconnectReq, _impl_.error_code_),
        PROTOBUF_FIELD_OFFSET(::agent::v1::DisconnectReq, _impl_.error_description_),
        2,
        3,
        1,
        4,
        0,
};

static const ::_pbi::MigrationSchema
    schemas[] ABSL_ATTRIBUTE_SECTION_VARIABLE(protodesc_cold) = {
        {0, sizeof(::agent::v1::DisconnectReq)},
};
static const ::_pbi::MessageGlobalsBase* PROTOBUF_NONNULL const
    file_message_globals[] = {
        &::agent::v1::DisconnectReq_globals_,
};
const char descriptor_table_protodef_agent_2fv1_2fdisconnect_2eproto[] ABSL_ATTRIBUTE_SECTION_VARIABLE(
    protodesc_cold) = {
    "\n\031agent/v1/disconnect.proto\022\010agent.v1\032\037g"
    "oogle/protobuf/timestamp.proto\"\241\001\n\rDisco"
    "nnectReq\022\037\n\027reconnect_after_seconds\030\001 \001("
    "\004\022\020\n\010permaban\030\002 \001(\010\022.\n\ncreated_at\030\003 \001(\0132"
    "\032.google.protobuf.Timestamp\022\022\n\nerror_cod"
    "e\030\004 \001(\r\022\031\n\021error_description\030\005 \001(\tB\020Z\016ag"
    "ent/v1;agentb\006proto3"
};
static const ::_pbi::DescriptorTable* PROTOBUF_NONNULL const
    descriptor_table_agent_2fv1_2fdisconnect_2eproto_deps[1] = {
        &::descriptor_table_google_2fprotobuf_2ftimestamp_2eproto,
};
static ::absl::once_flag descriptor_table_agent_2fv1_2fdisconnect_2eproto_once;
PROTOBUF_CONSTINIT const ::_pbi::DescriptorTable descriptor_table_agent_2fv1_2fdisconnect_2eproto = {
    false,
    false,
    260,
    descriptor_table_protodef_agent_2fv1_2fdisconnect_2eproto,
    "agent/v1/disconnect.proto",
    &descriptor_table_agent_2fv1_2fdisconnect_2eproto_once,
    descriptor_table_agent_2fv1_2fdisconnect_2eproto_deps,
    1,
    1,
    schemas,
    file_message_globals,
    TableStruct_agent_2fv1_2fdisconnect_2eproto::offsets,
    file_level_enum_descriptors_agent_2fv1_2fdisconnect_2eproto,
    file_level_service_descriptors_agent_2fv1_2fdisconnect_2eproto,
};
namespace agent {
namespace v1 {
// ===================================================================

void DisconnectReq::clear_created_at() {
  ::google::protobuf::internal::TSanWrite(&_impl_);
  if (_impl_.created_at_ != nullptr) _impl_.created_at_->Clear();
  ClearHasBit(_impl_._has_bits_[0], 0x00000002U);
}
DisconnectReq::DisconnectReq(::google::protobuf::Arena* PROTOBUF_NULLABLE arena)
#if defined(PROTOBUF_CUSTOM_VTABLE)
    : ::google::protobuf::Message(arena, DisconnectReq_get_class_data()) {
#else   // PROTOBUF_CUSTOM_VTABLE
    : ::google::protobuf::Message(arena) {
#endif  // PROTOBUF_CUSTOM_VTABLE
  SharedCtor(arena);
  // @@protoc_insertion_point(arena_constructor:agent.v1.DisconnectReq)
}
PROTOBUF_NDEBUG_INLINE DisconnectReq::Impl_::Impl_(
    [[maybe_unused]] ::google::protobuf::internal::InternalVisibility visibility,
    [[maybe_unused]] ::google::protobuf::Arena* PROTOBUF_NULLABLE arena, const Impl_& from,
    [[maybe_unused]] const ::agent::v1::DisconnectReq& from_msg)
      : _has_bits_{from._has_bits_},
        _cached_size_{0},
        error_description_(arena, from.error_description_) {}

DisconnectReq::DisconnectReq(
    ::google::protobuf::Arena* PROTOBUF_NULLABLE arena,
    const DisconnectReq& from)
#if defined(PROTOBUF_CUSTOM_VTABLE)
    : ::google::protobuf::Message(arena, DisconnectReq_get_class_data()) {

#else   // PROTOBUF_CUSTOM_VTABLE
    : ::google::protobuf::Message(arena) {
#endif  // PROTOBUF_CUSTOM_VTABLE
  DisconnectReq* const _this = this;
  (void)_this;
  _internal_metadata_.MergeFrom<::google::protobuf::UnknownFieldSet>(
      from._internal_metadata_);
  new (&_impl_) Impl_(internal_visibility(), arena, from._impl_, from);
  ::uint32_t cached_has_bits = _impl_._has_bits_[0];
  _impl_.created_at_ = (CheckHasBit(cached_has_bits, 0x00000002U))
                ? ::google::protobuf::Message::CopyConstruct(arena, *from._impl_.created_at_)
                : nullptr;
  ::memcpy(reinterpret_cast<char*>(&_impl_) +
               offsetof(Impl_, reconnect_after_seconds_),
           reinterpret_cast<const char*>(&from._impl_) +
               offsetof(Impl_, reconnect_after_seconds_),
           offsetof(Impl_, error_code_) -
               offsetof(Impl_, reconnect_after_seconds_) +
               sizeof(Impl_::error_code_));

  // @@protoc_insertion_point(copy_constructor:agent.v1.DisconnectReq)
}
PROTOBUF_NDEBUG_INLINE DisconnectReq::Impl_::Impl_(
    [[maybe_unused]] ::google::protobuf::internal::InternalVisibility visibility,
    [[maybe_unused]] ::google::protobuf::Arena* PROTOBUF_NULLABLE arena)
      : _cached_size_{0},
        error_description_(arena) {}

inline void DisconnectReq::SharedCtor(::_pb::Arena* PROTOBUF_NULLABLE arena) {
  new (&_impl_) Impl_(internal_visibility(), arena);
  ::memset(reinterpret_cast<char*>(&_impl_) +
               offsetof(Impl_, created_at_),
           0,
           offsetof(Impl_, error_code_) -
               offsetof(Impl_, created_at_) +
               sizeof(Impl_::error_code_));
}
DisconnectReq::~DisconnectReq() {
  // @@protoc_insertion_point(destructor:agent.v1.DisconnectReq)
  SharedDtor(*this);
}
inline void DisconnectReq::SharedDtor(MessageLite& self) {
  DisconnectReq& this_ = static_cast<DisconnectReq&>(self);
  if constexpr (::_pbi::DebugHardenCheckHasBitConsistency()) {
    this_.CheckHasBitConsistency();
  }
  this_._internal_metadata_.Delete<::google::protobuf::UnknownFieldSet>();
  ABSL_DCHECK(this_.GetArena() == nullptr);
  this_._impl_.error_description_.Destroy();
  delete this_._impl_.created_at_;
  this_._impl_.~Impl_();
}

#ifndef PROTOBUF_MESSAGE_GLOBALS
PROTOBUF_CONSTINIT PROTOBUF_ATTRIBUTE_INIT_PRIORITY1 const
    ::google::protobuf::internal::ClassDataFull DisconnectReq_class_data_ =
        DisconnectReq::InternalGenerateClassData_(DisconnectReq_globals_._default);

PROTOBUF_ATTRIBUTE_WEAK const ::google::protobuf::internal::ClassData* PROTOBUF_NONNULL
DisconnectReq::GetClassData() const {
  ::google::protobuf::internal::PrefetchToLocalCache(&DisconnectReq_class_data_);
  ::google::protobuf::internal::PrefetchToLocalCache(DisconnectReq_class_data_.tc_table);
  return DisconnectReq_class_data_.base();
}
#else
PROTOBUF_ATTRIBUTE_WEAK const ::google::protobuf::internal::ClassData* PROTOBUF_NONNULL
DisconnectReq::GetClassData() const {
  ::google::protobuf::internal::PrefetchToLocalCache(&DisconnectReq_globals_);
  ::google::protobuf::internal::PrefetchToLocalCache(
      ::google::protobuf::internal::MessageGlobalsBase::ToParseTableBase(&DisconnectReq_globals_));
  return DisconnectReq_globals_.GetClassData();
}
#endif  // !PROTOBUF_MESSAGE_GLOBALS
#ifndef PROTOBUF_MESSAGE_GLOBALS
PROTOBUF_CONSTINIT
PROTOBUF_ATTRIBUTE_INIT_PRIORITY1 const DisconnectReq::ParseTableT_
    DisconnectReq::_table_ =
        DisconnectReq::InternalGenerateParseTable_(DisconnectReq_class_data_.base());
#endif  // !PROTOBUF_MESSAGE_GLOBALS
PROTOBUF_NOINLINE void DisconnectReq::Clear() {
// @@protoc_insertion_point(message_clear_start:agent.v1.DisconnectReq)
  ::google::protobuf::internal::TSanWrite(&_impl_);
  ::uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void) cached_has_bits;

  cached_has_bits = _impl_._has_bits_[0];
  if (BatchCheckHasBit(cached_has_bits, 0x00000003U)) {
    if (CheckHasBit(cached_has_bits, 0x00000001U)) {
      _impl_.error_description_.ClearNonDefaultToEmpty();
    }
    if (CheckHasBit(cached_has_bits, 0x00000002U)) {
      ABSL_DCHECK(_impl_.created_at_ != nullptr);
      _impl_.created_at_->Clear();
    }
  }
  if (BatchCheckHasBit(cached_has_bits, 0x0000001cU)) {
    ::memset(&_impl_.reconnect_after_seconds_, 0, static_cast<::size_t>(
        reinterpret_cast<char*>(&_impl_.error_code_) -
        reinterpret_cast<char*>(&_impl_.reconnect_after_seconds_)) + sizeof(_impl_.error_code_));
  }
  _impl_._has_bits_.Clear();
  _internal_metadata_.Clear<::google::protobuf::UnknownFieldSet>();
}

#if defined(PROTOBUF_CUSTOM_VTABLE)
::uint8_t* PROTOBUF_NONNULL DisconnectReq::_InternalSerialize(
    const ::google::protobuf::MessageLite& base, ::uint8_t* PROTOBUF_NONNULL target,
    ::google::protobuf::io::EpsCopyOutputStream* PROTOBUF_NONNULL stream) {
  const DisconnectReq& this_ = static_cast<const DisconnectReq&>(base);
#else   // PROTOBUF_CUSTOM_VTABLE
::uint8_t* PROTOBUF_NONNULL DisconnectReq::_InternalSerialize(
    ::uint8_t* PROTOBUF_NONNULL target,
    ::google::protobuf::io::EpsCopyOutputStream* PROTOBUF_NONNULL stream) const {
  const DisconnectReq& this_ = *this;
#endif  // PROTOBUF_CUSTOM_VTABLE
  if constexpr (::_pbi::DebugHardenCheckHasBitConsistency()) {
    this_.CheckHasBitConsistency();
  }
  // @@protoc_insertion_point(serialize_to_array_start:agent.v1.DisconnectReq)
  ::uint32_t cached_has_bits = 0;
  (void)cached_has_bits;

  cached_has_bits = this_._impl_._has_bits_[0];
  // uint64 reconnect_after_seconds = 1;
  if (CheckHasBit(cached_has_bits, 0x00000004U)) {
    if (this_._internal_reconnect_after_seconds() != 0) {
      target = stream->EnsureSpace(target);
      target = ::_pbi::WireFormatLite::WriteUInt64ToArray(
          1, this_._internal_reconnect_after_seconds(), target);
    }
  }

  // bool permaban = 2;
  if (CheckHasBit(cached_has_bits, 0x00000008U)) {
    if (this_._internal_permaban() != 0) {
      target = stream->EnsureSpace(target);
      target = ::_pbi::WireFormatLite::WriteBoolToArray(
          2, this_._internal_permaban(), target);
    }
  }

  // .google.protobuf.Timestamp created_at = 3;
  if (CheckHasBit(cached_has_bits, 0x00000002U)) {
    target = ::google::protobuf::internal::WireFormatLite::InternalWriteMessage(
        3, *this_._impl_.created_at_, this_._impl_.created_at_->GetCachedSize(), target,
        stream);
  }

  // uint32 error_code = 4;
  if (CheckHasBit(cached_has_bits, 0x00000010U)) {
    if (this_._internal_error_code() != 0) {
      target = stream->EnsureSpace(target);
      target = ::_pbi::WireFormatLite::WriteUInt32ToArray(
          4, this_._internal_error_code(), target);
    }
  }

  // string error_description = 5;
  if (CheckHasBit(cached_has_bits, 0x00000001U)) {
    if (!this_._internal_error_description().empty()) {
      const ::std::string& _s = this_._internal_error_description();
      ::google::protobuf::internal::WireFormatLite::VerifyUtf8String(
          _s.data(), static_cast<int>(_s.length()), ::google::protobuf::internal::WireFormatLite::SERIALIZE, "agent.v1.DisconnectReq.error_description");
      target = stream->WriteStringMaybeAliased(5, _s, target);
    }
  }

  if (ABSL_PREDICT_FALSE(this_._internal_metadata_.have_unknown_fields())) {
    target =
        ::_pbi::WireFormat::InternalSerializeUnknownFieldsToArray(
            this_._internal_metadata_.unknown_fields<::google::protobuf::UnknownFieldSet>(::google::protobuf::UnknownFieldSet::default_instance), target, stream);
  }
  // @@protoc_insertion_point(serialize_to_array_end:agent.v1.DisconnectReq)
  return target;
}

#if defined(PROTOBUF_CUSTOM_VTABLE)
::size_t DisconnectReq::ByteSizeLong(const MessageLite& base) {
  const DisconnectReq& this_ = static_cast<const DisconnectReq&>(base);
#else   // PROTOBUF_CUSTOM_VTABLE
::size_t DisconnectReq::ByteSizeLong() const {
  const DisconnectReq& this_ = *this;
#endif  // PROTOBUF_CUSTOM_VTABLE
  // @@protoc_insertion_point(message_byte_size_start:agent.v1.DisconnectReq)
  ::size_t total_size = 0;

  ::uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void)cached_has_bits;

  ::_pbi::Prefetch5LinesFrom7Lines(&this_);
  cached_has_bits = this_._impl_._has_bits_[0];
  if (BatchCheckHasBit(cached_has_bits, 0x0000001fU)) {
    // string error_description = 5;
    if (CheckHasBit(cached_has_bits, 0x00000001U)) {
      if (!this_._internal_error_description().empty()) {
        total_size += 1 + ::google::protobuf::internal::WireFormatLite::StringSize(
                                        this_._internal_error_description());
      }
    }
    // .google.protobuf.Timestamp created_at = 3;
    if (CheckHasBit(cached_has_bits, 0x00000002U)) {
      total_size += 1 +
                    ::google::protobuf::internal::WireFormatLite::MessageSize(*this_._impl_.created_at_);
    }
    // uint64 reconnect_after_seconds = 1;
    if (CheckHasBit(cached_has_bits, 0x00000004U)) {
      if (this_._internal_reconnect_after_seconds() != 0) {
        total_size += ::_pbi::WireFormatLite::UInt64SizePlusOne(
            this_._internal_reconnect_after_seconds());
      }
    }
    // bool permaban = 2;
    if (CheckHasBit(cached_has_bits, 0x00000008U)) {
      if (this_._internal_permaban() != 0) {
        total_size += 2;
      }
    }
    // uint32 error_code = 4;
    if (CheckHasBit(cached_has_bits, 0x00000010U)) {
      if (this_._internal_error_code() != 0) {
        total_size += ::_pbi::WireFormatLite::UInt32SizePlusOne(
            this_._internal_error_code());
      }
    }
  }
  return this_.MaybeComputeUnknownFieldsSize(total_size,
                                             &this_._impl_._cached_size_);
}

void DisconnectReq::MergeImpl(::google::protobuf::MessageLite& to_msg,
                      const ::google::protobuf::MessageLite& from_msg) {
   auto* const _this = static_cast<DisconnectReq*>(&to_msg);
  auto& from = static_cast<const DisconnectReq&>(from_msg);
  if constexpr (::_pbi::DebugHardenCheckHasBitConsistency()) {
    from.CheckHasBitConsistency();
  }
  ::google::protobuf::Arena* arena = _this->GetArena();
  // @@protoc_insertion_point(class_specific_merge_from_start:agent.v1.DisconnectReq)
  ABSL_DCHECK_NE(&from, _this);
  ::uint32_t cached_has_bits = 0;
  (void)cached_has_bits;

  cached_has_bits = from._impl_._has_bits_[0];
  if (BatchCheckHasBit(cached_has_bits, 0x0000001fU)) {
    if (CheckHasBit(cached_has_bits, 0x00000001U)) {
      if (!from._internal_error_description().empty()) {
        _this->_internal_set_error_description(from._internal_error_description());
      } else {
        if (_this->_impl_.error_description_.IsDefault()) {
          _this->_internal_set_error_description("");
        }
      }
    }
    if (CheckHasBit(cached_has_bits, 0x00000002U)) {
      ABSL_DCHECK(from._impl_.created_at_ != nullptr);
      if (_this->_impl_.created_at_ == nullptr) {
        _this->_impl_.created_at_ = ::google::protobuf::Message::CopyConstruct(arena, *from._impl_.created_at_);
      } else {
        _this->_impl_.created_at_->MergeFrom(*from._impl_.created_at_);
      }
    }
    if (CheckHasBit(cached_has_bits, 0x00000004U)) {
      if (from._internal_reconnect_after_seconds() != 0) {
        _this->_impl_.reconnect_after_seconds_ = from._impl_.reconnect_after_seconds_;
      }
    }
    if (CheckHasBit(cached_has_bits, 0x00000008U)) {
      if (from._internal_permaban() != 0) {
        _this->_impl_.permaban_ = from._impl_.permaban_;
      }
    }
    if (CheckHasBit(cached_has_bits, 0x00000010U)) {
      if (from._internal_error_code() != 0) {
        _this->_impl_.error_code_ = from._impl_.error_code_;
      }
    }
  }
  _this->_impl_._has_bits_[0] |= cached_has_bits;
  _this->_internal_metadata_.MergeFrom<::google::protobuf::UnknownFieldSet>(
      from._internal_metadata_);
}

void DisconnectReq::CopyFrom(const DisconnectReq& from) {
  // @@protoc_insertion_point(class_specific_copy_from_start:agent.v1.DisconnectReq)
  if (&from == this) return;
  Clear();
  MergeFrom(from);
}


void DisconnectReq::InternalSwap(DisconnectReq* PROTOBUF_RESTRICT PROTOBUF_NONNULL other) {
  using ::std::swap;
  auto* arena = GetArena();
  ABSL_DCHECK_EQ(arena, other->GetArena());
  _internal_metadata_.InternalSwap(&other->_internal_metadata_);
  swap(_impl_._has_bits_[0], other->_impl_._has_bits_[0]);
  ::_pbi::ArenaStringPtr::InternalSwap(&_impl_.error_description_, &other->_impl_.error_description_, arena);
  ::google::protobuf::internal::memswap<
      PROTOBUF_FIELD_OFFSET(DisconnectReq, _impl_.error_code_)
      + sizeof(DisconnectReq::_impl_.error_code_)
      - PROTOBUF_FIELD_OFFSET(DisconnectReq, _impl_.created_at_)>(
          reinterpret_cast<char*>(&_impl_.created_at_),
          reinterpret_cast<char*>(&other->_impl_.created_at_));
}

::google::protobuf::Metadata DisconnectReq::GetMetadata() const {
  return ::google::protobuf::Message::GetMetadataImpl(GetClassData()->full());
}
// @@protoc_insertion_point(namespace_scope)
}  // namespace v1
}  // namespace agent
namespace google {
namespace protobuf {
}  // namespace protobuf
}  // namespace google
// @@protoc_insertion_point(global_scope)
PROTOBUF_ATTRIBUTE_INIT_PRIORITY2 static ::std::false_type
    _static_init2_ [[maybe_unused]] =
        (::_pbi::AddDescriptors(&descriptor_table_agent_2fv1_2fdisconnect_2eproto),
         ::std::false_type{});
#include "google/protobuf/port_undef.inc"
