// Generated by the protocol buffer compiler.  DO NOT EDIT!
// NO CHECKED-IN PROTOBUF GENCODE
// source: context/v1/stream.proto
// Protobuf C++ Version: 7.35.1

#include "context/v1/stream.pb.h"

#include <algorithm>
#include <type_traits>
#include "google/protobuf/io/coded_stream.h"
#include "google/protobuf/generated_message_tctable_impl.h"
#include "google/protobuf/internal_visibility.h"
#include "google/protobuf/extension_set.h"
#include "google/protobuf/generated_message_util.h"
#include "google/protobuf/wire_format_lite.h"
#include "google/protobuf/descriptor.h"
#include "google/protobuf/generated_message_reflection.h"
#include "google/protobuf/reflection_ops.h"
#include "google/protobuf/wire_format.h"
// @@protoc_insertion_point(includes)

// Must be included last.
#include "google/protobuf/port_def.inc"
PROTOBUF_PRAGMA_INIT_SEG
namespace _pb = ::google::protobuf;
namespace _pbi = ::google::protobuf::internal;
namespace _fl = ::google::protobuf::internal::field_layout;
#ifdef PROTOBUF_MESSAGE_GLOBALS
namespace {
PROTOBUF_CONSTINIT ::google::protobuf::internal::ReflectionData
    file_reflection_data[] = {
        // ::context::v1::ContextsCheckpoint
        {&::_pbi::kDescriptorMethods, &::descriptor_table_context_2fv1_2fstream_2eproto, /* tracker*/ nullptr,},
        // ::context::v1::StopStreamingContexts
        {&::_pbi::kDescriptorMethods, &::descriptor_table_context_2fv1_2fstream_2eproto, /* tracker*/ nullptr,},
};
}  // namespace
#endif
namespace context {
namespace v1 {
class StopStreamingContexts::_Internal {
 public:
  using HasBits = decltype(::std::declval<StopStreamingContexts>()._impl_._has_bits_);
  static constexpr ::int32_t kHasBitsOffset =
      8 * PROTOBUF_FIELD_OFFSET(StopStreamingContexts, _impl_._has_bits_);
};

constexpr StopStreamingContexts::ParseTableT_ StopStreamingContexts::InternalGenerateParseTable_(const ::_pbi::ClassData* class_data) {
  return ParseTableT_{
    {
      PROTOBUF_FIELD_OFFSET(StopStreamingContexts, _impl_._has_bits_),
      0, // no _extensions_
      3, 24,  // max_field_number, fast_idx_mask
      offsetof(ParseTableT_, field_lookup_table),
      4294967288,  // skipmap
      offsetof(ParseTableT_, field_entries),
      3,  // num_field_entries
      0,  // num_aux_entries
      offsetof(ParseTableT_, field_names),  // no aux_entries
      class_data,
      nullptr,  // post_loop_handler
      ::_pbi::TcParser::GenericFallback,  // fallback
      #ifdef PROTOBUF_PREFETCH_PARSE_TABLE
      ::_pbi::TcParser::GetTable<::context::v1::StopStreamingContexts>(),  // to_prefetch
      #endif  // PROTOBUF_PREFETCH_PARSE_TABLE
    }, {{
      {::_pbi::TcParser::MiniParse, {}},
      // string claim_id = 1;
      {::_pbi::TcParser::FastUS1,
       {10, 0, 0,
        PROTOBUF_FIELD_OFFSET(StopStreamingContexts, _impl_.claim_id_)}},
      // string node_id = 2;
      {::_pbi::TcParser::FastUS1,
       {18, 1, 0,
        PROTOBUF_FIELD_OFFSET(StopStreamingContexts, _impl_.node_id_)}},
      // .context.v1.StopStreamingContextsReason reason = 3;
      {::_pbi::TcParser::SingularVarintNoZag1<::uint32_t, offsetof(StopStreamingContexts, _impl_.reason_), 2>(),
       {24, 2, 0,
        PROTOBUF_FIELD_OFFSET(StopStreamingContexts, _impl_.reason_)}},
    }}, {{
      65535, 65535
    }}, {{
      // string claim_id = 1;
      {PROTOBUF_FIELD_OFFSET(StopStreamingContexts, _impl_.claim_id_), _Internal::kHasBitsOffset + 0, 0, (0 | ::_fl::kFcOptional | ::_fl::kUtf8String | ::_fl::kRepAString)},
      // string node_id = 2;
      {PROTOBUF_FIELD_OFFSET(StopStreamingContexts, _impl_.node_id_), _Internal::kHasBitsOffset + 1, 0, (0 | ::_fl::kFcOptional | ::_fl::kUtf8String | ::_fl::kRepAString)},
      // .context.v1.StopStreamingContextsReason reason = 3;
      {PROTOBUF_FIELD_OFFSET(StopStreamingContexts, _impl_.reason_), _Internal::kHasBitsOffset + 2, 0, (0 | ::_fl::kFcOptional | ::_fl::kOpenEnum)},
    }},
    // no aux_entries
    {{
      "\40\10\7\0\0\0\0\0"
      "context.v1.StopStreamingContexts"
      "claim_id"
      "node_id"
    }},
  };
}


inline constexpr StopStreamingContexts::Impl_::Impl_(
    [[maybe_unused]] ::google::protobuf::internal::InternalVisibility visibility,
    ::_pbi::ConstantInitialized) noexcept
      : _cached_size_{0},
        claim_id_(
            &::google::protobuf::internal::fixed_address_empty_string,
            ::_pbi::ConstantInitialized()),
        node_id_(
            &::google::protobuf::internal::fixed_address_empty_string,
            ::_pbi::ConstantInitialized()),
        reason_{static_cast< ::context::v1::StopStreamingContextsReason >(0)} {}

template <typename>
constexpr StopStreamingContexts::StopStreamingContexts(::_pbi::ConstantInitialized,
                       const ::_pbi::ClassData* PROTOBUF_NONNULL class_data)
    : ::google::protobuf::Message(
#if defined(PROTOBUF_CUSTOM_VTABLE)
          class_data
#endif  // PROTOBUF_CUSTOM_VTABLE
          ),
      _impl_(internal_visibility(), ::_pbi::ConstantInitialized()) {
}
inline void* PROTOBUF_NONNULL StopStreamingContexts::PlacementNew_(
    const void* PROTOBUF_NONNULL, void* PROTOBUF_NONNULL mem,
    ::google::protobuf::Arena* PROTOBUF_NULLABLE arena) {
  return ::new (mem) StopStreamingContexts(arena);
}
constexpr auto StopStreamingContexts::InternalNewImpl_() {
  return ::google::protobuf::internal::MessageCreator::CopyInit(sizeof(StopStreamingContexts), alignof(StopStreamingContexts));
}
constexpr auto StopStreamingContexts::InternalGenerateClassData_(
    const MessageLite& prototype,
    const ::google::protobuf::internal::TcParseTableBase* tc_table) {
  return ::google::protobuf::internal::ClassDataFull{
      ::google::protobuf::internal::ClassData{
          &prototype,
#ifndef PROTOBUF_MESSAGE_GLOBALS
          &_table_.header,
#else
          tc_table,
#endif
          nullptr,  // IsInitialized
          &StopStreamingContexts::MergeImpl,
          ::google::protobuf::Message::GetNewImpl<StopStreamingContexts>(),
#if defined(PROTOBUF_CUSTOM_VTABLE)
          &StopStreamingContexts::SharedDtor,
          ::google::protobuf::Message::GetClearImpl<StopStreamingContexts>(), &StopStreamingContexts::ByteSizeLong,
              &StopStreamingContexts::_InternalSerialize,
#endif  // PROTOBUF_CUSTOM_VTABLE
          PROTOBUF_FIELD_OFFSET(StopStreamingContexts, _impl_._cached_size_),
          false,
      },
#ifdef PROTOBUF_MESSAGE_GLOBALS
      &file_reflection_data[1],
#else   // !PROTOBUF_MESSAGE_GLOBALS
      &::_pbi::kDescriptorMethods,
      &descriptor_table_context_2fv1_2fstream_2eproto,
      nullptr,  // tracker
#endif  // PROTOBUF_MESSAGE_GLOBALS
  };
}
struct StopStreamingContextsGlobalsTypeInternal : ::_pbi::MessageGlobalsBase {
  constexpr StopStreamingContextsGlobalsTypeInternal()
      :
#ifndef PROTOBUF_MESSAGE_GLOBALS
        _default(::_pbi::ConstantInitialized{},
                 StopStreamingContexts_class_data_.base())
#else   // !PROTOBUF_MESSAGE_GLOBALS
        MessageGlobalsBase(StopStreamingContexts::InternalGenerateClassData_(
            _default, &StopStreamingContexts_globals_._table.header)),
        _default(::_pbi::ConstantInitialized{}, GetClassData()),
        _table(::_pbi::PrivateAccess::GenerateParseTable<StopStreamingContexts>(
            GetClassData()))
#endif  // PROTOBUF_MESSAGE_GLOBALS
  {
  }
  ~StopStreamingContextsGlobalsTypeInternal() {}
  union {
    alignas(::_pbi::kMaxMessageAlignment) StopStreamingContexts _default;
  };
#ifdef PROTOBUF_MESSAGE_GLOBALS
  decltype(::_pbi::PrivateAccess::GenerateParseTable<StopStreamingContexts>(
      ::std::declval<const ::_pbi::ClassData*>())) _table;
#endif
};
#ifdef PROTOBUF_MESSAGE_GLOBALS
static_assert(PROTOBUF_FIELD_OFFSET(StopStreamingContextsGlobalsTypeInternal, _default) ==
              ::_pbi::MessageGlobalsBase::OffsetToDefault());
#endif  // PROTOBUF_MESSAGE_GLOBALS

PROTOBUF_ATTRIBUTE_NO_DESTROY PROTOBUF_CONSTINIT
    PROTOBUF_ATTRIBUTE_INIT_PRIORITY1 PROTOBUF_MESSAGE_GLOBALS_CONST StopStreamingContextsGlobalsTypeInternal StopStreamingContexts_globals_
        PROTOBUF_MESSAGE_GLOBALS_SECTION(.data.rel.ro);
#if defined(PROTOBUF_CUSTOM_VTABLE)
namespace {
const ::_pbi::ClassData* StopStreamingContexts_get_class_data() {
#ifdef PROTOBUF_MESSAGE_GLOBALS
  return StopStreamingContexts_globals_.GetClassData();
#else
  return StopStreamingContexts_class_data_.base();
#endif  // PROTOBUF_MESSAGE_GLOBALS
}
}  // namespace
#endif  // PROTOBUF_CUSTOM_VTABLE
class ContextsCheckpoint::_Internal {
 public:
  using HasBits = decltype(::std::declval<ContextsCheckpoint>()._impl_._has_bits_);
  static constexpr ::int32_t kHasBitsOffset =
      8 * PROTOBUF_FIELD_OFFSET(ContextsCheckpoint, _impl_._has_bits_);
};

constexpr ContextsCheckpoint::ParseTableT_ ContextsCheckpoint::InternalGenerateParseTable_(const ::_pbi::ClassData* class_data) {
  return ParseTableT_{
    {
      PROTOBUF_FIELD_OFFSET(ContextsCheckpoint, _impl_._has_bits_),
      0, // no _extensions_
      3, 24,  // max_field_number, fast_idx_mask
      offsetof(ParseTableT_, field_lookup_table),
      4294967288,  // skipmap
      offsetof(ParseTableT_, field_entries),
      3,  // num_field_entries
      0,  // num_aux_entries
      offsetof(ParseTableT_, field_names),  // no aux_entries
      class_data,
      nullptr,  // post_loop_handler
      ::_pbi::TcParser::GenericFallback,  // fallback
      #ifdef PROTOBUF_PREFETCH_PARSE_TABLE
      ::_pbi::TcParser::GetTable<::context::v1::ContextsCheckpoint>(),  // to_prefetch
      #endif  // PROTOBUF_PREFETCH_PARSE_TABLE
    }, {{
      {::_pbi::TcParser::MiniParse, {}},
      // string claim_id = 1;
      {::_pbi::TcParser::FastUS1,
       {10, 0, 0,
        PROTOBUF_FIELD_OFFSET(ContextsCheckpoint, _impl_.claim_id_)}},
      // string node_id = 2;
      {::_pbi::TcParser::FastUS1,
       {18, 1, 0,
        PROTOBUF_FIELD_OFFSET(ContextsCheckpoint, _impl_.node_id_)}},
      // uint64 version_hash = 3;
      {::_pbi::TcParser::SingularVarintNoZag1<::uint64_t, offsetof(ContextsCheckpoint, _impl_.version_hash_), 2>(),
       {24, 2, 0,
        PROTOBUF_FIELD_OFFSET(ContextsCheckpoint, _impl_.version_hash_)}},
    }}, {{
      65535, 65535
    }}, {{
      // string claim_id = 1;
      {PROTOBUF_FIELD_OFFSET(ContextsCheckpoint, _impl_.claim_id_), _Internal::kHasBitsOffset + 0, 0, (0 | ::_fl::kFcOptional | ::_fl::kUtf8String | ::_fl::kRepAString)},
      // string node_id = 2;
      {PROTOBUF_FIELD_OFFSET(ContextsCheckpoint, _impl_.node_id_), _Internal::kHasBitsOffset + 1, 0, (0 | ::_fl::kFcOptional | ::_fl::kUtf8String | ::_fl::kRepAString)},
      // uint64 version_hash = 3;
      {PROTOBUF_FIELD_OFFSET(ContextsCheckpoint, _impl_.version_hash_), _Internal::kHasBitsOffset + 2, 0, (0 | ::_fl::kFcOptional | ::_fl::kUInt64)},
    }},
    // no aux_entries
    {{
      "\35\10\7\0\0\0\0\0"
      "context.v1.ContextsCheckpoint"
      "claim_id"
      "node_id"
    }},
  };
}


inline constexpr ContextsCheckpoint::Impl_::Impl_(
    [[maybe_unused]] ::google::protobuf::internal::InternalVisibility visibility,
    ::_pbi::ConstantInitialized) noexcept
      : _cached_size_{0},
        claim_id_(
            &::google::protobuf::internal::fixed_address_empty_string,
            ::_pbi::ConstantInitialized()),
        node_id_(
            &::google::protobuf::internal::fixed_address_empty_string,
            ::_pbi::ConstantInitialized()),
        version_hash_{::uint64_t{0u}} {}

template <typename>
constexpr ContextsCheckpoint::ContextsCheckpoint(::_pbi::ConstantInitialized,
                       const ::_pbi::ClassData* PROTOBUF_NONNULL class_data)
    : ::google::protobuf::Message(
#if defined(PROTOBUF_CUSTOM_VTABLE)
          class_data
#endif  // PROTOBUF_CUSTOM_VTABLE
          ),
      _impl_(internal_visibility(), ::_pbi::ConstantInitialized()) {
}
inline void* PROTOBUF_NONNULL ContextsCheckpoint::PlacementNew_(
    const void* PROTOBUF_NONNULL, void* PROTOBUF_NONNULL mem,
    ::google::protobuf::Arena* PROTOBUF_NULLABLE arena) {
  return ::new (mem) ContextsCheckpoint(arena);
}
constexpr auto ContextsCheckpoint::InternalNewImpl_() {
  return ::google::protobuf::internal::MessageCreator::CopyInit(sizeof(ContextsCheckpoint), alignof(ContextsCheckpoint));
}
constexpr auto ContextsCheckpoint::InternalGenerateClassData_(
    const MessageLite& prototype,
    const ::google::protobuf::internal::TcParseTableBase* tc_table) {
  return ::google::protobuf::internal::ClassDataFull{
      ::google::protobuf::internal::ClassData{
          &prototype,
#ifndef PROTOBUF_MESSAGE_GLOBALS
          &_table_.header,
#else
          tc_table,
#endif
          nullptr,  // IsInitialized
          &ContextsCheckpoint::MergeImpl,
          ::google::protobuf::Message::GetNewImpl<ContextsCheckpoint>(),
#if defined(PROTOBUF_CUSTOM_VTABLE)
          &ContextsCheckpoint::SharedDtor,
          ::google::protobuf::Message::GetClearImpl<ContextsCheckpoint>(), &ContextsCheckpoint::ByteSizeLong,
              &ContextsCheckpoint::_InternalSerialize,
#endif  // PROTOBUF_CUSTOM_VTABLE
          PROTOBUF_FIELD_OFFSET(ContextsCheckpoint, _impl_._cached_size_),
          false,
      },
#ifdef PROTOBUF_MESSAGE_GLOBALS
      &file_reflection_data[0],
#else   // !PROTOBUF_MESSAGE_GLOBALS
      &::_pbi::kDescriptorMethods,
      &descriptor_table_context_2fv1_2fstream_2eproto,
      nullptr,  // tracker
#endif  // PROTOBUF_MESSAGE_GLOBALS
  };
}
struct ContextsCheckpointGlobalsTypeInternal : ::_pbi::MessageGlobalsBase {
  constexpr ContextsCheckpointGlobalsTypeInternal()
      :
#ifndef PROTOBUF_MESSAGE_GLOBALS
        _default(::_pbi::ConstantInitialized{},
                 ContextsCheckpoint_class_data_.base())
#else   // !PROTOBUF_MESSAGE_GLOBALS
        MessageGlobalsBase(ContextsCheckpoint::InternalGenerateClassData_(
            _default, &ContextsCheckpoint_globals_._table.header)),
        _default(::_pbi::ConstantInitialized{}, GetClassData()),
        _table(::_pbi::PrivateAccess::GenerateParseTable<ContextsCheckpoint>(
            GetClassData()))
#endif  // PROTOBUF_MESSAGE_GLOBALS
  {
  }
  ~ContextsCheckpointGlobalsTypeInternal() {}
  union {
    alignas(::_pbi::kMaxMessageAlignment) ContextsCheckpoint _default;
  };
#ifdef PROTOBUF_MESSAGE_GLOBALS
  decltype(::_pbi::PrivateAccess::GenerateParseTable<ContextsCheckpoint>(
      ::std::declval<const ::_pbi::ClassData*>())) _table;
#endif
};
#ifdef PROTOBUF_MESSAGE_GLOBALS
static_assert(PROTOBUF_FIELD_OFFSET(ContextsCheckpointGlobalsTypeInternal, _default) ==
              ::_pbi::MessageGlobalsBase::OffsetToDefault());
#endif  // PROTOBUF_MESSAGE_GLOBALS

PROTOBUF_ATTRIBUTE_NO_DESTROY PROTOBUF_CONSTINIT
    PROTOBUF_ATTRIBUTE_INIT_PRIORITY1 PROTOBUF_MESSAGE_GLOBALS_CONST ContextsCheckpointGlobalsTypeInternal ContextsCheckpoint_globals_
        PROTOBUF_MESSAGE_GLOBALS_SECTION(.data.rel.ro);
#if defined(PROTOBUF_CUSTOM_VTABLE)
namespace {
const ::_pbi::ClassData* ContextsCheckpoint_get_class_data() {
#ifdef PROTOBUF_MESSAGE_GLOBALS
  return ContextsCheckpoint_globals_.GetClassData();
#else
  return ContextsCheckpoint_class_data_.base();
#endif  // PROTOBUF_MESSAGE_GLOBALS
}
}  // namespace
#endif  // PROTOBUF_CUSTOM_VTABLE
}  // namespace v1
}  // namespace context
static const ::_pb::EnumDescriptor* PROTOBUF_NONNULL
    file_level_enum_descriptors_context_2fv1_2fstream_2eproto[1];
static constexpr const ::_pb::ServiceDescriptor* PROTOBUF_NONNULL* PROTOBUF_NULLABLE
    file_level_service_descriptors_context_2fv1_2fstream_2eproto = nullptr;
const ::uint32_t
    TableStruct_context_2fv1_2fstream_2eproto::offsets[] ABSL_ATTRIBUTE_SECTION_VARIABLE(
        protodesc_cold) = {
        0x081, // bitmap
        PROTOBUF_FIELD_OFFSET(::context::v1::ContextsCheckpoint, _impl_._has_bits_),
        6, // hasbit index offset
        PROTOBUF_FIELD_OFFSET(::context::v1::ContextsCheckpoint, _impl_.claim_id_),
        PROTOBUF_FIELD_OFFSET(::context::v1::ContextsCheckpoint, _impl_.node_id_),
        PROTOBUF_FIELD_OFFSET(::context::v1::ContextsCheckpoint, _impl_.version_hash_),
        0,
        1,
        2,
        0x081, // bitmap
        PROTOBUF_FIELD_OFFSET(::context::v1::StopStreamingContexts, _impl_._has_bits_),
        6, // hasbit index offset
        PROTOBUF_FIELD_OFFSET(::context::v1::StopStreamingContexts, _impl_.claim_id_),
        PROTOBUF_FIELD_OFFSET(::context::v1::StopStreamingContexts, _impl_.node_id_),
        PROTOBUF_FIELD_OFFSET(::context::v1::StopStreamingContexts, _impl_.reason_),
        0,
        1,
        2,
};

static const ::_pbi::MigrationSchema
    schemas[] ABSL_ATTRIBUTE_SECTION_VARIABLE(protodesc_cold) = {
        {0, sizeof(::context::v1::ContextsCheckpoint)},
        {9, sizeof(::context::v1::StopStreamingContexts)},
};
static const ::_pbi::MessageGlobalsBase* PROTOBUF_NONNULL const
    file_message_globals[] = {
        &::context::v1::ContextsCheckpoint_globals_,
        &::context::v1::StopStreamingContexts_globals_,
};
const char descriptor_table_protodef_context_2fv1_2fstream_2eproto[] ABSL_ATTRIBUTE_SECTION_VARIABLE(
    protodesc_cold) = {
    "\n\027context/v1/stream.proto\022\ncontext.v1\"M\n"
    "\022ContextsCheckpoint\022\020\n\010claim_id\030\001 \001(\t\022\017\n"
    "\007node_id\030\002 \001(\t\022\024\n\014version_hash\030\003 \001(\004\"s\n\025"
    "StopStreamingContexts\022\020\n\010claim_id\030\001 \001(\t\022"
    "\017\n\007node_id\030\002 \001(\t\0227\n\006reason\030\003 \001(\0162\'.conte"
    "xt.v1.StopStreamingContextsReason*6\n\033Sto"
    "pStreamingContextsReason\022\027\n\023RATE_LIMIT_E"
    "XCEEDED\020\000B\024Z\022context/v1;contextb\006proto3"
};
static ::absl::once_flag descriptor_table_context_2fv1_2fstream_2eproto_once;
PROTOBUF_CONSTINIT const ::_pbi::DescriptorTable descriptor_table_context_2fv1_2fstream_2eproto = {
    false,
    false,
    319,
    descriptor_table_protodef_context_2fv1_2fstream_2eproto,
    "context/v1/stream.proto",
    &descriptor_table_context_2fv1_2fstream_2eproto_once,
    nullptr,
    0,
    2,
    schemas,
    file_message_globals,
    TableStruct_context_2fv1_2fstream_2eproto::offsets,
    file_level_enum_descriptors_context_2fv1_2fstream_2eproto,
    file_level_service_descriptors_context_2fv1_2fstream_2eproto,
};
namespace context {
namespace v1 {
[[nodiscard]] const ::google::protobuf::EnumDescriptor* PROTOBUF_NONNULL
StopStreamingContextsReason_descriptor() {
  ::google::protobuf::internal::AssignDescriptors(&descriptor_table_context_2fv1_2fstream_2eproto);
  return file_level_enum_descriptors_context_2fv1_2fstream_2eproto[0];
}
PROTOBUF_CONSTINIT const uint32_t StopStreamingContextsReason_internal_data_[] = {
    65536u, 0u, };
// ===================================================================

ContextsCheckpoint::ContextsCheckpoint(::google::protobuf::Arena* PROTOBUF_NULLABLE arena)
#if defined(PROTOBUF_CUSTOM_VTABLE)
    : ::google::protobuf::Message(arena, ContextsCheckpoint_get_class_data()) {
#else   // PROTOBUF_CUSTOM_VTABLE
    : ::google::protobuf::Message(arena) {
#endif  // PROTOBUF_CUSTOM_VTABLE
  SharedCtor(arena);
  // @@protoc_insertion_point(arena_constructor:context.v1.ContextsCheckpoint)
}
PROTOBUF_NDEBUG_INLINE ContextsCheckpoint::Impl_::Impl_(
    [[maybe_unused]] ::google::protobuf::internal::InternalVisibility visibility,
    [[maybe_unused]] ::google::protobuf::Arena* PROTOBUF_NULLABLE arena, const Impl_& from,
    [[maybe_unused]] const ::context::v1::ContextsCheckpoint& from_msg)
      : _has_bits_{from._has_bits_},
        _cached_size_{0},
        claim_id_(arena, from.claim_id_),
        node_id_(arena, from.node_id_) {}

ContextsCheckpoint::ContextsCheckpoint(
    ::google::protobuf::Arena* PROTOBUF_NULLABLE arena,
    const ContextsCheckpoint& from)
#if defined(PROTOBUF_CUSTOM_VTABLE)
    : ::google::protobuf::Message(arena, ContextsCheckpoint_get_class_data()) {

#else   // PROTOBUF_CUSTOM_VTABLE
    : ::google::protobuf::Message(arena) {
#endif  // PROTOBUF_CUSTOM_VTABLE
  ContextsCheckpoint* const _this = this;
  (void)_this;
  _internal_metadata_.MergeFrom<::google::protobuf::UnknownFieldSet>(
      from._internal_metadata_);
  new (&_impl_) Impl_(internal_visibility(), arena, from._impl_, from);
  _impl_.version_hash_ = from._impl_.version_hash_;

  // @@protoc_insertion_point(copy_constructor:context.v1.ContextsCheckpoint)
}
PROTOBUF_NDEBUG_INLINE ContextsCheckpoint::Impl_::Impl_(
    [[maybe_unused]] ::google::protobuf::internal::InternalVisibility visibility,
    [[maybe_unused]] ::google::protobuf::Arena* PROTOBUF_NULLABLE arena)
      : _cached_size_{0},
        claim_id_(arena),
        node_id_(arena) {}

inline void ContextsCheckpoint::SharedCtor(::_pb::Arena* PROTOBUF_NULLABLE arena) {
  new (&_impl_) Impl_(internal_visibility(), arena);
  _impl_.version_hash_ = {};
}
ContextsCheckpoint::~ContextsCheckpoint() {
  // @@protoc_insertion_point(destructor:context.v1.ContextsCheckpoint)
  SharedDtor(*this);
}
inline void ContextsCheckpoint::SharedDtor(MessageLite& self) {
  ContextsCheckpoint& this_ = static_cast<ContextsCheckpoint&>(self);
  if constexpr (::_pbi::DebugHardenCheckHasBitConsistency()) {
    this_.CheckHasBitConsistency();
  }
  this_._internal_metadata_.Delete<::google::protobuf::UnknownFieldSet>();
  ABSL_DCHECK(this_.GetArena() == nullptr);
  this_._impl_.claim_id_.Destroy();
  this_._impl_.node_id_.Destroy();
  this_._impl_.~Impl_();
}

#ifndef PROTOBUF_MESSAGE_GLOBALS
PROTOBUF_CONSTINIT PROTOBUF_ATTRIBUTE_INIT_PRIORITY1 const
    ::google::protobuf::internal::ClassDataFull ContextsCheckpoint_class_data_ =
        ContextsCheckpoint::InternalGenerateClassData_(ContextsCheckpoint_globals_._default);

PROTOBUF_ATTRIBUTE_WEAK const ::google::protobuf::internal::ClassData* PROTOBUF_NONNULL
ContextsCheckpoint::GetClassData() const {
  ::google::protobuf::internal::PrefetchToLocalCache(&ContextsCheckpoint_class_data_);
  ::google::protobuf::internal::PrefetchToLocalCache(ContextsCheckpoint_class_data_.tc_table);
  return ContextsCheckpoint_class_data_.base();
}
#else
PROTOBUF_ATTRIBUTE_WEAK const ::google::protobuf::internal::ClassData* PROTOBUF_NONNULL
ContextsCheckpoint::GetClassData() const {
  ::google::protobuf::internal::PrefetchToLocalCache(&ContextsCheckpoint_globals_);
  ::google::protobuf::internal::PrefetchToLocalCache(
      ::google::protobuf::internal::MessageGlobalsBase::ToParseTableBase(&ContextsCheckpoint_globals_));
  return ContextsCheckpoint_globals_.GetClassData();
}
#endif  // !PROTOBUF_MESSAGE_GLOBALS
#ifndef PROTOBUF_MESSAGE_GLOBALS
PROTOBUF_CONSTINIT
PROTOBUF_ATTRIBUTE_INIT_PRIORITY1 const ContextsCheckpoint::ParseTableT_
    ContextsCheckpoint::_table_ =
        ContextsCheckpoint::InternalGenerateParseTable_(ContextsCheckpoint_class_data_.base());
#endif  // !PROTOBUF_MESSAGE_GLOBALS
PROTOBUF_NOINLINE void ContextsCheckpoint::Clear() {
// @@protoc_insertion_point(message_clear_start:context.v1.ContextsCheckpoint)
  ::google::protobuf::internal::TSanWrite(&_impl_);
  ::uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void) cached_has_bits;

  cached_has_bits = _impl_._has_bits_[0];
  if (BatchCheckHasBit(cached_has_bits, 0x00000003U)) {
    if (CheckHasBit(cached_has_bits, 0x00000001U)) {
      _impl_.claim_id_.ClearNonDefaultToEmpty();
    }
    if (CheckHasBit(cached_has_bits, 0x00000002U)) {
      _impl_.node_id_.ClearNonDefaultToEmpty();
    }
  }
  _impl_.version_hash_ = ::uint64_t{0u};
  _impl_._has_bits_.Clear();
  _internal_metadata_.Clear<::google::protobuf::UnknownFieldSet>();
}

#if defined(PROTOBUF_CUSTOM_VTABLE)
::uint8_t* PROTOBUF_NONNULL ContextsCheckpoint::_InternalSerialize(
    const ::google::protobuf::MessageLite& base, ::uint8_t* PROTOBUF_NONNULL target,
    ::google::protobuf::io::EpsCopyOutputStream* PROTOBUF_NONNULL stream) {
  const ContextsCheckpoint& this_ = static_cast<const ContextsCheckpoint&>(base);
#else   // PROTOBUF_CUSTOM_VTABLE
::uint8_t* PROTOBUF_NONNULL ContextsCheckpoint::_InternalSerialize(
    ::uint8_t* PROTOBUF_NONNULL target,
    ::google::protobuf::io::EpsCopyOutputStream* PROTOBUF_NONNULL stream) const {
  const ContextsCheckpoint& this_ = *this;
#endif  // PROTOBUF_CUSTOM_VTABLE
  if constexpr (::_pbi::DebugHardenCheckHasBitConsistency()) {
    this_.CheckHasBitConsistency();
  }
  // @@protoc_insertion_point(serialize_to_array_start:context.v1.ContextsCheckpoint)
  ::uint32_t cached_has_bits = 0;
  (void)cached_has_bits;

  cached_has_bits = this_._impl_._has_bits_[0];
  // string claim_id = 1;
  if (CheckHasBit(cached_has_bits, 0x00000001U)) {
    if (!this_._internal_claim_id().empty()) {
      const ::std::string& _s = this_._internal_claim_id();
      ::google::protobuf::internal::WireFormatLite::VerifyUtf8String(
          _s.data(), static_cast<int>(_s.length()), ::google::protobuf::internal::WireFormatLite::SERIALIZE, "context.v1.ContextsCheckpoint.claim_id");
      target = stream->WriteStringMaybeAliased(1, _s, target);
    }
  }

  // string node_id = 2;
  if (CheckHasBit(cached_has_bits, 0x00000002U)) {
    if (!this_._internal_node_id().empty()) {
      const ::std::string& _s = this_._internal_node_id();
      ::google::protobuf::internal::WireFormatLite::VerifyUtf8String(
          _s.data(), static_cast<int>(_s.length()), ::google::protobuf::internal::WireFormatLite::SERIALIZE, "context.v1.ContextsCheckpoint.node_id");
      target = stream->WriteStringMaybeAliased(2, _s, target);
    }
  }

  // uint64 version_hash = 3;
  if (CheckHasBit(cached_has_bits, 0x00000004U)) {
    if (this_._internal_version_hash() != 0) {
      target = stream->EnsureSpace(target);
      target = ::_pbi::WireFormatLite::WriteUInt64ToArray(
          3, this_._internal_version_hash(), target);
    }
  }

  if (ABSL_PREDICT_FALSE(this_._internal_metadata_.have_unknown_fields())) {
    target =
        ::_pbi::WireFormat::InternalSerializeUnknownFieldsToArray(
            this_._internal_metadata_.unknown_fields<::google::protobuf::UnknownFieldSet>(::google::protobuf::UnknownFieldSet::default_instance), target, stream);
  }
  // @@protoc_insertion_point(serialize_to_array_end:context.v1.ContextsCheckpoint)
  return target;
}

#if defined(PROTOBUF_CUSTOM_VTABLE)
::size_t ContextsCheckpoint::ByteSizeLong(const MessageLite& base) {
  const ContextsCheckpoint& this_ = static_cast<const ContextsCheckpoint&>(base);
#else   // PROTOBUF_CUSTOM_VTABLE
::size_t ContextsCheckpoint::ByteSizeLong() const {
  const ContextsCheckpoint& this_ = *this;
#endif  // PROTOBUF_CUSTOM_VTABLE
  // @@protoc_insertion_point(message_byte_size_start:context.v1.ContextsCheckpoint)
  ::size_t total_size = 0;

  ::uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void)cached_has_bits;

  ::_pbi::Prefetch5LinesFrom7Lines(&this_);
  cached_has_bits = this_._impl_._has_bits_[0];
  if (BatchCheckHasBit(cached_has_bits, 0x00000007U)) {
    // string claim_id = 1;
    if (CheckHasBit(cached_has_bits, 0x00000001U)) {
      if (!this_._internal_claim_id().empty()) {
        total_size += 1 + ::google::protobuf::internal::WireFormatLite::StringSize(
                                        this_._internal_claim_id());
      }
    }
    // string node_id = 2;
    if (CheckHasBit(cached_has_bits, 0x00000002U)) {
      if (!this_._internal_node_id().empty()) {
        total_size += 1 + ::google::protobuf::internal::WireFormatLite::StringSize(
                                        this_._internal_node_id());
      }
    }
    // uint64 version_hash = 3;
    if (CheckHasBit(cached_has_bits, 0x00000004U)) {
      if (this_._internal_version_hash() != 0) {
        total_size += ::_pbi::WireFormatLite::UInt64SizePlusOne(
            this_._internal_version_hash());
      }
    }
  }
  return this_.MaybeComputeUnknownFieldsSize(total_size,
                                             &this_._impl_._cached_size_);
}

void ContextsCheckpoint::MergeImpl(::google::protobuf::MessageLite& to_msg,
                      const ::google::protobuf::MessageLite& from_msg) {
   auto* const _this = static_cast<ContextsCheckpoint*>(&to_msg);
  auto& from = static_cast<const ContextsCheckpoint&>(from_msg);
  if constexpr (::_pbi::DebugHardenCheckHasBitConsistency()) {
    from.CheckHasBitConsistency();
  }
  // @@protoc_insertion_point(class_specific_merge_from_start:context.v1.ContextsCheckpoint)
  ABSL_DCHECK_NE(&from, _this);
  ::uint32_t cached_has_bits = 0;
  (void)cached_has_bits;

  cached_has_bits = from._impl_._has_bits_[0];
  if (BatchCheckHasBit(cached_has_bits, 0x00000007U)) {
    if (CheckHasBit(cached_has_bits, 0x00000001U)) {
      if (!from._internal_claim_id().empty()) {
        _this->_internal_set_claim_id(from._internal_claim_id());
      } else {
        if (_this->_impl_.claim_id_.IsDefault()) {
          _this->_internal_set_claim_id("");
        }
      }
    }
    if (CheckHasBit(cached_has_bits, 0x00000002U)) {
      if (!from._internal_node_id().empty()) {
        _this->_internal_set_node_id(from._internal_node_id());
      } else {
        if (_this->_impl_.node_id_.IsDefault()) {
          _this->_internal_set_node_id("");
        }
      }
    }
    if (CheckHasBit(cached_has_bits, 0x00000004U)) {
      if (from._internal_version_hash() != 0) {
        _this->_impl_.version_hash_ = from._impl_.version_hash_;
      }
    }
  }
  _this->_impl_._has_bits_[0] |= cached_has_bits;
  _this->_internal_metadata_.MergeFrom<::google::protobuf::UnknownFieldSet>(
      from._internal_metadata_);
}

void ContextsCheckpoint::CopyFrom(const ContextsCheckpoint& from) {
  // @@protoc_insertion_point(class_specific_copy_from_start:context.v1.ContextsCheckpoint)
  if (&from == this) return;
  Clear();
  MergeFrom(from);
}


void ContextsCheckpoint::InternalSwap(ContextsCheckpoint* PROTOBUF_RESTRICT PROTOBUF_NONNULL other) {
  using ::std::swap;
  auto* arena = GetArena();
  ABSL_DCHECK_EQ(arena, other->GetArena());
  _internal_metadata_.InternalSwap(&other->_internal_metadata_);
  swap(_impl_._has_bits_[0], other->_impl_._has_bits_[0]);
  ::_pbi::ArenaStringPtr::InternalSwap(&_impl_.claim_id_, &other->_impl_.claim_id_, arena);
  ::_pbi::ArenaStringPtr::InternalSwap(&_impl_.node_id_, &other->_impl_.node_id_, arena);
  swap(_impl_.version_hash_, other->_impl_.version_hash_);
}

::google::protobuf::Metadata ContextsCheckpoint::GetMetadata() const {
  return ::google::protobuf::Message::GetMetadataImpl(GetClassData()->full());
}
// ===================================================================

StopStreamingContexts::StopStreamingContexts(::google::protobuf::Arena* PROTOBUF_NULLABLE arena)
#if defined(PROTOBUF_CUSTOM_VTABLE)
    : ::google::protobuf::Message(arena, StopStreamingContexts_get_class_data()) {
#else   // PROTOBUF_CUSTOM_VTABLE
    : ::google::protobuf::Message(arena) {
#endif  // PROTOBUF_CUSTOM_VTABLE
  SharedCtor(arena);
  // @@protoc_insertion_point(arena_constructor:context.v1.StopStreamingContexts)
}
PROTOBUF_NDEBUG_INLINE StopStreamingContexts::Impl_::Impl_(
    [[maybe_unused]] ::google::protobuf::internal::InternalVisibility visibility,
    [[maybe_unused]] ::google::protobuf::Arena* PROTOBUF_NULLABLE arena, const Impl_& from,
    [[maybe_unused]] const ::context::v1::StopStreamingContexts& from_msg)
      : _has_bits_{from._has_bits_},
        _cached_size_{0},
        claim_id_(arena, from.claim_id_),
        node_id_(arena, from.node_id_) {}

StopStreamingContexts::StopStreamingContexts(
    ::google::protobuf::Arena* PROTOBUF_NULLABLE arena,
    const StopStreamingContexts& from)
#if defined(PROTOBUF_CUSTOM_VTABLE)
    : ::google::protobuf::Message(arena, StopStreamingContexts_get_class_data()) {

#else   // PROTOBUF_CUSTOM_VTABLE
    : ::google::protobuf::Message(arena) {
#endif  // PROTOBUF_CUSTOM_VTABLE
  StopStreamingContexts* const _this = this;
  (void)_this;
  _internal_metadata_.MergeFrom<::google::protobuf::UnknownFieldSet>(
      from._internal_metadata_);
  new (&_impl_) Impl_(internal_visibility(), arena, from._impl_, from);
  _impl_.reason_ = from._impl_.reason_;

  // @@protoc_insertion_point(copy_constructor:context.v1.StopStreamingContexts)
}
PROTOBUF_NDEBUG_INLINE StopStreamingContexts::Impl_::Impl_(
    [[maybe_unused]] ::google::protobuf::internal::InternalVisibility visibility,
    [[maybe_unused]] ::google::protobuf::Arena* PROTOBUF_NULLABLE arena)
      : _cached_size_{0},
        claim_id_(arena),
        node_id_(arena) {}

inline void StopStreamingContexts::SharedCtor(::_pb::Arena* PROTOBUF_NULLABLE arena) {
  new (&_impl_) Impl_(internal_visibility(), arena);
  _impl_.reason_ = {};
}
StopStreamingContexts::~StopStreamingContexts() {
  // @@protoc_insertion_point(destructor:context.v1.StopStreamingContexts)
  SharedDtor(*this);
}
inline void StopStreamingContexts::SharedDtor(MessageLite& self) {
  StopStreamingContexts& this_ = static_cast<StopStreamingContexts&>(self);
  if constexpr (::_pbi::DebugHardenCheckHasBitConsistency()) {
    this_.CheckHasBitConsistency();
  }
  this_._internal_metadata_.Delete<::google::protobuf::UnknownFieldSet>();
  ABSL_DCHECK(this_.GetArena() == nullptr);
  this_._impl_.claim_id_.Destroy();
  this_._impl_.node_id_.Destroy();
  this_._impl_.~Impl_();
}

#ifndef PROTOBUF_MESSAGE_GLOBALS
PROTOBUF_CONSTINIT PROTOBUF_ATTRIBUTE_INIT_PRIORITY1 const
    ::google::protobuf::internal::ClassDataFull StopStreamingContexts_class_data_ =
        StopStreamingContexts::InternalGenerateClassData_(StopStreamingContexts_globals_._default);

PROTOBUF_ATTRIBUTE_WEAK const ::google::protobuf::internal::ClassData* PROTOBUF_NONNULL
StopStreamingContexts::GetClassData() const {
  ::google::protobuf::internal::PrefetchToLocalCache(&StopStreamingContexts_class_data_);
  ::google::protobuf::internal::PrefetchToLocalCache(StopStreamingContexts_class_data_.tc_table);
  return StopStreamingContexts_class_data_.base();
}
#else
PROTOBUF_ATTRIBUTE_WEAK const ::google::protobuf::internal::ClassData* PROTOBUF_NONNULL
StopStreamingContexts::GetClassData() const {
  ::google::protobuf::internal::PrefetchToLocalCache(&StopStreamingContexts_globals_);
  ::google::protobuf::internal::PrefetchToLocalCache(
      ::google::protobuf::internal::MessageGlobalsBase::ToParseTableBase(&StopStreamingContexts_globals_));
  return StopStreamingContexts_globals_.GetClassData();
}
#endif  // !PROTOBUF_MESSAGE_GLOBALS
#ifndef PROTOBUF_MESSAGE_GLOBALS
PROTOBUF_CONSTINIT
PROTOBUF_ATTRIBUTE_INIT_PRIORITY1 const StopStreamingContexts::ParseTableT_
    StopStreamingContexts::_table_ =
        StopStreamingContexts::InternalGenerateParseTable_(StopStreamingContexts_class_data_.base());
#endif  // !PROTOBUF_MESSAGE_GLOBALS
PROTOBUF_NOINLINE void StopStreamingContexts::Clear() {
// @@protoc_insertion_point(message_clear_start:context.v1.StopStreamingContexts)
  ::google::protobuf::internal::TSanWrite(&_impl_);
  ::uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void) cached_has_bits;

  cached_has_bits = _impl_._has_bits_[0];
  if (BatchCheckHasBit(cached_has_bits, 0x00000003U)) {
    if (CheckHasBit(cached_has_bits, 0x00000001U)) {
      _impl_.claim_id_.ClearNonDefaultToEmpty();
    }
    if (CheckHasBit(cached_has_bits, 0x00000002U)) {
      _impl_.node_id_.ClearNonDefaultToEmpty();
    }
  }
  _impl_.reason_ = 0;
  _impl_._has_bits_.Clear();
  _internal_metadata_.Clear<::google::protobuf::UnknownFieldSet>();
}

#if defined(PROTOBUF_CUSTOM_VTABLE)
::uint8_t* PROTOBUF_NONNULL StopStreamingContexts::_InternalSerialize(
    const ::google::protobuf::MessageLite& base, ::uint8_t* PROTOBUF_NONNULL target,
    ::google::protobuf::io::EpsCopyOutputStream* PROTOBUF_NONNULL stream) {
  const StopStreamingContexts& this_ = static_cast<const StopStreamingContexts&>(base);
#else   // PROTOBUF_CUSTOM_VTABLE
::uint8_t* PROTOBUF_NONNULL StopStreamingContexts::_InternalSerialize(
    ::uint8_t* PROTOBUF_NONNULL target,
    ::google::protobuf::io::EpsCopyOutputStream* PROTOBUF_NONNULL stream) const {
  const StopStreamingContexts& this_ = *this;
#endif  // PROTOBUF_CUSTOM_VTABLE
  if constexpr (::_pbi::DebugHardenCheckHasBitConsistency()) {
    this_.CheckHasBitConsistency();
  }
  // @@protoc_insertion_point(serialize_to_array_start:context.v1.StopStreamingContexts)
  ::uint32_t cached_has_bits = 0;
  (void)cached_has_bits;

  cached_has_bits = this_._impl_._has_bits_[0];
  // string claim_id = 1;
  if (CheckHasBit(cached_has_bits, 0x00000001U)) {
    if (!this_._internal_claim_id().empty()) {
      const ::std::string& _s = this_._internal_claim_id();
      ::google::protobuf::internal::WireFormatLite::VerifyUtf8String(
          _s.data(), static_cast<int>(_s.length()), ::google::protobuf::internal::WireFormatLite::SERIALIZE, "context.v1.StopStreamingContexts.claim_id");
      target = stream->WriteStringMaybeAliased(1, _s, target);
    }
  }

  // string node_id = 2;
  if (CheckHasBit(cached_has_bits, 0x00000002U)) {
    if (!this_._internal_node_id().empty()) {
      const ::std::string& _s = this_._internal_node_id();
      ::google::protobuf::internal::WireFormatLite::VerifyUtf8String(
          _s.data(), static_cast<int>(_s.length()), ::google::protobuf::internal::WireFormatLite::SERIALIZE, "context.v1.StopStreamingContexts.node_id");
      target = stream->WriteStringMaybeAliased(2, _s, target);
    }
  }

  // .context.v1.StopStreamingContextsReason reason = 3;
  if (CheckHasBit(cached_has_bits, 0x00000004U)) {
    if (this_._internal_reason() != 0) {
      target = stream->EnsureSpace(target);
      target = ::_pbi::WireFormatLite::WriteEnumToArray(
          3, this_._internal_reason(), target);
    }
  }

  if (ABSL_PREDICT_FALSE(this_._internal_metadata_.have_unknown_fields())) {
    target =
        ::_pbi::WireFormat::InternalSerializeUnknownFieldsToArray(
            this_._internal_metadata_.unknown_fields<::google::protobuf::UnknownFieldSet>(::google::protobuf::UnknownFieldSet::default_instance), target, stream);
  }
  // @@protoc_insertion_point(serialize_to_array_end:context.v1.StopStreamingContexts)
  return target;
}

#if defined(PROTOBUF_CUSTOM_VTABLE)
::size_t StopStreamingContexts::ByteSizeLong(const MessageLite& base) {
  const StopStreamingContexts& this_ = static_cast<const StopStreamingContexts&>(base);
#else   // PROTOBUF_CUSTOM_VTABLE
::size_t StopStreamingContexts::ByteSizeLong() const {
  const StopStreamingContexts& this_ = *this;
#endif  // PROTOBUF_CUSTOM_VTABLE
  // @@protoc_insertion_point(message_byte_size_start:context.v1.StopStreamingContexts)
  ::size_t total_size = 0;

  ::uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void)cached_has_bits;

  ::_pbi::Prefetch5LinesFrom7Lines(&this_);
  cached_has_bits = this_._impl_._has_bits_[0];
  if (BatchCheckHasBit(cached_has_bits, 0x00000007U)) {
    // string claim_id = 1;
    if (CheckHasBit(cached_has_bits, 0x00000001U)) {
      if (!this_._internal_claim_id().empty()) {
        total_size += 1 + ::google::protobuf::internal::WireFormatLite::StringSize(
                                        this_._internal_claim_id());
      }
    }
    // string node_id = 2;
    if (CheckHasBit(cached_has_bits, 0x00000002U)) {
      if (!this_._internal_node_id().empty()) {
        total_size += 1 + ::google::protobuf::internal::WireFormatLite::StringSize(
                                        this_._internal_node_id());
      }
    }
    // .context.v1.StopStreamingContextsReason reason = 3;
    if (CheckHasBit(cached_has_bits, 0x00000004U)) {
      if (this_._internal_reason() != 0) {
        total_size += 1 +
                      ::_pbi::WireFormatLite::EnumSize(this_._internal_reason());
      }
    }
  }
  return this_.MaybeComputeUnknownFieldsSize(total_size,
                                             &this_._impl_._cached_size_);
}

void StopStreamingContexts::MergeImpl(::google::protobuf::MessageLite& to_msg,
                      const ::google::protobuf::MessageLite& from_msg) {
   auto* const _this = static_cast<StopStreamingContexts*>(&to_msg);
  auto& from = static_cast<const StopStreamingContexts&>(from_msg);
  if constexpr (::_pbi::DebugHardenCheckHasBitConsistency()) {
    from.CheckHasBitConsistency();
  }
  // @@protoc_insertion_point(class_specific_merge_from_start:context.v1.StopStreamingContexts)
  ABSL_DCHECK_NE(&from, _this);
  ::uint32_t cached_has_bits = 0;
  (void)cached_has_bits;

  cached_has_bits = from._impl_._has_bits_[0];
  if (BatchCheckHasBit(cached_has_bits, 0x00000007U)) {
    if (CheckHasBit(cached_has_bits, 0x00000001U)) {
      if (!from._internal_claim_id().empty()) {
        _this->_internal_set_claim_id(from._internal_claim_id());
      } else {
        if (_this->_impl_.claim_id_.IsDefault()) {
          _this->_internal_set_claim_id("");
        }
      }
    }
    if (CheckHasBit(cached_has_bits, 0x00000002U)) {
      if (!from._internal_node_id().empty()) {
        _this->_internal_set_node_id(from._internal_node_id());
      } else {
        if (_this->_impl_.node_id_.IsDefault()) {
          _this->_internal_set_node_id("");
        }
      }
    }
    if (CheckHasBit(cached_has_bits, 0x00000004U)) {
      if (from._internal_reason() != 0) {
        _this->_impl_.reason_ = from._impl_.reason_;
      }
    }
  }
  _this->_impl_._has_bits_[0] |= cached_has_bits;
  _this->_internal_metadata_.MergeFrom<::google::protobuf::UnknownFieldSet>(
      from._internal_metadata_);
}

void StopStreamingContexts::CopyFrom(const StopStreamingContexts& from) {
  // @@protoc_insertion_point(class_specific_copy_from_start:context.v1.StopStreamingContexts)
  if (&from == this) return;
  Clear();
  MergeFrom(from);
}


void StopStreamingContexts::InternalSwap(StopStreamingContexts* PROTOBUF_RESTRICT PROTOBUF_NONNULL other) {
  using ::std::swap;
  auto* arena = GetArena();
  ABSL_DCHECK_EQ(arena, other->GetArena());
  _internal_metadata_.InternalSwap(&other->_internal_metadata_);
  swap(_impl_._has_bits_[0], other->_impl_._has_bits_[0]);
  ::_pbi::ArenaStringPtr::InternalSwap(&_impl_.claim_id_, &other->_impl_.claim_id_, arena);
  ::_pbi::ArenaStringPtr::InternalSwap(&_impl_.node_id_, &other->_impl_.node_id_, arena);
  swap(_impl_.reason_, other->_impl_.reason_);
}

::google::protobuf::Metadata StopStreamingContexts::GetMetadata() const {
  return ::google::protobuf::Message::GetMetadataImpl(GetClassData()->full());
}
// @@protoc_insertion_point(namespace_scope)
}  // namespace v1
}  // namespace context
namespace google {
namespace protobuf {
}  // namespace protobuf
}  // namespace google
// @@protoc_insertion_point(global_scope)
PROTOBUF_ATTRIBUTE_INIT_PRIORITY2 static ::std::false_type
    _static_init2_ [[maybe_unused]] =
        (::_pbi::AddDescriptors(&descriptor_table_context_2fv1_2fstream_2eproto),
         ::std::false_type{});
#include "google/protobuf/port_undef.inc"
